/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Attr;


/**
 * Propagates request parameters to links so that navigation from link to link
 * maintains state without resorting to sessions.  Configurable with respect to
 * {@link #setLimitToHosts(String[]) hosts}, {@link #setPersistParameters(String[] parameters},
 * {@link #setMutuallyExclusiveParameters(Map) parameter combinations}, and
 * {@link #setMutuallyExclusiveParametersForPaths(Map) parameter/path combinations}. 
 */
public class ParamPropagatorNodeFilter extends AbstractPathModifierNodeFilter {
    private final Map/*<String, String[]>*/ requestParameterMap;
    private final String requestContextPath, requestServletPath;
    private String[] hosts;
    private Map/*<String, String[]>*/ mutuallyExclusiveParameters, mutuallyExclusiveParametersForPaths;
    private boolean invertPathMatchFormutuallyExclusiveParametersForPaths;

    /**
     * @param requestParameterMap the request parameter map
     * @param requestContextPath the request context path
     * @param requestServletPath the request servlet path
     * @throws IllegalArgumentException if any of the provided arguments are null
     */
    ParamPropagatorNodeFilter(final Map/*<String, String[]>*/ requestParameterMap, final String requestContextPath, final String requestServletPath) {
        if (requestParameterMap == null || requestContextPath == null || requestServletPath == null)
            throw new IllegalArgumentException("Constructor arguments must all be non-null");
        this.requestParameterMap = new HashMap/*<String, String[]>*/(requestParameterMap);
        this.requestContextPath = requestContextPath;
        this.requestServletPath = requestServletPath;
    }

    /**
     * Optionally use to explicitly limit propagation of request parameters to
     * certain hosts for links that are fully qualified, e.g., "http://somehost/somepath/somepage.html".
     * If not explicitly set, propagation occurs by default.  This setting has
     * no effect upon relative links.
     * <p>
     * If it is known that no links in a document are fully qualified, don't
     * bother setting a value here as any non-null value will result in slight
     * overhead to check whether the host part exists on each link.
     * </p>
     * <p>
     * If some fully qualified links may exist and the intent is to disable request
     * parameter propagation on all of them, simply provide an empty array.
     * </p>
     * 
     * @param hosts explicit hosts upon which to propagate request parameters
     * @return the current object, fluent style
     */
    public ParamPropagatorNodeFilter setLimitToHosts(final String[] hosts) {
        this.hosts = hosts;
        return this;
    }

    /**
     * Optionally use to explicitly limit request parameters to persist.  This modifies
     * the parameterMap provided in the constructor to only include those parameters
     * provided here.
     * <p>
     * If not explicitly limited, all parameters are candidates for persisting,
     * though still subject to configured mutually exclusive 
     * {@link #setMutuallyExclusiveParameters(Map) parameter combinations} and
     * {@link #setMutuallyExclusiveParametersForPaths(Map) parameters/path combinations}.
     * </p>
     * 
     * @param persistParameters explicit request parameters to persist, serving
     *                          to exclude request parameters not provided here
     * @return the current object, fluent style
     */
    public ParamPropagatorNodeFilter setPersistParameters(final String[] persistParameters) {
        if (persistParameters != null && persistParameters.length > 0) {
            requestParameterMap.keySet().retainAll(Arrays.asList(persistParameters));
        }
        return this;
    }

    /**
     * Optionally use to specify parameters that shouldn't exist together on a given
     * link (where params on a given link override those by the same name on the request).
     * <p>
     * The Map structure is the same as the request parameter map, i.e., Map<String, String[]>.
     * That is, keyed by parameter name with values being one, or more, other
     * parameter names (not values) that should NOT exist along with the key parameter.
     * <p>
     * 
     * @param mutuallyExclusiveParameters parameter combinations that shouldn't
     *                                    exist together on a given link
     * @return the current object, fluent style
     */
    public ParamPropagatorNodeFilter setMutuallyExclusiveParameters(final Map/*<String, String[]>*/ mutuallyExclusiveParameters) {
        this.mutuallyExclusiveParameters = new HashMap/*<String, String[]>*/(mutuallyExclusiveParameters);
        return this;
    }

    /**
     * Same as {@link #setMutuallyExclusiveParametersForPaths(Map, boolean) setMutuallyExclusiveParametersForPaths(mutuallyExclusiveParametersForPaths, false)}
     * 
     * @param mutuallyExclusiveParametersForPaths parameter/path combinations
     *                                            where specified parameters
     *                                            shouldn't be propagated on
     *                                            links starting with specified paths
     * @return the current object, fluent style
     */
    public ParamPropagatorNodeFilter setMutuallyExclusiveParametersForPaths(final Map/*<String, String[]>*/ mutuallyExclusiveParametersForPaths) {
        return setMutuallyExclusiveParametersForPaths(mutuallyExclusiveParametersForPaths, false);
    }

    /**
     * Optionally use to specify request parameters that shouldn't be propagated
     * on links starting with specified servlet paths.  Also applies to links
     * that are empty, e.g., href="", or contain only request parameters, e.g.,
     * href="?foo=bar", in which case the specified paths are checked to see
     * whether they start with the the requestServletPath provided in the constructor
     * (as empty paths inherit the current requestServletPath).
     * <p>
     * The Map structure is the same as the request parameter map, i.e., Map<String, String[]>.
     * That is, keyed by parameter name with values being one, or more, servlet
     * paths that should NOT have the key parameter propagated upon links starting
     * with the specified servlet paths, nor on empty links where specified servlet
     * paths start with the constructor-provided requestServletPath.
     * <p>
     * <p>
     * Note that specified servlet paths should NOT include the context path which,
     * if it exists on a link, is stripped off prior to checking whether the link
     * starts with one of the specified servlet paths, e.g., use "/MyServletPath"
     * instead of "/MyContext/MyServletPath". 
     * </p>
     * <p>
     * Also note that this exclusivity only applies to propagation of <strong>request<strong>
     * parameters.  Parameters specified on the links themselves will remain on
     * the link.
     * </p>
     * 
     * @param mutuallyExclusiveParametersForPaths parameter/path combinations
     *                                            where specified parameters
     *                                            shouldn't be propagated on
     *                                            links starting with specified paths
     * @param invertPathMatch when true path matching is inverted, i.e., specified
     *                        paths are checked for NOT matching, in other words
     *                        "match anything but the specified paths"
     * @return the current object, fluent style
     */
    public ParamPropagatorNodeFilter setMutuallyExclusiveParametersForPaths(final Map/*<String, String[]>*/ mutuallyExclusiveParametersForPaths, boolean invertPathMatch) {
        this.mutuallyExclusiveParametersForPaths = new HashMap/*<String, String[]>*/(mutuallyExclusiveParametersForPaths);
        this.invertPathMatchFormutuallyExclusiveParametersForPaths = invertPathMatch;
        return this;
    }

    protected boolean didModifyPathAttr(final Set/*<Attr>*/ pathAttrs) {
        boolean retVal = false;
        if (requestParameterMap == null || requestParameterMap.isEmpty()) return retVal;

        final boolean hostsIsNotNullButEmpty = hosts != null && hosts.length == 0
            , mutuallyExclusiveParameterMapIsEmpty = mutuallyExclusiveParameters == null || mutuallyExclusiveParameters.isEmpty()
            , mutuallyExclusiveParametersForPathMapIsEmpty = mutuallyExclusiveParametersForPaths == null || mutuallyExclusiveParametersForPaths.isEmpty();

        for (final Iterator/*<Attr>*/ iter = pathAttrs.iterator(); iter.hasNext();) {
            final Attr pathAttr = (Attr) iter.next();
            final Map/*<String, String[]>*/ mergedParameterMap = new HashMap/*<String, String[]>*/(requestParameterMap);
            final String path = pathAttr.getValue().trim()
                , pathLessContext = path.startsWith(requestContextPath) ? path.substring(requestContextPath.length()) : path;

            if (hosts != null) {
                final String host = getHost(path);
                if (host != null) {
                    if (hostsIsNotNullButEmpty) continue;
                    boolean isModifiableHost = false;
                    for (int i = 0; i < hosts.length; i++) {
                        if (host.equals(hosts[i])) {
                            isModifiableHost = true;
                            break;
                        }
                    }
                    if (!isModifiableHost) continue;
                }
            }

            final Map/*<String, String[]>*/ pathParameterMap;
            int pathQueryIdx = path.indexOf("?");
            if (pathQueryIdx > -1) {
                final String query = path.substring(pathQueryIdx + 1);
                pathParameterMap = new URIParamParser(query, null).getParameterMap();
                //TODO - is there a more clean way to do this, like using Map/Set intersections in some such way?
                if (!mutuallyExclusiveParameterMapIsEmpty && !mergedParameterMap.isEmpty()) {
                    for (final Iterator innerIter = pathParameterMap.keySet().iterator(); innerIter.hasNext();) {
                        final String key = (String) innerIter.next();
                        final String[] exclusiveParams = (String[]) mutuallyExclusiveParameters.get(key);
                        if (exclusiveParams != null) {
                            for (int i = 0; i < exclusiveParams.length; i++)
                                mergedParameterMap.remove(exclusiveParams[i]);
                        }
                    }
                }
            } else {
                pathParameterMap = null;
                pathQueryIdx = path.length();
            }

            if (!mutuallyExclusiveParametersForPathMapIsEmpty && !mergedParameterMap.isEmpty()) {
                final boolean pathIsEmpty = pathQueryIdx == 0;
                for (final Iterator innerIter = mutuallyExclusiveParametersForPaths.entrySet().iterator(); innerIter.hasNext();) {
                    final Map.Entry entry = (Map.Entry) innerIter.next();
                    final String key = (String) entry.getKey();
                    final String[] exclusivePaths = (String[]) entry.getValue();
                    if (exclusivePaths != null) {
                        for (int i = 0; i < exclusivePaths.length; i++) {
                            if ((pathIsEmpty && exclusivePaths[i].startsWith(requestServletPath) == !invertPathMatchFormutuallyExclusiveParametersForPaths) || (!pathIsEmpty && pathLessContext.startsWith(exclusivePaths[i]) == !invertPathMatchFormutuallyExclusiveParametersForPaths))
                                mergedParameterMap.remove(key);
                        }
                    }
                }
            }

            if (pathParameterMap != null) {
                //override request parameters with path parameters
                mergedParameterMap.putAll(pathParameterMap);
            }

            pathAttr.setValue(mergedParameterMap.isEmpty() ? path : path.substring(0, pathQueryIdx) + "?" + URIParamParser.formatNonEncoded(mergedParameterMap));
            retVal = true;
        }
        return retVal;
    }

}
