/*
 * Copyright (C) 2003, 2010 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.io.IOException;
import java.util.Locale;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.enhydra.xml.dom.DOMOps;
import org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;
import org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory;
import org.enhydra.xml.xmlc.html.HTMLObject;
import org.enhydra.xml.xmlc.misc.I18nUtil;
import org.enhydra.xml.xmlc.servlet.XMLCContext;
import org.w3c.dom.Document;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;

/**
 * Demonstrates loading documents using request path info as the document path
 * rather than parameters, i.e., it uses the path after '/Preview/*'.
 */
public class Preview extends HttpServlet {

    private static final String dynDomFactoryClassName = XercesHTMLDomFactory.class.getName();
    private static final XMLCCreateOptions 
        frameworkXmlcCreateOptions = new XMLCCreateOptions()
                                         .setDynMarkupFilePath("demo/framework.html")
                                         .setDynDomFactoryClassName(dynDomFactoryClassName)
        , demoLinkXmlcCreateOptions = new XMLCCreateOptions()
                                          .setDynMarkupFilePath("xmlc/demo/demoLinks.ssi")
                                          .setDynDomFactoryClassName(dynDomFactoryClassName);

    private XMLCContext context = null;

    public void init() throws ServletException {
        context = XMLCContext.getContext(this);
    }

    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
        doGet(req, resp);
    }

    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
        XMLCDeferredParsingFactory dpFactory = context.getXMLCDeferredParsingFactory();

        String docPath = req.getPathInfo();
        if (!(docPath.endsWith("ml") || docPath.endsWith("htm"))) {
            //ensure we don't process .js/.css/.gif/.jpg/etc... files here
            try {
                super.doGet(req, resp);
                return;
            } catch (IOException ioe) {
                throw new ServletException(ioe);
            }
        }

        Locale locale = I18nUtil.locale(req.getParameter("locale"));
        if (locale == null) locale = req.getLocale();

        HTMLObject htmlObj = null;
        if (docPath != null) {
            htmlObj = (HTMLObject) dpFactory.create(frameworkXmlcCreateOptions);
            HTMLObject mainMlObj = (HTMLObject) dpFactory.create(new XMLCCreateOptions()
                                                                     .setDynMarkupFilePath(docPath)
                                                                     .setDynDomFactoryClassName(dynDomFactoryClassName)
                                                                     .setUserLocale(locale)
                                                                     //.setFallbackLocale(Locale.US)
                                                                 )
                , subMlObj = (HTMLObject) dpFactory.create(demoLinkXmlcCreateOptions);

            htmlObj.setTitle(mainMlObj.getTitle());
            new DOMDecorator()
                .replaceSectionWith("header", DOMDecorator.newEl(mainMlObj, "div")
                                                  .setAttribute("id", "outer")
                                                  .appendChild(DOMDecorator.newEl(mainMlObj, "h1")
                                                                   .setTextContent("Welcome To The XMLC Project Demo")
                                                                   .setAttribute("style", "color:white;")
                                                               )
                                                  .getElement())
                .replaceSectionWithContents("main", mainMlObj.getBody())
                .surroundSectionWith("main", DOMDecorator.newEl(mainMlObj, "div")
                                                 .setAttribute("id", "supermain")
                                                 .getElement())
                .replaceSectionWith("local", htmlObj.createCDATASection("<h1>\"'Fun With The CDATA Hack'\"</h1>"))
                //.clearSection("sub")
                .replaceSectionWithContents("sub", subMlObj.getElementById("demoLinks"))
                .replaceSectionWith("footer", DOMDecorator.newEl(mainMlObj, "p")
                                                  .setTextContent(" Copyright 2010 XMLC Project")
                                                  .getElement())
                .decorate(htmlObj);
        }

        if (htmlObj != null) {
            postprocessDocument(htmlObj, req);

            try {
                org.enhydra.xml.io.OutputOptions oo = context.createOutputOptions(req, resp, htmlObj);
                oo.setUseAposEntity(false);
                oo.setUseHTML4Entities(true);
                oo.setPrettyPrinting(true);
                oo.setIndentSize(2);
                oo.setPreserveSpace(false);
                oo.setURLRewriteAttributes(new String[]{"href"});
                oo.setOmitDocType(false);
                oo.setFormat(org.enhydra.xml.io.OutputOptions.FORMAT_XHTML);
                oo.setEnableXHTMLCompatibility(true);
                oo.setEnableXMLCDATAHack(true);
                //oo.setPublicId("-//W3C//DTD XHTML 1.0 Transitional//EN");
                //oo.setSystemId("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd");
                //oo.setMIMEType("application/xhtml+xml");
                context.writeDOM(req, resp, oo, htmlObj);
            } catch (IOException ioe) {
                throw new ServletException(ioe);
            }
        } else {
            try {
                if (docPath == null) {
                    resp.sendError(HttpServletResponse.SC_NOT_FOUND, "Must provide request path info for document.");
                } else {
                    resp.sendError(HttpServletResponse.SC_NOT_FOUND, "Document '" + docPath + "' not found");
                }
            } catch (IOException ioe) {
                throw new ServletException(ioe);
            }
        }
    }

    private static void postprocessDocument(final Document root, final HttpServletRequest req) {
        //Fully qualify relative links, otherwise they will end up relative
        //to contextPath + pathInfo where they will not be found
        //Do below or just create a <base> tag with the href being the path
        //prefix, though below only modifies relevant attributes so it is less
        //intrusive/imposing.
        //Propagate request parameters on relevant links while taking care to
        //allow pre-existing query parameters on links to take precedence.
        boolean expandEntityReferences = false;
        NodeIterator iter =
            ((DocumentTraversal) DOMOps.getDocument(root)).createNodeIterator(
                 root,
                 NodeFilter.SHOW_ELEMENT,
                 new CompoundNodeFilter(
                     new NodeFilter[] {
                         new PathQualifierNodeFilter(req.getContextPath() + "/"),
                         ReloadTest.newParamPropagatorNodeFilter(req)
                     }
                 ),
                 expandEntityReferences);
        //Must iterate in order to apply NodeFilter changes
        while (iter.nextNode() != null);
        iter.detach();
    }

}

