/*
 * Copyright (C) 2003, 2010 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.io.IOException;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.enhydra.xml.dom.DOMOps;
import org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;
import org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory;
import org.enhydra.xml.xmlc.html.HTMLObject;
import org.enhydra.xml.xmlc.misc.I18nUtil;
import org.enhydra.xml.xmlc.servlet.XMLCContext;
import org.w3c.dom.Document;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeFilter;
import org.w3c.dom.traversal.NodeIterator;

/**
 * Demonstrates loading documents using servlet parameters; either a fully
 * qualified class name ('docClass') name or a relative document path ('doc').
 */
public class ReloadTest extends HttpServlet {

    private static final String dynDomFactoryClassName = XercesHTMLDomFactory.class.getName(),
        DOC_CLASS_PARAM = "docClass",
        DOC_PARAM = "doc";
    private static final String[] parameterPropagationHosts = new String[]{};
    private static final Map/*<String, String[]>*/ mutuallyExclusiveParameterMap, mutuallyExclusiveParametersForPathMap;
    static {
        mutuallyExclusiveParameterMap = new HashMap/*<String, String[]>*/();
        mutuallyExclusiveParameterMap.put(DOC_PARAM, new String[] {DOC_CLASS_PARAM});
        mutuallyExclusiveParameterMap.put(DOC_CLASS_PARAM, new String[] {DOC_PARAM});

        mutuallyExclusiveParametersForPathMap = new HashMap/*<String, String[]>*/();
        mutuallyExclusiveParametersForPathMap.put(ReloadTest.DOC_PARAM, new String[] {"/ReloadTest"});
        mutuallyExclusiveParametersForPathMap.put(ReloadTest.DOC_CLASS_PARAM, new String[] {"/ReloadTest"});
    }

    private XMLCContext context = null;

    public void init() throws ServletException {
        context = XMLCContext.getContext(this);
    }

    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
        doGet(req, resp);
    }

    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws ServletException {
        XMLCDeferredParsingFactory dpFactory = context.getXMLCDeferredParsingFactory();

        String docPath = req.getParameter(DOC_PARAM);
        String docClass = req.getParameter(DOC_CLASS_PARAM);
        Locale locale = I18nUtil.locale(req.getParameter("locale"));
        if (locale == null) locale = req.getLocale();

        HTMLObject htmlObj = null;
        if (docPath != null) {
            htmlObj = (HTMLObject) dpFactory.create(new XMLCCreateOptions()
                                                        .setDynMarkupFilePath(docPath)
                                                        .setDynDomFactoryClassName(dynDomFactoryClassName)
                                                        .setUserLocale(locale)
                                                        //.setFallbackLocale(Locale.US)
                                                    );
        } else if (docClass != null) {
            htmlObj = (HTMLObject) dpFactory.create(new XMLCCreateOptions()
                                                        .setXmlcClassName(docClass)
                                                        .setUserLocale(locale)
                                                        //.setFallbackLocale(Locale.US)
                                                    );
        }

        if (htmlObj != null) {
            postprocessDocument(htmlObj, req);

            try {
                org.enhydra.xml.io.OutputOptions oo = context.createOutputOptions(req, resp, htmlObj);
                oo.setUseAposEntity(false);
                oo.setUseHTML4Entities(true);
                oo.setPrettyPrinting(true);
                oo.setIndentSize(2);
                oo.setPreserveSpace(false);
                oo.setURLRewriteAttributes(new String[]{"href"});
                oo.setOmitDocType(false);
                oo.setFormat(org.enhydra.xml.io.OutputOptions.FORMAT_HTML);
                oo.setForceHTMLLowerCase(true);
                //oo.setPublicId("-//W3C//DTD HTML 4.01 Transitional//EN");
                //oo.setSystemId("http://www.w3.org/TR/html401/loose.dtd");
                //oo.setMIMEType("text/html");
                context.writeDOM(req, resp, oo, htmlObj);
            } catch (IOException ioe) {
                throw new ServletException(ioe);
            }
        } else {
            try {
                if (docPath == null && docClass == null) {
                    resp.sendError(HttpServletResponse.SC_NOT_FOUND, "Must set '" + DOC_PARAM + "' or '" + DOC_CLASS_PARAM + "' request parameter");
                } else if (docPath != null) {
                    resp.sendError(HttpServletResponse.SC_NOT_FOUND, DOC_PARAM + " '" + docPath + "' not found");
                } else {
                    resp.sendError(HttpServletResponse.SC_NOT_FOUND, DOC_CLASS_PARAM + " '" + docClass + "' not found");
                }
            } catch (IOException ioe) {
                throw new ServletException(ioe);
            }
        }
    }

    private static void postprocessDocument(final Document root, final HttpServletRequest req) {
        //Propagate request parameters on relevant links while taking care to
        //allow pre-existing query parameters on links to take precedence.
        boolean expandEntityReferences = false;
        NodeIterator iter =
            ((DocumentTraversal) DOMOps.getDocument(root)).createNodeIterator(
                 root,
                 NodeFilter.SHOW_ELEMENT,
                 newParamPropagatorNodeFilter(req),
                 expandEntityReferences);
        //Must iterate in order to apply NodeFilter changes
        while (iter.nextNode() != null);
        iter.detach();
    }

    static ParamPropagatorNodeFilter newParamPropagatorNodeFilter(final HttpServletRequest req) {
        return new ParamPropagatorNodeFilter(req.getParameterMap(), req.getContextPath(), req.getServletPath())
            .setLimitToHosts(parameterPropagationHosts)
            .setMutuallyExclusiveParameters(mutuallyExclusiveParameterMap)
            .setMutuallyExclusiveParametersForPaths(mutuallyExclusiveParametersForPathMap, true);
    }

}

