/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jregex.Pattern;

/**
 * A Stand-alone URI parameter parser mimicking HttpServletRequest parameter
 * parsing behavior.
 */
public class URIParamParser
{
    private static final String //DEFAULT_ENCODING = "UTF-8",
        PARAM_SEP = "&",
        NAME_VALUE_SEP = "=";
    private static final Pattern RE_PARAM_SPLITTER = new Pattern(PARAM_SEP);

    final Map/*<String, String[]>*/ parameters;

    /**
     * Note that the query string is expected to exclude the '?' prefix.  If it
     * exists, it will be treated as part of the first parameter name which is
     * most likely not what is intended.
     *
     * @param query    the query string, may be null or empty which will result
     *                 in an empty {@link #getParameterMap() parameter map}
     * @param encoding the encoding to use for performing URL-encoding, defaults
     *                 to "UTF-8" if null
     */
    public URIParamParser(String query, final String encoding)
    {
        parameters = new HashMap/*<String, String[]>*/();
        if (query == null || (query = trimRight(query)).length() == 0) return;

        final Map/*<String, List<String>>*/ listParameters = new HashMap/*<String, List<String>>*/();

        final String params[] = RE_PARAM_SPLITTER.tokenizer(query).split();
        for (int i=0; i < params.length; i++)
        {
            final String param = params[i];
            final String name;
            final String value;
            final int pos = param.indexOf(NAME_VALUE_SEP);
            if (pos == -1)
            {
                name = param;
                if (name.length() == 0) continue; //ignore && to mimic Tomcat's behavior
                value = "";
            }
            else
            {
                name = decode(param.substring(0, pos), encoding);
                if (name.length() == 0) continue; //ignore &= to mimic Tomcat's behavior
                value = decode(param.substring(pos + 1, param.length()), encoding);
            }
            List/*<String>*/ list = (List) listParameters.get(name);
            if (list == null)
            {
                list = new ArrayList/*<String>*/();
                listParameters.put(name, list);
            }
            list.add(value);
        }

        for (final Iterator iter = listParameters.entrySet().iterator(); iter.hasNext();)
        {
            final Map.Entry/*<String, List<String>>*/ entry = (Map.Entry) iter.next();
            final List/*<String>*/ list = (List) entry.getValue();
            parameters.put(entry.getKey(), list.toArray(new String[list.size()]));
        }
    }

    /**
     * @param name the parameter name
     * @return the parameter value (the first value if multiple exist) or null
     *         if the parameter name does not exist
     */
    public String getParameter(final String name)
    {
        final String[] values = getParameterValues(name);
        return values != null ? values[0] : null;
    }

    /**
     * @param name the parameter name
     * @return the parameter values or null if the parameter name does not exist
     */
    public String[] getParameterValues(final String name)
    {
        return (String[]) parameters.get(name);
    }

    /**
     * @return the parameters names, empty if no parameters exist
     */
    public Enumeration/*<String>*/ getParameterNames()
    {
        return Collections.enumeration(parameters.keySet());
    }

    /**
     * @return the parameter map, empty if no parameters exist
     */
    public Map/*<String, String[]>*/ getParameterMap()
    {
        return new HashMap/*<String, String[]>*/(parameters);
    }

    /**
     * Formats a parameter map as a, non URL-encoded, query string useful for
     * appending to link attributes in documents where the browser, itself, will
     * perform the encoding when the links are followed.
     *
     * @param parameters the parameters, i.e., as provided by HttpRequest.getParamterMap()),
     *                   may be null or empty in which case an empty string will be returned
     * @return a non-URL encoded string of parameters separated by '&amp;', with
     *         name/value pairs separated by '=', NOT prefixed by '?'
     */
    public static final String formatNonEncoded(final Map/*<String, String[]>*/ parameters)
    {
        return formatImpl(parameters, null, false);
    }

    /**
     * Formats a parameter map as a, URL-encoded, query string useful for appending
     * to URLs passed directly into HTTP clients.
     *
     * @param parameters the parameters, i.e., as provided by HttpRequest.getParamterMap()),
     *                   may be null or empty in which case an empty string will be returned
     * @param encoding   preferred encoding, defaults to "UTF-8" if null
     * @return a URL encoded string of parameters separated by '&amp;', with
     *         name/value pairs separated by '=', NOT prefixed by '?'
     */
    public static final String format(final Map/*<String, String[]>*/ parameters, final String encoding)
    {
        return formatImpl(parameters, encoding, true);
    }

    private static String formatImpl(final Map/*<String, String[]>*/ parameters, final String encoding, final boolean useEncoding)
    {
        if (parameters == null || parameters.isEmpty()) return "";
        final StringBuffer result = new StringBuffer();
        for (final Iterator iter = parameters.keySet().iterator(); iter.hasNext();)
        {
            final String name = (String) iter.next();
            String[] values = (String[]) parameters.get(name);
            //Ensure at least one element exists in the array so all parameters get
            //appended to the query string, regardless of value availability
            if (values == null || values.length == 0) values = new String[]{""};
            final String encodedName = useEncoding ? encode(name, encoding) : name;
            for (int i = 0; i < values.length; i++)
            {
                final String value = values[i];
                result.append(PARAM_SEP);
                result.append(encodedName);
                result.append(NAME_VALUE_SEP);
                result.append(useEncoding ? encode(value, encoding) : value);
            }
        }
        //Must append '&' in order to maintain right-most white space value
        //because URI's are right-trimmed by most user agents.  Note, if '+'
        //exists as a parameter, URLEncoder.encode() encodes it as "%2B".
        //However, when followed by '=', URLEncoder.encode() treats it as plain
        //old whitespace.  This inconsistency seems like a bug in
        //URLEncoder.encode().  Then again, maybe '+' is just not supported as
        //a literal value in URI parameters (though Tomcat always treats it as
        //whitespace using its own encoder)?
        char lastChar = result.charAt(result.length() - 1);
        if (lastChar == '+' || Character.isWhitespace(lastChar))
        {
            result.append(PARAM_SEP);
        }
        //Delete the first parameter separator instead of checking for buffer
        //length each time through the loop
        return result.deleteCharAt(0).toString();
    }

    private static String decode(final String content, final String encoding)
    {
//        try
//        {
            return URLDecoder.decode(content);//, encoding != null ? encoding : DEFAULT_ENCODING);
//        }
//        catch (final UnsupportedEncodingException uee)
//        {
//            throw new IllegalArgumentException(uee.getMessage());
//        }
    }

    private static String encode(final String content, final String encoding)
    {
//        try
//        {
            return URLEncoder.encode(content);//, encoding != null ? encoding : DEFAULT_ENCODING);
//        }
//        catch (final UnsupportedEncodingException uee)
//        {
//            throw new IllegalArgumentException(uee.getMessage());
//        }
    }

    private static String trimRight(final String text)
    {
        int idx = text.length() - 1;
        if (idx >= 0)
        {
            while (text.charAt(idx) == ' ')
            {
                idx--;
            }

            if (idx < 0)
            {
                return "";
            }
            else
            {
                return text.substring(0, idx + 1);
            }
        }
        else
        {
            return text;
        }
    }

}
