/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package xmlc.demo;

import java.net.MalformedURLException;
import java.net.URL;

import javax.servlet.ServletContext;

import org.enhydra.xml.xmlc.deferredparsing.ResourceLoader;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;
import org.enhydra.xml.xmlc.servlet.ServletDocumentLoaderImpl;
import org.enhydra.xml.xmlc.servlet.ServletResourceLoaderImpl;

/**
 * A Servlet Document Loader which provides a Resource Loader that prevents certain
 * paths from being loaded by certain contexts for security reasons, as paths in
 * the demo app are user-defined and could, otherwise, be used to load arbitrary
 * resources, e.g., "/META-INF/somePrivateConfig.xml" from the classpath or
 * "WEB-INF/web.xml" from the servlet context.
 * 
 * @author Jacob Kjome
 */
public class ValidatingDocumentLoader extends ServletDocumentLoaderImpl {

    private ResourceLoader fResourceLoader;

    /**
     * Provides a custom resource loader implementation that ignores paths
     * considered in relevant contexts.
     * 
     * @return a resource loader implementation
     */
    public ResourceLoader getResourceLoader() {
        if (fResourceLoader == null) {
            fResourceLoader = new ValidatingResourceLoader(getFactory(), fServletContext);
        }
        return fResourceLoader;
    }

    /**
     * A {@link ResourceLoader resource loader} implementation that provides the
     * capability of loading resources from the ServletContext while preventing
     * illegitimate access to user-defined resource paths.
     */
    private static class ValidatingResourceLoader extends ServletResourceLoaderImpl {

        /**
         * @param factory the XMLC deferred parsing factory
         * @param servletContext the servlet context
         */
        public ValidatingResourceLoader(final XMLCDeferredParsingFactory factory, final ServletContext servletContext) {
            super(factory, servletContext);
        }

        /**
         * Candidate paths starting with "META-INF/" are ignored.  This is to
         * prevent illegitimate access to protected resources, as this demo app
         * allows for user-defined paths.  Imagine a user defined path of
         * "/META-INF/somePrivateConfig.xml"; allowing loading of this path would
         * be a security hole.
         * 
         * @param candidatePath a candidate path from which to load a resource
         * @return a URL corresponding the candidatePath or null if not found
         * @see #getPathURLFromClasspath(String)
         */
        protected URL getPathURLFromClasspath(final String candidatePath) {
            return !candidatePath.startsWith("META-INF/") ? super.getPathURLFromClasspath(candidatePath) : null;
        }

        /**
         * Candidate paths starting with "WEB-INF/" and "META-INF/" are ignored.
         * This is to prevent illegitimate access to protected resources, as
         * this demo app allows for user-defined paths.  Imagine a user defined
         * path of "/WEB-INF/web.xml"; allowing loading of this path would be a
         * security hole.
         * <p>
         * If the candidate path is found to be legitimate, it is first attempted
         * literally.  If not found, it is attempted after prefixing with
         * "/WEB-INF/xmlc".  This allows for serving markup files from within
         * "/WEB-INF/", while keeping hidden the structural details of the webapp.
         * </p>
         * 
         * @param candidatePath a candidate path from which to load a resource
         *                      (with a leading slash)
         * @return a URL corresponding the candidatePath or null if not found
         * @see #getPathURLFromServletContext(String)
         */
        protected URL getPathURLFromServletContext(final String candidatePath) throws MalformedURLException {
            URL url = null;
            if (!(candidatePath.startsWith("/WEB-INF/") || candidatePath.startsWith("/META-INF/"))) {
                url = super.getPathURLFromServletContext(candidatePath);
                if (url == null) {
                    url = super.getPathURLFromServletContext("/WEB-INF/xmlc" + candidatePath);
                }
            }
            return url;
        }
    }

}
