/*
 * Copyright (C) 2009  Jacob Kjome [hojuATvisi.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.enhydra.xml.xmlc.taskdef;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.text.CollationKey;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import jregex.Matcher;
import jregex.Pattern;
import jregex.REFlags;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.types.DataType;
import org.apache.tools.ant.util.JavaEnvUtils;
import org.enhydra.xml.io.InputSourceOps;
import org.enhydra.xml.io.XmlReader;
import org.enhydra.xml.io.XmlReaderException;
import org.enhydra.xml.xmlc.misc.I18nUtil;
import org.enhydra.xml.xmlc.misc.SSIReader;
import org.xml.sax.InputSource;

/**
 * This Ant task reads base, non-localized, template files, along with associated
 * L10n property resource bundle files, and generates localized templates using
 * text substitution.  Applying a bit of a twist to standard resource bundle usage,
 * values from base property resource bundle files are used as keys to match text
 * for replacement in the base template file.  Then, the key associated with the
 * matched value is noted and used to look up the localized value from the resource
 * bundle.  At that point, the text is replaced and written to disk using the base
 * template file name, the localized resource bundle locale string, and the base
 * template file extension as the name of the file, e.g., foo.html, foo.properties,
 * and foo_en_US.properties results in foo_en_US.html.
 * <p>
 * Existing localized template files are skipped as candidates for base template
 * files even when matched by MatchingTask includes.  Likewise, Generation of
 * localized template files is avoided when they already exist and are newer than
 * corresponding base template files and property resource bundle files.  This can
 * be overridden by setting {@link #setForce} to <code>true</code>.
 * </p>
 * <p>
 * {@link #setSsi Server Side Includes} are also supported, allowing mockups to
 * use includes, as always, with resulting localized templates having included
 * content localized right along with content in the immediate template.  Localized
 * templates will actually no longer have includes, as they will all have been
 * resolved, just as XMLC does at runtime.  Setting the {@link #setSsiBase ssibase}
 * is supported, just as it is in XMLC.
 * </p>
 * <h2>Perl5.6, Java 1.3+ compatible, Regex Text Substitution Capabilities Provided By <a href="http://jregex.sourceforge.net/">JRegex</a></h2>
 * <p>
 * This task provides much more than basic String matching.  Strings provided by
 * base property resource bundle values are used to match text in templates in
 * a sophisticated manner.  Base bundle values are pre-processed, making them
 * literal regex patterns; e.g., "foo (bar)" becomes "foo\s+\(bar\)" which will
 * match the target text regardless of whitespace within it in the template.
 * Also notice that regex meta characters are escaped, ensuring all characters
 * in the string are treated literally (other than the added \s+ between words
 * to match any number of spaces, including line breaks).  The only caveat is
 * that line breaks in the target text of original template are lost upon
 * replacement, though non-localized text remains unaffected.
 * </p>
 * <p>
 * Of course, there are some cases where one might want text replaced in one part
 * of the document, but not another.  For instance, attributes often contain
 * meaningful values such as Ids and class names for CSS/Javascript or form data.
 * It was for this purpose that the {@link #addConfiguredSkipReplace &lt;skipReplace/&gt;}
 * feature was added.  This feature allows users significant control over what
 * text gets localized and what doesn't.  Regular expressions are used to match
 * contexts where L10n should be avoided.  But because regular expressions can
 * be tricky to write, and since skipping attribute L10n is such a common
 * use-case, <code>&lt;skipReplace idref="markup.attrs.all"/&gt;</code>,
 * <code>&lt;skipReplace idref="markup.attrs.except" idrefSubPrefix="[optional
 * pipe-delimited attribute names with no spaces here, defaults to [X]HTML spec
 * defined <a href="http://www.w3.org/TR/html401/sgml/dtd.html#Text">%Text;</a>
 * <a href="http://www.w3.org/TR/html401/index/attributes.html">attributes</a>]"/&gt;</code>,
 * and <code>&lt;skipReplace idref="markup.attrs" idrefSubPrefix="[required
 * pipe-delimited attribute names with no spaces here]"/&gt;</code> have been
 * pre-defined for convenience.  For more info, see usage samples below as well
 * as {@link #addConfiguredSkipReplace &lt;skipReplace/&gt; documentation}.
 * </p>
 * <h2>Designing Property Resource Bundles</h2>
 * <p>
 * Property resource bundle files are written as usual, with special care being
 * taken to match <em>values</em> in base bundle files with words/phrases in base
 * template files.  Localized bundle files need only match keys to the base bundle
 * files, as the values will serve as the replacement text.  To prevent a brittle
 * template localization infrastructure, it is important to understand why even
 * though XMLC provides fallback to the base template, a localized bundle should
 * be defined {@link #setAutoGenLocalizedBaseBundle even for the base bundle locale}.
 * </p>
 * <p>
 * Because of special base bundle processing, it is highly recommended to set the
 * {@link #setBaseBundleLocale baseBundleLocale} to the Locale string representing the
 * {@link java.util.Locale} of the base bundle property values.  Without this, the
 * system default Locale will be used, making the build environment specific.
 * This is especially important when setting
 * {@link #setAutoGenLocalizedBaseBundle autoGenLocalizedBaseBundle} to true.
 * Also, note that property resource bundle files must use the ".properties" file
 * extension, per {@link java.util.ResourceBundle} and the conventions of this task.
 * </p>
 * <p>
 * Normally, resource bundle encoding must be "ISO-8859-1".  However, since Java6,
 * it is possible to write resource bundles using any encoding supported by Java.
 * To take advantage of this feature, this task must be run under Java6 or greater
 * with {@link #setBundleEncoding bundleEncoding} set to the desired encoding.
 * This setting is ignored when running under JVMs older than Java6.  Java6+
 * capabilities are reflected upon to allow for building/running under Java1.3
 * and greater, while providing access to features provided only in Java6+.
 * </p>
 * <p>
 * By default, property resource bundle file names are inferred using the base
 * template file name (less extension) and presumed to be located in the same
 * directory as the base template file (unless the {@link #setBundleDir bundleDir}
 * attribute is configured and set to a different directory than {@link #setSrcDir srcDir}).
 * While this behavior is fairly intuitive, it does not account for reuse of a
 * single set of bundle files across multiple base template files.  To accomplish
 * this, use the {@link #setMasterBundle masterBundle} attribute to point to a master
 * base property resource bundle file.  Template associated bundle files will also
 * be used and override values in the master bundle.  This behavior can be modified
 * using two options.  First, using {@link #setMasterBundleDominant setMasterBundleDominant(true)}.
 * ensures that properties found in the master bundle override properties in
 * template associated bundles.  Second, using {@link #setMasterBundleAppliesToAll setMasterBundleAppliesToAll(true)}
 * will make it so that the master bundle applies to all base templates, whether
 * or not template associated bundles exist.
 * </p>
 * <h2>Example Usage</h2>
 * <pre>
 * &lt;taskdef name="l10n"
 *             classname="org.enhydra.xml.xmlc.taskdef.L10nTask"
 *             classpathref="project.classpath"/&gt;
 * 
 * &lt;l10n srcdir="${templates.dir}" includes="*.*ml"/&gt;
 * 
 * &lt;l10n srcdir="${templates.dir}"
 *          includes="*.*ml"
 *          destdir="${dest.dir}"
 *          bundleDir="${bundle.dir}"
 *          bundleDirFlattened="true"
 *          masterBundle="{some.dir}/master.properties"
 *          masterBundleDominant="true"
 *          masterBundleAppliesToAll="false"
 *          ssi="true"
 *          ssiBase="[path to some directory, no need to set if SSI is already relative to including markup file]"
 *          verbose="true"
 *          force="false"
 *          defaultTemplateEncoding="UTF-8"
 *          bundleEncoding="ISO-8859-1" &lt;!-- May use "UTF-8" if building/running under Java6+, negating the need for the native2ascii tool --&gt;
 *          baseBundleLocale="en_US"
 *          autoGenLocalizedBaseBundle="true"&gt;
 *   &lt;skipReplace description="skip replacement of target text inside attributes, except for those defined here"
 *                  idref="markup.attrs.except"
 *                  idRefSubPrefix="abbr|alt|label|prompt|standby|summary|title|value"/&gt;
 *   &lt;!--skipReplace description="skip replacement of target text inside all attributes, no exceptions"
 *                idref="markup.attrs.all"/--&gt;
 *   &lt;!--skipReplace description="skip replacement of target text inside those attributes defined here"
 *                idref="markup.attrs"
 *                idrefSubPrefix="class|id|name"/--&gt;
 *   &lt;!--skipReplace description="[some description of what your custom regex prefix/suffix does]"
 *                prefix="[some custom regex prefix to the target text]"
 *                suffix="[some custom regex suffix to the target text]"/--&gt;
 * &lt;/l10n&gt;
 * </pre>
 * <h2>Acknowledgment</h2>
 * <p>
 * This task is inspired by the BarracudaMVC Localize task.  Many thanks go to
 * the Localize task developers for coming up with this unique localization approach,
 * which, in concert with XMLC, enables mockups to be deployed as live dynamic
 * pages as is (other than optional automated pre-processing, such as performed by
 * this task and XMLC), maintaining designer/developer separation of concerns.
 * </p>
 * <p>
 * While the general flow for discovering and generating templates is maintained,
 * the details are nearly 100% rewritten from scratch.  This task supports many
 * more features and avoids a number of bugs and deficiencies inherent in the
 * Localize task.  As such, it is recommended that Localize users update their
 * build scripts to use this new task.  Enjoy!
 * </p>
 */
public class L10nTask extends org.apache.tools.ant.taskdefs.MatchingTask {

    static final boolean IS_JAVA6_PLUS;
    static {
        final Float javaVersion = Float.valueOf(JavaEnvUtils.getJavaVersion())
            , java6 = Float.valueOf("1.6");
        IS_JAVA6_PLUS = javaVersion.compareTo(java6) >= 0;
    }

    static final String GRP_pathBaseNameL10nPattern = "pathBaseNameL10nPattern"
    	, GRP_pathBaseName = "pathBaseName"
        , GRP_path = "path"
        , GRP_baseName = "baseName"
        , GRP_l10nPattern = "l10nPattern"
        , GRP_extension = "extension";
    static final Pattern RE_LENIENT_PATH_GROUPER = new Pattern("^({" + GRP_pathBaseNameL10nPattern + "}({" + GRP_pathBaseName + "}({" + GRP_path + "}.*[/\\\\])?(?:$|({" + GRP_baseName + "}.+?)))({" + GRP_l10nPattern + "}_[a-z]{2}(?:_+?[A-Z]{2}(?:_.+?)?)?)?)(?:({" + GRP_extension + "}\\.[^.]*$)|$)");
    private static final Pattern RE_SPACES = new Pattern("\\s+");

    protected File masterBundleFile //the master properties file for localization, may be null
        , srcDir //must not be null
        , destDir //may be null, defaults to srcDir
        , bundleDir; //may be null, defaults to srcDir
    protected int logLevel = Project.MSG_VERBOSE;
    protected boolean masterBundleDominant, masterBundleAppliesToAll, bundleDirFlattened, autoGenLocalizedBaseBundle, ssi, verbose, force;
    protected String ssiBase
        , bundleEncoding = "ISO-8859-1";
    protected Locale baseBundleLocale = Locale.getDefault();

    private Set embeddedSkipReplaceTypes = new HashSet();

    /**
     * Required - This is the one and only required attribute.  The srcDir must
     * be both configured and exist or the task will fail fast.
     * 
     * @param srcDir the directory containing the template files
     */
    public void setSrcDir(final File srcDir) {
        this.srcDir = srcDir;
    }

    /**
     * Optional
     * 
     * @param destDir the directory into which the localized template files are
     *                to be written, defaults to srcDir if not defined
     */
    public void setDestDir(final File destDir) {
        this.destDir = destDir;
    }

    /**
     * Optional - It's only useful to define this if template associated bundle
     * files are stored in a separate directory from the template files.  If
     * defined, and different from srcDir, the relative path of the template
     * files from srcDir is hierarchy in which bundles in the bundleDir must be
     * stored... except when using {@link #setBundleDirFlattened setBundleDirFlattened(true)},
     * in which case the bundleDir is flattened so that all bundles are searched
     * for in the root of bundleDir.
     * <p>
     * Keep in mind that phrases common to all templates are better placed in a
     * {@link #setMasterBundle master bundle}.
     * </p>
     * 
     * @param bundleDir the directory containing the template file associated
     *                  property resource bundle files, e.g, foo.html would
     *                  have a foo.properties base bundle and one or more localized
     *                  bundles, such as foo_en_US.properties, default is srcDir
     *                  if not defined
     * @see #setBundleDirFlattened
     * @see #setMasterBundle
     */
    public void setBundleDir(final File bundleDir) {
        this.bundleDir = bundleDir;
    }

    /**
     * Optional - Localized master bundle files are presumed to be located in
     * the same directory as the base master bundle.
     * 
     * @param masterBundle the base master L10n properties file
     * @see #setMasterBundleDominant
     * @see #setMasterBundleAppliesToAll
     */
    public void setMasterBundle(final File masterBundle) {
        this.masterBundleFile = masterBundle;
    }
    
    /**
     * Optional - only relevant if a master bundle has been defined
     * 
     * @param masterBundleDominant true to make the master bundle dominant to,
     *                             or override, corresponding template associated
     *                             bundles, default is false
     * @see #setMasterBundle
     */
    public void setMasterBundleDominant(final boolean masterBundleDominant) {
        this.masterBundleDominant = masterBundleDominant;
    }
    
    /**
     * Optional - only relevant if a {@link #setMasterBundle master bundle} has
     * been defined
     * <p>
     * Caution: if this is set to true includes must be carefully set, otherwise
     * localized templates may be generated for more base templates than you may
     * have planned on!
     * </p>
     * 
     * @param masterBundleAppliesToAll true to make the master bundle apply to
     *                                 all templates, not just those having
     *                                 corresponding template associated bundles,
     *                                 default is false
     * @see #setMasterBundle
     */
    public void setMasterBundleAppliesToAll(final boolean masterBundleAppliesToAll) {
    	this.masterBundleAppliesToAll = masterBundleAppliesToAll;
    }

    /**
     * Optional - Only applicable when the {@link #setBundleDir bundleDir} is
     * set to a location other than the srcDir, otherwise ignored.  Note that
     * if <code>false</code>, the bundleDir is presumed to store bundles in the
     * same relative directory hierarchy as the relative paths to the corresponding
     * base template files from srcDir.  Setting <code>bundleDirFlattened</code>
     * to <code>true</code> short-circuits this relative path interdependency,
     * allowing changes to srcDir without requiring changes to the directory
     * hierarchy within bundleDir.  The caveat, however, is that base template
     * files of the same name, in different directory hierarchies, will not be
     * able to use distinct bundles, e.g., srcDir/index.html and srcDir/foo/index.html
     * will necessarily share the same bundleDir/index.*properties bundle.  Then
     * again, this behavior may be desirable, as it reduces duplication... though
     * this can be achieved even more effectively using a {@link #setMasterBundle masterBundle}.
     * 
     * @param bundleDirFlattened true to treat bundlDir as a flattened directory
     *                           with no hierarchy so all bundle files are
     *                           presumed to exist in the root of bundleDir,
     *                           default is false
     */
    public void setBundleDirFlattened(final boolean bundleDirFlattened) {
        this.bundleDirFlattened = bundleDirFlattened;
    }

    /**
     * Optional - This means that localized templates won't have Server Side
     * includes to parse at runtime, as they will have been pre-resolved at build
     * time.  It also means that phrases in includes files will be localized in
     * the template, where they wouldn't at runtime... unless, of course, one
     * created includes template associated resource bundles or applied a master
     * bundle and set {@link #masterBundleAppliesToAll masterBundleAppliesToAll(true)}.
     * 
     * @param ssi true to enable parsing Server Side Includes, default is false
     * @see #setSsiBase
     */
    public void setSsi(final boolean ssi) {
        this.ssi = ssi;
    }
    
    /**
     * Optional
     * 
     * @param ssiBase the base directory of the SSI files, only applicable when
     *                SSI parsing is enabled, may be null, in which case the
     *                base directory will be that of the including template
     * @see #setSsi
     */
    public void setSsiBase(final String ssiBase) {
        this.ssiBase = ssiBase;
    }

    /**
     * Optional
     * 
     * @param verbose true to see verbose output, default is false
     */
    public void setVerbose(final boolean verbose) {
        this.verbose = verbose;
        //Set the project log level to INFO if the "verbose" argument is used
        //in the Ant invocation of this taskdef, otherwise log at the lower
        //verbose level where logging would only occur if the user manually
        //lowered Ant's logging level to verbose.
        if (this.verbose) {
            this.logLevel = Project.MSG_INFO;
        } else {
            this.logLevel = Project.MSG_VERBOSE;
        }
    }

    /**
     * Optional
     * 
     * @param force true to ignore template and bundle timestamps, thus forcing
     *              generation of L10n templates, default is false
     */
    public void setForce(final boolean force) {
        this.force = force;
    }

    /**
     * Optional
     * 
     * @param bundleEncoding the encoding of the resource bundles, ignored unless
     *                       running under Java version 1.6 or greater, default
     *                       is "ISO-8859-1", as required by java.util.ResourceBundle
     *                       prior to Java 1.6.
     */
    public void setBundleEncoding(final String bundleEncoding) {
        if (IS_JAVA6_PLUS) {
            this.bundleEncoding = bundleEncoding;
        }
    }

    /**
     * Optional - Note: because the default is the system default Locale, neglecting
     * to set this means the build is platform specific.  You've been warned!  E.g...
     * <p>
     * "_en", "en", "en_", "en__", "_en__", "en_US", "en__US", "en_POSIX",
     * "en__POSIX", "en___POSIX", "en_US_POSIX", "en__US__POSIX", "US", "US_",
     * "_US", "_US_", "US_POSIX", "US__POSIX", "_US_POSIX", "_US__POSIX", "_POSIX",
     * "__POSIX"
     * </p>
     * 
     * @param baseBundleLocaleStr a String representing the Locale of the text
     *                            in the base resource bundle (values rather than
     *                            keys, actually), default is the current system
     *                            default Locale
     * @see #setAutoGenLocalizedBaseBundle
     * @see I18nUtil#locale(String)
     */
    public void setBaseBundleLocale(final String baseBundleLocaleStr) {
        final Locale usrLocale = I18nUtil.locale(baseBundleLocaleStr);
        if (usrLocale != null) {
            this.baseBundleLocale = usrLocale;
        }
    }

    /**
     * Optional - a convenience to reduce the tedium of generating locale-named
     * copies of the base bundle manually.  Works in concert with the
     * {@link #setBaseBundleLocale base bundle locale setting}.  Having locale
     * specific bundles for each locale, especially the base locale, frees up the
     * base bundle to use alternate values for template keys - recall that base
     * bundle values are actually the keys used to match strings in the base
     * template.  This ameliorates phrasing duplication among base template, base
     * bundle, and locale-named bundle files.
     * <p>
     * For instance, a base bundle key/value pair may go from "key.1=hello citizens"
     * to "key.1=key.1" and now the base template would contain "key.1" instead of
     * "hello world".  By the same token the base bundle and base template could
     * continue to use "hello citizens" while the locale-named template for the base
     * bundle locale may use "hello U.S. citizens" (assuming a base bundle locale
     * of en_US).  Either way, the localized phrase is now entirely driven by the
     * locale-named bundle(s) and may vary from the base bundle and base template
     * as much or as little as disired.  Phrasing is centralized, thus avoiding
     * a highly coupled, brittle, template localization infrastructure.
     * </p>
     * <p>
     * Note that when using this feature, it is recommended that your application
     * code set both
     * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions#setUserLocale(Locale) XMLCCreateOptions.setUserLocale(Locale)}
     * and
     * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions#setFallbackLocale(Locale) XMLCCreateOptions.setFallbackLocale(Locale)}
     * (and/or {@link org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory#setDefaultFallbackLocale(Locale) setDefaultFallbackLocale(Locale)})
     * when calling the
     * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory#create(org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions) create(XMLCCreateOptions)}
     * method.  For instance, let's say the user locale is "fr", but no French
     * translation has been created.  Without specifying a fallback locale, the base
     * template will serve as the default fallback.  But if the base template uses keys
     * like "key.1" instead of phrases like "hello citizens" (the benefits of which are
     * detailed above), it won't make much sense to the user.  Assuming we've
     * {@link #setBaseBundleLocale set the base bundle locale} to "en_US" and either
     * manually or auto-generated the localized base bundle, by setting the current
     * feature to <code>true</code>, then we can set {@link Locale#US} as the fallback
     * locale and the en_US template will be used as fallback instead of the base template.
     * In effect, your preferred localized template acts as the defacto base template.
     * Of course if neither the user locale template nor the preferred fallback locale
     * template happen to exist, the base template continues to serve as the final fallback.
     * </p>
     * 
     * @param autoGenLocalizedBaseBundle true to generate a copy of the base bundle
     *                                   named in accordance with the base bundle
     *                                   locale, but only if it doesn't exist already,
     *                                   applies to both template and master bundles,
     *                                   default is false
     * @see #setBaseBundleLocale
     */
    public void setAutoGenLocalizedBaseBundle(boolean autoGenLocalizedBaseBundle) {
        this.autoGenLocalizedBaseBundle = autoGenLocalizedBaseBundle;
    }

    /**
     * Optional
     * 
     * @param defaultTemplateEncoding the encoding to use to read/write template
     *                                files if none can be auto-detected, default
     *                                is "UTF-8"
     */
    public void setDefaultTemplateEncoding(final String defaultTemplateEncoding) {
        XmlReader.setDefaultEncoding(defaultTemplateEncoding);
    }

    /**
     * Optional - zero or more...
     * <pre>
     * &lt;skipReplace
     *     description="[some description]"
     *     prefix="[some regex prefix]"
     *     suffix="[some regex suffix]"/&gt;
     * --OR--
     * &lt;skipReplace
     *     description="[some description]"
     *     idref="[some predefined id reference]"
     *     idrefSubPrefix="[some predefined regex sub prefix]"
     *     idrefSubSuffix="[some predefined regex sub suffix]"/&gt;
     * </pre>
     * <p>
     * The <code>prefix/suffix</code> combinations are applied as the first and
     * third group of a regular expression, i.e.,
     * <code>"(skipReplace.prefix)(target text)(skipReplace.suffix)".</code>
     * The prefix and suffix are used to search for instances of the target text
     * surrounded by, or embedded within, the defined pattern.  In these cases
     * L10n template replacement that would otherwise take place is skipped.  In
     * other words, matched patterns are flagged for skipping replacement (with
     * the flag removed prior to final output of the localized template).  When
     * defining more than one <code>skipReplace</code> element, keep in mind that
     * matches trump non-matches.
     * </p>
     * <p>
     * Any regular expression may be used, though caution will need to be taken
     * when using parenthesis.  If you need to use them, make sure to define them
     * in such a way that they won't define a group, e.g., <code>(?:regex)</code>.
     * All bets are off if you define a group because the three predefined groups
     * are referenced in the replacement.  This is a power feature.  And with power
     * comes responsibility.  How well this feature works is dependent on how well
     * you write your regular expressions.  Using available predefined skipReplace
     * elements will help ensure that the regular expressions work, as they have
     * been well tested.
     * </p>
     * <p>
     * The <code>idref</code> attribute is used to reference a predefined
     * <code>skipReplace</code> element, which alleviates most (sometimes all) of
     * the need to write, potentially, complex regular expressions.  When
     * <code>idref</code> is defined, user-defined <code>prefix</code> and
     * <code>suffix</code> attributes are ignored.  The attributes
     * <code>idrefSubPrefix</code> and <code>idrefSubSuffix</code>, if utilized
     * by the respective predefined skipReplace elements, may have defaults that
     * can be overridden.  When overriding, use patterns documented by the
     * respective predefined skipReplace element.  The following predefined
     * elements may be used...
     * </p>
     * <dl>
     * <dt><code>idref="markup.attrs.all"</code></dt>
     * <dd>Skips all cases where the target text exists within attributes</dd>
     * <dd>All other <code>skipReplace</code> attributes are ignored</dd>
     * <dd>Example: <code>&lt;skipReplace idref="markup.attrs.all"/&gt;</code></dd>
     * <dt><code>idref="markup.attrs.except"</code></dt>
     * <dd>Skips all cases where the target text exists within attributes, **except** those defined in <code>idrefSubPrefix</code></dd>
     * <dd>Optionally provide <code>idrefSubPrefix</code> to override defaults (user-defined <code>idrefSubSuffix</code> is ignored) - expects pipe-delimited attribute names with no spaces</dd>
     * <dd>Default is <code>idrefSubPrefix="abbr|alt|label|prompt|standby|summary|title"</code> ([X]HTML spec defined <a href="http://www.w3.org/TR/html401/sgml/dtd.html#Text">%Text;</a> <a href="http://www.w3.org/TR/html401/index/attributes.html">attributes</a>)</dd>
     * <dd>Examples: <code>&lt;skipReplace idref="markup.attrs.except" idrefSubPrefix="alt|title|value"/&gt;</code>, or simply <code>&lt;skipReplace idref="markup.attrs.except"/&gt;</code></dd>
     * <dt><code>idref="markup.attrs"</code></dt>
     * <dd>Skips all cases where the target text exists within user-defined attributes</dd>
     * <dd>Must provide <code>idrefSubPrefix</code> (no default, user-defined <code>idrefSubSuffix</code> is ignored) - expects pipe-delimited attribute names with no spaces</dd>
     * <dd>Example: <code>&lt;skipReplace idref="markup.attrs" idrefSubPrefix="class|id|name"/&gt;</code></dd>
     * </dl>
     * <p>
     * Note that for the predefined <code>markup.attrs*</code> elements, attributes
     * are recognized with or without namespaces.  So, even though, e.g., "id" is
     * defined by the user without a namespace, both "id" and "[somePrefix]:id"
     * are matched.
     * </p>
     *
     * @param skipReplace a configured &lt;skipReplace/&gt; element
     * @throws BuildException if the defined skipReplace is improperly configured,
     *                        meaning both prefix and suffix are empty, or undefined,
     *                        and no idref has been provided or an idref is provided
     *                        and doesn't match a predefined id or idref is provided
     *                        and matches a predefined id, but a required
     *                        idrefSubPrefix/Suffix is not provided
     */
    public void addConfiguredSkipReplace(final EmbededSkipReplaceType skipReplace) {
        final String skipReplaceValidationMsg = skipReplace.validate();
        if (skipReplaceValidationMsg == null) {
            embeddedSkipReplaceTypes.add(skipReplace);
        } else {
            throw new BuildException("Improperly configured <skipReplace/>: " + skipReplaceValidationMsg);
        }
    }

    /**
     * Executes the task
     * 
     * @throws BuildException if the build is improperly configured or enters a
     *                        bad state where it cannot proceed
     */
    public void execute() {
    	if (srcDir == null) {
            throw new BuildException("The 'srcDir' attribute must be set!");
        } else if (!srcDir.exists()) {
            throw new BuildException("The 'srcDir' location does not exist!  " + srcDir);
        }

        if (destDir == null) {
            destDir = srcDir;
        } else if (!destDir.exists()) {
            throw new BuildException("The 'destDir' location does not exist!  " + destDir);
        }

        boolean bundleDirDistinctFromSrcDir = false;
        if (bundleDir == null) {
            bundleDir = srcDir;
        } else if (!bundleDir.equals(srcDir)) {
            bundleDirDistinctFromSrcDir = true;
        }

        log("srcDir    = " + srcDir, logLevel);
        log("destDir   = " + destDir, logLevel);
        log("bundleDir = " + bundleDir, logLevel);

        File masterBundleDir = null;
        String masterBundlePath = null
            , masterBundleBaseName = null;
        final boolean masterBundleDefined = masterBundleFile != null
            , masterBundleFileExists = masterBundleDefined && masterBundleFile.exists();

        if (masterBundleDefined && !masterBundleFileExists) {
            log("Master bundle [" + masterBundleFile + "] defined, but does not exist.  Using only template-associated bundles...", Project.MSG_WARN);
        } else if (masterBundleFileExists) {
            final Matcher m = RE_LENIENT_PATH_GROUPER.matcher(masterBundleFile.toString());
            m.find();
            masterBundlePath = emptyIfNull(m.group(GRP_path));
            masterBundleBaseName = m.group(GRP_baseName);
            masterBundleDir = new File(masterBundlePath);
            log("masterBundle = " + masterBundleFile, logLevel);
            if (autoGenLocalizedBaseBundle) {
                final File localizedBaseBundleFile = new File(masterBundleDir, I18nUtil.canonicalLocaleString(masterBundleBaseName, ".properties", baseBundleLocale));
                try {
                    if (copyIfDestMissing(masterBundleFile, localizedBaseBundleFile)) {
                        log("Auto-generated localized master base bundle: " + localizedBaseBundleFile, logLevel);
                    }
                } catch (final Exception e) {
                    System.err.println("Error auto-generating localized master base bundle: " + localizedBaseBundleFile);
                    e.printStackTrace(System.err);
                }
            }
        }

        int baseTemplateCount = 0, l10nTemplateCount = 0;
        final String[] templatePathNames = this.getDirectoryScanner(srcDir).getIncludedFiles();

        for (int i=0; i<templatePathNames.length; i++) {
            final String templatePathName = templatePathNames[i];

            final Matcher m = RE_LENIENT_PATH_GROUPER.matcher(templatePathName);
            m.find();

            //Skip existing localized template files.  We only want to process the base files.
            final String l10nPattern = m.group(GRP_l10nPattern); //will be null if not already a localized file
            if (l10nPattern != null) {
                continue;
            }

            final String
//                pathComplete = m.group(0) //the complete matched string, same as 'templatePathName' var above
//                , pathBaseNameL10nPattern = m.group(GRP_pathBaseNameL10nPattern)
                pathBaseName = m.group(GRP_pathBaseName)
//                , path = emptyIfNull(m.group(GRP_path))
                , baseName = m.group(GRP_baseName)
                , extension = emptyIfNull(m.group(GRP_extension))
                , bundlePathBaseName = bundleDirDistinctFromSrcDir && bundleDirFlattened ? baseName : pathBaseName;

            /*
            System.out.println();
            System.out.print("pathComplete: ");System.out.println(pathComplete);
            System.out.print(GRP_pathBaseNameL10nPattern);System.out.print(": ");System.out.println(pathBaseNameL10nPattern);
            System.out.print(GRP_pathBaseName);System.out.print(": ");System.out.println(pathBaseName);
            System.out.print(GRP_path);System.out.print(": ");System.out.println(path);
            System.out.print(GRP_baseName);System.out.print(": ");System.out.println(baseName);
            System.out.print(GRP_l10nPattern);System.out.print(": ");System.out.println(l10nPattern);
            System.out.print(GRP_extension);System.out.print(": ");System.out.println(extension);
            */

            final File templateBundleFile = new File(bundleDir, bundlePathBaseName + ".properties");
            final boolean templateBundleFileExists = templateBundleFile.exists();

            final File primaryBundleFile, primaryBundleDir, secondaryBundleFile, secondaryBundleDir;
            final String primaryBundlePathBaseName, secondaryBundlePathBaseName;
            if (!masterBundleFileExists) {
            	if (!templateBundleFileExists) continue; //neither master, nor template, bundle files exist so continue
            	primaryBundleFile = templateBundleFile;
            	primaryBundleDir = bundleDir;
            	primaryBundlePathBaseName = bundlePathBaseName;
            	secondaryBundleFile = null;
            	secondaryBundleDir = null;
            	secondaryBundlePathBaseName = null;
            } else {
            	if (!masterBundleAppliesToAll && !templateBundleFileExists) continue; //no template bundle, but master bundle not configured to apply in the latter condition so continue
            	if (masterBundleDominant || !templateBundleFileExists) {
            	    primaryBundleFile = masterBundleFile;
            	    primaryBundleDir = masterBundleDir;
            	    primaryBundlePathBaseName = masterBundleBaseName;
            	    secondaryBundleFile = templateBundleFileExists ? templateBundleFile : null;
            	    secondaryBundleDir = bundleDir;
            	    secondaryBundlePathBaseName = bundlePathBaseName;
            	} else {
            	    primaryBundleFile = templateBundleFile;
            	    primaryBundleDir = bundleDir;
            	    primaryBundlePathBaseName = bundlePathBaseName;
            	    secondaryBundleFile = masterBundleFile;
            	    secondaryBundleDir = masterBundleDir;
            	    secondaryBundlePathBaseName = masterBundleBaseName;
            	}
            }

            if (templateBundleFileExists && autoGenLocalizedBaseBundle) {
                final File localizedBaseBundleFile = new File(bundleDir, I18nUtil.canonicalLocaleString(bundlePathBaseName, ".properties", baseBundleLocale));
                try {
                    if (copyIfDestMissing(templateBundleFile, localizedBaseBundleFile)) {
                        log("Auto-generated localized template base bundle: " + localizedBaseBundleFile, logLevel);
                    }
                } catch (final Exception e) {
                    System.err.println("Error auto-generating localized template base bundle: " + localizedBaseBundleFile);
                    e.printStackTrace(System.err);
                }
            }

            final File templateFile = new File(srcDir, templatePathName);

            //Check for resource bundle files associated with the current base
            //template file, and whether any of the latter are new or modified.
            //If none are found, continue to the next template file.
            final List propResourceBundles;
            try {
            	propResourceBundles = buildPropResourceBundles(templateFile, pathBaseName, extension
            	    , destDir
            	    , primaryBundleFile, primaryBundleDir, primaryBundlePathBaseName
            	    , secondaryBundleFile, secondaryBundleDir, secondaryBundlePathBaseName
            	    , bundleEncoding
            	    , force);
                if (propResourceBundles.isEmpty()) continue;
            } catch (final IOException ioe) {
            	System.err.println("Error processing resource bundle files...");
            	ioe.printStackTrace(System.err);
                continue;
            }

            //load the current template file and turn it into a String
            final String templateFileEncoding, masterTemplateStr;
            try {
                final InputSource inputSource = buildInputSource(templateFile);
                templateFileEncoding = inputSource.getEncoding();
                masterTemplateStr = loadTemplate(inputSource, ssi, ssiBase);
                baseTemplateCount++;
                log("Loaded template: " + templatePathName, logLevel);
            } catch (final IOException ioe) {
                System.err.println("Error reading template file: " + templateFile);
                ioe.printStackTrace(System.err);
                continue;
            }

            //load the base bundles to get the bundle keys and values needed for
            //replacement, ensuring they are sorted in descending natural order
            //by length, then value.  This way, we replace longer phrases before
            //shorter ones that could be subsets of the former.
            final Map sortedPropsMap;
            try {
            	final Properties props = props(primaryBundleFile, secondaryBundleFile, bundleEncoding);
            	sortedPropsMap = descNatOrderByLenThenValMap(props, baseBundleLocale);
            } catch (final IOException ioe) {
                System.err.println("Error reading bundle file: " + templateBundleFile);
                ioe.printStackTrace(System.err);
                continue;
            }

            log("Generating localized templates in...", logLevel);
            log(destDir.toString(), logLevel);

            //ensure output subdirectories of destDir exist in preparation for
            //writing localized templates
            new File(destDir, pathBaseName).getParentFile().mkdirs();

            //process resource bundles and generate the localized templates
            for (final Iterator rbIter = propResourceBundles.iterator(); rbIter.hasNext();) {
            	final PropertyResourceBundleWrapper rb = (PropertyResourceBundleWrapper) rbIter.next();
            	final String rbL10nPattern = rb.getId();

                //clone the master template
                String templateStr = new String(masterTemplateStr);

                //iterate over the non-localized, base bundle, key/value pairs
                //and obtain the localized value for replacement.
                final Map locHashValMap = new HashMap(sortedPropsMap.size());
                for (final Iterator iter = sortedPropsMap.entrySet().iterator(); iter.hasNext();) {
                    final Map.Entry entry = (Map.Entry) iter.next();
                    final String bundleKey = (String) entry.getKey();
                    final String nonLocVal = (String) entry.getValue();
                    final String locVal = rb.getString(bundleKey);

                    //use a string representation of the localized value's
                    //hashcode as the replacement text of the non-localized value,
                    //which will prevent already localized text from getting
                    //replaced on subsequent iterations
                    final String locHash = "@" + Integer.toHexString(locVal.hashCode());
                    templateStr = qualifiedStrReplace(templateStr, nonLocVal, locHash, embeddedSkipReplaceTypes);
                    locHashValMap.put(locHash, locVal);
                }

                //once replacement is complete, iterate over the localized
                //hash/value map and replace all the hashes with the real
                //localized values
                for (final Iterator iter = locHashValMap.entrySet().iterator(); iter.hasNext();) {
                    final Map.Entry entry = (Map.Entry) iter.next();
                    final String locHash = (String) entry.getKey();
                    final String locVal = (String) entry.getValue();
                    templateStr = strReplace(templateStr, locHash, locVal);
                }

                //write the localized template
                String locTemplate = null;
                try {
                    locTemplate = pathBaseName + rbL10nPattern + extension;
                    final BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(new File(destDir, locTemplate)), templateFileEncoding));
                    bw.write(templateStr);
                    bw.close();
                    l10nTemplateCount++;
                    log(locTemplate, logLevel);
                } catch (final IOException ioe) {
                    System.err.println("Error writing localized template: " + locTemplate);
                    ioe.printStackTrace(System.err);
                    continue;
                }
            }
        }

        if (baseTemplateCount > 0 || l10nTemplateCount > 0 || verbose) log("Processed " + templatePathNames.length + " source files, found " + baseTemplateCount + " base template(s), created "+l10nTemplateCount+" new localized template(s)" + (baseTemplateCount >= 0 && l10nTemplateCount == 0 ? " (all files up-to-date)" : ""));
    }

    static String emptyIfNull(final String str) {
    	return str != null ? str : "";
    }

    /**
     * Replace occurrences of strPattern in str with replacement, but only if
     * the strPattern is qualified, i.e., not embedded within user-defined
     * skipReplace prefix and suffix patterns.
     */
     private static String qualifiedStrReplace(String str, final String strPattern, final String replacement, final Set embeddedSkipReplaceTypes) {
        //TODO - REFlags.UNICODE seems to break regexp replace for some reason

        //Prepare strPattern by making it a literal regex pattern; e.g., "foo (bar)"
        //becomes "foo\s+\(bar\)" which will match the target text regardless
        //of whitespace within it in the template.  Also notice that regex meta
        //characters are escaped, ensuring all characters in the string are treated
        //literally (other than the added \s+ between words to match any number of
        //spaces, including line breaks).  The only caveat is that line breaks in
        //the target text of the original template are lost upon replacement.
    	final String escapedStrPattern = strReplace(I18nUtil.escapeRegexMetaChars(strPattern)
            , RE_SPACES
            , "\\\\s+");

    	//use 3 passes to avoid modification of embedded strings...

        //pass #1 prefixes embedded cases of the target pattern with a flag pattern,
        //indicating that we don't want the target pattern replaced in step #2
        for (final Iterator iter = embeddedSkipReplaceTypes.iterator(); iter.hasNext();) {
            final EmbededSkipReplaceType esrt = (EmbededSkipReplaceType) iter.next();
            str = strReplace(str
                , "(" + esrt.getPrefix() + ")(" + escapedStrPattern + ")(" + esrt.getSuffix() + ")"
                , "$1" + EmbededSkipReplaceType.UNIQUE_STR_PATTERN + "$2$3"
                );//, REFlags.IGNORE_CASE);
        }

        //pass #2 replaces the target pattern with the replacement text, but with
        //a negative lookbehind that avoids replacement in cases where the target
        //pattern is prefixed by the flag pattern (added in step #1)
        str = strReplace(str
            , "(?<!" + EmbededSkipReplaceType.UNIQUE_STR_PATTERN + ")" + escapedStrPattern
            , replacement
            );//, REFlags.IGNORE_CASE);

        //pass #3 replaces the flag pattern with an empty string, leaving embedded
        //cases of the target pattern unchanged, maintaining line breaks, etc....
        str = strReplace(str
            , EmbededSkipReplaceType.UNIQUE_STR_PATTERN
            , "");

        return str;
    }

    private static String strReplace(final String str, final String strPattern, final String replacement) {
    	return strReplace(str, strPattern, replacement, REFlags.DEFAULT);
    }

    private static String strReplace(final String str, final String strPattern, final String replacement, final int reFlags) {
    	return strReplace(str, new Pattern(strPattern, reFlags), replacement);
    }

    /**
     * Note that any escaping must be completed prior to calling this method
     */
    private static String strReplace(final String str, final Pattern pattern, final String replacement) {
    	return pattern.replacer(replacement).replace(str);
    }

    static InputSource buildInputSource(final File file) throws IOException {
    	final InputSource inputSource = new InputSource(file.toString());
        try {
            final XmlReader reader = new XmlReader(file);
            inputSource.setCharacterStream(reader);
            inputSource.setEncoding(reader.getEncoding());
        } catch (final XmlReaderException xre) {
            //This should never happen using lenient encoding detection according to the XMLReader doc as well as a review of the code
            throw new IOException("Unexpected failure during lenient encoding detection by XMLReader of file "+file);
        }
        return inputSource;
    }

    static String loadTemplate(final InputSource inputSource, final boolean ssi, final String ssiBase) throws IOException {
        final BufferedReader br;
        if (ssi) {
            br = new BufferedReader(new SSIReader(inputSource, ssiBase));
        } else {
            final Reader r = InputSourceOps.open(inputSource);
            br = r instanceof BufferedReader ? (BufferedReader) r : new BufferedReader(r);
        }

        final char[] buff = new char[4096];
        final StringWriter sw = new StringWriter(buff.length * 4);
        while (true)
        {
            final int read;
            if ((read = br.read(buff)) == -1) break;
            sw.write(buff, 0, read);
        }
        br.close();
        return sw.toString(); //note that sw.flush() and sw.close() do nothing, so no point in calling them
    }

    private static List buildPropResourceBundles(
        final File templateFile, final String templatePathBaseName, final String templateFileExtension
        , final File localeTemplateFileDestDir
        , final File primaryBundleFile, final File primaryBundleDir, final String primaryBundlePathBaseName
        , final File secondaryBundleFile, final File secondaryBundleDir, final String secondaryBundlePathBaseName
        , final String bundleEncoding
        , final boolean force
    ) throws IOException {
    	//now see if there are other properties files associated with the template file that require processing
        final DirectoryScanner primaryDSLocalePropFiles = new DirectoryScanner();
        primaryDSLocalePropFiles.setBasedir(primaryBundleDir);
        primaryDSLocalePropFiles.setIncludes(new String[]{primaryBundlePathBaseName+"_*.properties"});
        primaryDSLocalePropFiles.scan();
        final String[] primaryLocalePropFilePaths = primaryDSLocalePropFiles.getIncludedFiles();

        final Map primaryLocalePropFilesByL10nPattern = new HashMap(primaryLocalePropFilePaths.length)
            , secondaryLocalePropFilesByL10nPattern;
        final Set allL10Patterns = new TreeSet(Collections.reverseOrder());
        final boolean mergePrimarySecondary;
        final int primarySpos = primaryBundlePathBaseName.length();

        for (int i=0; i<primaryLocalePropFilePaths.length; i++) {
            final String localePropFilePath = primaryLocalePropFilePaths[i];
            primaryLocalePropFilesByL10nPattern.put(propertiesL10nPattern(localePropFilePath, primarySpos), new File(primaryBundleDir, localePropFilePath));
        }
        allL10Patterns.addAll(primaryLocalePropFilesByL10nPattern.keySet());

        if (secondaryBundleFile != null) {
            final DirectoryScanner secondaryDSLocalePropFiles = new DirectoryScanner();
            secondaryDSLocalePropFiles.setBasedir(secondaryBundleDir);
            secondaryDSLocalePropFiles.setIncludes(new String[]{secondaryBundlePathBaseName+"_*.properties"});
            secondaryDSLocalePropFiles.scan();
            final String[] secondaryLocalePropFilePaths = secondaryDSLocalePropFiles.getIncludedFiles();

            secondaryLocalePropFilesByL10nPattern = new HashMap(secondaryLocalePropFilePaths.length);
            final int secondarySpos = secondaryBundlePathBaseName.length();
            for (int i=0; i<secondaryLocalePropFilePaths.length; i++) {
            	final String localePropFilePath = secondaryLocalePropFilePaths[i];
                secondaryLocalePropFilesByL10nPattern.put(propertiesL10nPattern(localePropFilePath, secondarySpos), new File(secondaryBundleDir, localePropFilePath));
            }
            allL10Patterns.addAll(secondaryLocalePropFilesByL10nPattern.keySet());
            mergePrimarySecondary = true;
        } else {
            secondaryLocalePropFilesByL10nPattern = null;
            mergePrimarySecondary = false;
        }

        final List propBundles = new ArrayList();
        String lastStr = null;
        PropertyResourceBundleWrapper lastRb = null
            , defaultRb = propRb(primaryBundleFile, bundleEncoding, "base(primary)");
        final long templateLastModfied = templateFile.lastModified()
            , primaryBundleFileLastModified = primaryBundleFile.lastModified()
            , secondaryBundleFileLastModified;
        if (secondaryBundleFile != null) {
            defaultRb.setParent(propRb(secondaryBundleFile, bundleEncoding, "base(secondary)"));
            secondaryBundleFileLastModified = secondaryBundleFile.lastModified();
        } else {
            secondaryBundleFileLastModified = 0L;
        }

        for (final Iterator iter = allL10Patterns.iterator(); iter.hasNext();) {
            final String currentL10nPattern = (String) iter.next();

            final File primaryLocalePropFile = (File) primaryLocalePropFilesByL10nPattern.get(currentL10nPattern);
            long primaryLocalePropFileLastModified = 0L, secondaryLocalePropFileLastModified = 0L;
            PropertyResourceBundleWrapper rb = null;
            if (primaryLocalePropFile != null) {
            	rb = propRb(primaryLocalePropFile, bundleEncoding, currentL10nPattern);
            	primaryLocalePropFileLastModified = primaryLocalePropFile.lastModified();
            }
            if (mergePrimarySecondary) {
            	final File secondaryLocalePropFile = (File) secondaryLocalePropFilesByL10nPattern.get(currentL10nPattern);
            	if (rb != null) {
            	    if (secondaryLocalePropFile != null) {
            	        final PropertyResourceBundleWrapper rb2 = propRb(secondaryLocalePropFile, bundleEncoding, currentL10nPattern + "(secondary)");
            	        rb2.setParent(defaultRb);
            	        rb.setParent(rb2);
            	        secondaryLocalePropFileLastModified = secondaryLocalePropFile.lastModified();
            	    }
            	} else {
            	    //if primary rb was null, then secondary file must exist or it wouldn't be in the set of L10n patterns
            	    rb = propRb(secondaryLocalePropFile, bundleEncoding, currentL10nPattern);
            	    secondaryLocalePropFileLastModified = secondaryLocalePropFile.lastModified();
            	}
            }

            if (lastStr != null && !lastRb.hasParent()) {
                lastRb.setParent(lastStr.startsWith(currentL10nPattern) ? rb : defaultRb);
            }
            lastStr = currentL10nPattern;
            lastRb = rb;

            if (!force) {
                final File localeTemplateFile = new File(localeTemplateFileDestDir, templatePathBaseName+currentL10nPattern+templateFileExtension);
                if (localeTemplateFile.exists()) {
                    final long localeTemplateFileLastModified = localeTemplateFile.lastModified();
                    if (localeTemplateFileLastModified > templateLastModfied
                            && localeTemplateFileLastModified > primaryBundleFileLastModified
                            && localeTemplateFileLastModified > primaryLocalePropFileLastModified
                            && localeTemplateFileLastModified > secondaryBundleFileLastModified
                            && localeTemplateFileLastModified > secondaryLocalePropFileLastModified) {
                        continue; //skip cases where the existing localized template is newer than the source files
                    }
                }
            }

            propBundles.add(rb);
        }
        if (lastRb != null && !lastRb.hasParent()) lastRb.setParent(defaultRb); //The last one through may not have had a chance to get a parent, so set it to the default
        return propBundles;
    }

    private static String propertiesL10nPattern(final String fullPathName, final int basePathNameLength) {
    	final int epos = fullPathName.length()-11;
        return fullPathName.substring(basePathNameLength,epos);
    }

    /**
     * A Map with contents identical to the provided Properties object, but sorted
     * in, locale-sensitive, length, then value, descending natural order so, when
     * iterating over the entry set, more specific vals occur first, e.g., "foobar"
     * then "foo".  Values are compared only as fallback to guarantee a consistent
     * order when the length is the same, e.g., always "The" then "the".
     */
    private static Map descNatOrderByLenThenValMap(final Properties props, final Locale locale) {
        final Collator c = Collator.getInstance(locale);
        c.setDecomposition(Collator.FULL_DECOMPOSITION); //slower than canonical (and none), but most correct
        c.setStrength(Collator.IDENTICAL); //consider differences in letters (primary) and same letters with different accents (secondary), case (tertiary) and exact equivalence (identical)
        final Map mpck = new HashMap(props.size());
        for (final Iterator iter = props.entrySet().iterator(); iter.hasNext();) {
            final Map.Entry entry = (Map.Entry) iter.next();
            mpck.put(entry.getKey(), c.getCollationKey((String) entry.getValue()));
        }

        final Map m = new TreeMap(new Comparator() {
            public int compare(final Object o1, final Object o2) {
                final CollationKey ck1 = (CollationKey) mpck.get(o1);
                final CollationKey ck2 = (CollationKey) mpck.get(o2);                
                final int lenCheckRetVal = new Integer(ck1.getSourceString().length()).compareTo(new Integer(ck2.getSourceString().length())) * -1;
                if (lenCheckRetVal != 0) return lenCheckRetVal;
                return ck1.compareTo(ck2) * -1;
            }
        });
        m.putAll(props);
        return m;
    }

    private static Properties props(final File primaryPropFile, final File secondaryPropFile, final String bundleEncoding) throws IOException {
    	final Java6PlusProperties props = new Java6PlusProperties();
    	props.load(new FileInputStream(primaryPropFile), bundleEncoding);
    	if (secondaryPropFile != null) {
    	    final Java6PlusProperties secondaryProps = new Java6PlusProperties();
    	    secondaryProps.load(new FileInputStream(secondaryPropFile), bundleEncoding);
    	    secondaryProps.keySet().removeAll(props.keySet());
    	    props.putAll(secondaryProps);
    	}
    	return props;
    }

    private static PropertyResourceBundleWrapper propRb(final File propFile, final String bundleEncoding, final String id) throws IOException {
    	return new PropertyResourceBundleWrapper(new FileInputStream(propFile), bundleEncoding, id);
    }

    private static class Java6PlusProperties extends Properties {
        private static Method propsReaderLoadMethod;
    	static {
    	    if (IS_JAVA6_PLUS) {
    	        try {
    	            propsReaderLoadMethod = Properties.class.getMethod("load", new Class[] {Reader.class});
    	        } catch (final Throwable th) {
    	            //squash
    	        }
    	    }
    	}

    	public Java6PlusProperties() {
    	    super();
    	}

    	public Java6PlusProperties(final Properties defaults) {
    	    super(defaults);
    	}

    	public void load(final InputStream stream, final String encoding) throws IOException {
    	    if (propsReaderLoadMethod != null) {
    	        try {
    	            propsReaderLoadMethod.invoke(this, new Object[] {new InputStreamReader(stream, encoding)});
    	        } catch (final UnsupportedEncodingException uee) {
    	            throw uee;
    	        } catch (final IllegalArgumentException iae) {
    	            throw iae;
    	        } catch (final Exception e) {
    	            throw new IllegalStateException("Unable to invoke Java6+ Properties.load(Reader) method via reflection.");
    	        }
    	    } else {
    	        super.load(stream);
    	    }
    	}
    }

    private static class PropertyResourceBundleWrapper {
        private static final Method rbSetParentMethod;
        private static Constructor prbReaderConstructor;
        static {
            try {
                rbSetParentMethod = ResourceBundle.class.getDeclaredMethod("setParent", new Class[] {ResourceBundle.class});
                rbSetParentMethod.setAccessible(true);
            } catch (final NoSuchMethodException nsme) {
                throw new ExceptionInInitializerError(nsme);
            }
            if (IS_JAVA6_PLUS) {
                try {
                    prbReaderConstructor = PropertyResourceBundle.class.getConstructor(new Class[] {Reader.class});
                } catch (final Throwable th) {
                    //squash
                }
            }
        }

        private final ResourceBundle rb;
        private final String id;
        private boolean hasParent;

        /**
         * @param stream an InputStream that represents a property file to read from
         * @param encoding the desired encoding, ignored unless Java version >= 1.6
         * @param id optionally provide a custom string used to identify this
         *           resource bundle.  Useful for mapping, as well as used in
         *           {@link #toString()} to display fallback bundle order.  A
         *           default is used if null.
         */
        public PropertyResourceBundleWrapper(final InputStream stream, final String encoding, final String id) throws IOException {
            if (prbReaderConstructor != null) {
                try {
            	    //TODO - how do I add my toString() implementation to the reflected instantiated object.  Maybe via dynamic proxy, except that wouldn't work because PropertyResourceBundle doesn't have a commmon interface, which is requried for dynamic proxies???
            	    rb = (ResourceBundle) prbReaderConstructor.newInstance(new Object[] {new InputStreamReader(stream, encoding)});
                } catch (final UnsupportedEncodingException uee) {
                    throw uee;
                } catch (final IllegalArgumentException iae) {
                    throw iae;
                } catch (final Exception e) {
                    throw new IllegalStateException("Unable to instantiate using Java6+ PropertyResourceBundle(Reader) constructor via reflection");
                }
            } else {
                rb = new PropertyResourceBundle(stream) {
                    /**
                     * Displays fallback bundle order (child-&gt;parent relationships)
                     */
                    public String toString() {
                        final StringBuffer b = new StringBuffer();
                        b.append(id);
                        if (parent != null) {
                            b.append("->");
                            b.append(parent.toString());
                        }
                        return b.toString();
                    }
                };
            }
            this.id = id != null ? id : getClass().getName()+"@"+hashCode();
        }

        /**
         * @return the id of this resource bundle, always non-null
         * @see #PropertyResourceBundleWrapper(InputStream, String, String)
         */
        public String getId() {
            return id;
        }

        /**
         * Publicizes access to protected ResourceBundle.setParent(ResourceBundle) method
         * 
         * @see ResourceBundle#setParent
         */
        public void setParent(final PropertyResourceBundleWrapper rbw) {
            try {
                rbSetParentMethod.invoke(this.rb, new Object[] {rbw.rb});
                this.hasParent = true;
            } catch (Exception e) {
                throw new IllegalStateException("Unable to invoke protected ResourceBundle.setParent(ResourceBundle) method via reflection.");
            }
        }

        public boolean hasParent() {
            return hasParent;
        }

        /**
         * Delegates to ResourceBundle.getString(String)
         */
        public String getString(final String key) {
            return rb.getString(key);
        }

        /**
         * Delegates to overridden PropertyResourceBundle.toString()
         */
        public String toString() {
            return rb.toString();
        }
    }

    private static final boolean copyIfDestMissing(final File src, final File dest) throws FileNotFoundException, IOException {
        if (dest.createNewFile()) {
            InputStream in = new FileInputStream(src);
            OutputStream out = new FileOutputStream(dest);
            byte[] buf = new byte[1024];
            int len;
            while ((len = in.read(buf)) > 0){
                out.write(buf, 0, len);
            }
            in.close();
            out.close();
            return true;
        }
        return false;
    }

    public static final class EmbededSkipReplaceType extends DataType {
        public static final String UNIQUE_STR_PATTERN = "@@u@n@i@q@u@e@s@t@r@p@a@t@t@e@r@n@@";
        private static final Pattern RE_IDREF_SUB_PATTERN = new Pattern(UNIQUE_STR_PATTERN);
        private static final Map embeddedSkipReplaceTypesPredefined;
        static {
            embeddedSkipReplaceTypesPredefined = new HashMap(3);
            new EmbededSkipReplaceType("markup.attrs.all"
                    , "\\S+\\s*=\\s*[\"']?\\s*"
                    , "\\s*[\"']?");
            //See http://www.w3.org/TR/html401/index/attributes.html
            //Default to matching all attributes except those defined by the HTML
            //spec as free-form text (arguably this could include "value"?).
            //Only caveat with this regex is that it also skips, e.g., "-title",
            //"title-", and "-title-" (an exception attr with non-word characters
            //on either side).  Interestingly, java.util.regex behaves the same
            //same with or without the (?=\\W?) positive lookahead where JRegex
            //requires it to match, e.g., "a-title" and "a-".  It provides parity.
            new EmbededSkipReplaceType("markup.attrs.except"
            	    , "\\b(?!:?(?:" + EmbededSkipReplaceType.UNIQUE_STR_PATTERN + ")\\b)(?=\\W?)\\W?\\w+\\W?\\s*=\\s*[\"']?\\s*"
            	    , "\\s*[\"']?"
            	    , "abbr|alt|label|prompt|standby|summary|title"
            	    , null
            	    , true
            	    , false);
            //Any default attribute list would be presumptuous and almost certainly
            //incomplete.  Forcing the user to provide it is the only reasonable
            //approach.
            new EmbededSkipReplaceType("markup.attrs"
                    , "\\b(?:" + EmbededSkipReplaceType.UNIQUE_STR_PATTERN + ")\\b\\s*=\\s*[\"']?\\s*"
                    , "\\s*[\"']?"
                    , null //"accesskey|class|coords|for|id|media|method|name|shape|style|tabindex|target|type|usemap"
                    , null
                    , true
                    , false);
        }

        private EmbededSkipReplaceType idRefEsrt;
    	private String idRef, idRefSubPrefix, idRefSubSuffix, prefix, suffix;
    	private boolean configured, idRefSubPrefixNeeded, idRefSubSuffixNeeded;

        public EmbededSkipReplaceType(){}

        private EmbededSkipReplaceType(final String id, final String prefix, final String suffix) {
            this(id, prefix, suffix, null, null, false, false);
        }

        private EmbededSkipReplaceType(final String id, final String prefix, final String suffix, final String idRefSubPrefixDefault, final String idRefSubSuffixDefault, final boolean idRefSubPrefixNeeded, final boolean idRefSubSuffixNeeded) {
            this.prefix = prefix;
            this.suffix = suffix;
            this.idRefSubPrefix = idRefSubPrefixDefault;
            this.idRefSubSuffix = idRefSubSuffixDefault;
            this.idRefSubPrefixNeeded = idRefSubPrefixNeeded;
            this.idRefSubSuffixNeeded = idRefSubSuffixNeeded;
            if (id != null) {
                embeddedSkipReplaceTypesPredefined.put(id, this);
            }
        }

        /**
         * @return the regex skip replacement prefix, guaranteed non-null
         */
        public String getPrefix() {
            return prefix != null ? prefix : "";
        }

        public void setPrefix(final String prefix) {
            this.prefix = prefix;
        }

        /**
         * @return the regex skip replacement suffix, guaranteed non-null
         */
        public String getSuffix() {
            return suffix != null ? suffix : "";
        }

        public void setSuffix(final String suffix) {
            this.suffix = suffix;
        }

        public void setIdRef(final String idRef) {
            this.idRef = idRef;
            idRefEsrt = (EmbededSkipReplaceType) embeddedSkipReplaceTypesPredefined.get(idRef);
        }

        public void setIdRefSubPrefix(final String idRefSubPrefix) {
            this.idRefSubPrefix = idRefSubPrefix;
        }

        public void setIdRefSubSuffix(final String idRefSubSuffix) {
            this.idRefSubSuffix = idRefSubSuffix;
        }

        /**
         * @return an error message if configuration if found to be invalid or
         *         null if valid
         */
        public String validate() {
            if (!configured) configure();
            if (idRef != null && idRefEsrt == null) return "No idref found with value '" + idRef + "'.";
            return (getPrefix().trim().length() > 0 || getSuffix().trim().length() > 0) ? null : "A valid idref, with applicable idrefSubPrefix/Suffix, or at least one of prefix/suffix must be defined.";
        }

        private void configure() {
            if (!configured && idRefEsrt != null) {
                final String predefinedPrefix = idRefEsrt.prefix
                    , predefinedIdRefSubPrefix = idRefEsrt.idRefSubPrefix
                    , predefinedSuffix = idRefEsrt.suffix
                    , predefinedIdRefSubSuffix = idRefEsrt.idRefSubSuffix;
                final boolean idrefSubPrefixProvided = idRefSubPrefix != null || predefinedIdRefSubPrefix != null
                    , idrefSubSuffixProvided = idRefSubSuffix != null || predefinedIdRefSubSuffix != null
                    , predefinedIdRefPrefixNeeded = idRefEsrt.idRefSubPrefixNeeded
                    , predefinedIdRefSuffixNeeded = idRefEsrt.idRefSubSuffixNeeded;
                //get rid of irrelevant user-defined prefix/suffix when idref is defined
                prefix = null;
                suffix = null;
                if (predefinedIdRefPrefixNeeded && idrefSubPrefixProvided) {
                    prefix = RE_IDREF_SUB_PATTERN.replacer(idRefSubPrefix != null ? idRefSubPrefix : predefinedIdRefSubPrefix).replace(predefinedPrefix);
                } else if (!predefinedIdRefPrefixNeeded && (!predefinedIdRefSuffixNeeded || (predefinedIdRefSuffixNeeded && idrefSubSuffixProvided))) {
                    prefix = predefinedPrefix;
                }
                if (predefinedIdRefSuffixNeeded && idrefSubSuffixProvided) {
                    suffix = RE_IDREF_SUB_PATTERN.replacer(idRefSubSuffix != null ? idRefSubSuffix : predefinedIdRefSubSuffix).replace(predefinedSuffix);
                } else if (!predefinedIdRefSuffixNeeded && (!predefinedIdRefPrefixNeeded || (predefinedIdRefPrefixNeeded && idrefSubPrefixProvided))) {
                    suffix = predefinedSuffix;
                }
            }
            configured = true;
        }
    }

}

