/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site (http://www.enhydra.org/).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *  @author     simon.tuffs@lutris.com
 *  @author     christos@dynamic.net.au
 *  @author     kevin@bluedog.com.au
 *  @author     ericb@interactiveportal.com
 *  @author     Christian Cryder [christianc@granitepeaks.com]
 *  @author     Stephen Peterson [stephen_peterson@agilent.com]
 *  @author     Jacob Kjome [hoju@visi.com]
 *  @author     David Li [taweili@yahoo.com]
 *  @author     Stefan Armbruster [sarmbruster@web.de]
 *
 * $Id: Xmlc.java,v 1.11 2008-04-11 04:09:19 jkjome Exp $
 */
package org.enhydra.xml.xmlc.taskdef;

import java.io.File;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Locale;

import org.apache.tools.ant.AntClassLoader;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.types.Commandline;
import org.apache.tools.ant.types.EnumeratedAttribute;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;


/**
 * ANT task to invoke the XMLC compiler to generate Java code.
 *
 *  <p>Implements an Ant taskdef to perform XMLC compilations. Please refer
 *  to the <a href="http://ant.apache.org/">Ant website</a>
 *  for more information. XMLCTask extends MatchingTask which means that
 *  it has the ability to include or exclude files based on pattern matching.
 *  The supported tags are as follows:</p>
 *
 *  <table border="1" cellspacing="3" cellpadding="3" width="100%">
 *    <tr>
 *      <th>tag</th>
 *      <th>required</th>
 *      <th>default value</th>
 *      <th>description</th>
 *    </tr>
 *    <tr>
 *      <td>srcdir</td>
 *      <td>yes</td>
 *      <td>&nbsp;</td>
 *      <td>The source directory root</td>
 *    </tr>
 *    <tr>
 *      <td>sourceout</td>
 *      <td>no</td>
 *      <td>value of srcdir</td>
 *      <td>The destination directory into which the generated source *ML.java files are
 *          written, if keep=&quot;yes&quot;. If sourceout is not specified it is set to the value
 *          of the srcdir. This will put the *ML.java files in the same location as the .*ml files.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>destdir</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>The destination directory root. Only really useful if you want XMLC to
 *          do the compilation, so we can check dependencies.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>packagedir</td>
 *      <td>no</td>
 *      <td>relative path from srcdir</td>
 *      <td>Package prefix for generated classes, expressed as a directory path (relative to srcdir).  Eg...  my/package.
 *          If not specified, the relative path from srcdir becomes the package.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>packagename</td>
 *      <td>no</td>
 *      <td>blank package</td>
 *      <td>Package prefix for generated classes, while preserving any source file directory paths.
 *          For example, given &lt;xmlc srcdir=&quot;foo/resources&quot; sourceout=&quot;foo/source&quot; packagename=&quot;a.b&quot; /&gt;, packagename=&quot;a.b&quot;,
 *          markup file located at &quot;foo/resources/c/d/Foo.ml&quot; will be generated to &quot;foo/source/a/b/c/d/FooML.java&quot; and will start with a &quot;package a.b.c.d;&quot; statement.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>suffix</td>
 *      <td>no</td>
 *      <td>The markup file's extension, UPPER-cased</td>
 *      <td>The class name suffix for all XMLC generated classes.  Traditionally, this value is used to denote the markup type, such as
 *          "HTML", "XHTML", "WML", "VoiceXML", "XML", etc...  However, it is, in fact, entirely optional, arbitrary, and may set to an
 *          empty String or left entirely unset.  In the latter case, the suffix defaults to the file extension (UPPER-cased) of each markup file, respectively.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>verbose</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Displays the XMLC command and all of the arguments</td>
 *    </tr>
 *    <tr>
 *      <td>compile</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Have XMLC compile the resulting java files? Note: The default is to not
 *          invoke the Java compiler through XMLC, since this is really slow. Any
 *          java files created are better handled by the ANT <javac> rule.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>force</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>Force XMLC to always re-process the ML code whether it has changed or not</td>
 *    </tr>
 *    <tr>
 *      <td>keep</td>
 *      <td>no</td>
 *      <td>yes</td>
 *      <td>Keep the Java files generated by XMLC? To not keep them, you must allow XMLC to
 *          compile them first. However, since this is slower, the default is to keep them
 *          and let ANT compile the files using a seperate task.
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>performExec</td>
 *      <td>no</td>
 *      <td>no</td>
 *      <td>XMLC is invoked by performing a Runtime.exec()<br><i>(kept for historical reasons by the author)</i></td>
 *    </tr>
 *    <tr>
 *      <td>options</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>File name of XMLC options or a glob style pattern.</td>
 *    </tr>
 *    <tr>
 *      <td>&lt;arg&gt;</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>Nested element representing additional command line arguments to be passed to XMLC.  See the XMLC documentation
 *          for all the possible commands. Note: these are applied after the other XMLC arguments derived from the options listed above.
 *          <p>For example, to specify an URL mapping to XMLC:
 *          <code>
 *            <pre>
 *            &lt;xmlc source="." includes="*.html"&gt;
 *              &lt;arg value="-urlmapping" /&gt;
 *              &lt;arg value="Edit.html" /&gt;
 *              &lt;arg value="To.po" /&gt;
 *            &lt;/xmlc&gt;
 *            </pre>
 *          </code>
 *          <b>Caution:</b> Do not be tempted to surround the URL's with single-quotes, as is common
 *          when calling XMLC from a shell command line, for example
 *            <pre>
 *            &lt;xmlc source="." includes="*.html"&gt;
 *              &lt;arg line="-urlmapping 'Edit.html' 'To.po'" /&gt;
 *            &lt;/xmlc&gt;
 *            </pre>
 *          The single-quotes will be taken as part of the URL to replace, and this is probably not what
 *          you intend
 *      </td>
 *    </tr>
 *    <tr>
 *      <td>&lt;resourcePath&gt; | resourcePath | resourcePathRef</td>
 *      <td>no</td>
 *      <td>&nbsp;</td>
 *      <td>Nested element or attributes representing additional classpath entries providing XMLC access to user-defined resources such as XCatalog or DTD files,
 *          which are looked up via the classloader when given values such as 'file:/org/mypackage/resources/custom.xcat' or 'file:/org/mypackage/resources/custom.dtd'.
 *          Though these look like file system URL's, XMLC's XMLEntityResolver strips the leading 'file:/' and looks these up on the classpath.
 *          Without configuring the resource path, XMLC will only be able to look up resources in the classpath provided to the taskdef for &lt;xmlc/&gt;, which should only be used
 *          to provide the XMLC and XMLC taskdef jars (and any other dependencies) to load the task itself.  The classloader's classpath is manipulated only temporarily while
 *          in use by XMLC.  It is then restored to its original state before exiting the task so that the classloader is not polluted for uses external to the current
 *          &lt;xmlc/&gt; task invocation.
 *          
 *          <p>One caveat to note is that this feature will only work if the XMLC task is defined using a taskdef with a custom classpath rather than loaded from Ant's primordial
 *             classloader, which would be the case if one put the jar (and dependencies) in ${ant.home}/lib or ${user.home}/.ant/lib.  The issue is that when this is the case, the
 *             classloader is a plain old URLClassLoader, which doesn't provide any easy way to manipulate, then restore, its internal classpath.  This should not be a problem in
 *             normal usage, since I can't imagine anyone wanting to put the XMLC taskdef, XMLC itself, and all XMLC's dependencies up at that level.  In any case, now you know for
 *             sure not to do that.</p>
 *          
 *          <p>Note: If the resource can't be found on the classpath, an attempt will be made to look it up on the file system, so you can provide a fully qualified file system URL
 *             (or any other supported URL protocol, for that matter).  However, it is recommended that the resources be loaded from the classpath for consistent behavior between
 *             compile and run-time environments, with no extra effort required for packaging the application for runtime.</p>
 *          
 *          <p>For example...</p>
 *          <code>
 *            <pre>
 *            &lt;xmlc ...&gt;
 *              &lt;resourcePath&gt;
 *                &lt;pathelement location="${resources}"/&gt;
 *              &lt;/resourcePath&gt;
 *            &lt;/xmlc&gt;
 *            
 *            OR
 *            
 *            &lt;xmlc resourcePath="${resources}" .../&gt;
 *            
 *            OR
 *            
 *            &lt;path id="xmlcResources"&gt;
 *   		    &lt;pathelement location="${resources}"/&gt;
 *  		  &lt;/path&gt;
 *            &lt;xmlc resourcePathRef="xmlcResources" .../&gt;
 *            </pre>
 *          </code>
 *      </td>
 *    </tr>
 *
 *  </table>
 * <p>
 *
 * @author     simon.tuffs@lutris.com
 * @author     christos@dynamic.net.au
 * @author     kevin@bluedog.com.au
 * @author     ericb@interactiveportal.com
 * @author     Christian Cryder [christianc@granitepeaks.com]
 * @author     Stephen Peterson [stephen_peterson@agilent.com]
 * @author     Jacob Kjome [hoju@visi.com]
 * @author     David Li [taweili@yahoo.com]
 * @author     Stefan Armbruster [sarmbruster@web.de]
 * @version    %I%, %G%
 * @since      2.2
 */
public class Xmlc extends org.apache.tools.ant.taskdefs.MatchingTask {

    protected static String logHdr = "     [xmlc] ";
    protected File srcDir, destDir, srcOutDir;
    protected String classPath, packageDir, packageName, classNameSuffix, options, xmlcArgs, performExec;
    protected boolean verbose, forceBuild, compile;
    protected boolean keep = true;
    protected int logLevel = Project.MSG_VERBOSE;
    protected int upToDate = 0;
    protected Commandline cmdl = new Commandline();
    protected Path resourcePath;

    /**
     * Kept for backward compatibility only.  
     * 
     * @deprecated use {@link #srcDir} and {@link srcDestDir} instead.
     */
    protected String srcDirName, srcOutDirName;

    /**
     * Kept for backward compatibility only.
     * 
     * @deprecated use {@link #classNameSuffix} instead
     */
    protected String ML = "HTML";

    /**
     * Sets the resource path used to look up resources such as XCatalogs
     * and custom DTDs.
     *
     * @param resourcePath an Ant Path object containing the resource path
     */
    public void setResourcePath(Path resourcePath) {
        if (this.resourcePath == null) {
            this.resourcePath = resourcePath;
        }
        else {
            this.resourcePath.append(resourcePath);
        }
    }

    /**
     * Gets the resource path used to look up resources such as XCatalogs
     * and custom DTDs
     * 
     * @return an Ant Path object containing the resource path
     */
    public Path getResourcePath() {
        return resourcePath;
    }

    /**
     * Adds a path to the resource path
     * 
     * @return a resource path to be configured
     */
    public Path createResourcePath() {
        if (resourcePath == null) {
            resourcePath = new Path(getProject());
        }
        return resourcePath.createPath();
    }

    /**
     * Adds a reference to a resource path defined elsewhere
     * 
     * @param ref a reference to a path
     */
    public void setResourcePathRef(Reference ref) {
        createResourcePath().setRefid(ref);
    }

    /**
     * Set the source directory containing the HTML files.
     *
     * @param srcdirname The name of the directory containing the HTML files
     */
    public void setSrcdir(String srcdirname) {
        this.srcDirName = srcdirname; //Only left here for backward compatibility.  Do not actually use!!!
        this.srcDir = getProject().resolveFile(srcdirname);
    }

    /**
     * Set the destination directory into which the generated classes are
     * written.
     *
     * @param destDirName The destination directory into which the generated classes are written.
     */
    public void setDestdir(String destDirName) {
        this.destDir = getProject().resolveFile(destDirName);
    }

    /**
     * Allow general XMLC options to be passed through unchanged.  Appended to the
     * options lists.
     *
     * @param args The XMLC options to be passed through.
     * @deprecated Use &lt;arg&gt; tags for now, this will eventually be removed.
     */
    public void setArgs(String args) {
        this.xmlcArgs = args;
        getProject().log(logHdr+"DEPRECATED, The 'args' attribute.  Use <arg> tags instead.", Project.MSG_WARN);
    }

    /**
     * Set the classpath for the XMLC compiler.
     *
     * @param classpath The classpath for XMLC to use.
     * @deprecated Use nested <arg> elements or "args" option instead.
     */
    public void setClasspath(String classpath) {
        this.classPath = classpath;
        getProject().log(logHdr+"DEPRECATED, The 'classpath' attribute.  Use <arg> tags instead.", Project.MSG_WARN);
    }

    /**
     * Set the destination directory into which the generated source .java files are
     * written, if -keep is used.
     *
     * @param sourceout The destination directory into which the generated sources are written.
     *
     * Note: Currently -keep is used by default since the .java files are best
     * compiled by the built-in Ant javac command as a separate rule.
     */
    public void setSourceout(String sourceout) {
        this.srcOutDirName = sourceout; //Only left here for backward compatibility.  Do not actually use!!!
        this.srcOutDir = getProject().resolveFile(sourceout);
    }

    /**
     * Force XMLC to always re-process the ML code (or not)
     *
     * @param force whether to force re-processing of markup or not
     */
    public void setForce(BooleanAttribute force) {
        this.forceBuild = force.booleanValue();
    }

    /**
     * Run XMLC in verbose mode (-verbose flag)
     *
     * @param verbose Set "true" for -verbose, anything else (e.g. "false") for
     * quiet mode.
     */
    public void setVerbose(BooleanAttribute verbose) {
        this.verbose = verbose.booleanValue();
        // Set the project log level to INFO if the "verbose" argument is used in the Ant invocation
        // of this taskdef.
        if (this.verbose) {
            this.logLevel = Project.MSG_INFO;
        } else {
            this.logLevel = Project.MSG_VERBOSE;
        }
    }

    /**
     * Set whether or not to keep the generated Java file.
     *
     * @param keep Set to "true" or "false"
     */
    public void setKeep(BooleanAttribute keep) {
        this.keep = keep.booleanValue();
    }

    /**
     * Set whether or not to have XMLC compile the generated Java file.
     *
     * @param compile Set to "true" or "false"
     */
    public void setCompile(BooleanAttribute compile) {
        this.compile = compile.booleanValue();
    }

    /**
     * Set to a specific options file to add parameters to the XMLC command. Or,
     * use a glob style pattern to base the options file name off the name of each
     * source file to be processed (e.g. *.xmlc).
     *
     *  @param options Set to file name or glob pattern
     */
    public void setOptions(String options) {
        this.options = options;
    }

    /**
     * Set the Package prefix for generated classes.  Expressed as a directory path.
     *
     * <p>Note: packagedir is used to set the package name of the generated class, <b>and</b>
     * to specify the directory where the file will be generated (relative to sourceout).
     * As such, any directory prefix on a source file is stripped before the packagedir
     * is applied.  This means that all files from the source tree will end up in the
     * same directory, which may not be what you intend, if you are using Xmlc to
     * process files in a directory hierarchy.</p>
     *
     * @param packagedir Expressed as a directory path, either '/' or '\' are accepted.
     * @see #setPackagename(String)
     */
    public void setPackagedir(String packagedir) {
        this.packageDir = packagedir;
    }

    /**
     * Set the package prefix for generated classes, while preserving any source
     * file directory paths.  This is useful if you want to take source (markup
     * language) files from one tree, but generate the Java code into another
     * tree, preserving relative paths.
     * <p>
     * For example, suppose that you have .html files under a directory called
     * "foo/resources", and you want to generate Java code under "foo/source",
     * but have the package name be "a.b". The following Xmlc taskdef will
     * achieve this:
     * <p>
     * &lt;xmlc srcdir=&quot;foo/resources&quot; sourceout=&quot;foo/source&quot; packagename=&quot;a.b&quot; /&gt;
     * <p>
     * A markup file located at "foo/resources/c/d/Foo.ml" will be generated to
     * "foo/source/a/b/c/d/FooML.java" and will start with a "package a.b.c.d"
     * statement.
     *
     * @param packagename Expressed as a package path, [package].[package]
     * @see #setPackagedir(String)
     */
    public void setPackagename(String packagename) {
        this.packageName = packagename;
    }

    /**
     * Defaults to "HTML" if left unset.  Only applicable to subclasses referencing
     * the deprecated {@link #ML} field.  See {@link #setSuffix(String)} for new,
     * more intelligent, default behavior.
     * 
     * @param markup an arbitrary class name suffix, traditionally denoting a
     *                  particular markup type
     * @deprecated use {@link #setSuffix(String)} instead
     * @see #setSuffix(String)
     */
    public void setMarkup(String markup) {
        setSuffix(markup);
        getProject().log(logHdr+"DEPRECATED, The 'markup' attribute.  Use the 'suffix' attribute instead.", Project.MSG_WARN);
    }

    /**
     * Optionaly set the class name suffix for all XMLC generated classes.
     * Traditionally, this value is used to denote the markup type, such as
     * "HTML", "XHTML", "WML", "VoiceXML", "XML", etc...  However, it is, in
     * fact, entirely optional, arbitrary, and may set to an empty String or left
     * entirely unset.  In the latter case, the suffix defaults to the file
     * extension (UPPER-cased) of each markup file, respectively.
     * <p>
     * Left unset, a markup file named mypage.xhtml will end up as the class name
     * <code>mypageXHTML</code>.  If &lt;xmlc suffix="FoO" .../&gt; is applied,
     * then the class name will be <code>mypageFoO</code>.  This example uses a
     * multi-cased suffix to illustrate that the value set here will be used
     * unmodified (other than trimming).  Set to an empty String if no suffix is
     * desired.
     * </p>
     * 
     * @param suffix the String to append to the class name, traditionally
     *               denoting a particular markup type
     */
    public void setSuffix(String suffix) {
        if (suffix != null) {
            this.classNameSuffix = suffix.trim();
            this.ML = this.classNameSuffix; //Only left here for backward compatibility.  Do not actually use!!!
        }
    }

    /**
     *  Set whether or not to invoke XMLC by using Runtime.exec() to
     *  create a seperate JVM. This capability was originally written to get
     *  around a problem with name-space clashes between packages. Set this
     *  to the command to execute.
     *
     *  @param performExec Set to "true" or "false"
     */
    public void setPerformexec(String performExec) {
        this.performExec = performExec;
    }

    /**
     * Check if the given target file is newer than all of the files
     * @param traget the file to be checked
     * @param files the files
     */
    private boolean isNewerThanFiles(File target, File[] files) {
        boolean newer = true;
        for (int i = 0 ; i < files.length ; i++) {
            if (files[i].lastModified() > target.lastModified()) {
                newer = false;
                getProject().log(logHdr+"    " + files[i].getPath() + " is newer then " + target.getPath(), logLevel);
                break;
            }
        }
        return newer;
    }

    /**
     *  Required by ant framework - implements execute() in Task.
     *  Called by the project to let the task do its work.
     *
     *  @exception BuildException All exceptions are thrown as BuildException.
     *
     * Note: most of the logic here was cloned from the Javac command. Unfortunately
     * this was the only way to 'reuse' it.
     */
    public void execute() throws org.apache.tools.ant.BuildException {
        getProject().log(logHdr+"Xmlc srcdir=" + srcDir, logLevel);
        getProject().log(logHdr+"     destdir=" + destDir, logLevel);
        getProject().log(logHdr+"     sourceout=" + srcOutDir, logLevel);
        getProject().log(logHdr+"     packagedir=" + packageDir, logLevel);
        getProject().log(logHdr+"     packagename=" + packageName, logLevel);
        // first off, make sure that we've got a srcdir

        if (srcDir == null) {
            throw new BuildException("srcdir attribute must be set!");
        }
        if (!srcDir.exists()) {
            throw new BuildException("srcdir does not exist!");
        }
        if (!compile && !keep) {
            throw new BuildException("If you do not keep the source, you must have XMLC compile them!");
        }

        XmlcUtils xmlcUtils = XmlcUtils.create();

        // scan source and dest dirs to build up both copy lists and
        // compile lists
        DirectoryScanner ds = this.getDirectoryScanner(srcDir);
        String[] files = ds.getIncludedFiles();

        getProject().log(logHdr+"XMLC processing " + files.length + " files: ", logLevel);

        ClassLoader xmlcLoader = getClass().getClassLoader();
        ClassLoader origCtxLoader = Thread.currentThread().getContextClassLoader();
        String origClasspath = null;
        try {
            //Reset the context classloader to the current xmlcLoader so that
            //Xerces, sitting in a parent classloader (eg.. ANT_HOME/lib), can
            //see libraries added to the classpath of this taskdef using
            //<classpath>.  We'll reset it back before we return from this
            //method in a "finally" block.
            Thread.currentThread().setContextClassLoader(xmlcLoader);

            // Add the user-defined resource path to the current classloader so
            // that XCatalog and DTD resources can be found by XMLC during
            // compilation.  Store the original classpath so it can be reset.
            // This avoids pollution of the classpath outside of the current
            // <xmlc/> task invocation.
            if (getResourcePath() != null && xmlcLoader instanceof AntClassLoader) {
                origClasspath = ((AntClassLoader)xmlcLoader).getClasspath();
                ((AntClassLoader)xmlcLoader).addPathElement(getResourcePath().toString());
            }

            for (int i=0; i<files.length; i++) {
                String file = files[i];
                String suffix;
                if (classNameSuffix != null) {
                    suffix = classNameSuffix;
                } else {
                    int dotIdx = file.lastIndexOf('.');
                    if (dotIdx > -1 && file.length() > dotIdx + 1) {
                        suffix = file.substring(dotIdx + 1).toUpperCase(Locale.ENGLISH);
                    } else {
                        suffix = "";
                    }
                }

                String fileName = xmlcUtils.buildFullBaseFileName(packageName, packageDir, file);

                String className = xmlcUtils.buildClassName(fileName, suffix);
                String javaName = className.replace('.', '/');

                // If srcOutDir is not specified, set it to srcDir.  This will put the
                // ML.java files in the same location as the .ml files, which is
                // desirable.
                if (srcOutDir == null) {
                    srcOutDir = srcDir;
                    srcOutDirName = srcDirName; //Only left here for backward compatibility.  Do not actually use!!!  
                }

                File mlFile = new File(srcDir, file);
                //csc        File javaFile = new File(srcOutDir, fileName + ML + ".java");
                File javaFile = new File(srcOutDir, javaName + ".java");  //csc
                File classFile = new File((destDir != null) ? destDir : srcOutDir, fileName + ".class");

                java.util.ArrayList args = new java.util.ArrayList();

                // Quick test: if the .java file is newer than the .ml file, don't need
                // to run XMLC.  This is a rats-nest of logic, but hard to simplify.
                File target = null;

                // The 'options' arg must  precede the name of the file to be compiled
                // Should really have -compile and -nokeep options since -nocompile
                // and -keep are the default.

                // sam, 2003-01-26
                // add recursive search
                ArrayList al = new ArrayList();
                File f = javaFile;
                while ((f = f.getParentFile()) != null) {
                    al.add(f);
                }
                f = mlFile;
                while ((f = f.getParentFile()) != null) {
                    al.add(f);
                }
                File[] possibleOptionFilesDirs = new File[al.size()];
                for (int ii=0; ii<al.size(); ii++) {
                    // saw, 2006-04-10
                    // Search files in child-first order instead of parent-first (start at bottom of hierarchy)
                    // With this change, the options.xmlc found closest to the source file (searching up the directory tree) is used.

                    //possibleOptionFilesDirs[al.size() - ii - 1] = (File)al.get(ii);
                    //getProject().log("possibleOptionFilesDirs :"  + possibleOptionFilesDirs[al.size() - ii - 1].getAbsolutePath(), logLevel);
                    possibleOptionFilesDirs[ii] = (File)al.get(ii);
                    getProject().log("possibleOptionFilesDirs :"  + possibleOptionFilesDirs[ii].getAbsolutePath(), logLevel);
                    // END saw, 2006-04-10
                }
                //File[] possibleOptionFilesDirs = { javaFile.getParentFile(), mlFile.getParentFile()};
                // END sam, 2003-01-26

                String[] possibleOptionFileNames = { fileName, options};
                String[] foundOptionsFiles =
                    xmlcUtils.getOptionFiles(possibleOptionFilesDirs, possibleOptionFileNames);

                File[] foundOptions = new File[foundOptionsFiles.length];
                for (int j = 0 ; j < foundOptions.length ; j++) {
                    foundOptions[j] = new File(foundOptionsFiles[j]);
                }

                /*
                 * Since we can't update the last modified time of the
                 * ML, an option checkfile is created along with the
                 * XMLC output files and this files is checked against
                 * the option files.
                 */
                File optionCheck = null;
                if (keep) {
                    optionCheck = new File(javaFile.getPath() + ".optionCheck");
                } else {
                    optionCheck = new File(classFile.getPath() + ".optionCheck");
                }

                getProject().log(logHdr+"    option check to created " + optionCheck.getPath(), logLevel);
                if (keep) {
                    if (!javaFile.exists()) {
                        target = javaFile;
                        // If we are going to keep the generated file but don't have one, then better create it
                        getProject().log(logHdr+"    " + javaFile + " does not exist, invoking XMLC", logLevel);
                    } else if (compile) {
                        target = classFile;
                        // If we are going to keep the generated file and compile it, then we have two dependencies to verify
                        if (mlFile.lastModified() <= classFile.lastModified() && mlFile.lastModified() <= javaFile.lastModified()) {
                            if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                                if (!forceBuild) {
                                    getProject().log(logHdr+"    " + classFile.getName() + " and " + javaFile.getName() + " are newer than " + mlFile.getName(), logLevel);
                                    upToDate++;
                                    continue;
                                }
                            } else {
                                getProject().log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                            }
                        } else {
                            getProject().log(logHdr+"    " + classFile.getName() + " or " + javaFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                        }
                    } else {
                        // If we are going to keep the generated file but not compile it, then we only care if the java
                        // file is out of date. We'll let someone else check if the class file is out of date.
                        target = javaFile;
                        if (mlFile.lastModified() <= javaFile.lastModified()) {
                            if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                                if (!forceBuild) {
                                    getProject().log(logHdr+"    " + javaFile.getName() + " is newer than " + mlFile.getName(), logLevel);
                                    upToDate++;
                                    continue;
                                }
                            } else {
                                getProject().log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                            }
                        } else {
                            getProject().log(logHdr+"    " + javaFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                            target = javaFile;
                        }
                    }
                } else {
                    // Not "keep" so target must be classFile.
                    target = classFile;
                    if (mlFile.lastModified() <= classFile.lastModified()) {
                        if (optionCheck.exists() && isNewerThanFiles(optionCheck, foundOptions)) {
                            if (!forceBuild) {
                                // If we are not going to keep the generated file then we must compile it
                                getProject().log(logHdr+"    " + classFile.getName() + " is newer than " + mlFile.getName(), logLevel);
                                upToDate++;
                                continue;
                            }
                        } else {
                            getProject().log(logHdr+"    " + "One of the option files are newer than " + mlFile.getName() + " invoking XMLC" , logLevel);
                        }
                    } else {
                        getProject().log(logHdr+"    " + classFile.getName() + " is older than " + mlFile.getName() + " invoking XMLC", logLevel);
                    }
                }

                getProject().log(logHdr+"    " + mlFile.getName() + " --> " + target.getName(), logLevel);
                if (classPath != null) {
                    args.add("-classpath");
                    args.add(classPath);
                }
                // Don't do the Java compilation here, it's much faster to do it with
                // the Javac task which invokes the Javac compiler without forking a
                // new process.
                if (!compile) args.add("-nocompile");
                if (keep) args.add("-keep");
                if (verbose) args.add("-verbose");

                args.add("-class");
                args.add(className);

                args.add("-sourceout");
                args.add(srcOutDir.toString());

                if (xmlcArgs != null) {
                    // We have to tokenize the arguments.  Assumpions: URL's will not contain
                    // whitespace, and if they did, they should use %20 or &space entities.
                    //
                    java.util.StringTokenizer tok = new java.util.StringTokenizer(xmlcArgs, " ");
                    while (tok.hasMoreElements()) {
                        args.add(tok.nextToken());
                    }
                }

                String[] additionalArgs = cmdl.getArguments();
                for (int idx=0; idx < additionalArgs.length; idx++) {
                    args.add(additionalArgs[idx]);
                }

                // Option files to be added right before the ML file
                if (foundOptionsFiles.length > 0) {
                    for (int optionsIdx = 0; optionsIdx < foundOptionsFiles.length; optionsIdx++) {
                        args.add(foundOptionsFiles[optionsIdx]);
                        getProject().log(logHdr+"    Using options file " + foundOptionsFiles[optionsIdx], logLevel);
                    }
                } else {
                    getProject().log(logHdr+"No options file could not be located for " + file +
                        " in paths " + javaFile.getParentFile() + ", " +
                        mlFile.getParentFile(), Project.MSG_WARN);
                }
                // *** WARNING DO NOT SEPARATE THE FOLLOWING CODE FROM THE BLOCK ABOVE! ***
                args.add(new File(srcDir, file).getAbsolutePath());

                getProject().log(logHdr+"args==" + args, logLevel);

                getProject().log(logHdr+"Invoke XMLC on " + mlFile.getPath(), Project.MSG_INFO);
                if (performExec != null) {
                    args.add(0, performExec);
                    String [] _args = (String[])args.toArray(new String[args.size()]);
                    Runtime runtime = Runtime.getRuntime();
                    Process process = runtime.exec(_args);
                    int status = process.waitFor();
                    if (status != 0) {
                        throw new BuildException("exit status of exec = "+status + toString(_args));
                    }
                } else {
                    String [] _args = (String[])args.toArray(new String[args.size()]);

                    // Execute XMLC directly.  This requires that enhydra.jar, or an equivalent
                    // path containing the needed DOM code is on the CLASSPATH of ANT.
                    String xmlcClassName = "org.enhydra.xml.xmlc.commands.xmlc.XMLC";

                    Method m = null;
                    Class  c = null;

                    try {
                        c = Class.forName(xmlcClassName, true, xmlcLoader);
                        m = c.getMethod("main", new Class[] { String[].class});
                        m.invoke(null,new Object[] { _args});
                    } catch (Exception e) {
                        // Reflection errors.  Class or method does not exist or can't be invoked!
                        e.printStackTrace();
                        throw new BuildException(e);
                    }
                }

                /*
                 * Create the option file
                 */
                if (optionCheck.exists()) {
                    optionCheck.delete();
                }
                optionCheck.createNewFile();
            }
        } catch (Exception x) {
            x.printStackTrace();
            throw new BuildException(x);
        } finally {
            //reset the context classloader back to the original one
            Thread.currentThread().setContextClassLoader(origCtxLoader);

        	if (origClasspath != null && xmlcLoader instanceof AntClassLoader) {
        		// Restore the original classpath to avoid leaking user-defined
        		// resource path information outside the current invocation of
        		// the <xmlc/> task.
                ((AntClassLoader)xmlcLoader).setClassPath(new Path(getProject(), origClasspath));
        	}
        }

        if (upToDate > 0) {
            getProject().log(logHdr+"XMLC " + upToDate + " files were up-to-date, consider using force=\"true\"", logLevel);
        }
    }

    /**
     *  Simple utility to concatenate an array of Strings into one single
     *  String.
     *
     *  @param      args  The array of Strings
     *  @return     The resulting String
     */
    static private String toString(String[] args) {
        StringBuffer  buffer = new StringBuffer();
        for (int i=0; i<args.length; i++) buffer.append(args[i]).append(" ");
        return buffer.toString();
    }

    /**
     *  Set nested arguments to the XMLC command to be executed.
     *
     */
    public Commandline.Argument createArg() {
        return cmdl.createArgument();
    }

    public static class BooleanAttribute extends EnumeratedAttribute {
        public String[] getValues() {
            return new String[] {"yes", "no", "on", "off", "true", "false"};
        }

        public boolean booleanValue() {
            return (this.getValue().equals("yes") || this.getValue().equals("on") || this.getValue().equals("true"));
        }
    }

}
