/*
 * Copyright 1999,2000,2004 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.enhydra.xml.xhtml.dom.xerces;
    
import org.enhydra.xml.xhtml.dom.XHTMLDOMImplementation;
import org.enhydra.xml.xhtml.dom.XHTMLDocument;

import org.apache.xerces.dom.DOMImplementationImpl;
import org.apache.xerces.dom.DOMMessageFormatter;
import org.w3c.dom.DOMException;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.html.HTMLDOMImplementation;
import org.w3c.dom.html.HTMLDocument;


/**
 * Provides number of methods for performing operations that are independent
 * of any particular instance of the document object model. This class is
 * unconstructable, the only way to obtain an instance of a DOM implementation
 * is by calling the static method {@link #getDOMImplementation}.
 * 
 * @xerces.internal
 * 
 * @version $Revision: 1.5 $ $Date: 2008-04-13 19:47:46 $
 * @author <a href="mailto:arkin@exoffice.com">Assaf Arkin</a>
 * @see org.w3c.dom.DOMImplementation
 */
public class XHTMLDOMImplementationImpl
    extends DOMImplementationImpl
    implements XHTMLDOMImplementation
{


    /**
     * Holds a reference to the single instance of the DOM implementation.
     * Only one instance is required since this class is multiple entry.
     */
    private static final XHTMLDOMImplementation _instance = new XHTMLDOMImplementationImpl();


    /**
     * Private constructor assures that an object of this class cannot
     * be created. The only way to obtain an object is by calling {@link
     * #getDOMImplementation}.
     */
    private XHTMLDOMImplementationImpl()
    {
    }


    /**
     * @see org.w3c.dom.DOMImplementation
     */
    public Document createDocument(String namespaceURI, 
				   String qualifiedName, 
				   DocumentType doctype) throws DOMException {
    	if (doctype != null && doctype.getOwnerDocument() != null) {
            throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, 
                                   DOMMessageFormatter.formatMessage(
                                   DOMMessageFormatter.XML_DOMAIN, 
                                   "WRONG_DOCUMENT_ERR", null));
        }
        Document doc = new XHTMLDocumentImpl(doctype);
        // If namespaceURI and qualifiedName are null return a Document with no document element.
        if (qualifiedName != null || namespaceURI != null) {
            Element e = doc.createElementNS(namespaceURI, qualifiedName);
            doc.appendChild(e);
        }
        return doc;
    }


    /**
     * @see #createDocument
     */
    public XHTMLDocument createXHTMLDocument(String namespaceURI, 
               String qualifiedName, 
               DocumentType doctype) throws DOMException {
        return (XHTMLDocument) createDocument(namespaceURI, qualifiedName, doctype);
    }


    public final XHTMLDocument createXHTMLDocument( String title )
        throws DOMException
    {
	XHTMLDocument doc;

	if ( title == null )
	    throw new NullPointerException( "HTM014 Argument 'title' is null." );
	doc = new XHTMLDocumentImpl();
	doc.setTitle( title );
	return doc;
    }


    /**
     * Create a new XHTML document of the specified <TT>title</TT> text.
     *
     * @param title The document title text
     * @return New HTML document
     * @see #createXHTMLDocument(String)
     */
    public final HTMLDocument createHTMLDocument( String title )
        throws DOMException
    {
        return createXHTMLDocument( title );
    }


    /**
     * Returns an instance of a {@link XHTMLDOMImplementation} that can be
     * used to perform operations that are not specific to a particular
     * document instance, e.g. to create a new document.
     *
     * @return Reference to a valid DOM implementation
     * @see #getXHTMLDOMImplementation
     */
    public static HTMLDOMImplementation getHTMLDOMImplementation()
    {
	return _instance;
    }


    /**
     * Returns an instance of a {@link XHTMLDOMImplementation} that can be
     * used to perform operations that are not specific to a particular
     * document instance, e.g. to create a new document.
     *
     * @return Reference to a valid DOM implementation
     */
    public static XHTMLDOMImplementation getXHTMLDOMImplementation()
    {
	return _instance;
    }


    /**
     * Returns an instance of a {@link XHTMLDOMImplementation} that can be
     * used to perform operations that are not specific to a particular
     * document instance, e.g. to create a new document.
     *
     * @return Reference to a valid DOM implementation
     */
    public static DOMImplementation getDOMImplementation()
    {
	return _instance;
    }


}
