/*
 * Copyright 1999,2000,2004-2006 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.enhydra.xml.xhtml.dom.xerces;

/*
 * Derived from Xerces HTMLDocument implementation.
 * Author: Assaf Arkin <arkin@exoffice.com>
 */
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Locale;

import org.enhydra.xml.xhtml.dom.XHTMLDocument;
import org.enhydra.xml.xhtml.dom.XHTMLElement;
import org.enhydra.xml.xhtml.dom.XHTMLTitleElement;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.dom.ElementImpl;
import org.enhydra.xml.xmlc.XMLCError;
import org.w3c.dom.DOMException;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.UserDataHandler;
import org.w3c.dom.html.HTMLCollection;
import org.w3c.dom.html.HTMLElement;

/**
 * Base class used to implements documents defined in terms of XHTML.  This
 * includes the full XHTML and XHTML modular documents. This defined all of
 * the HTMLDocument methods plus a framework for creating elements, but no
 * element definitions themselves.
 * 
 * @see org.w3c.dom.html.HTMLDocument
 */
public class XHTMLDocumentBase
    extends DocumentImpl
    implements XHTMLDocument
{


    /**
     * Holds <code>XHTMLCollectionImpl</code> object with live collection of all
     * anchors in document. This reference is on demand only once.
     */
    private XHTMLCollectionImpl fAnchors;


    /**
     * Holds <code>XHTMLCollectionImpl</code> object with live collection of all
     * forms in document. This reference is on demand only once.
     */
    private XHTMLCollectionImpl fForms;


    /**
     * Holds <code>XHTMLCollectionImpl</code> object with live collection of all
     * images in document. This reference is on demand only once.
     */
    private XHTMLCollectionImpl fImages;


    /**
     * Holds <code>XHTMLCollectionImpl</code> object with live collection of all
     * links in document. This reference is on demand only once.
     */
    private XHTMLCollectionImpl fLinks;


    /**
     * Holds <code>XHTMLCollectionImpl</code> object with live collection of all
     * applets in document. This reference is on demand only once.
     */
    private XHTMLCollectionImpl fApplets;


    /**
     * Holds string writer used by direct manipulation operation ({@link #open}.
     * {@link #write}, etc) to write new contents into the document and parse
     * that text into a document tree.
     */
    private StringWriter fWriter;


    /**
     * Holds names and classes of HTML element types. When an element with a
     * particular tag name is created, the matching {@link java.lang.Class}
     * is used to create the element object. For example, &lt;A&gt; matches
     * {@link XHTMLAnchorElementImpl}. This static table is shared across all
     * HTML documents.
     *
     * @see #createElement
     */
    private static HashMap fElementTypes;


    /**
     * Signature used to locate constructor of HTML element classes. This
     * static array is shared across all HTML documents.
     *
     * @see #createElement
     */
    private static final Class[] fElemConstructorSig = new Class[] {
	XHTMLDocumentBase.class, String.class, String.class
    };

    /**
     * Constructor without document type.
     *
     */
    public XHTMLDocumentBase()
    {
	this(null);
    }

    /**
     * Constructor with document type.
     */
    public XHTMLDocumentBase(DocumentType doctype)
    {
	super(doctype);
        populateElementTypes();
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getDocumentElement
     */
    public synchronized Element getDocumentElement()
    {
        // Enhydra modified: Original Xerces code tried to reorder nodes to
        // make things right, which moved around comments to weird locations.

	Element html = (Element)getDirectChildElement("html", this);
	if (html == null) {
            throw new DOMException(DOMException.HIERARCHY_REQUEST_ERR,
                                   "<html> document element not found");
        }
        return html;
    }


    /**
     * Find the direct child element of a node given its name.
     */
    private Node getDirectChildElement(String name,
                                       Node root) {
        for (Node child = root.getFirstChild(); child != null;
             child = child.getNextSibling()) {
	    if ((child instanceof Element)
                && child.getNodeName().equals(name)) {
                return child;
            }
        }
        return null;
    }


    /**
     * Obtains the &lt;HEAD&gt; element in the document, creating one if does
     * not exist before. The &lt;HEAD&gt; element is the first element in the
     * &lt;HTML&gt; in the document. The &lt;HTML&gt; element is obtained by
     * calling {@link #getDocumentElement}. If the element does not exist, one
     * is created.
     * <P>
     * Called by {@link #getTitle}, {@link #setTitle}, {@link #getBody} and
     * {@link #setBody} to assure the document has the &lt;HEAD&gt; element
     * correctly placed.
     *
     * @return The &lt;HEAD&gt; element
     */
    public synchronized XHTMLElement getHead()
    {
        // Enhydra modified: Original Xerces code tried to reorder nodes to
        // make things right, which moved around comments to weird locations.
        // Throwing an error would be more appropriate, but we were afraid of
        // breaking existing code, so just get the node.

        // Call getDocumentElement() to get the HTML element that is also the
        // top-level element in the document. Get the first element in the
        // document that is called HEAD. Work with that.
        Element html = getDocumentElement();
	XHTMLElement head
            = (XHTMLElement)getDirectChildElement("head", html);
        if (head == null) {
            // Head does not exist, create a new one, place it at the top of the
            // HTML element and return it.
	    head = new XHTMLHeadElementImpl(this, html.getNamespaceURI(), "head");
            html.insertBefore( head, html.getFirstChild() );
        }
        return head;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getTitle
     */
    public synchronized String getTitle()
    {
        // Enhydra modified: Original Xerces code is some what strange, it
        // called getElementsByTagName() twice, but only used the second
        // result.  We assume it' a direct child of HEAD (although more 
        // error checking might be better).

	XHTMLTitleElement title
	    = (XHTMLTitleElement)getDirectChildElement("title", getHead());
        if (title == null) {
            return ""; // No TITLE found, return an empty string.
        } else {
            return title.getText();
        }
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#setTitle
     */
    public synchronized void setTitle( String newTitle )
    {
        // Enhydra modified: Original Xerces code used getElementsByTagName()
        // to find the title.  We assume it' a direct child of HEAD (although
        // more error checking might be better).

	XHTMLElement head = getHead();
	XHTMLTitleElement title
	    = (XHTMLTitleElement)getDirectChildElement("title", head);
        if (title == null) {
	    title = new XHTMLTitleElementImpl(this, head.getNamespaceURI(), "title");
        }
        title.setText(newTitle);
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getBody
     */
    public synchronized HTMLElement getBody()
    {
        // Enhydra modified: Original Xerces code tried to reorder nodes to
        // make things right, which moved around comments to weird locations.
        // Throwing an error would be more appropriate, but we were afraid of
        // breaking existing code, so just get the node.

        // Find BODY or FRAMESET
        Element html = getDocumentElement();
	XHTMLElement body = findBody(html);
        if (body == null) {
            // BODY does not exist, create a new one, place it in the HTML element
            // right after the HEAD and return it.
	    body = new XHTMLBodyElementImpl(this, html.getNamespaceURI(), "body");
            html.appendChild( body );
        }
        return body;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#setBody
     */
    public synchronized void setBody( XHTMLElement newBody )
    {
        // Enhydra modified: Original Xerces code tried to reorder nodes to
        // make things right, which moved around comments to weird locations.
        // Throwing an error would be more appropriate, but we were afraid of
        // breaking existing code, so just get the node.

        // Find BODY or FRAMESET
        Element html = getDocumentElement();
	XHTMLElement body = findBody(html);
        if (body == null) {
            html.appendChild(newBody);
        } else {
            html.replaceChild(newBody, body);
        }
    }

    public synchronized void setBody( HTMLElement newBody )
    {
        setBody((XHTMLElement) newBody);
    }

    /**
     * Find a BODY or FRAMESET element.
     */
    private XHTMLElement findBody(Element html) {
	XHTMLElement body = (XHTMLElement)getDirectChildElement("body", html);
        if (body == null) {
	    body = (XHTMLElement)getDirectChildElement("frameset", html);
        }
        return body;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getElementById
     */
    public synchronized Element getElementById( String elementId )
    {
        Element idElement = super.getElementById(elementId);
        if (idElement != null) {
            return idElement;
        }
        return getElementById( elementId, this );
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getElementsByName
     */
    public NodeList getElementsByName( String elementName )
    {
        return new NameNodeListImpl( this, elementName );
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getElementsByTagName
     */
    public final NodeList getElementsByTagName( String tagName )
    {
	return super.getElementsByTagName( tagName );
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getElementsByTagNameNS
     */
    public final NodeList getElementsByTagNameNS( String namespaceURI,
                                                  String localName )
    {
        if ( namespaceURI != null && namespaceURI.length() > 0 ) {
	    return super.getElementsByTagNameNS( namespaceURI, localName );
        }
        return super.getElementsByTagName( localName );
    }


    /**
     * Xerces-specific constructor. "localName" is passed in, so we don't need
     * to create a new String for it.
     *
     * @param namespaceURI The namespace URI of the element to
     *                     create.
     * @param qualifiedName The qualified name of the element type to
     *                      instantiate.
     * @param localpart     The local name of the element to instantiate.
     * @return Element A new Element object with the following attributes:
     * @throws DOMException INVALID_CHARACTER_ERR: Raised if the specified
     *                      name contains an invalid character.
     */
    public Element createElementNS(String namespaceURI, String qualifiedName,
                                   String localpart)
        throws DOMException
    {
        return createElementNS(namespaceURI, qualifiedName);
    }

    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#createElementNS
     */
    public Element createElementNS( String namespaceURI, String qualifiedName )
    {
        // FIXME: should be a common method for doing this
        int index = qualifiedName.indexOf(':');
        String tagName;
        if (index < 0) {
            if (namespaceURI == null) qualifiedName = qualifiedName.toLowerCase(Locale.ENGLISH);
            tagName =  qualifiedName;
        } else {
            tagName = qualifiedName.substring(index+1);
        }
        Class elemClass = (Class)fElementTypes.get(tagName);
        if (elemClass != null) {
            try	{
                Constructor cnst = elemClass.getConstructor(fElemConstructorSig);
                return (Element)cnst.newInstance(new Object[] {this, namespaceURI, qualifiedName});
            } catch (Exception except) {
                // Need more specific error..
                throw new XMLCError("failed to construct element for \""
                                    + qualifiedName + "\"", except);
            }
        } else {
            return new XHTMLElementImpl(this, namespaceURI, qualifiedName);
        }
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#createElement
     */
    public Element createElement( String tagName )
        throws DOMException
    {
        return createElementNS(null, tagName);
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getReferrer
     */
    public String getReferrer()
    {
        // Information not available on server side.
        return null;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getDomain
     */
    public String getDomain()
    {
        // Information not available on server side.
        return null;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getURL
     */
    public String getURL()
    {
        // Information not available on server side.
        return null;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getCookie
     */
    public String getCookie()
    {
        // Information not available on server side.
        return null;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#setCookie
     */
    public void setCookie( String cookie )
    {
        // Information not available on server side.
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getImages
     */
    public HTMLCollection getImages()
    {
        // For more information see HTMLCollection#collectionMatch
	if ( fImages == null )
	    fImages = new XHTMLCollectionImpl( (XHTMLElement) getBody(), XHTMLCollectionImpl.IMAGE );
	return fImages;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getApplets
     */
    public HTMLCollection getApplets()
    {
        // For more information see HTMLCollection#collectionMatch
	if ( fApplets == null )
	    fApplets = new XHTMLCollectionImpl( (XHTMLElement) getBody(), XHTMLCollectionImpl.APPLET );
	return fApplets;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getLinks
     */
    public HTMLCollection getLinks()
    {
        // For more information see HTMLCollection#collectionMatch
	if ( fLinks == null )
	    fLinks = new XHTMLCollectionImpl( (XHTMLElement) getBody(), XHTMLCollectionImpl.LINK );
	return fLinks;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getForms
     */
    public HTMLCollection getForms()
    {
        // For more information see HTMLCollection#collectionMatch
	if ( fForms == null )
	    fForms = new XHTMLCollectionImpl( (XHTMLElement) getBody(), XHTMLCollectionImpl.FORM );
	return fForms;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#getAnchors
     */
    public HTMLCollection getAnchors()
    {
        // For more information see HTMLCollection#collectionMatch
	if ( fAnchors == null )
	    fAnchors = new XHTMLCollectionImpl( (XHTMLElement) getBody(), XHTMLCollectionImpl.ANCHOR );
	return fAnchors;
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#open
     */
    public void open()
    {
        // When called an in-memory is prepared. The document tree is still
        // accessible the old way, until this writer is closed.
	if ( fWriter == null )
	    fWriter = new StringWriter();
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#close
     */
    public void close()
    {
        // ! NOT IMPLEMENTED, REQUIRES PARSER !
	if ( fWriter != null )
        {
	    fWriter = null;
        }
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#write
     */
    public void write( String text )
    {
        // Write a string into the in-memory writer.
	if ( fWriter != null )
	    fWriter.write( text );
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#writeln
     */
    public void writeln( String text )
    {
        // Write a line into the in-memory writer.
	if ( fWriter != null )
	    fWriter.write( text + "\n" );
    }


    /**
     * @see org.apache.html.dom.HTMLDocumentImpl#cloneNode
     */
    public Node cloneNode( boolean deep )
    {
        XHTMLDocumentImpl newdoc = new XHTMLDocumentImpl();
        callUserDataHandlers(this, newdoc, UserDataHandler.NODE_CLONED);
        cloneNode(newdoc, deep);

        // experimental
        newdoc.mutationEvents = mutationEvents;

        return newdoc;
    }

    /**
     * override in order to return the correct dom implementation for XHTML
     * @see org.apache.xerces.dom.DocumentImpl#getImplementation
     */
    public DOMImplementation getImplementation() {
        return XHTMLDOMImplementationImpl.getDOMImplementation();
    }


    /* (non-Javadoc)
     * @see CoreDocumentImpl#canRenameElements()
     */
    protected boolean canRenameElements(String newNamespaceURI, String newNodeName, ElementImpl el) {
        if (el.getNamespaceURI() != null) {
            // element is not HTML:
            // can be renamed if not changed to HTML
            return newNamespaceURI != null;
        }
        
        // check whether a class change is required
        Class newClass = (Class) fElementTypes.get(newNodeName.toUpperCase(Locale.ENGLISH));
        Class oldClass = (Class) fElementTypes.get(el.getTagName());
        return newClass == oldClass;
    }


    /**
     * Recursive method retreives an element by its <code>id</code> attribute.
     * Called by {@link #getElementById(String)}.
     *
     * @param elementId The <code>id</code> value to look for
     * @return The node in which to look for
     */
    private Element getElementById( String elementId, Node node )
    {
        Node    child;
        Element    result;

        child = node.getFirstChild();
        while ( child != null )
        {
            if ( child instanceof Element )
            {
                if ( elementId.equals( ( (Element) child ).getAttribute( "id" ) ) )
                    return (Element) child;
                result = getElementById( elementId, child );
                if ( result != null )
                    return result;
            }
            child = child.getNextSibling();
        }
        return null;
    }


    /**
     * Called by the constructor to populate the element types list (see {@link
     * #_elementTypesHTML}). Will be called multiple times but populate the list
     * only the first time. Replacement for static constructor.
     */
    private synchronized static void populateElementTypes()
    {
        // This class looks like it is due to some strange
        // (read: inconsistent) JVM bugs.
        // Initially all this code was placed in the static constructor,
        // but that caused some early JVMs (1.1) to go mad, and if a
        // class could not be found (as happened during development),
        // the JVM would die.
        // Bertrand Delacretaz <bdelacretaz@worldcom.ch> pointed out
        // several configurations where HTMLAnchorElementImpl.class
        // failed, forcing me to revert back to Class.forName().

        if ( fElementTypes != null )
            return;
        fElementTypes = new HashMap( 91 );
        populateElementType( "basefont", "XHTMLBaseFontElementImpl" );
        populateElementType( "button", "XHTMLButtonElementImpl" );
        populateElementType( "textarea", "XHTMLTextAreaElementImpl" );
        populateElementType( "em", "XHTMLEmElementImpl" );
        populateElementType( "small", "XHTMLSmallElementImpl" );
        populateElementType( "area", "XHTMLAreaElementImpl" );
        populateElementType( "noframes", "XHTMLNoframesElementImpl" );
        populateElementType( "bdo", "XHTMLBdoElementImpl" );
        populateElementType( "form", "XHTMLFormElementImpl" );
        populateElementType( "link", "XHTMLLinkElementImpl" );
        populateElementType( "label", "XHTMLLabelElementImpl" );
        populateElementType( "dt", "XHTMLDtElementImpl" );
        populateElementType( "span", "XHTMLSpanElementImpl" );
        populateElementType( "isindex", "XHTMLIsIndexElementImpl" );
        populateElementType( "title", "XHTMLTitleElementImpl" );
        populateElementType( "strong", "XHTMLStrongElementImpl" );
        populateElementType( "script", "XHTMLScriptElementImpl" );
        populateElementType( "div", "XHTMLDivElementImpl" );
        populateElementType( "dl", "XHTMLDListElementImpl" );
        populateElementType( "blockquote", "XHTMLQuoteElementImpl" );
        populateElementType( "kbd", "XHTMLKbdElementImpl" );
        populateElementType( "menu", "XHTMLMenuElementImpl" );
        populateElementType( "body", "XHTMLBodyElementImpl" );
        populateElementType( "dir", "XHTMLDirectoryElementImpl" );
        populateElementType( "ins", "XHTMLModElementImpl" );
        populateElementType( "map", "XHTMLMapElementImpl" );
        populateElementType( "dd", "XHTMLDdElementImpl" );
        populateElementType( "fieldset", "XHTMLFieldSetElementImpl" );
        populateElementType( "head", "XHTMLHeadElementImpl" );
        populateElementType( "col", "XHTMLTableColElementImpl" );
        populateElementType( "base", "XHTMLBaseElementImpl" );
        populateElementType( "big", "XHTMLBigElementImpl" );
        populateElementType( "meta", "XHTMLMetaElementImpl" );
        populateElementType( "code", "XHTMLCodeElementImpl" );
        populateElementType( "tbody", "XHTMLTbodyElementImpl" );
        populateElementType( "option", "XHTMLOptionElementImpl" );
        populateElementType( "u", "XHTMLUElementImpl" );
        populateElementType( "s", "XHTMLSElementImpl" );
        populateElementType( "q", "XHTMLQuoteElementImpl" );
        populateElementType( "p", "XHTMLParagraphElementImpl" );
        populateElementType( "ol", "XHTMLOListElementImpl" );
        populateElementType( "thead", "XHTMLTheadElementImpl" );
        populateElementType( "ul", "XHTMLUListElementImpl" );
        populateElementType( "i", "XHTMLIElementImpl" );
        populateElementType( "pre", "XHTMLPreElementImpl" );
        populateElementType( "optgroup", "XHTMLOptGroupElementImpl" );
        populateElementType( "img", "XHTMLImageElementImpl" );
        populateElementType( "caption", "XHTMLTableCaptionElementImpl" );
        populateElementType( "b", "XHTMLBElementImpl" );
        populateElementType( "a", "XHTMLAnchorElementImpl" );
        populateElementType( "frame", "XHTMLFrameElementImpl" );
        populateElementType( "br", "XHTMLBRElementImpl" );
        populateElementType( "style", "XHTMLStyleElementImpl" );
        populateElementType( "hr", "XHTMLHRElementImpl" );
        populateElementType( "param", "XHTMLParamElementImpl" );
        populateElementType( "table", "XHTMLTableElementImpl" );
        populateElementType( "applet", "XHTMLAppletElementImpl" );
        populateElementType( "tt", "XHTMLTtElementImpl" );
        populateElementType( "tr", "XHTMLTableRowElementImpl" );
        populateElementType( "th", "XHTMLTableCellElementImpl" );
        populateElementType( "center", "XHTMLCenterElementImpl" );
        populateElementType( "td", "XHTMLTableCellElementImpl" );
        populateElementType( "samp", "XHTMLSampElementImpl" );
        populateElementType( "tfoot", "XHTMLTfootElementImpl" );
        populateElementType( "font", "XHTMLFontElementImpl" );
        populateElementType( "dfn", "XHTMLDfnElementImpl" );
        populateElementType( "noscript", "XHTMLNoscriptElementImpl" );
        populateElementType( "object", "XHTMLObjectElementImpl" );
        populateElementType( "colgroup", "XHTMLTableColElementImpl" );
        populateElementType( "sup", "XHTMLSupElementImpl" );
        populateElementType( "html", "XHTMLHtmlElementImpl" );
        populateElementType( "h6", "XHTMLHeadingElementImpl" );
        populateElementType( "h5", "XHTMLHeadingElementImpl" );
        populateElementType( "h4", "XHTMLHeadingElementImpl" );
        populateElementType( "h3", "XHTMLHeadingElementImpl" );
        populateElementType( "frameset", "XHTMLFrameSetElementImpl" );
        populateElementType( "h2", "XHTMLHeadingElementImpl" );
        populateElementType( "h1", "XHTMLHeadingElementImpl" );
        populateElementType( "iframe", "XHTMLIFrameElementImpl" );
        populateElementType( "strike", "XHTMLStrikeElementImpl" );
        populateElementType( "sub", "XHTMLSubElementImpl" );
        populateElementType( "acronym", "XHTMLAcronymElementImpl" );
        populateElementType( "select", "XHTMLSelectElementImpl" );
        populateElementType( "del", "XHTMLModElementImpl" );
        populateElementType( "li", "XHTMLLIElementImpl" );
        populateElementType( "cite", "XHTMLCiteElementImpl" );
        populateElementType( "var", "XHTMLVarElementImpl" );
        populateElementType( "legend", "XHTMLLegendElementImpl" );
        populateElementType( "abbr", "XHTMLAbbrElementImpl" );
        populateElementType( "input", "XHTMLInputElementImpl" );
        populateElementType( "address", "XHTMLAddressElementImpl" );
    }


    private static void populateElementType( String tagName, String className )
    {
        try {
            fElementTypes.put( tagName,
                ObjectFactory.findProviderClass("org.enhydra.xml.xhtml.dom.xerces." + className,
                    XHTMLDocumentBase.class.getClassLoader(), true) );
        } catch ( Exception except ) {
            throw new RuntimeException( "HTM019 OpenXML Error: Could not find or execute class " + className + " implementing XHTML element " + tagName
                                  + "\n" + className + "\t" + tagName);
        }
    }


}

