/*
 * Copyright 1999,2000,2004,2005 The Apache Software Foundation.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.enhydra.xml.xhtml.dom.xerces;

import java.util.Locale;

import org.enhydra.xml.xhtml.dom.XHTMLElement;
import org.enhydra.xml.xhtml.dom.XHTMLFormElement;

import org.apache.xerces.dom.ElementNSImpl;
import org.w3c.dom.Attr;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
//import org.w3c.dom.html.HTMLElement;
import org.w3c.dom.html.HTMLFormElement;

/*
 * Derived from Xerces HTMLDocument implementation.
 * Author: Assaf Arkin <arkin@exoffice.com>
 */

/**
 * Implements an HTML-specific element, an {@link org.w3c.dom.Element} that
 * will only appear inside HTML documents. This element extends {@link
 * org.apache.xerces.dom.ElementImpl} by adding methods for directly
 * manipulating HTML-specific attributes. All HTML elements gain access to
 * the <code>id</code>, <code>title</code>, <code>lang</code>,
 * <code>dir</code> and <code>class</code> attributes. Other elements
 * add their own specific attributes.
 * 
 * @xerces.internal
 * 
 * @version $Revision: 1.9 $ $Date: 2007-04-22 05:40:58 $
 * @see org.w3c.dom.html.HTMLElement
 */
public class XHTMLElementImpl
    extends ElementNSImpl
    implements XHTMLElement
{


    /**
     * Constructor required owner document and element tag name. Will be called
     * by the constructor of specific element types but with a known tag name.
     * Assures that the owner document is an HTML element.
     * 
     * @param owner The owner HTML document
     * @param tagName The element's tag name
     */
    XHTMLElementImpl( XHTMLDocumentBase owner, String namespaceURI, String tagName ) {
        super( owner, namespaceURI, tagName );
    }
    
    public String getId() {
        return getAttribute( "id" );
    }
    
    public void setId( String id ) {
        setAttribute( "id", id );
    }   
    
    public String getTitle() {
        return getAttribute( "title" );
    }
    
    public void setTitle( String title ) {
        setAttribute( "title", title );
    }
    
    public String getLang() {
        return getAttribute( "lang" );
    }  
    
    public void setLang( String lang ) {
        setAttribute( "lang", lang );
    }
    
    
    public void setXmlLang( String xmlLang ) {
	setAttribute( "xml:lang" , xmlLang );
    }


    public String getXmlLang() {
	return getAttribute( "xml:lang" );
    }


    public String getDir() {
        return getAttribute( "dir" );
    }
    
    public void setDir( String dir ) {
        setAttribute( "dir", dir );
    }
 
    public String getClassName() {
        return getAttribute( "class" );
    }
    
    public void setClassName( String className ) {
        setAttribute( "class", className );
    }
    
    /**
     * Convenience method used to translate an attribute value into an integer
     * value. Returns the integer value or zero if the attribute is not a
     * valid numeric string.
     * 
     * @param value The value of the attribute
     * @return The integer value, or zero if not a valid numeric string
     */
    int getInteger( String value ) {
        try {
            return Integer.parseInt( value );
        }
        catch ( NumberFormatException except ) {
            return 0;
        }
    }


    /*
     * Set the value of a integer attribute.
     */
    void setAttribute(String attr, int value) {
        setAttribute( attr, String.valueOf( value ) );
    }
    
    /**
     * Convenience method used to translate an attribute value into a boolean
     * value. If the attribute has an associated value (even an empty string),
     * it is set and true is returned. If the attribute does not exist, false
     * is returend.
     * 
     * @param value The value of the attribute
     * @return True or false depending on whether the attribute has been set
     */
    boolean getBinary( String name ) {
        return ( getAttributeNode( name ) != null );
    }
    
    /**
     * Convenience method used to set a boolean attribute. If the value is true,
     * the attribute is set to an empty string. If the value is false, the attribute
     * is removed. HTML 4.0 understands empty strings as set attributes.
     * 
     * @param name The name of the attribute
     * @param value The value of the attribute
     */
    void setAttribute( String name, boolean value ) {
        if ( value ) {
            setAttribute( name, name );
        }
        else {
            removeAttribute( name );
        }
    }

    public Attr getAttributeNode( String attrName ) {
        if (attrName == null) return null;
        return super.getAttributeNode( attrName.toLowerCase(Locale.ENGLISH) );
    }

    public Attr getAttributeNodeNS( String namespaceURI,
            String localName ) {
        if (localName == null) return null;
        if ( namespaceURI != null && namespaceURI.length() > 0 ) {
            return super.getAttributeNodeNS( namespaceURI, localName );
        }
        return super.getAttributeNode( localName.toLowerCase(Locale.ENGLISH) );
    }
    
    public String getAttribute( String attrName ) {
        if (attrName == null) return null;
        return super.getAttribute( attrName.toLowerCase(Locale.ENGLISH) );
    }

    public String getAttributeNS( String namespaceURI,
            String localName ) {
        if (localName == null) return null;
        if ( namespaceURI != null && namespaceURI.length() > 0 ) {
            return super.getAttributeNS( namespaceURI, localName );
        }
        return super.getAttribute( localName.toLowerCase(Locale.ENGLISH) );
    }

    public final NodeList getElementsByTagName( String tagName ) {
	return super.getElementsByTagName( tagName.toLowerCase(Locale.ENGLISH) );
    }

    public final NodeList getElementsByTagNameNS( String namespaceURI,
            String localName ) {
        if ( namespaceURI != null && namespaceURI.length() > 0 ) {
	    return super.getElementsByTagNameNS( namespaceURI, localName );
        }
        return super.getElementsByTagName( localName.toLowerCase(Locale.ENGLISH) );
    }

    /**
     * Convenience method returns the form in which this form element is contained.
     * This method is exposed for form elements through the DOM API, but other
     * elements have no access to it through the API.
     */
    public HTMLFormElement getForm() {
        Node parent = getParentNode(); 
        while ( parent != null ) {
            if ( parent instanceof XHTMLFormElement ) {
                return (XHTMLFormElement) parent;
            }
            parent = parent.getParentNode();
        }
        return null;
    }


    public void setOnKeyUp (String newValue) {
	setAttribute("onkeyup", newValue);
    }

    public String getOnKeyUp () {
	return getAttribute ("onkeyup");
    }

    public void setStyle (String newValue) {
	setAttribute("style", newValue);
    }

    public String getStyle () {
	return getAttribute ("style");
    }

    public void setOnMouseDown (String newValue) {
	setAttribute("onmousedown", newValue);
    }

    public String getOnMouseDown () {
	return getAttribute ("onmousedown");
    }

    public void setOnKeyPress (String newValue) {
	setAttribute("onkeypress", newValue);
    }

    public String getOnKeyPress () {
	return getAttribute ("onkeypress");
    }

    public void setOnDblClick (String newValue) {
	setAttribute("ondblclick", newValue);
    }

    public String getOnDblClick () {
	return getAttribute ("ondblclick");
    }

    public void setOnKeyDown (String newValue) {
	setAttribute("onkeydown", newValue);
    }

    public String getOnKeyDown () {
	return getAttribute ("onkeydown");
    }

    public void setOnMouseMove (String newValue) {
	setAttribute("onmousemove", newValue);
    }

    public String getOnMouseMove () {
	return getAttribute ("onmousemove");
    }

    public void setOnMouseUp (String newValue) {
	setAttribute("onmouseup", newValue);
    }

    public String getOnMouseUp () {
	return getAttribute ("onmouseup");
    }

    public void setOnMouseOut (String newValue) {
	setAttribute("onmouseout", newValue);
    }

    public String getOnMouseOut () {
	return getAttribute ("onmouseout");
    }

    public void setOnClick (String newValue) {
	setAttribute("onclick", newValue);
    }

    public String getOnClick () {
	return getAttribute ("onclick");
    }

    public void setOnMouseOver (String newValue) {
	setAttribute("onmouseover", newValue);
    }

    public String getOnMouseOver () {
	return getAttribute ("onmouseover");
    }

    public void setOnBlur (String newValue) {
	setAttribute("onblur", newValue);
    }

    public String getOnBlur () {
	return getAttribute ("onblur");
    }

    public void setOnFocus (String newValue) {
	setAttribute("onfocus", newValue);
    }

    public String getOnFocus () {
	return getAttribute ("onfocus");
    }


}
