/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DTDInfo.java,v 1.3 2007-04-22 05:40:58 jkjome Exp $
 */

package tests.org.enhydra.xml.xhtml.dominfo;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.apache.xerces.framework.XMLAttrList;
import org.apache.xerces.framework.XMLContentSpec;
import org.apache.xerces.framework.XMLDocumentHandler;
import org.apache.xerces.framework.XMLParser;
import org.apache.xerces.utils.QName;
import org.enhydra.error.FatalExceptionError;
import org.enhydra.xml.io.ErrorReporter;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Parse a DTD using the Xerces parser. A subset of the DTD information is stored in this
 * object.
 */
class DTDInfo {
    /**
     * Element information.
     */
    public class ElementInfo {
        /* Element properties */
        private final String fRawName;
        private Map fAttrs = new TreeMap();
           

        /** Constructor */
        ElementInfo(String rawName) {
            fRawName = rawName;
        }

        /** Add an attribute */
        public void addAttr(AttrInfo attr) {
            fAttrs.put(attr.getRawName(), attr);
        }
        
        /* Get the element name */
        public String getRawName() {
            return fRawName;
        }

        /* Get the attribute names */
        public Set getAttrNames() {
            return fAttrs.keySet();
        }

        /* Get an attribute */
        public AttrInfo getAttr(String name) {
            AttrInfo attr = (AttrInfo)fAttrs.get(name);
            if (attr == null) {
                throw new Error("Can't find attr \"" + attr + "\" of element \"" + fRawName + "\"");
            }
            return attr;
        }

        /* Dump the element */
        public void dump(PrintWriter out,
                         int level) {
            PrintUtils.printIndent(out, level);
            out.println(fRawName);
            Iterator attrNames = getAttrNames().iterator();
            while (attrNames.hasNext()) {
                getAttr((String)attrNames.next()).dump(out, level+1);
            }
        }

        /** Get a string representation */
        public String toString() {
            return fRawName;
        }
    }

    /**
     * Attribute information.   
     */
    public class AttrInfo {
        /** Element we are associated with */
        private ElementInfo fElement;
        
        /** Attribute properties */
        private String fRawName;

        /** Constructor */
        AttrInfo(ElementInfo element,
                 String rawName) {
            fElement = element;
            fRawName = rawName;
        }

        /* Get the element we are associated with */
        public ElementInfo getElement() {
            return fElement;
        }

        /* Get the raw attr name */
        public String getRawName() {
            return fRawName;
        }

        /* Dump the attribute */
        public void dump(PrintWriter out,
                         int level) {
            PrintUtils.printIndent(out, level);
            out.println(fRawName);
        }

        /** Get a string representation */
        public String toString() {
            return fRawName;
        }
    }

    /**
     * The DTD file that was parsed.
     */
    private String fDTDPath;
    
    /**
     * Table of elements
     */
    private Map fElements = new TreeMap();

    /**
     * Xerces handlers to parse the DTD.  Parse the DTD.  Since there isn't a
     * direct `DTD' parser in Xerces, we fake it by constructing a document to
     * parse.
     */
    private class DTDParser extends XMLParser
        implements EntityResolver, XMLDocumentHandler, XMLDocumentHandler.DTDHandler {

        /**
         * Constructor.
         */
        public DTDParser() throws IOException, SAXException {
            ErrorReporter errorReporter = new ErrorReporter();

            // Configure parser.
            initHandlers(true, this, this);
            setEntityResolver(this);
            setErrorHandler(errorReporter);
            setAllowJavaEncodings(true);
            setNamespaces(true);
            setValidation(true);
        }

        /**
         * Parse a DTD.
         */
        public void parse(String dtdPath) throws IOException, SAXException {
            String xmlDoc = "<?xml version=\"1.0\"?>"
                + "<!DOCTYPE html SYSTEM \"" +dtdPath + "\">"
                + "<html></html>";
            StringReader reader = new StringReader(xmlDoc);

            InputSource input = new InputSource();
            input.setCharacterStream(reader);
            input.setSystemId(dtdPath);
            super.parse(input);
        }

        /**
         * Get a string from the string pool.
         */
        private String getString(int index) {
            return fStringPool.toString(index);
        }

        /**
         * Resolve an entity.
         * @see EntityResolver#resolveEntity
         */
        public InputSource resolveEntity(String publicId,
                                         String systemId) throws SAXException, IOException {
            return null;
        }

        /** 
         * Handle start of document.
         * @see XMLDocumentHandler#startDocument
         */
        public void startDocument() throws Exception {
        }

        /**
         * Handle end document.
         * @see XMLDocumentHandler#endDocument
         */
        public void endDocument() throws Exception {
        }

        /**
         * @see XMLDocumentHandler#xmlDecl
         */
        public void xmlDecl(int version, int encoding, int standalone) throws Exception {
        }

        /**
         * @see XMLDocumentHandler#textDecl
         * @see XMLDocumentHandler.DTDHandler#textDecl
         */
        public void textDecl(int version, int encoding) throws Exception {
        }

        /** 
         * Handle start of a namespace declaration scope. 
         *
         * @see XMLDocumentHandler#startNamespaceDeclScope
         */
        public void startNamespaceDeclScope(int prefix,
                                            int uri) throws Exception {
        }

        /** 
         * Handle end of a namespace declaration scope. 
         *
         * @see XMLDocumentHandler#startNamespaceDeclScope
         */
        public void endNamespaceDeclScope(int prefix) throws Exception {
        }

        /**
         * Handle start of element.
         * @see XMLDocumentHandler#startElement
         */
        public void startElement(QName element,
                                 XMLAttrList attrList,
                                 int attrListHandle) throws Exception {
        }

        /**
         * Handle end element.
         * @see XMLDocumentHandler#endElement
         */
        public void endElement(QName element) throws Exception {
        }

        /**
         * Handle the start of an entity reference.  If it's is one of the
         * standard character entity references, we don't push the create
         * the node, we just let the child be appended directly in its place.
         *
         * @see XMLDocumentHandler#startEntityReference
         */
        public void startEntityReference(int entityName,
                                         int entityType,
                                         int entityContext) throws Exception {
        }

        /**
         * Handle the end of an entity reference.
         * @see XMLDocumentHandler#endEntityReference
         */
        public void endEntityReference(int entityName,
                                       int entityType,
                                       int entityContext) throws Exception {
        }

        /**
         * Not used.
         * @see XMLDocumentHandler#characters
         */
        public void characters(int data) throws Exception {
            throw new Error("fatal error: method that should not be invoked called");
        }

        /**
         * Not used.
         * @see XMLDocumentHandler#ignorableWhitespace
         */
        public void ignorableWhitespace(int data) throws Exception {
            throw new Error("fatal error: method that should not be invoked called");
        }

        /** 
         * Handle start of CDATA section. 
         * @see XMLDocumentHandler#startCDATA
         */
        public void startCDATA() {
        }

        /** 
         * Handle for end of CDATA section.
         * @see XMLDocumentHandler#endCDATA
         */
        public void endCDATA() {
        }

        /** 
         * Handle processing instruction.
         * @see XMLDocumentHandler#processingInstruction
         * @see XMLDocumentHandler.DTDHandler#processingInstruction
         */
        public void processingInstruction(int target,
                                          int data) throws Exception {
        }

        /**
         * Handle a comment.
         * @see XMLDocumentHandler#comment
         */
        public void comment(int comment) throws Exception {
        }

        /**
         * Handle characters.
         * @see XMLDocumentHandler#characters
         */
        public void characters(char ch[],
                               int start,
                               int length) throws Exception {
        }

        /**
         * Handle ignorable whitespace.
         * @see XMLDocumentHandler#ignorableWhitespace
         */
        public void ignorableWhitespace(char ch[],
                                        int start,
                                        int length) throws Exception {
        }

        /**
         * @see XMLDocumentHandler.DTDHandler#startDTD
         */
        public void startDTD(QName rootElement,
                             int publicId,
                             int systemId) {
        }

        /**
         * @see XMLDocumentHandler.DTDHandler#internalSubset
         */
        public void internalSubset(int internalSubset) {
        }

        /** 
         * @see XMLDocumentHandler.DTDHandler#endDTD
         */
        public void endDTD() {
        }

        /**
         * <!ELEMENT Name contentspec>
         *
         * @see XMLDocumentHandler.DTDHandler#elementDecl
         */
        public void elementDecl(QName elementDecl,
                                int contentSpecType, 
                                int contentSpecIndex,
                                XMLContentSpec.Provider contentSpecProvider) throws Exception {
            if (fElements.containsKey(getString(elementDecl.rawname))) {
                throw new Error("Duplicate element definintion: " + elementDecl.rawname);
            }
            ElementInfo element = new ElementInfo(getString(elementDecl.rawname));
            fElements.put(element.getRawName(), element);
        }

        /**
         * <!ATTLIST Name AttDef>
         *
         * @see XMLDocumentHandler.DTDHandler#attlistDecl
         */
        public void attlistDecl(QName elementDecl,
                                QName attributeDecl,
                                int attType,
                                boolean attList,
                                String enumString,
                                int attDefaultType,
                                int attDefaultValue) throws Exception {
            ElementInfo element = (ElementInfo)fElements.get(getString(elementDecl.rawname));
            if (!fElements.containsKey(getString(elementDecl.rawname))) {
                throw new Error("Element definintion not found when adding attribute: " + elementDecl.rawname);
            }
            element.addAttr(new AttrInfo(element, getString(attributeDecl.rawname)));
        }

        /**
         * <!ENTITY % Name EntityValue>
         *
         * @see XMLDocumentHandler.DTDHandler#internalPEDecl
         */
        public void internalPEDecl(int entityName,
                                   int entityValue) {
        }

        /**
         * <!ENTITY % Name ExternalID>
         *
         * @see XMLDocumentHandler.DTDHandler#externalPEDecl
         */
        public void externalPEDecl(int entityName,
                                   int publicId,
                                   int systemId) {
        }

        /**
         * <!ENTITY Name EntityValue>
         *
         * @see XMLDocumentHandler.DTDHandler#internalEntityDecl
         */
        public void internalEntityDecl(int entityName,
                                       int entityValue) {
        }

        /**
         * <!ENTITY Name ExternalID>
         *
         * @see XMLDocumentHandler.DTDHandler#externalEntityDecl
         */
        public void externalEntityDecl(int entityName,
                                       int publicId,
                                       int systemId) {
        }

        /**
         * <!ENTITY Name ExternalID NDataDecl>
         *
         * @see XMLDocumentHandler.DTDHandler#unparsedEntityDecl
         */
        public void unparsedEntityDecl(int entityName,
                                       int publicId,
                                       int systemId,
                                       int notationName) {
        }

        /**
         * <!NOTATION Name ExternalId>
         *
         * @see XMLDocumentHandler.DTDHandler#notationDecl
         */
        public void notationDecl(int notationName,
                                 int publicId,
                                 int systemId) {
        }
    }

    /**
     * Constructor.
     */
    public DTDInfo(String dtdPath) {
        try {
            fDTDPath = dtdPath;
            DTDParser parser = new DTDParser();
            parser.parse(dtdPath);
        } catch (IOException except) {
            throw new FatalExceptionError(except);
        } catch (SAXException except) {
            throw new FatalExceptionError(except);
        }
    }

    /**
     * Get the DTD path.
     */
    public String getDTDPath() {
        return fDTDPath;
    }

    /* Get the element names */
    public Set getElementNames() {
        return fElements.keySet();
    }

    /* Get an element */
    public ElementInfo getElement(String name) {
        ElementInfo element = (ElementInfo)fElements.get(name);
        if (element == null) {
            throw new Error("Can't find element \"" + name + "\"");
        }
        return element;
    }

    /* Get the set of ElementInfo objects */
    public Set getElementInfoSet() {
        HashSet set = new HashSet();

        Iterator elementNames = getElementNames().iterator();
        while (elementNames.hasNext()) {
            set.add(getElement((String)elementNames.next()));
        }
        return set;
    }

    /**
     * Dump the elements in the tree.
     */
    public void dump(PrintWriter out) {
        Iterator elementNames = getElementNames().iterator();
        while (elementNames.hasNext()) {
            getElement((String)elementNames.next()).dump(out, 0);
        }
    }

    /**
     * Main used for testing.
     */
    public static void main(String[] args) throws IOException, SAXException {
        if (args.length != 1) {
            System.err.println("Wrong # args: DTDInfo dtdfile");
            System.exit(1);
        }
        DTDInfo dtdInfo = new DTDInfo(args[0]);
        dtdInfo.dump(new PrintWriter(System.out, true));
    }
}
