/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DOMInfo.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.dom;

import java.io.OutputStream;
import java.io.PrintWriter;

import org.enhydra.xml.lazydom.LazyDOMInfoPrinter;
import org.enhydra.xml.lazydom.LazyDocument;
import org.w3c.dom.Node;

/*
 * NB: A lot of this code is duplicated in LazyDOMInfo using template nodes,
 * keep them in-sync.
 */

// FIXME: need to implement no-recursion; the organization of this class is
// to fixed.  It would be useful to have methods to just format a node
// (without recursing) into a string, then the traversal to tie it together.

/**
 * Methods to print information out about a DOM.  Used for debugging, etc.
 * Has special handling for Lazy DOMs to prevent expansion.
 */
public class DOMInfo {
    /**
     * Print attributes as the Nodes that represent them.
     */
    public static final int PRINT_ATTR_DETAILS = 0x01;

    /**
     * Print all attributes, even unspecified ones.
     */
    public static final int ALL_ATTRS = 0x02;

    /**
     * Print newlines and carriage returns as \n and \r, avoiding
     * linebreaks in text
     */
    public static final int SYM_TEXT_LINEBREAKS = 0x04;

    /**
     * Don't recurse, just print the specified node.
     */
    private static final int NO_RECURSION = 0x08;  //FIXME: private because not implemented

    /**
     * Truncate text, comments etc, to 16 chars. Implys SYM_TEXT_LINEBREAKS.
     */
    public static final int TRUNCATE_TEXT = 0x10;

    /**
     * All Print options, except NO_RECURSION.
     */
    public static final int PRINT_ALL = 0xffff & ~NO_RECURSION;

    /**
     * Default print options.
     */
    public static final int PRINT_DEFAULT = 0;

    /**
     * Options to use.
     */
    protected int fOptions;

    /**
     * Write for output.
     */
    protected PrintWriter fOut;

    /**
     * DOM traversal object and handler.
     */
    protected DOMTraversal fTraverser;
    protected DOMInfoPrinter fInfoPrinter;

    /**
     * Create a DOMInfo object to print data about the DOM.  Sets up the
     * traverser and traversal handlers based on the document type.
     */
    private static DOMInfo createDOMInfo(int options,
                                         Node node,
                                         PrintWriter out) {

        // This will create a traverser and handler based on document type.
        // Much check real document, as it could be an XMLC object..
        DOMTraversal traverser
            = DOMTraversal.getTraverser(null, 
                                        getTraverserOptions(options),
                                        node);

        DOMInfoPrinter handler;
        if (DOMOps.getDocument(node) instanceof LazyDocument) {
            handler = new LazyDOMInfoPrinter(options, traverser, out);
        } else {
            handler = new DOMInfoPrinter(options, traverser, out);
        }
        traverser.setHandler(handler);
        return new DOMInfo(options, traverser, handler, out);
    }

    /**
     * Constructor for derived objects.
     */
    protected DOMInfo(int options,
                      DOMTraversal traverser,
                      DOMInfoPrinter infoPrinter,
                      PrintWriter out) {
        fOptions = options;
        fTraverser = traverser;
        fInfoPrinter = infoPrinter;
        fOut = out;
    }

    /**
     * Traverse the DOM.
     */
    private void traverse(Node node) {
        fTraverser.traverse(node);
    }

    /**
     * Get traverser options from DOMInfo options.
     */
    public static int getTraverserOptions(int options) {
        int travOptions = DOMTraversal.SORT_ATTRIBUTES;
        if ((options & ALL_ATTRS) != 0) {
            travOptions |= DOMTraversal.ALL_ATTRIBUTES;
        }
        return travOptions;
    }


    /**
     * Convert a node type to a name.
     */
    public static String nodeTypeToName(short type) {
        switch (type) {
        case Node.ELEMENT_NODE:
            return "Element";
        case Node.ATTRIBUTE_NODE:
            return "Attr";
        case Node.TEXT_NODE:
            return "Text";
        case Node.CDATA_SECTION_NODE:
            return "CDATASection";
        case Node.ENTITY_REFERENCE_NODE:
            return "EntityReference";
        case Node.ENTITY_NODE:
            return "Entity";
        case Node.PROCESSING_INSTRUCTION_NODE:
            return "ProcessingInstruction";
        case Node.COMMENT_NODE:
            return "Comment";
        case Node.DOCUMENT_NODE:
            return "Document";
        case Node.DOCUMENT_TYPE_NODE:
            return "DocumentType";
        case Node.DOCUMENT_FRAGMENT_NODE:
            return "DocumentFragment";
        case Node.NOTATION_NODE:
            return "Notation";
        default:
            throw new IllegalArgumentException("Unknown node type: " + type);
        }
    }

    /**
     * Print a DOM node and its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param root Top of the tree.
     * @param options Set of print options.
     * @param out Output writer.  If null, stderr will be used.
     */
    public static void printTree(String msg,
                                 Node root,
                                 int options,
                                 PrintWriter out) {
        // Default options
        if((options & TRUNCATE_TEXT) != 0) {
            options |= SYM_TEXT_LINEBREAKS;
        }

        if (out == null) {
            out = new PrintWriter(System.err, true);
        }
        if ((msg != null) && (msg.length() > 0)) {
            out.println(msg + ":");
        }
        if (root == null) {
            out.println(" null node");
        } else {
            DOMInfo domInfo = createDOMInfo(options, root, out);
            domInfo.traverse(root);
        }
        out.flush();
    }

    /**
     * Print a DOM node an its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param node Top of the tree.
     * @param out Output writer.  If null, stderr will be used.
     */
    public static void printTree(String msg,
                                 Node root,
                                 PrintWriter out) {
        printTree(msg, root, PRINT_DEFAULT, out);
    }

    /**
     * Print a DOM node an its children.
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param root Top of the tree.
     * @param out Output stream.  If null, stderr will be used.
     */
    public static void printTree(String msg,
                                 Node root,
                                 OutputStream out) {
        PrintWriter writer = new PrintWriter((out != null) ? out : System.err, true);
        printTree(msg, root, PRINT_DEFAULT, writer);
    }

    /**
     * Print a DOM node an its children to stderr..
     *
     * @param msg A message to print at the start.  If null or empty,
     *  don't print a message.
     * @param root Top of the tree.
     */
    public static void printTree(String msg,
                                 Node root) {
        PrintWriter writer = new PrintWriter(System.err, true);
        printTree(msg, root, PRINT_DEFAULT, writer);
    }
}
