/*
 * Copyright (C) 2003, 2005, 2006, 2007, 2010 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.io;

import java.io.IOException;
import java.util.Locale;

import org.enhydra.xml.dom.DOMAccess;
import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

/*
 * FIXME:
 *  - Need to check for optional attributes.
 */


/**
 * Formatter for outputting a HTML DOM as a HTML text document.
 */
final class HTMLFormatter extends BaseDOMFormatter implements Formatter {
    /**
     * Default XML encoding.
     */
    private static final String DEFAULT_XML_ENCODING = "ISO-8859-1";

    /**
     * Table use to optimized checking for characters that should be
     * represented as entity references.
     */
    private static final boolean[] ENTITY_QUICK_CHECK
        = new boolean[MAX_ENTITY_QUICK_CHECK_CHAR+1];

    /**
     * Should SPAN ID attributes be dropped?
     */
    private final boolean fDropSpanIds;

    /**
     * Flag that indicates the ID attribute should be dropped for the current
     * element.
     */
    private boolean fDropThisId;

    /**
     * Flag that indicates whether to use all named entites for all HTML 4.0 character
     * entities or not.
     */
    private boolean fUseHTML4Entities;

    /**
     * Flag that indicates whether to serialize elements, which are normally
     * stored in the HTML DOM as upper-case, in lower-case.  Only applies when
     * using the HTML DOM.
     */
    private boolean fForceHTMLLowerCase;

    /**
     * Indicates that the DOM is the XHTML DOM so special rules are to be applied
     * to output as HTML, eg.. uppercasing element names and replacing the
     * provided document type with the HTML document type
     */
    private final boolean fXHTMLFormattedAsHTML;

    /**
     * Nesting count for script and style elements.  This goes one step beyond
     * fNoFormatNestCount, in that it allows us to avoid resolving special
     * characters or entities, which may exist inside script or style elements,
     * but be meant literally.  Essentially, it's a CDATA Section without being
     * marked as such officially.
     */
    private int fNoFormatScriptStyleNestCount;

    /**
     * Indicates that the next sibling is a Text node
     */
    private boolean fNextSiblingText;

    /**
     * Static constructor.
     */
    static {
        for (char ch = 0; ch <= MAX_ENTITY_QUICK_CHECK_CHAR; ch++) {
            ENTITY_QUICK_CHECK[ch] = (HTMLEntities.charToEntity(ch) != null);
        }
    }

    /**
     * Constructor.
     */
    public HTMLFormatter(Node node,
                         OutputOptions outputOptions,
                         boolean forPreFormatting) {
        super(node, outputOptions, forPreFormatting, DEFAULT_XML_ENCODING, ENTITY_QUICK_CHECK);
        Element docElement =  DOMAccess.accessDocumentElement(fDocument);
        fXHTMLFormattedAsHTML = DOMFormatter.isXHTMLDocument(docElement);
        if (fXHTMLFormattedAsHTML || docElement.hasAttribute("xml:lang") || docElement.hasAttribute("xmlns")) {
            //root element is always expanded, so no need to do it explicitly as is done via DOMAccess.getExpandedElement()
            if (docElement.hasAttribute("xml:lang")) {
                docElement.setAttribute("lang", docElement.getAttribute("xml:lang"));
                docElement.removeAttribute("xml:lang");
            }
            docElement.removeAttribute("xmlns");
        }
        if (fOptions.getPublicId() == null && fOptions.getSystemId() == null) {
            //Re-map DTD only when not overridden by fOptions and processing a
            //[HTML or XML parsed] document with a DocType declared as XHTML
            if (fPublicId != null && fPublicId.indexOf("XHTML") != -1) {
                if (fPublicId.indexOf("Frameset") != -1) {
                    fPublicId = "-//W3C//DTD HTML 4.01 Frameset//EN";
                    if (fSystemId != null) fSystemId = "http://www.w3.org/TR/html401/frameset.dtd";
                } else if (fPublicId.indexOf("Transitional") != -1) {
                    fPublicId = "-//W3C//DTD HTML 4.01 Transitional//EN";
                    if (fSystemId != null) fSystemId = "http://www.w3.org/TR/html401/loose.dtd";
                } else {
                    fPublicId = "-//W3C//DTD HTML 4.01//EN";
                    if (fSystemId != null) fSystemId = "http://www.w3.org/TR/html401/strict.dtd";
                }
            } else if (fSystemId != null && fSystemId.indexOf("xhtml") != -1) {
                if (fSystemId.indexOf("frameset.dtd") != -1) {                
                    fPublicId = "-//W3C//DTD HTML 4.01 Frameset//EN";
                    fSystemId = "http://www.w3.org/TR/html401/frameset.dtd";
                } else if (fSystemId.indexOf("transitional.dtd") != -1) {
                    fPublicId = "-//W3C//DTD HTML 4.01 Transitional//EN";
                    fSystemId = "http://www.w3.org/TR/html401/loose.dtd";
                } else {
                    fPublicId = "-//W3C//DTD HTML 4.01//EN";
                    fSystemId = "http://www.w3.org/TR/html401/strict.dtd";
                }
            }
        }
        fDropSpanIds = fOptions.getDropHtmlSpanIds();
        fUseHTML4Entities = fOptions.getUseHTML4Entities();
        fForceHTMLLowerCase = !fXHTMLFormattedAsHTML && fOptions.getForceHTMLLowerCase();
    }

    /**
     * Get the default OutputOptions for a document formatter with this
     * formatter.  The encoding will not be set, which signals to use the
     * default encoding.
     */
    static OutputOptions getDefaultOutputOptions() {
        return new OutputOptions();  // Nothing special
    }

    /**
     * @see BaseDOMFormatter#getCharacterEntity
     */
    protected final String getCharacterEntity(char textChar) {
        if (fUseHTML4Entities) { 
            return HTMLEntities.charToEntity4(textChar);
        }
        return HTMLEntities.charToEntity(textChar);
    }

    /**
     * Determine if an attribute's value should be printer. Those that don't
     * normally have values only get them if one was explictly supplied.
     */
    private boolean printableAttrValue(Attr attr) {
        return (!(HTMLElements.isBooleanAttr(attr.getName())));
    }

    /**
     * Handler called for Document nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocument
     */
    public void handleDocument(Document document) throws IOException {
        if (fDocType == null) {
            handleDocumentType(null);
        } else {
            fTraverser.processDocumentType(document);
        }
        fTraverser.processChildren(document);
    }

    /**
     * Handler called for Document nodes; writes the DOCTYPE specification.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocumentType
     */
    public void handleDocumentType(DocumentType documentType) throws IOException {
        if (fOptions.getOmitDocType()) {
            return; // Don't output DOCTYPE.
        }

        if (documentType == null && fPublicId == null && fSystemId == null && fOptions.getPublicId() == null && fOptions.getSystemId() == null) {
            return; // No DOCTYPE - neither provided by document nor fOptions
        }
        //Must get element with out expanding if a LazyDOM.
        Element docElement =  DOMAccess.accessDocumentElement(fDocument);
        if (docElement == null) {
            throw new XMLIOError("Document has DocumentType, with out having a root element"); 
        }

        String tagName = docElement.getTagName();
        if (!fUsePreFormattedElements) { //we can't control tag names of preformatted elements, can we?
            if (fForceHTMLLowerCase) {
                tagName = tagName.toLowerCase(Locale.ENGLISH);
            } else if (fXHTMLFormattedAsHTML) {
                tagName = tagName.toUpperCase(Locale.ENGLISH);
            }
        }

        fOut.write("<!DOCTYPE ");
        fOut.write(tagName);
        
        if (fPublicId != null) {
            fOut.write(" PUBLIC \"");
            fOut.write(fPublicId);
            fOut.write("\"");
        }
        if (fSystemId != null) {
            if (fPublicId == null) {
                fOut.write(" SYSTEM");
            }
            fOut.write(" \"");
            fOut.write(fSystemId);
            fOut.write("\"");
        }

        fOut.write('>');
        writeln();
    }

    /**
     * Handler called for Attr nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleAttr
     */
    public void handleAttr(Attr attr) throws IOException {
        String name = attr.getName();
        if (!(fDropThisId && name.equals("id"))) {
            fOut.write(' ');
            fOut.write(name);
            if (printableAttrValue(attr)) {
                writeAttributeValue(attr);
            }
        }
    }

    /**
     * Write an element open tag.
     */
    protected final void writeOpenTag(Element element, 
                                      String tagName,
                                      boolean hasChildren) throws IOException {
        String formattedTag = null;
        if (fPrettyPrinting) {
            //Revisit - excluding this newline when formatting HTML
            //and preserve space is true because of differences, presumably, in
            //the NekoHTML parser where seems to store extra newlines between
            //elements.  Until a better solution is found, this will have to
            //suffice.
            if ((fXHTMLFormattedAsHTML || (!fXHTMLFormattedAsHTML && !fPreserveSpace)) && fNextSiblingText && fNoFormatNestCount == 0) {
                writeln();
            }
            fNextSiblingText = !isNotTextOrIsTextButAllWhitespace(DOMAccess.accessNextSibling(fDocument, element));
        }

        if (fUsePreFormattedElements && (element instanceof PreFormattedText)) {
            formattedTag = ((PreFormattedText)element).getPreFormattedText();
        }
        if (formattedTag != null) {
            fOut.write(formattedTag);
            fPreFormattedElementCount++;
        } else {
            if (fPrettyPrinting && isNotTextOrIsTextButAllWhitespace(element.getPreviousSibling())) {
                printIndent();
            }

            fDropThisId = fDropSpanIds && tagName.equalsIgnoreCase("SPAN");
            fOut.write('<');
            fOut.write(tagName);
            fTraverser.processAttributes(element);
            if (!hasChildren) {
                fHandledText = true; //spoof to avoid unnecessary indent later
            }
            fOut.write('>');
            fDynamicFormattedElementCount++;
            if (fPrettyPrinting && hasChildren && isNotTextOrIsTextButAllWhitespace(element.getFirstChild())) {
                writeln();
            }
        }
    }

    /**
     * Handler called for Element nodes.
     * 
     * <p>
     * This optionally corrects problem cases for browsers:
     * </p>
     * <ul>
     * <li>
     * ID attributes are dropped from SPAN tags.  This cause Internet Explorer
     * 4.0 to get confused on keep-alive connections.
     * </li>
     * </ul>
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleElement
     */
    public void handleElement(Element element) throws IOException {
        String tagName = element.getTagName();
        String tagNameCompare = tagName;
        if (!fUsePreFormattedElements) { //we can't control tag names of preformatted elements, can we?
            if (fForceHTMLLowerCase) {
                tagName = tagName.toLowerCase(Locale.ENGLISH);
            } else if (fXHTMLFormattedAsHTML) {
                tagName = tagName.toUpperCase(Locale.ENGLISH);
            }
        }
        boolean hasChildren = element.hasChildNodes();
        boolean isPreserveSpaceTag = false;
        boolean isScriptStyleTag = false;

        if (fXHTMLFormattedAsHTML) {
            tagNameCompare = tagNameCompare.toUpperCase(Locale.ENGLISH);
            if (HTMLElements.isPreserveSpaceTag(tagNameCompare)) {
                element = DOMAccess.getExpandedElement(fDocument, element);
                element.removeAttribute("xml:space");
                isPreserveSpaceTag = true;
            }
        }

        // Start Barracuda Kludge ======
        // check the element to see if it contains a
        // attribute "visdom".
        // (org.enhydra.barracuda.core.comp.BComponent.VISIBILITY_MARKER)
        // This controls DOM visibility. If this value exists and does not
        // match to true, don't print this particular node.
        // Note: This should be made generic, but for now...
        //TODO - consider getting rid of this, I don't think it's used much or at all in Barracuda
        Attr attr = DOMAccess.accessAttribute(fDocument, element, null, "visdom");
        if (attr != null && !(Boolean.valueOf(DOMAccess.accessAttributeValue(fDocument, attr)).booleanValue())) return;        
        // End Barracuda Kludge ======

        if (hasChildren) {
            if (!fXHTMLFormattedAsHTML || fNoFormatNestCount == 0) {
                isPreserveSpaceTag = isPreserveSpaceTag || HTMLElements.isPreserveSpaceTag(tagNameCompare);
                if (isPreserveSpaceTag) {
                    fNoFormatNestCount++;
                    isScriptStyleTag = HTMLElements.isScriptStyleTag(tagNameCompare);
                    if (isScriptStyleTag) fNoFormatScriptStyleNestCount++;
                }
            }
        }

        writeOpenTag(element, tagName, hasChildren);

        // Process children and close.
        boolean processedChildren = false;
        boolean processedLastChildElement = false;
        if (hasChildren) {
            fTraverser.processChildren(element);
            processedChildren = true;
            //below only matters when using pretty printing AND preserving space (nothing to do with xml:space).  If preserve space is false, avoid the hit of the nextSibling check (checking if the current element is a last child)
            if (fPrettyPrinting && fPreserveSpace && DOMAccess.accessNextSibling(fDocument, element) == null) {
                processedLastChildElement = true;
            }
        }

        if (hasChildren && isPreserveSpaceTag) {
            fNoFormatNestCount--;
            if (isScriptStyleTag) fNoFormatScriptStyleNestCount--;
        }

        if (fHandledText) {
            fHandledText = false;
        } else {
            printIndent();
        }

        if (processedChildren || HTMLElements.hasCloseTag(tagNameCompare)) {
            fOut.write("</");
            fOut.write(tagName);
            fOut.write('>');
        }

        if (fPrettyPrinting && (!fNextSiblingText || processedLastChildElement)) {
            writeln();
        }
    }

    /**
     * Handler called for ProcessingInstruction nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleProcessingInstruction
     */
    public void handleProcessingInstruction(ProcessingInstruction pi) throws IOException {
        if (!fXHTMLFormattedAsHTML) throw new XMLIOError("Unexpected call to handleProcessingInstruction");
    }

    /**
     * Handler called for CDATASection nodes.
     * <p>Non-standard extension: outputs data as-is.</p>
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleCDATASection
     */
    public void handleCDATASection(CDATASection cdata) throws IOException {
        fOut.write(cdata.getData());
    }

    /**
     * Handler called for Text nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleText
     */
    public final void handleText(Text text) throws IOException {
        if (fNoFormatScriptStyleNestCount > 0) {
            fOut.write(text.getData());
            fHandledText = true;
            return;
        }
        super.handleText(text);
    }

}
