/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InputSourceOps.java,v 1.4 2007-09-06 05:17:33 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

import org.xml.sax.InputSource;

/**
 * Various operations on InputSources.
 */
public final class InputSourceOps {
    /**
     * XML document header prefix.
     */
    private static String XML_DOCUMENT_HEADER_PREFIX = "<?xml";

    /**
     * Disallow instantiation.
     */
    private InputSourceOps() {
    }

    /**
     * Open a bytestream given a system id.
     */
   public static InputStream openSystemId(String systemId) throws IOException {
        if (systemId == null) {
            throw new IllegalArgumentException("no open stream or system id specified in InputSource");
        }

        try {
            URL url = new URL(systemId);
            URLConnection conn = url.openConnection();
            conn.setUseCaches(false); //avoid file locking on Windows
            return conn.getInputStream();
        } catch (MalformedURLException mue) {
            return new FileInputStream(systemId);
        }
    }

    /**
     * Open a character input stream for an input source if one is not
     * already open.
     *
     * @param inputSource Specification of the document to open.
     * @return A pointer to the character stream.
     */
    public static Reader open(InputSource inputSource) throws IOException {
        if (inputSource.getCharacterStream() != null) {
            return inputSource.getCharacterStream();
        }
        
        InputStream in = inputSource.getByteStream();
        if (in == null) {
            in = openSystemId(inputSource.getSystemId());
        }
        String encoding = inputSource.getEncoding();
        if (encoding != null) {
            return new BufferedReader(new InputStreamReader(in, encoding));
        } else {
            return new BufferedReader(new InputStreamReader(in));
        }
    }

    /**
     * Determine if an input source has an open byte or character stream.
     */
    public static boolean isOpen(InputSource input) {
        return (input.getCharacterStream() != null)
            || (input.getByteStream() != null);
    }

    /**
     * Close an InputSource, if open.
     */
    public static void close(InputSource input) throws IOException {
        if (input.getCharacterStream() != null) {
            input.getCharacterStream().close();
        }
        if (input.getByteStream() != null) {
            input.getByteStream().close();
        }
    }

    /**
     * Close a byte stream returned by <code>open()</code>, only if it was
     * actually opened by open.
     *
     * @param inputSource Specification of the document that was opened.
     * @param reader The character stream returned by open.
     */
    public static void closeIfOpened(InputSource inputSource,
                                     Reader reader) throws IOException {
        if ((inputSource.getCharacterStream() == null) 
            && (inputSource.getByteStream() == null)) {
            reader.close();
        }
    }

    /**
     * Get a description of an input source.
     */
    public static String getName(InputSource input) {
        if (input == null) {
            return "<null input source>";
        }

        String name = null;
        if (input.getSystemId() != null) {
            name = input.getSystemId();
        } else if (input.getPublicId() != null) {
            name = input.getPublicId();
        }

        // If byte or character stream is open, indicate that with the name.
        if (input.getByteStream() != null) {
            if (name == null) {
                return "<unnamed byte stream>";
            } else {
                return name +"<byte stream>";
            }
        }
        if (input.getCharacterStream() != null) {
            if (name == null) {
                return "<unnamed character stream>";
            } else {
                return name +"<character stream>";
            }
        }
        if (name != null) {
            return name;
        } else {
            return "<empty input source>";
        }
    }

    /**
     * Determine an document reader is attached to an XML document
     * by checking the first few bytes.  Reader must be positioned
     * at the start.
     * 
     * @param reader Character stream.
     */
    public static boolean isXMLDocument(Reader reader) throws IOException {
        char[] buf = new char[XML_DOCUMENT_HEADER_PREFIX.length()];
        //ensure mark() is supported by using a BufferedReader if need be
        Reader br = reader.markSupported() ? reader : new BufferedReader(reader);
        br.mark(XML_DOCUMENT_HEADER_PREFIX.length()+1);
        if (br.read(buf) != buf.length) {
            return false;
        }
        return String.copyValueOf(buf).equals(XML_DOCUMENT_HEADER_PREFIX);
    }

    /**
     * Determine an input source points to an XML document.
     * 
     * @param inputSource Specification of the document to examine.
     */
    public static boolean isXMLDocument(InputSource inputSource) throws IOException {
        Reader reader = open(inputSource);
        try {
            return isXMLDocument(reader);
        } finally {
            closeIfOpened(inputSource, reader);
        }
    }
}
