/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLEntityResolver.java,v 1.7 2007-09-06 05:17:33 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Vector;

import org.apache.xml.resolver.Catalog;
import org.apache.xml.resolver.CatalogException;
import org.apache.xml.resolver.CatalogManager;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * A flexable entity resolver for use with parsing XML files.
 * This implements:
 * <UL>
 * <LI> XCatalog support.
 * <LI> Resolving entities from class loaders.
 * <LI> Resolving entities from the file system
 * <LI> Resolving standard URL entities.
 * <LI> A debug output stream, useful for debugging resolution problems.
 * </UL>
 * <P>
 * A list of class loaders maybe associated with this object.  If so
 * an attempt will be made to resolve <code>file:</code> URLs from
 * the class path.
 * 
 * @see org.apache.xerces.readers.XCatalog
 */
public class XMLEntityResolver implements EntityResolver {

    //FIXME - remove commented methods and make sure all previous catalog functionality continues to work

    /**
     * Protocol to use in specifying a URL pointing to the CLASSPATH or a file
     * on the file system.
     */
    public static final String CLASSPATH_PROTOCOL = "file";

    /**
     * Protocol prefix.  In the case of loading from the classpath, this
     * protocol prefix and any leading slash is stripped to provide a path
     * expected by <code>classloader.getResourceAsStream(String)</code>.
     */
    public static final String CLASSPATH_PROTOCOL_PREFIX
        = CLASSPATH_PROTOCOL + ":";

    /**
     * URL for Enhydra default entity catalog.
     * @see #setDefaultResolving
     */
    public static final String DEFAULT_ENTITY_CATALOG
        = CLASSPATH_PROTOCOL_PREFIX + "/org/enhydra/xml/catalog/default.xcat";

    /**
     * URL for Lutrus Enhydra default entity catalog.  Only loaded if
     * available.
     * @see #setDefaultResolving
     */
    public static final String LUTRIS_ENTITY_CATALOG
        = CLASSPATH_PROTOCOL_PREFIX + "/org/enhydra/xml/catalog/lutris.xcat";

    /**
     * If not null, write debug information to this writer.
     */
    private PrintWriter fDebugWriter = null;

    /**
     * XCatalog catalog and entity resolver..
     */
    private Catalog fXCatalog;

    /**
     * List of ClassLoaders use to resolve entities.
     */
    private Vector fClassLoaders = new Vector();

    public XMLEntityResolver() {        
        CatalogManager manager = new CatalogManager();
        manager.setAllowOasisXMLCatalogPI(false);
        manager.setCatalogClassName("org.apache.xml.resolver.Catalog");
        manager.setCatalogFiles("");
        manager.setIgnoreMissingProperties(true);
        manager.setPreferPublic(true);
        manager.setRelativeCatalogs(false);
        manager.setUseStaticCatalog(false);
        manager.setVerbosity(0);

        fXCatalog = new Catalog(manager);
        fXCatalog.setupReaders();
    }
    
    /**
     * Set a writer used to output debug messages.
     */
    public void setDebugWriter(PrintWriter writer) {
        fDebugWriter = writer;
    }
    
    /**
     * Get the debug writer.
     */
    public PrintWriter getDebugWriter() {
        return fDebugWriter;
    }

    /**
     * Loads a catalog specified by the given input source and
     * appends the contents to the catalog object.
     *
     * @param source The catalog source.
     *
     * @exception org.xml.sax.SAXException Throws exception on SAX error.
     * @exception java.io.IOException Throws exception on i/o error.
     */
    public void loadCatalog(InputSource source)
        throws SAXException, IOException {
        if (fDebugWriter != null) {
            fDebugWriter.println("loadCatalog(" + InputSourceOps.getName(source) + ")");
        }
        // Check for catalog on the CLASSPATH
        InputSource loadSource = searchClassPathOrFilePath(source);
        if (loadSource == null) {
            loadSource = source;
        }
        if (fDebugWriter != null) {
            fDebugWriter.println("loading catalog from(" + InputSourceOps.getName(loadSource) + ")");
        }
        //System.out.println("loading catalog from(" + InputSourceOps.getName(loadSource) + ")");
        try {
            fXCatalog.parseCatalog("application/xml", loadSource.getByteStream());
        } catch (CatalogException ce) {
            System.out.println("Can't load catalog" + ce);
        }
    }

    /**
     * Load the default entity catalog and enable loading entities from the
     * classloader for the classloader that loaded this class.
     */
    public void setDefaultResolving() throws SAXException, IOException {
        // JDeveloper JVM doesn't correctly return system class loader
        // if it loaded this class, so we do it explictly.
        ClassLoader classLoader = XMLEntityResolver.class.getClassLoader();
        if (classLoader == null) {
            classLoader = ClassLoader.getSystemClassLoader();
        }
        addClassLoader(classLoader);
        loadCatalog(new ClosingInputSource(DEFAULT_ENTITY_CATALOG));

        // Check for optional default catalog
        try {
            InputSource optSource = searchClassPathOrFilePath(LUTRIS_ENTITY_CATALOG);      
            loadCatalog(optSource);
        } catch (IOException ioe) {
            ; //FIXME - should this old Lutris specific functionality be removed?
        }
    }

    /**
     * Adds a public to system identifier mapping.
     *
     * @param publicId The public identifier, or "key".
     * @param systemId The system identifier, or "value".
     */
    //public void addPublicMapping(String publicId, String systemId) {
        //fXCatalog.addEntry(new CatalogEntry())addPublicMapping(publicId, systemId);
    //}

    /**
     * Removes a public identifier mapping.
     *
     * @param publicId The public identifier to remove.
     */
    //public void removePublicMapping(System publicId) {
        //fXCatalog.removePublicMapping(publicId);
    //}

    /** Returns an enumeration of public identifier mapping keys. */
    //public Enumeration getPublicMappingKeys() {
        //return fXCatalog.getPublicMappingKeys();
        //return Collections.enumeration(Collections.EMPTY_LIST);
    //}

    /**
     * Returns a public identifier mapping.
     *
     * @param publicId The public identifier, or "key".
     *
     * @return Returns the system identifier value or null if there
     *         is no mapping defined.
     */
    public String getPublicMapping(String publicId, String systemId) {
        try {
            return fXCatalog.resolvePublic(publicId, systemId);
        } catch (Exception e) {
            return null;
        }
    }

    /**
     * Adds a system identifier alias.
     *
     * @param publicId The system identifier "key".
     * @param systemId The system identifier "value".
     */
    //public void addSystemMapping(String systemId1, String systemId2) {
        //fXCatalog.addSystemMapping(systemId1, systemId2);
    //}

    /**
     * Removes a system identifier alias.
     *
     * @param systemId The system identifier to remove.
     */
    //public void removeSystemMapping(String systemId) {
        //fXCatalog.removeSystemMapping(systemId);
    //}

    /** Returns an enumeration of system identifier mapping keys. */
    //public Enumeration getSystemMappingKeys() {
        //return fXCatalog.getSystemMappingKeys();
        //return Collections.enumeration(Collections.EMPTY_LIST);
    //}

    /**
     * Returns a system identifier alias.
     *
     * @param systemId The system identifier "key".
     *
     * @return Returns the system identifier alias value or null if there
     *         is no alias defined.
     */
    public String getSystemMapping(String systemId) {
        try {
            return fXCatalog.resolveSystem(systemId);
        } catch (Exception e) {
            return null;
        }
    }

    /**
     * Add a classloader to the end of the order list of classloaders to
     * search for entities.  Attempts to set a null classloader value are ignored.
     */
    public void addClassLoader(ClassLoader classLoader) {
        if (classLoader != null) fClassLoaders.add(classLoader);
    }

    /**
     * Get an enumeration of the class loaders.
     */
    public Enumeration getClassLoaders() {
        return fClassLoaders.elements();
    }

    /**
     * Attempt to load a systemId from a classloader.
     */
    private InputSource loadFromClassLoader(ClassLoader loader,
                                            String resourceId) {
        // Drop leading `/' from path.
        if (resourceId.startsWith("/")) {
            resourceId = resourceId.substring(1);
        }
        InputStream in = loader.getResourceAsStream(resourceId);
        if (in != null) {
            return new ClosingInputSource(in);
        } else {
            return null;
        }
    }

    /**
     * Attempt to load a systemId from the file system
     */
    private InputSource loadFromFileURL(String systemId) {
        try {
            return new ClosingInputSource(InputSourceOps.openSystemId(systemId));
        } catch (IOException ioe) {
            return null;
        }
    }

    /**
     * Determine if a a system id is a CLASSPATH or File path URL.
     */
    private boolean isClassPathOrFilePathURL(String systemId) {
        return ((systemId != null) && systemId.startsWith(CLASSPATH_PROTOCOL_PREFIX));
    }

    /**
     * Get the file name from a CLASSPATH URL.
     */
    private String parseClassPathURL(String url) {
        return url.substring(CLASSPATH_PROTOCOL_PREFIX.length());
    }

    /**
     * Attempt to load a systemId from the classpath or file path (in that order).
     * Will return an InputSource with a contained byte stream
     * (doesn't need to be closed).  If systemId is null or doesn't have the
     * "file:" protocal, then null is returned.
     * 
     * @throws IOException if the systemId does start with the "file:" protocol,
     *                     but the file cannot be found on either the classpath
     *                     or the file system.
     */
    private InputSource searchClassPathOrFilePath(String systemId)
            throws IOException {
        InputSource source = null;
        if ((systemId != null) && isClassPathOrFilePathURL(systemId)) {
            String resourceName =  parseClassPathURL(systemId);
            int len = fClassLoaders.size();
            for (int idx = 0; (idx < len) && (source == null); idx++) {
                source = loadFromClassLoader((ClassLoader)fClassLoaders.get(idx),
                                             resourceName);
            }

            if (source == null) {
                try {
                    source = loadFromFileURL(systemId);
                } catch (java.security.AccessControlException ace) {
                    //Don't allow file system lookup under a strict security manager,
                    //such as Google AppEngine, to bomb the application.
                    if (fDebugWriter != null) {
                        fDebugWriter.println("Warning: " + ace.getLocalizedMessage());
                    }
                }
            }

            if (fDebugWriter != null) {
                fDebugWriter.println("Entity " + ((source == null) ? "not " : "")
                                     + "found searching " + fClassLoaders.size()
                                     + " class loader(s) or file system: " + systemId
                                     + " => " + InputSourceOps.getName(source));
            }

            if (source == null) {
                throw new IOException("Entity not found on classpath or file system: " + systemId);
            }

            source.setSystemId(systemId);
        } else {
            if (fDebugWriter != null) {
                fDebugWriter.println("No attempt made to search for Entity because systemId does not begin with expected protocol prefix '" + CLASSPATH_PROTOCOL_PREFIX + "': " + systemId);
            }
        }

        return source;
    }


    /**
     * @see #searchClassPathOrFilePath(String)
     */
    private InputSource searchClassPathOrFilePath(InputSource source)
            throws IOException {
        if ((source.getByteStream() != null)
            || (source.getCharacterStream() != null)) {
            return null; // Already open.
        } else {
            return searchClassPathOrFilePath(source.getSystemId());
        }
    }


    /**
     * @see org.xml.sax.EntityResolver
     */
    public InputSource resolveEntity(String publicId,
                                     String systemId)
    throws SAXException, IOException {
        if (fDebugWriter != null) {
            fDebugWriter.println("resolveEntity(" + publicId + ", " + systemId + ")");
        }

        InputSource source;
        //InputSource catalogSource = new CatalogResolver(fXCatalog.getCatalogManager()).resolveEntity(publicId, systemId);
        InputSource catalogSource = null;
        String sourceURI = fXCatalog.resolveEntity(null, publicId, systemId);
        //System.out.println("publicId: " + publicId);
        //System.out.println("systemId: " + systemId);
        //System.out.println("sourceURI: " + sourceURI);
        if (sourceURI != null) {
            catalogSource = new ClosingInputSource(sourceURI);
            catalogSource.setPublicId(publicId);
        }
        
        if (catalogSource != null) {
            if (fDebugWriter != null) {
                fDebugWriter.println("XCatalog maps entity to: " + catalogSource.getSystemId());
            }
            source = searchClassPathOrFilePath(catalogSource);
            if (source == null) {
                source = catalogSource;  // Use unchanged.
            }
        } else {
            source = searchClassPathOrFilePath(systemId);
            if (source != null) source.setPublicId(publicId);
        }
        if (fDebugWriter != null) {
            fDebugWriter.println("Resolved: " + InputSourceOps.getName(source));
        }

        return source;
    }

}
