/*
 * Copyright (C) 2003, 2005, 2006, 2007, 2009, 2010 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.io;

import java.io.IOException;
import java.util.Arrays;
import java.util.Locale;

import org.enhydra.xml.dom.DOMAccess;
import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

/*
 * FIXME: 
 *  - Maybe create a separate XHTMLFormatter class
 *  - Add option to encode entity references from the DTD.
 *  - namespaces
 *  - writespace
 *  - Need to handle entities for 8bit encodings
 */

/**
 * Formatter for outputting a HTML DOM as a HTML text document.
 */
final class XMLFormatter extends BaseDOMFormatter implements Formatter {
    /**
     * Default XML encoding.
     */
    private static final String DEFAULT_XML_ENCODING = "UTF-8";

    /**
     * XML version (only 1.0 supported, as thats all there is now).
     */
    private static final String XML_VERSION = "1.0";

    /**
     * Table use to optimized checking for characters that should be
     * represented as entity references.
     */
    private static final boolean[] ENTITY_QUICK_CHECK
        = new boolean[MAX_ENTITY_QUICK_CHECK_CHAR+1];

    /**
     * Array of XHTML tags that have an empty content model.  Used to avoid the
     * minimized form for tags without children, but not having an empty content
     * model.  In alphabetical order so a proper binary search can be performed.
     * See {@link http://www.w3.org/TR/xhtml1/#C_2} and
     * {@link http://www.w3.org/TR/xhtml1/#C_3}.
     */
    private static final String[] XHTML_EMPTY_TAGS = {"area", "base", "br", "col", "hr", "img", "input", "link", "meta", "param"};

    /**
     * Array of XHTML tags that get special treatment, as if the contents are
     * part of a CDATASection.  No resolution of special characters or entities
     * should be performed on the content of these tags.
     */
    private static final String[] XHTML_SCRIPT_STYLE_TAGS = {"script", "style"};

    /**
     * Only used when formatting HTML as XHTML.  Array of lower-case tag names
     * which, in XHTML, would have xml:space="preserve" applied to them by
     * default.  For an HTML parse tree, xml:space="preserve" needs to be added
     * manually.
     */
    private static final String[] HTML_PRESERVE_SPACE_TAGS = {"pre", "script", "style"};

    /**
     * Indicates that the DOM being processed is not just XML, but XHTML, so
     * special rules can be applied that wouldn't apply to basic XML
     */
    private final boolean fXHTMLFormattingMode;

    /**
     * Indicates that the DOM is the HTML DOM so special rules are to be applied
     * to output as XHTML, eg.. lowercasing element names and replacing the
     * provided document type with the XHTML document type
     */
    private final boolean fHTMLFormattedAsXHTML;

    /**
     * Nesting count for script and style elements.  This goes one step beyond
     * fNoFormatNestCount, in that it allows us to avoid resolving special
     * characters or entities, which may exist inside script or style elements,
     * but be meant literally.  Essentially, it's a CDATA Section without being
     * marked as such officially.
     */
    private int fNoFormatScriptStyleNestCount;

    /**
     * Indicates that the next sibling is a Text node
     */
    private boolean fNextSiblingText;

    /**
     * Static constructor.
     */
    static {
        ENTITY_QUICK_CHECK['<'] = true;
        ENTITY_QUICK_CHECK['>'] = true;
        ENTITY_QUICK_CHECK['"'] = true;
        ENTITY_QUICK_CHECK['\''] = true;
        ENTITY_QUICK_CHECK['&'] = true;
    }

    /**
     * Calls 4 argument constructor with forXHTMLFormatting as false
     */
    public XMLFormatter(Node node,
                        OutputOptions outputOptions,
                        boolean forPreFormatting) {
        this(node, outputOptions, forPreFormatting, false);
    }

    /**
     * Constructor that provides for declaring if this is an XHTML document
     */
    public XMLFormatter(Node node,
                        OutputOptions outputOptions,
                        boolean forPreFormatting,
                        boolean forXHTMLFormatting) {
        super(node, outputOptions, forPreFormatting, DEFAULT_XML_ENCODING, ENTITY_QUICK_CHECK);
        fXHTMLFormattingMode = forXHTMLFormatting;
        boolean htmlFormattedAsXHTML = false;
        if (fXHTMLFormattingMode) {
            //Must get element with out expanding if a LazyDOM.
            Element docElement =  DOMAccess.accessDocumentElement(fDocument);
            if (docElement != null) {
                String rootTag = docElement.getTagName();
                htmlFormattedAsXHTML = Character.isUpperCase(rootTag.charAt(0));
                if (htmlFormattedAsXHTML) {
                    //root element is always expanded, so no need to do it explicitly as is done via DOMAccess.getExpandedElement()
                    docElement.setAttribute("xmlns", "http://www.w3.org/1999/xhtml");
                    String lang = docElement.getAttribute("lang");
                    if (lang != null) {
                        //see http://www.w3.org/TR/xhtml1/#C_7
                        docElement.setAttribute("xml:lang", lang);
                    }
                    if (fOptions.getPublicId() == null && fOptions.getSystemId() == null) {
                        //Re-map DTD only when not overridden by fOptions and
                        //processing an HTML parsed document (i.e., DOM tag names
                        //upper-case) with one of the 3 standard HTML DocTypes
                        if ((fPublicId != null && fPublicId.indexOf("Frameset") != -1) || (fSystemId != null && fSystemId.indexOf("frameset.dtd") != -1)) {
                            fPublicId = "-//W3C//DTD XHTML 1.0 Frameset//EN";
                            fSystemId = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd";
                        } else if ((fPublicId != null && fPublicId.indexOf("Transitional") != -1) || (fSystemId != null && fSystemId.indexOf("loose.dtd") != -1)) {
                            fPublicId = "-//W3C//DTD XHTML 1.0 Transitional//EN";
                            fSystemId = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd";
                        } else if ((fPublicId != null && fPublicId.indexOf("Strict") != -1) || (fSystemId != null && fSystemId.indexOf("strict.dtd") != -1)) {
                            fPublicId = "-//W3C//DTD XHTML 1.0 Strict//EN";
                            fSystemId = "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd";
                        }
                    }
                }
            }
        }
        fHTMLFormattedAsXHTML = htmlFormattedAsXHTML;
    }

    /**
     * Get the default OutputOptions for a document formatter with this
     * formatter.  The encoding will not be set, which signals to use the
     * default encoding.
     */
    static OutputOptions getDefaultOutputOptions() {
        return new OutputOptions();  // Nothing special
    }

    /**
     * @see BaseDOMFormatter#getCharacterEntity
     */
    protected final String getCharacterEntity(char textChar) {
        if (fXHTMLFormattingMode) {
            return HTMLEntities.charToEntity4(textChar);
        }
        switch (textChar) {
            case '<':  return "lt";
            case '>':  return "gt";
            case '"':  return "quot";
            case '\'': return "apos";
            case '&':  return "amp";
            default:   return null;
        }        
    }

    /**
     * Write the XML header.
     */
    private void writeXMLHeader(Document document) throws IOException {
        fOut.write("<?xml version=\"");
        fOut.write(XML_VERSION);
        fOut.write("\"");

        // Encoding
        if (!fOptions.getOmitEncoding()) {
            fOut.write(" encoding=\"");
            fOut.write(getMIMEEncoding());
            fOut.write('"');
        }

        // Standalone - can only determine this if document
        // supports the method. If not, default to "no"
        if (document.getXmlStandalone()) {
            fOut.write(" standalone=\"yes\"");
            // Ignore else case - we assume 'standalone="no"' in this case...
        }

        fOut.write("?>");
        writeln();
    }

    /**
     * Handler called for Document nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocument
     */
    public void handleDocument(Document document) throws IOException {
        //see http://www.w3.org/TR/xhtml1/#C_1
        if (!(fXHTMLFormattingMode && fOptions.getEnableXHTMLCompatibility()) && !fOptions.getOmitXMLHeader()) {
            writeXMLHeader(document);
        }
        if (fDocType == null) {
            handleDocumentType(null);
        } else {
            fTraverser.processDocumentType(document);
        }
        fTraverser.processChildren(document);
    }

    /**
     * Handler called for Document nodes; writes the DOCTYPE specification.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocumentType
     */
    public void handleDocumentType(DocumentType documentType) throws IOException {
        if (fOptions.getOmitDocType()) {
            return; // Don't output DOCTYPE.
        }

        String internalSubset = documentType != null ? documentType.getInternalSubset() : null;
        if (documentType == null && fPublicId == null && fSystemId == null && internalSubset == null && fOptions.getPublicId() == null && fOptions.getSystemId() == null) {
            return; // No DOCTYPE - neither provided by document nor fOptions
        }
        //Must get element with out expanding if a LazyDOM.
        Element docElement =  DOMAccess.accessDocumentElement(fDocument);
        if (docElement == null) {
            throw new XMLIOError("Document has DocumentType, with out having a root element"); 
        }

        String tagName = docElement.getTagName();
        if (!fUsePreFormattedElements && fHTMLFormattedAsXHTML) { //we can't control tag names of preformatted elements, can we?
            tagName = tagName.toLowerCase(Locale.ENGLISH);
        }

        fOut.write("<!DOCTYPE ");
        fOut.write(tagName);
        
        if (fPublicId != null) {
            // Public requires system.
            if (fSystemId == null) {
                throw new XMLIOError("No SYSTEM id to accompany PUBLIC id: "
                                     + fPublicId);
            }
            fOut.write(" PUBLIC \"");
            fOut.write(fPublicId);
            fOut.write("\"");
        }
        if (fSystemId != null) {
            if (fPublicId == null) {
                fOut.write(" SYSTEM");
            }
            fOut.write(" \"");
            fOut.write(fSystemId);
            fOut.write("\"");
        }

        if (internalSubset != null) {
            writeln();
            fOut.write(" [");
            fOut.write(internalSubset);
            fOut.write("]");
        }
        fOut.write('>');
        writeln();
    }

    /**
     * Handler called for Attr nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleAttr
     */
    public void handleAttr(Attr attr) throws IOException {
        fOut.write(' ');
        fOut.write(attr.getName());
        writeAttributeValue(attr);
    }

    /**
     * Write an element open tag.
     */
    protected final void writeOpenTag(Element element, 
                                      String tagName,
                                      boolean hasChildren) throws IOException {
        String formattedTag = null;
        if (fPrettyPrinting) {
            //Revisit - excluding this newline when formatting HTML
            //and preserve space is true because of differences, presumably, in
            //the NekoHTML parser where seems to store extra newlines between
            //elements.  Until a better solution is found, this will have to
            //suffice.
            if ((!fHTMLFormattedAsXHTML || (fHTMLFormattedAsXHTML && !fPreserveSpace)) && fNextSiblingText && fNoFormatNestCount == 0) {
                writeln();
            }
            fNextSiblingText = !isNotTextOrIsTextButAllWhitespace(DOMAccess.accessNextSibling(fDocument, element));
        }

        if (fUsePreFormattedElements && (element instanceof PreFormattedText)) {
            formattedTag = ((PreFormattedText)element).getPreFormattedText();
        }
        if (formattedTag != null) {
            fOut.write(formattedTag);
            fPreFormattedElementCount++;
        } else {
            if (fPrettyPrinting && isNotTextOrIsTextButAllWhitespace(element.getPreviousSibling())) {
                printIndent();
            }

            fOut.write('<');
            fOut.write(tagName);
            fTraverser.processAttributes(element);
            if (!hasChildren && !(fXHTMLFormattingMode && fOptions.getEnableXHTMLCompatibility() && !isXHTMLContentModelEmpty(tagName))) {
                //see http://www.w3.org/TR/xhtml1/#C_2
                if (fOptions.getEnableXHTMLCompatibility()) fOut.write(" /");
                else fOut.write('/');
                fHandledText = true; //spoof to avoid unnecessary indent later
            }
            fOut.write('>');
            fDynamicFormattedElementCount++;
            if (fPrettyPrinting && hasChildren && isNotTextOrIsTextButAllWhitespace(element.getFirstChild())) {
                writeln();
            }
        }
    }

    /**
     * Handler called for Element nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleElement
     */
    public void handleElement(Element element) throws IOException {
        String tagName = element.getTagName();
        String tagNameCompare = tagName;
        boolean hasChildren = element.hasChildNodes();
        boolean isPreserveSpaceTag = false;
        boolean isScriptStyleTag = false;

        if (fHTMLFormattedAsXHTML) {
            if (!fUsePreFormattedElements) { //we can't control tag names of preformatted elements, can we?
                tagName = tagName.toLowerCase(Locale.ENGLISH);
            }
            tagNameCompare = tagNameCompare.toLowerCase(Locale.ENGLISH);
            if (hasChildren && isHTMLPreserveSpaceTag(tagNameCompare)) {
                //this is done automatically when parsed as XHTML, so duplicate this behavior manually when parsed as HTML
                element = DOMAccess.getExpandedElement(fDocument, element);
                element.setAttribute("xml:space", "preserve");
                isPreserveSpaceTag = true;
            }
        }

        // Start Barracuda Kludge ======
        // check the element to see if it contains a
        // attribute "visdom".
        // (org.enhydra.barracuda.core.comp.BComponent.VISIBILITY_MARKER)
        // This controls DOM visibility. If this value exists and does not
        // match to true, don't print this particular node.
        // Note: This should be made generic, but for now...
        //TODO - consider getting rid of this, I don't think it's used much or at all in Barracuda
        Attr attr = DOMAccess.accessAttribute(fDocument, element, null, "visdom");
        if (attr != null && !(Boolean.valueOf(DOMAccess.accessAttributeValue(fDocument, attr)).booleanValue())) return;        
        // End Barracuda Kludge ======

        if (hasChildren) {
            if (fHTMLFormattedAsXHTML || fNoFormatNestCount == 0) {
                isPreserveSpaceTag = isPreserveSpaceTag || isXMLSpacePreserveElement(fDocument, element);
                if (isPreserveSpaceTag) {
                    fNoFormatNestCount++;
                    if (fXHTMLFormattingMode) {
                        isScriptStyleTag = isXHTMLScriptStyleTag(tagNameCompare);
                        if (isScriptStyleTag) fNoFormatScriptStyleNestCount++;
                    }
                }
            }
        }

        //pass tagNameCompare because it's guaranteed to be lower-case and will
        //be ignored if preformatted text is used in writeOpenTag()
        writeOpenTag(element, tagNameCompare, hasChildren);

        // Process children and close.
        boolean processedChildren = false;
        boolean processedLastChildElement = false;
        if (hasChildren || (fXHTMLFormattingMode && fOptions.getEnableXHTMLCompatibility() && !isXHTMLContentModelEmpty(tagNameCompare))) {
            fTraverser.processChildren(element);
            processedChildren = true;
            //below only matters when using pretty printing AND preserving space (nothing to do with xml:space).  If preserve space is false, avoid the hit of the nextSibling check (checking if the current element is a last child)
            if (fPrettyPrinting && fPreserveSpace && DOMAccess.accessNextSibling(fDocument, element) == null) {
                processedLastChildElement = true;
            }
        }

        if (hasChildren && isPreserveSpaceTag) {
            fNoFormatNestCount--;
            if (isScriptStyleTag) fNoFormatScriptStyleNestCount--;
        }

        if (fHandledText) {
            fHandledText = false;
        } else {
            printIndent();
        }

        if (processedChildren) {
            fOut.write("</");
            fOut.write(tagName);
            fOut.write('>');
        }

        if (fPrettyPrinting && (!fNextSiblingText || processedLastChildElement)) {
            writeln();
        }
    }

    /**
     * Handler called for ProcessingInstruction nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleProcessingInstruction
     */
    public void handleProcessingInstruction(ProcessingInstruction pi) throws IOException {
        fOut.write("<?");
        fOut.write(pi.getTarget());
        String data = pi.getData(); //FIXME - shouldn't this check for preformatted text?  But it doesn't seem to implement PreformattedText
        if (data != null) {
            fOut.write(' ');
            fOut.write(data);
        }
        fOut.write("?>");
    }

    /**
     * Handler called for CDATASection nodes.
     * <p>Non-standard extension: outputs data as-is if formatting HTML as XHTML.</p>
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleCDATASection
     */
    public void handleCDATASection(CDATASection cdata) throws IOException {
        //FIXME: Should we handle embedded ]]>???
        if (!fOptions.getEnableXMLCDATAHack() || fNoFormatScriptStyleNestCount > 0) fOut.write("<![CDATA[");
        fOut.write(cdata.getData()); //FIXME - shouldn't this check for preformatted text?  But it doesn't seem to implement PreformattedText
        if (!fOptions.getEnableXMLCDATAHack() || fNoFormatScriptStyleNestCount > 0) fOut.write("]]>");
    }

    /**
     * Handler called for Text nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleText
     */
    public final void handleText(Text text) throws IOException {
        if (fNoFormatScriptStyleNestCount > 0) {
            fOut.write(text.getData());
            fHandledText = true;
            return;
        }
        super.handleText(text);
    }

    /**
     * Utility method to tell us which XHTML elements contain an empty
     * content model allowing us to, in the converse, encapsulate the 
     * logic needed to deal with XHTML tags that, if backward compatibility
     * is requested, should be written with an explicit close tag rather than
     * use the default minimized syntax when the element has no children.
     * 
     * <p>
     * See {@link http://www.w3.org/TR/xhtml1/#C_2} and
     * {@link http://www.w3.org/TR/xhtml1/#C_3}
     * </p>
     * 
     * @param tagName expected to be a tag name with no prefixes and all
     *                lower-case
     * @return true if the content model for the provided tagName is empty,
     *         false if not
     * @version 2.2
     */
    private static boolean isXHTMLContentModelEmpty(String tagName) {
        return Arrays.binarySearch(XHTML_EMPTY_TAGS, tagName) >= 0;
    }

    private static boolean isXHTMLScriptStyleTag(String tagName) {
        return Arrays.binarySearch(XHTML_SCRIPT_STYLE_TAGS, tagName) >= 0;
    }

    private static boolean isHTMLPreserveSpaceTag(String tagName) {
        return Arrays.binarySearch(HTML_PRESERVE_SPACE_TAGS, tagName) >= 0;
    }

    /**
     * @param document not technically necessary since DOMAccess doesn't do
     *                 anything with it, but it's part of the interface.
     *                 Something to figure out later.
     * @param element the element in question
     * @return true if the element has an "xml:space" attribute with a value of
     *         "preserve", otherwise false
     */
    private static boolean isXMLSpacePreserveElement(Document document, Element element) {
        Attr xmlSpaceAttr = DOMAccess.accessAttribute(document, element, null, "xml:space");
        return (xmlSpaceAttr != null && DOMAccess.accessAttributeValue(document, xmlSpaceAttr).equals("preserve"));
    }

}
