/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLCStdFactory.java,v 1.1.1.1 2003-03-10 16:36:19 taweili Exp $
 */

package org.enhydra.xml.xmlc;

/**
 * XMLC standard factory class. Used for creating instances of
 * XMLC-generated classes.  This is a default factory used when
 * no special support, such as auto-recompilation is required.
 */
public class XMLCStdFactory implements XMLCFactory {
    /**
     * Class loader that is used to load the first reference to a class.
     */
    private ClassLoader fDefaultClassLoader;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Constructor.
     * @param classLoader Classloader used to load classes when a class name
     *  is specified.  If null, the system classload is used.
     * @param logger XMLC logger or null for no logging.
     */
    public XMLCStdFactory(ClassLoader classLoader,
                          XMLCLogger logger) {
        fDefaultClassLoader = classLoader;
        if (logger != null) {
            fLogger = logger;
        } else {
            fLogger = new StreamXMLCLogger();  // all will be disabled.
        }
    }

    /**
     * Handle all errors, logging if enabled.  Also handles logging
     * error output of the compiler.
     * 
     * @param className Class being loaded.
     * @param except Can be any error or exception.
     * @return Returns runtime exceptions to be rethrown.  Could throw
     * everything, but catch blocks appear to fall through to the compiler so
     * we would have to put in dummy returns.
     */
    protected Error handleError(String className,
                              Throwable except) {
// ole        fLogger.logError("Factory creation of an XMLC class failed: \""
//                         + className + "\"", except);
        
        // Errors just continue on.
        if (except instanceof Error) {
            return (Error)except;
        }
        
        // All other exceptions get wrapped.
        return new XMLCError("Couldn't load XMLC class", except);
    }

    /**
     * Do actualy work of creating a new object.
     */
    protected XMLObject doCreate(Class xmlcBasedClass)
            throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        if (xmlcBasedClass.isInterface()) {
            // Convert interface to class name.
            return doCreate(xmlcBasedClass.getName() + "Impl",
                            fDefaultClassLoader);
        } else {
            if (!XMLObject.class.isAssignableFrom(xmlcBasedClass)) {
                throw new XMLCError("class \"" + xmlcBasedClass.getName()
                                    + "\" does not implement \""
                                    + XMLObject.class.getName() + "\"");
            }
            return (XMLObject)(xmlcBasedClass.newInstance());
        }
    }

    /**
     * Do actualy work of creating a new object given a name.
     */
    protected XMLObject doCreate(String xmlcClassName,
                               ClassLoader classLoader)
            throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        if (classLoader != null) {
            return doCreate(classLoader.loadClass(xmlcClassName));
        } else {
            return doCreate(Class.forName(xmlcClassName));
        }
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived from
     * one, given either a class or an interface object, using the associated
     * classloader
     */
    public XMLObject create(String xmlcClassName) {
        try {
            return doCreate(xmlcClassName, fDefaultClassLoader);
        } catch (Throwable except) {
            throw handleError(xmlcClassName, except);
        }
    }

    /**
     * Create an instance of a XMLC-generated class or a class derived
     * from one, given the name of the class, using the associated
     * classloader
     */
    public XMLObject create(Class xmlcBasedClass) {
        try {
            if (xmlcBasedClass.isInterface()) {
                // Convert interface to class name.
                return doCreate(xmlcBasedClass.getName() + "Impl",
                                xmlcBasedClass.getClassLoader());
            } else {
                return doCreate(xmlcBasedClass);
            }
        } catch (Throwable except) {
            throw handleError(xmlcBasedClass.getName(), except);
        }
    }

    /** Get the default class loader */
    protected ClassLoader getDefaultClassLoader() {
	return fDefaultClassLoader;
    }
    /**
     * Get the logger.
     */
    public XMLCLogger getLogger() {
        return fLogger;
    }

}
