/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: CompileOptsCmdOptions.java,v 1.6 2007-11-09 05:34:10 jkjome Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.commands.options.BooleanOption;
import org.enhydra.xml.xmlc.commands.options.FlagOption;
import org.enhydra.xml.xmlc.commands.options.Option;
import org.enhydra.xml.xmlc.commands.options.OptionSet;
import org.enhydra.xml.xmlc.metadata.DocumentFormat;
import org.enhydra.xml.xmlc.metadata.MetaData;

/**
 * Command line options parsing for options that control the compilation
 * process.
 */
class CompileOptsCmdOptions {
    /**
     * Class for -verbose option
     */
    private class VerboseOption extends FlagOption {
        /**
         * Constructor.
         */
        public VerboseOption() {
            super("-verbose",
                  "- Generate useful output about the compiler process.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setVerbose(true);
        }
    }

    /**
     * Class for -keep option
     */
    private class KeepOption extends FlagOption {
        /**
         * Constructor.
         */
        public KeepOption() {
            super("-keep",
                  "- Keep the generate java source, do not delete it.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setKeepGeneratedSource(true);
        }
    }

    /**
     * Class for - option
     */
    private class NoCompileOption extends FlagOption {
        /**
         * Constructor.
         */
        public NoCompileOption() {
            super("-nocompile",
                  "- If specified, don't compile the generate java file.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setCompileSource(false);
        }
    }

    /**
     * Class for -info option
     */
    private class InfoOption extends FlagOption {
        /**
         * Constructor.
         */
        public InfoOption() {
            super("-info",
                  "- Print useful information about the page object, such as ids and URLs.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setPrintDocumentInfo(true);
        }
    }

    /**
     * Class for -dump option
     */
    private class DumpOption extends FlagOption {
        /**
         * Constructor.
         */
        public DumpOption() {
            super("-dump",
                  "- Dump out the DOM tree for the page.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setPrintDOM(true);
        }
    }

    /**
     * Class for -parseinfo option
     */
    private class ParseInfoOption extends FlagOption {
        /**
         * Constructor.
         */
        public ParseInfoOption() {
            super("-parseinfo",
                  "- Print detailed information about the page parsing.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setPrintParseInfo(true);
        }
    }

    /**
     * Class for -methods option
     */
    private class MethodsOption extends FlagOption {
        /**
         * Constructor.
         */
        public MethodsOption() {
            super("-methods",
                  "- Print the signature of each generated access method.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setPrintAccessorInfo(true);
        }
    }

    /**
     * Class for -version option
     */
    private class VersionOption extends FlagOption {
        /**
         * Constructor.
         */
        public VersionOption() {
            super("-version",
                  "- Print the XMLC version number.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setPrintVersion(true);
        }
    }

    /**
     * Class for -ssi option
     */
    private class SSIOption extends FlagOption {
        /**
         * Constructor.
         */
        public SSIOption() {
            super("-ssi",
                  "- Process server-side include directives");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getInputDocument().setProcessSSI(true);
        }
    }

    // dbr_20020128.1_start
    /**
     * Class for -ssibase option
     */
    private class SSIBaseOption extends Option {
        /**
         * Constructor.
         */
        public SSIBaseOption() {
            super("-ssibase", 1, false,
                  "basedir - Process absolute given ssi directives relative from the specified dir");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            //System.out.println("ssibase-arg:" + args[0]);
            ((MetaData)clientData).getInputDocument().setSSIBase(args[0]);
        }
    }
    // dbr_20020128.1_end

    /**
     * Class for -docformat option
     */
    private class DocFormatOption extends Option {
        /**
         * Constructor.
         */
        public DocFormatOption() {
            super("-docformat", 1, false,
                  "format - the format of the document[s] being processed (xml, html, unspecified)");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            try {
                ((MetaData)clientData).getInputDocument().setDocumentFormat(DocumentFormat.getType(args[0]));
            } catch (IllegalArgumentException iae) {
                throw new XMLCException("-docformat - " + iae.getMessage());
            }
        }
    }

    /**
     * Class for the -docout option.
     */
    private class DocOutOption extends Option {
        /**
         * Constructor.
         */
        public DocOutOption() {
            super("-docout", 1, false,
                  "outfile - Write a static document to the outfile instead of generating a class");
        }

        /**
         * Parse an instance of the option.
         */
        public void parse(String[] args,
                          ErrorReporter errorReporter,
                          Object clientData) throws XMLCException {
            MetaData metaData = (MetaData)clientData;
            metaData.getCompileOptions().setDocumentOutput(args[0]);
            metaData.getCompileOptions().setCreateSource(false);
        }
    }

    /**
     * Class for the -sourceout option.
     */
    private class SourceOutputOption extends Option {
        /**
         * Constructor.
         */
        public SourceOutputOption() {
            super("-sourceout", 1, false,
                  "sourceout - Root directory for generated source files.");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setSourceOutputRoot(args[0]);
        }
    }

    /**
     * Class for the -warnings option.
     */
    private class WarningsOption extends BooleanOption {
        /**
         * Constructor.
         */
        public WarningsOption() {
            super("-warnings",
                  "yes|no|true|false - Enable or disable warnings");
        }

        /**
         * Set the value.
         */
        protected void set(boolean value,
                           Object clientData) throws XMLCException {
            // sets both parse and compile warning flags
            ((MetaData)clientData).getParser().setWarnings(value);
            ((MetaData)clientData).getCompileOptions().setWarnings(value);
        }
    }

    /**
     * Constructor.  Add options to option set.
     */
    public CompileOptsCmdOptions(OptionSet optionSet) {
        optionSet.addOption(new VersionOption());
        optionSet.addOption(new KeepOption());
        optionSet.addOption(new NoCompileOption());
        optionSet.addOption(new InfoOption());
        optionSet.addOption(new DumpOption());
        optionSet.addOption(new VerboseOption());
        optionSet.addOption(new ParseInfoOption());
        optionSet.addOption(new MethodsOption());
        optionSet.addOption(new SSIOption());
        // dbr_20020128.2_start
        optionSet.addOption(new SSIBaseOption());
        // dbr_20020128.2_end
        optionSet.addOption(new DocFormatOption());
        optionSet.addOption(new DocOutOption());
        optionSet.addOption(new SourceOutputOption());
        optionSet.addOption(new WarningsOption());
    }
}
