/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: CompilerCmdOptions.java,v 1.3 2006-01-22 05:57:14 jkjome Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.commands.options.Option;
import org.enhydra.xml.xmlc.commands.options.OptionSet;
import org.enhydra.xml.xmlc.metadata.JavacOption;
import org.enhydra.xml.xmlc.metadata.MetaData;

/**
 * Command line options parsing for Java compiler-related options.
 */
class CompilerCmdOptions {
    /**
     * Class for the -d option.
     */
    private class DestDirOption extends Option {
        /**
         * Constructor.
         */
        public DestDirOption() {
            super("-d", 1, false,
                  "dir - Destintation directory, passed to javac");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getCompileOptions().setClassOutputRoot(args[0]);
        }
    }

    /**
     * Class for directly supported javac flag or option.
     */
    private class JavacStdArg extends Option {
        /**
         * Constructor.
         */
        public JavacStdArg(String option, int numArgs, String optHelp) {
            super(option, numArgs, false, optHelp);
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            JavacOption javacOption = ((MetaData)clientData).getJavaCompilerSection().addJavacOption();
            javacOption.setName(getName());
            if (args.length > 0) {
                javacOption.setValue(args[0]);
            }
        }
    }

    /**
     * Class for arbitrary javac flag or option.
     */
    private class JavacArbitraryArg extends Option {
        /**
         * Constructor.
         */
        public JavacArbitraryArg(String option, int numArgs, String optHelp) {
            super(option, numArgs, true, optHelp);
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            JavacOption javacOption = ((MetaData)clientData).getJavaCompilerSection().addJavacOption();
            javacOption.setName(args[0]);
            if (args.length > 1) {
                javacOption.setValue(args[1]);
            }
        }
    }

    /**
     * Class for the javac program option.
     */
    private class JavacProgOption extends Option {
        /**
         * Constructor.
         */
        public JavacProgOption() {
            super("-javac", 1, false,
                  "prog - Specify the java compiler to use.");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getJavaCompilerSection().setJavac(args[0]);
        }
    }

    /**
     * Constructor.  Add options to option set.
     */
    public CompilerCmdOptions(OptionSet optionSet) {
        optionSet.addOption(new DestDirOption());
        optionSet.addOption(new JavacStdArg("-g", 0,
                                            "- Passed to javac"));
        optionSet.addOption(new JavacStdArg("-O", 0,
                                            "- Passed to javac"));
        optionSet.addOption(new JavacStdArg("-classpath", 1,
                                            "path - Passed to javac"));
        optionSet.addOption(new JavacArbitraryArg("-javacflag", 1,
                                                  "flag - Aribitrary flag to pass to the javac program; including leading `-' or `+' character"));
        optionSet.addOption(new JavacArbitraryArg("-javacopt", 2,
                                                  "opt value - Aribitrary option and value to pass to the javac program; including leading `-' or `+' character"));
        optionSet.addOption(new JavacProgOption());
    }
}
