/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: AccessMethods.java,v 1.5 2006-10-03 02:42:15 jkjome Exp $
 */

package org.enhydra.xml.xmlc.compiler;

import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Iterator;

import org.enhydra.xml.xmlc.codegen.IndentWriter;
import org.enhydra.xml.xmlc.codegen.JavaClass;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.codegen.JavaMethod;
import org.enhydra.xml.xmlc.codegen.JavaModifiers;
import org.enhydra.xml.xmlc.codegen.JavaParameter;
import org.enhydra.xml.xmlc.dom.AccessorGenerator;
import org.enhydra.xml.xmlc.metadata.MetaData;

/**
 * Data collection and code generation for element access methods.
 */
class AccessMethods {
    /**
     * Information collected about each element.
     */
    private ElementTable fElementTable;

    /**
     * Should delegation support be create.
     */
    private boolean fCreateDelegation;

    /**
     * Class being generated.
     */
    private JavaClass fDocClass;

    /**
     * DOM Accessor generator.
     */
    private AccessorGenerator fAccessorGenerator;

    /**
     * List of accessor methods, indexed by element info object.
     */
    private HashMap fAccessorMethods = new HashMap();

    /**
     * Parameter for syncWithDocument methods.
     * A static since in can be shared.
     */
    private static final JavaParameter[] syncWithDocParam = {
        new JavaParameter("node", "Node",
                          "Root of subtree to sync.")
    };

    /**
     * Constructor.
     */
    public AccessMethods(MetaData metaData,
                         ElementTable elementTable,
                         AccessorGenerator accessorGenerator) {
        fCreateDelegation = metaData.getDocumentClass().getDelegateSupport();
        fElementTable = elementTable;
        fAccessorGenerator = accessorGenerator;
    }

    /**
     * Create the access methods and field definitions for an element.
     */
    private void createAccessMethods(ElementInfo elementInfo) {
        ElementInfo.AccessorInfo[] accessors = elementInfo.getAccessors();
        int methodIdx = 0;
        JavaMethod[] methods = new JavaMethod[elementInfo.getNumAccessMethods()];
        if (accessors != null) {
            for (int idx = 0; idx < accessors.length; idx++) {
                methods[methodIdx++]
                    = fAccessorGenerator.createAccessorMethod(elementInfo,
                                                              accessors[idx],
                                                              fCreateDelegation,
                                                              fDocClass);
            }
        }
        if (elementInfo.createSetText()) {
            methods[methodIdx++]
                = fAccessorGenerator.createSetTextMethod(elementInfo,
                                                         fCreateDelegation,
                                                         fDocClass);
        }
        fAccessorMethods.put(elementInfo, methods);
    }


    /**
     * Generate access methods and field definitions.
     */
    private void createAccessMethods() {
        Iterator elements = fElementTable.getJavaIdElements();
        while (elements.hasNext()) {
            ElementInfo elementInfo = (ElementInfo)elements.next();
            if (elementInfo.getNumAccessMethods() > 0) {
                createAccessMethods(elementInfo);
            }
        }
    }

    /**
     * Create body of field initialization method
     */
    private void createlIdTagInit(JavaCode body) {
        body.addln("if (node.getNodeType() != Node.DOCUMENT_NODE) {");
        body.enter();
        body.addln("throw new org.enhydra.xml.xmlc.XMLCRuntimeException("
                    + JavaLang.createStringConst("Node must be a document node")
                    + ");");
        body.leave();
        body.addln("}");
        body.addln("Document doc = (Document) node;");
        body.addln("try {");
        body.enter();
        body.addln("java.lang.reflect.Field[] fs = this.getClass().getDeclaredFields();");
        body.addln("int substStart = "
                    + JavaLang.createStringConst("$element_")
                    + ".length();");
        body.addln("for (int i = 0; i < fs.length; i++) {");
        body.enter();
        body.addln("java.lang.reflect.Field f = fs[i];");
        body.addln("if (!f.getName().startsWith("
                    + JavaLang.createStringConst("$element_")
                    + ")) continue;");
        body.addln("String id = f.getName().substring(substStart);");
        body.addln("Node idNode = doc.getElementById(id);");
        body.addln("if (idNode == null) {");
        body.enter();
        body.addln("id = id.substring(0, 1).toLowerCase() + id.substring(1);");
        body.addln("idNode = doc.getElementById(id);");
        body.leave();
        body.addln("}");
        body.addln("if (idNode != null) f.set(this, idNode);");
//        body.addln("else System.out.println("
//                    + JavaLang.createStringConst("No node match for field: ")
//                    + " + f.getName());");
        body.leave();
        body.addln("}");
        body.leave();
        body.addln("} catch (Exception e) {");
        body.enter();
        body.addln("throw new org.enhydra.xml.xmlc.XMLCRuntimeException("
                    + JavaLang.createStringConst("Error reflecting on element access fields")
                    + ", e);");
        body.leave();
        body.addln("}");
    }

    /**
     * Generate access method field initialization function used in copy
     * construction or building the document, in the case of deferred parsing.
     * 
     */
    private void createAccessMethodSync() {
        String[] doc = {
            "Called to set access method fields from the DOM.",
            "Missing Ids have fields set to null.",
        };
        JavaMethod method
            = new JavaMethod("syncWithDocument",
                             "void",
                             JavaModifiers.PROTECTED,
                             syncWithDocParam, doc);
        fDocClass.addMethod(method);
        JavaCode body = method.getCode();

        // Generate body to initialize.
        createlIdTagInit(body);
    }

    /**
     * Generate the code for the access methods.
     * @param docClass Object containing the class being generated.
     */
    public void generateCode(JavaClass docClass) {
        fDocClass = docClass;
        createAccessMethodSync();
        createAccessMethods();
    }
    
    /**
     * Print the signatures of the access methods are created.
     */
    public void printAccessMethods(PrintWriter out) {
        IndentWriter iout = new IndentWriter(out);
        Iterator elements = fElementTable.getJavaIdElements();
        while (elements.hasNext()) {
            ElementInfo elementInfo = (ElementInfo)elements.next();
            JavaMethod[] methods = (JavaMethod[])fAccessorMethods.get(elementInfo);
            if (methods != null) {
                for (int idx = 0; idx < methods.length; idx++) {
                    methods[idx].printMethodSignature(iout, false);
                }
            }
        }
        iout.flush();
    }

    /**
     * Print ids that didn't have accessors generated due to being illegal
     * Java identifiers.  We keep them sorted so that the test always work.
     */
    public void printOmittedIds(PrintWriter out) {
        Iterator ids = fElementTable.getInvalidElementIds();
        if ((ids != null) && ids.hasNext()) {
            out.println("Access methods not created for these element IDs");
            out.println("since they are not legal Java identifiers:");
            while (ids.hasNext()) {
                out.print("`");
                out.print((String)ids.next());
                out.println("'");
            }
        }
    }
}
