/*
 * Copyright (C) 2003, 2005, 2006, 2008, 2010 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.compiler;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.InputSourceOps;
import org.enhydra.xml.io.XmlReader;
import org.enhydra.xml.io.XmlReaderException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;
import org.enhydra.xml.xmlc.dom.XMLCDomFactoryCache;
import org.enhydra.xml.xmlc.metadata.CompileOptions;
import org.enhydra.xml.xmlc.metadata.DocumentClass;
import org.enhydra.xml.xmlc.metadata.DocumentFormat;
import org.enhydra.xml.xmlc.metadata.InputDocument;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.metadata.Parser;
import org.enhydra.xml.xmlc.metadata.ParserType;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import org.enhydra.xml.xmlc.misc.SSIReader;
import org.enhydra.xml.xmlc.misc.SSISystemIdResolver;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.parsers.xerces.XercesDOMParser;
import org.enhydra.xml.xmlc.parsers.xerces.XercesHTMLDOMParser;
import org.w3c.dom.DOMConfiguration;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


/**
 * Parse a XML or HTML document into a DOM.
 */
public class Parse {

    //FIXME - remove commented code, leaving for now for posterity and reference

    /**
     * XML parser object.
     */
    private XMLCParser fXMLCParser;

    /**
     * Error output.
     */
    private ErrorReporter fErrorReporter;

    /**
     * Verbose output stream.
     */
    private PrintWriter fVerboseOut;

    /**
     * Print verbose messages.
     */
    private boolean fVerbose;

    private SSISystemIdResolver fSSISystemIdResolver;

    /**
     * Same as
     * {@link #Parse(ErrorReporter, PrintWriter, SSISystemIdResolver) Parse(errorReporter, verboseOut, null)}
     *
     * @param errorReporter object used to handle errors
     * @param verboseOut output stream for verbose and trace information, optional,
     *                   may be null
     */
    public Parse(ErrorReporter errorReporter,
                 PrintWriter verboseOut) {
        this(errorReporter, verboseOut, null);
    }

    /**
     * @param errorReporter object used to handle errors
     * @param verboseOut output stream for verbose and trace information, optional,
     *                   may be null
     * @param ssiSystemIdResolver a configured SSI system Id resolver, optional,
     *                            may be null
     */
    public Parse(ErrorReporter errorReporter,
            PrintWriter verboseOut,
            SSISystemIdResolver ssiSystemIdResolver) {
        fErrorReporter = errorReporter;
        fVerboseOut = verboseOut;
        fSSISystemIdResolver = ssiSystemIdResolver;
   }

    /*
     * Is this an the name of an HTML parser?
     */
    private static boolean isHtmlParser(ParserType parser) {
        return (parser == ParserType.SWING)
            || (parser == ParserType.TIDY
            || parser == ParserType.NEKOHTML);
    }

    /**
     * Determine the parser to use.
     */
    private void setupParser(MetaData metaData,
                             ParserType parser,
                             boolean isHtmlDocument)
        throws XMLCException, IOException {
        if (parser == null) {
            // Use default parser based on document type.
            if (isHtmlDocument) {
                parser = ParserType.NEKOHTML;
            } else {
                parser = ParserType.XERCES;
            }
        }
        
        // Check for conflicts with parser and document type.
        if (isHtmlDocument) {
            if (!isHtmlParser(parser)) {
                throw new XMLCException("Document appears to be an HTML document; the "
                                        + parser + " parser only supports XML"
                                        + " (does the document start with `<?xml ... ?>'?): "
                                        + getInputSourceDesc(metaData));
            }
        } else {
            if (isHtmlParser(parser)) {
                throw new XMLCException("Document appears to be an XML document; the "
                                        + parser + " parser only supports HTML: "
                                        + getInputSourceDesc(metaData));
            }
            if (metaData.getHTMLSection() != null) {
                //FIXME: need to include check for options being defaulted.
                //throw new XMLCException("HTML options may not be specified for a XML document");
            }
        }

        // Load and initialize parser.
        if (parser == ParserType.SWING
                || parser == ParserType.TIDY
                || parser == ParserType.NEKOHTML) {
            //fXMLCParser = new SwingHTMLParser();
            //fXMLCParser = new TidyHTMLParser();
            fXMLCParser = new XercesHTMLDOMParser();
        } else if (parser == ParserType.XERCES) {
            //fXMLCParser = new XercesParser();
            fXMLCParser = new XercesDOMParser();
        } else {
            throw new XMLCError("Unknown parser \"" + parser + "\"");
        }
    }

    /**
     * Determines if the file pointed at by the MetaData <code>&lt;inputDocument url=''/&gt;</code>
     * attribute is XML.  It's either explicitly specified using the MetaData
     * <code>&lt;inputDocument documentFormat=''/&gt;</code> attribute, set to 'xml' to assert
     * XML format or 'html' to assert NOT XML format, or, if not specified,
     * must be determined by {@link InputSourceOps#isXMLDocument(InputSource) looking at the file}
     * for an '<code>&lt;?xml</code>' prolog.
     * 
     * @param metaData the XMLC metadata
     * @return true if the document is declared, or found, to be XML, otherwise false
     * @since 2.3.2
     */
    public static boolean isXMLDocument(MetaData metaData) throws IOException {
        InputDocument inputDoc = metaData.getInputDocument();
        DocumentFormat docFormat = inputDoc.getDocumentFormat();

        if (docFormat == DocumentFormat.XML) {
            return true;
        } else if (docFormat == DocumentFormat.HTML) {
            return false;
        } else {
            InputSource inputSource = inputDoc.getInputSource("UTF-8");
            return inputSource != null ? InputSourceOps.isXMLDocument(inputSource) : false;
        }
    }

    /**
     * Get the input source, handling SSI filtering.
     */
    private static InputSource getInputSource(MetaData metaData, boolean isHtmlDocument, SSISystemIdResolver ssiSystemIdResolver) throws IOException {
        InputDocument inputDoc = metaData.getInputDocument();
        InputSource inputSource = inputDoc.getInputSource(!isHtmlDocument ? "UTF-8" : null);

        try {
            XmlReader reader = new XmlReader(InputSourceOps.openSystemId(inputSource.getSystemId()), true, inputSource.getEncoding());
            inputSource.setCharacterStream(reader);
            inputSource.setEncoding(reader.getEncoding());

            //System.out.println("___________________________");
            //System.out.print("url: ");
            //System.out.println(inputSource.getSystemId());
            //System.out.print("enc: ");
            //System.out.println(inputSource.getEncoding());
            //System.out.println("---------------------------");
        }
        catch (XmlReaderException xre) {
            //This should never happen using lenient encoding detection according to the XMLReader doc as well as a review of the code
            throw new XMLCError("Unexpected failure during lenient encoding detection by XMLReader of file "+inputSource.getSystemId(), xre);
        }

        if (inputDoc.getProcessSSI()) {
            return SSIReader.create(inputSource, inputDoc.getSSIBase(), ssiSystemIdResolver);
        } else {
            return inputSource;
        }
    }

    /** Get a description of the input source for error messages */
    private static String getInputSourceDesc(MetaData metaData) {
        InputDocument inputDoc = metaData.getInputDocument();
        return InputSourceOps.getName(inputDoc.getInputSource());
    }

    /**
     * Get the line number map to pass to the parse, or null
     * if one is not in uses.
     */
    private static LineNumberMap getLineNumberMap(InputSource input) {
        Reader reader = input.getCharacterStream();
        if (reader instanceof SSIReader) {
            return ((SSIReader)reader).getLineNumberMap();
        } else {
            return null;
        }
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     * <p>
     * The Encoding detection hierarchy goes like this...
     * </p>
     * <ol>
     * <li>Use encoding detected by {@link XMLReader} (BOM or XML prolog).  If
     * not detected, then...</li>
     * <li>Use encoding provided by {@link InputDocument#getInputSource(String)}
     * in order of...
     * <ol style="list-style-type:lower-alpha">
     * <li>Explicit encoding provided by the MetaData <code>&lt;html encoding=''/&gt;</code>
     * attribute or, if not specified, then...</li>
     * <li>"UTF-8" if the document is {@link #isXmlDocument(MetaData) determined to be XML},
     * otherwise...</li>
     * <li>The default system file encoding unless it's ASCII, in which case
     * ISO_8859-1 is used (converted to the MIME-preferred name).</li>
     * </ol></li>
     * </ol>
     *
     * @param metaData Document metadata.
     */
    public XMLCDocument parse(MetaData metaData)
        throws XMLCException, IOException {

        Parser parser = metaData.getParser();
        CompileOptions compileOptions = metaData.getCompileOptions();
        DocumentClass documentClass = metaData.getDocumentClass();

        // Setup tracing
        if (fVerboseOut != null) {
            // Only enable verbose output if a stream is available
            // and its requested.
            fVerbose = compileOptions.getVerbose();
        }
        boolean printParseInfo
            = (compileOptions.getPrintParseInfo() && (fVerboseOut != null));

        ParseTracer traceOut = new ParseTracer(printParseInfo ? fVerboseOut : null);

        boolean isHtmlDocument = !isXMLDocument(metaData);
        InputSource inputSource = getInputSource(metaData, isHtmlDocument, fSSISystemIdResolver);
        LineNumberMap lineNumberMap = getLineNumberMap(inputSource);

        XMLCDomFactory domFactory
            = XMLCDomFactoryCache.createFactory(documentClass.getDomFactoryClass(isHtmlDocument),
                                                isHtmlDocument);
        if (fVerbose) {
            fVerboseOut.println(">>> using DOM Factory class: " + domFactory.getClass().getName());
        }
        XMLCDomFactoryCache.checkForOutdatedClass(domFactory);

        setupParser(metaData, parser.getName(), isHtmlDocument);

        boolean saveWarnings = fErrorReporter.getPrintWarnings();
        fErrorReporter.setPrintWarnings(parser.getWarnings());

        XMLCDocument xmlcDoc;
        try {
            xmlcDoc = fXMLCParser.parse(inputSource, 
                                        lineNumberMap,
                                        domFactory,
                                        metaData,
                                        fErrorReporter,
                                        traceOut);
        } catch (SAXException except) {
            Exception useExcept = except.getException();
            if (useExcept == null) {
                useExcept = except;
            }
            throw new XMLCException("Parse of \"" + inputSource.getSystemId()
                                    + "\" failed: " + useExcept, useExcept);
        } finally {
            fErrorReporter.setPrintWarnings(saveWarnings);
        }
        int cnt = fErrorReporter.getErrorCnt();
        if (cnt > 0) {
            String errorOutput = fErrorReporter.getOutput();
            throw new XMLCException(cnt + " error" + ((cnt == 1) ? "" : "s")
                                    + " parsing document"
                                    + (errorOutput != null ? "... \n" + errorOutput.trim() : ""));
        }

        // Normalize the text nodes.
        /*Element root = xmlcDoc.getDocument().getDocumentElement();
        if (root != null) {
            root.normalize();
        }*/

        DOMConfiguration config = xmlcDoc.getDocument().getDomConfig();

        //below is an experiment for validating an HTML document in memory.  The
        //benefit of this would be to set "id" attributes as of type "ID",
        //effectively optimizing getElementById() lookups for the HTML DOM (in 
        //case the parser didn't do this already; normally it wouldn't).  Another
        //possible benefit is beter whitepace normalization (utilizing the DTD to
        //direct the behavior), though I didn't notice any difference.
        //Note: requires catalog entries for the HTML DTDs, and these HTML DTDs
        //really need to be mofified XHTML DTDs because the HTML DTDs themselves
        //are not well-formed (kind of defeats their purpose, no?).
        /*if (xmlcDoc.isHtmlDocument()) {
            DocumentType docType = DOMAccess.accessDocumentType(xmlcDoc.getDocument());
            if (docType != null) {
                String publicId = docType.getPublicId();
                if (publicId != null && publicId.indexOf("XHTML") == -1) {
                    XMLEntityResolver resolver = new XMLEntityResolver();
                    try {
                        resolver.setDefaultResolving();
                        String systemId = resolver.getPublicMapping(docType.getPublicId(), docType.getSystemId());
                        if (systemId != null) {
                            systemId = systemId.substring(XMLEntityResolver.CLASSPATH_PROTOCOL_PREFIX.length());
                            URL url = getClass().getResource(systemId);
                            if (url != null) {
                                config.setParameter("validate", Boolean.TRUE);
                                config.setParameter("schema-type", javax.xml.XMLConstants.XML_DTD_NS_URI);
                                config.setParameter("schema-location", url.toExternalForm());
                            }
                        }
                    } catch (Exception e) {
                        e.printStackTrace(); //remove when no longer experimental
                    }
                }
            }
        }*/

        //config.setParameter("normalize-characters", Boolean.TRUE); //Fully normalize the characters in the document - unmodifiable, Xerces reports errors if value is set
        //config.setParameter("element-content-whitespace", Boolean.FALSE); //Discard all Text nodes that contain whitespaces in element content - unmodifiable, Xerces reports errors if value is set
        //config.setParameter("entities", Boolean.FALSE); //Remove all EntityReference nodes from the document, putting the entity expansions directly in their place. Text nodes are normalized, as defined in Node.normalize. - ok to set, but not sure if desireable?  Revisit.
        config.setParameter("namespaces", Boolean.FALSE);  //the parser should have done this for us, avoid the overhead
        config.setParameter("well-formed", Boolean.FALSE); //the parser should have done this for us, avoid the overhead
        //1. Normalizes Text nodes, consolidating adjacent Text nodes into a single Text node
        //2. Updates the content of EntityReference nodes according to the entities they reference
        xmlcDoc.getDocument().normalizeDocument();

        return xmlcDoc;
    }    
}
