/*
 * Copyright (C) 2003, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Richard Kunze, Ole Arndt, David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.deferredparsing;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.net.URLConnection;
import java.util.Arrays;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.WeakHashMap;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.lazydom.LazyDocument;
import org.enhydra.xml.lazydom.TemplateDOM;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.XMLCRuntimeException;
import org.enhydra.xml.xmlc.compiler.EditDOM;
import org.enhydra.xml.xmlc.compiler.Parse;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.misc.I18nUtil;
import org.enhydra.xml.xmlc.misc.SSISystemIdResolver;
import org.w3c.dom.Document;

/**
 * Default implementation of the {@link DocumentLoader} interface.
 *
 * <p>This implementation maintains a global cache of loaded DOM
 * objects. If a DOM object is found in cache and the cached copy is
 * up to date with regard to the source document and XMLC metadata
 * object, a copy of the cached object is handed out. Otherwise, the
 * document is re-parsed from the source document, the DOM edits
 * specified in the document metadata are performed and the document
 * is stored in cache as a template. Note that the cache may be a
 * {@link WeakHashMap} or a map with similar semantics that frees up
 * entries if memory gets low.</p>
 *
 * <p>This class can be used as the basis for specialized
 * document loaders.</p>
 */
public class DocumentLoaderImpl implements DocumentLoader {

    /** The associated factory */
    private XMLCDeferredParsingFactory fFactory;
    private ResourceLoader fResourceLoader;

    /** Helper class for storing cached DOMs */
    protected static class CacheEntry {
        volatile long timestamp = -1;
        volatile Document document = null;
        volatile URL src = null;
        volatile boolean ready = false;
    }

    /** The map to use for caching parsed template DOMs */
    private final Map fTemplateCache;

    /**
     * Default constructor. Uses a {@link Cache} as template
     * DOM cache.
     * @see #DocumentLoaderImpl(Map)
     */
    public DocumentLoaderImpl() {
        this(null);
    }

    /**
     * Constructor. Uses the specified map as template cache.
     * 
     * @param cache the cache to use, defaults to {@link Cache} if null
     */
    public DocumentLoaderImpl(final Map cache) {
        if (cache == null) {
            fTemplateCache = new Cache();
        } else {
            fTemplateCache = cache;
        }
    }

   /**
    * Generates a cache key for the given document class.  May
    * be overridden by subclasses to enhance caching behavior.
    * 
    * @param docClass the class to generate the cache key for
    * @return the key used to identify this class in the internal parsed
    *         document cache
    */
    protected String makeCacheKey(Class docClass) {
        return docClass.getName();
    }

    /**
     * Get the cache entry for a document, initialize one of not
     * already exists
     * 
     * Note that the cache entry does not necessarily hold a template
     * - it may need to be initialized first. Subclasses overriding
     * this method must make sure that new cache entries are
     * constructed and put in the cache in a thread safe manner.
     *
     * @param docClass the class to get the cache entry for.
     * @return a cache entry. The returned object is suitable as mutex
     * for document initialization.
     */
    protected CacheEntry getCacheEntry(Class docClass) {
        String name = makeCacheKey(docClass);

        Locale[] locales = getFactory().getLocales();
        if (locales != null) {
            for (int i = 0; i < locales.length; i++) {
                Locale locale = locales[i];
                if (locale != null) {
                    name = I18nUtil.canonicalLocaleString(name, null, locale);
                    //break; //TODO - without this break, the name becomes a combination of all non-null locale strings appended to each other.  Verify that we want this (I think we do so far)
                }
            }
        }

        /* XXX: This assumes that concurrent unsynchronized accesses
         * to templateCache.get() with different keys are not
         * disturbed by the code in the synchronized block, and that
         * concurrent accesses with the same key return either null or
         * the entry that has been put into the map inside the
         * synchronized block. 
         *
         * The synchronization is broken down to two step when the
         * entry does not exist in the cache. First, lock the template
         * cache and create a empty entry and template cache is
         * unlocked.  Then the new empty entry is locked and
         * updated. Breaking down to two steps would avoid the entry
         * creating 
         */
        CacheEntry entry = (CacheEntry) fTemplateCache.get(name);
        if (entry == null) {
            synchronized (fTemplateCache) {
                entry = (CacheEntry) fTemplateCache.get(name);
                if (entry == null) {

                    if (getLogger().debugEnabled()) {
                        getLogger().logDebug(">>>Creating cache entry for "
                                             + docClass.getName());
                    }

                    entry = new CacheEntry();
                    fTemplateCache.put(name, entry);
                }
            }
        }

        if (!entry.ready) {
            synchronized (entry) {
                if (!entry.ready) {
                    entry.src = findSourceUrl(docClass);
                    entry.document = parseDocument(docClass, entry.src);
                    entry.timestamp = getCurrentTimestamp(docClass, entry.src);
                    entry.ready = true;
                }
            }
        }

        return entry;
    }

    /**
     * Check if the DOM instance is out of date
     */
    private CacheEntry checkCacheEntry(Class docClass, CacheEntry entry) {

        // Check if the document is out of date, and reload it if yes.
        long ts = getCurrentTimestamp(docClass, entry.src);

        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Checking time stamp ts="
                                 + new Date(ts));
            getLogger().logDebug(">>>       entry.timestamp="
                                 + new Date(entry.timestamp));
        }

        if (ts == -1 || ts > entry.timestamp) {
            synchronized (entry) {
                ts = getCurrentTimestamp(docClass, entry.src);
                // Check if some other thread did reparse the document
                // while this one has been waiting for the lock...
                if (ts == -1 || ts > entry.timestamp) {
                    if (getFactory().isEnableI18n()) entry.src = findSourceUrl(docClass);//recheck the source document in case a new localized version was introduced or removed
                    entry.document = parseDocument(docClass, entry.src);
                    entry.timestamp = getCurrentTimestamp(docClass, entry.src);
                } else {
                    if (getLogger().debugEnabled()) {
                        getLogger().logDebug(">>>DOM for "
                                             + docClass.getName()
                                             + " has been updated by another thread.");
                    }
                }
            }
        }

        return entry;
    }

    /**
     * See {@link ResourceLoaderImpl} for resource loading behavior.
     * 
     * @return a resource loader implementation
     * @see ResourceLoaderImpl
     * @since 2.3.2
     */
    public ResourceLoader getResourceLoader() {
        if (fResourceLoader == null) {
            fResourceLoader = new ResourceLoaderImpl(fFactory);
        }
        return fResourceLoader;
    }

    /** 
     * Get a DOM instance representing the specified document.
     * @param docClass the XMLC generated class to get the DOM for.
     * @return the DOM representing objects of <code>docClass</code>
     */
    public Document getDocument(Class docClass) throws XMLCRuntimeException {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>DOM instance requested for class "
                                 + docClass.getName());
        }

        CacheEntry entry = getCacheEntry(docClass); // get the cached entry

        entry = checkCacheEntry(docClass, entry);
        // make sure cache entry is up to date

        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Document class "
                                 + entry.document.getClass().getName());
        }

        return (Document) entry.document.cloneNode(true);
    }

    /**
     * Get one of the compiled-in constants in a XMLC-generated class (or
     * subclass).
     *
     * @deprecated use XMLCDeferredParsingFactory.getClassConstant(Class, String) instead
     */
    public static Object getClassConstant(Class xmlcBasedClass,
                                          String constName) {
        return XMLCDeferredParsingFactory.getClassConstant(xmlcBasedClass, constName);
    }

    /** 
     * Get the name of the source file associated with <code>docClass</code> 
     * <p>
     * Note: while not deprecated because it is public and has always been called
     * by findSourceUrl(Class), it might as well be since identical functionality
     * can be obtained via XMLCDeferredParsingFactory.getSourceFilePath(Class).
     * </p>
     */
    protected String getSourceFileName(Class docClass) {
        return XMLCDeferredParsingFactory.getSourceFilePath(docClass);
    }

    /**
     * Find the URL of the source document associated with <code>docClass</code>.
     * If the source file cannot be found, an exception is thrown.
     * Resource dirs are searched first, then the classloader.
     * Package prefix paths are taken into account.
     * 
     * @param docClass the class for which to load the source doc
     * @throws XMLCRuntimeException if the source doc cannot be found
     */
    protected URL findSourceUrl(Class docClass) {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Loading source doc for "
                                 + docClass.getName());
        }

        String srcPath = getSourceFileName(docClass);
        String[] paths = getFactory().getPathsFromPackagePrefixes(srcPath);
        URL srcURL = getResourceLoader().getResource(paths);

        if (srcURL == null) {
            throw new XMLCRuntimeException("Source path(s) '"
                                           + Arrays.asList(paths)
                                           + "' not found for "
                                           + docClass.getName());
        }
        return srcURL;
    }

    /**
     * Load the XMLC metadata file.  If the class-specific metadata file cannot
     * be found, the default metadata file searched for.  If the default
     * metadata file cannot be found (or has not been defined), an exception
     * is thrown.  Resource dirs are searched first, then the classloader.
     * Package prefix paths are taken into account.
     * 
     * @param docClass the class for which to load the metadata
     * @param errorReporter the error handler to use
     * @throws XMLCRuntimeException if the metadata cannot be found
     * @throws XMLCException if the found metadata cannot be parsed
     */
    protected MetaData loadMetaData(Class docClass,
                                    ErrorReporter errorReporter)
        throws XMLCException {
        if (getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Loading metadata for "
                                 + docClass.getName());
        }

        XMLCDeferredParsingFactory factory = getFactory();

        String[] paths = factory.getXmlcClassMetaDataFilePaths(docClass);
        URL metaDataURL = getResourceLoader().getResource(paths);
        if (metaDataURL == null) {
            metaDataURL = factory.getDefaultMetaDataURL();
        }

        if (metaDataURL == null) {
            throw new XMLCRuntimeException("Metadata path(s) '"
                                           + Arrays.asList(paths)
                                           + "' (as well as default metadata file paths) not found for "
                                           + docClass.getName());
        }

        return factory.loadMetaData(metaDataURL, errorReporter);
    }

    /**
     * Get the current timestamp of the *ML template for docClass.
     * @param docClass the class to get the information for
     * @param src the source URL for this class, as returned by
     * {@link #findSourceUrl(Class)}
     * @return the timestamp, or -1 to force a reload.
     */
    protected long getCurrentTimestamp(Class docClass, URL src) {
        if ("file".equals(src.getProtocol())) {
            long lastModified = new File(src.getPath()).lastModified();
            if (lastModified == 0) {
            	if (getLogger().infoEnabled()) {
            	    getLogger().logInfo(">>>Cannot read last modified time for " + src);
            	}
                return -1;
            }
            return lastModified;
        }

        URLConnection urlc = null;
        try {
            urlc = src.openConnection();
            urlc.setUseCaches(false);
            return urlc.getLastModified();
        } catch (IOException e) {
            if (getLogger().infoEnabled()) {
                getLogger().logInfo(">>>Cannot read last modified time for " + src, e);
            }
            return -1;
        } finally {
            if (urlc != null) {
                try {
                    urlc.getInputStream().close();
                } catch (Exception e) {
                    //squash
                }
            }
        }
    }

    /** Create a new template DOM for docClass
     * @param docClass the class to load the DOM for
     * @param src the source file for this class, as returned by
     * {@link #findSourceUrl(Class)}
     */
    protected Document parseDocument(Class docClass, URL src) {
        ErrorReporter errorReporter = new ErrorReporter();
        try {
            if (getLogger().debugEnabled()) {
                getLogger().logDebug(">>>Parsing DOM for "
                                     + docClass.getName()
                                     + " from source URL "
                                     + src.toString());
            }

            MetaData metaData = loadMetaData(docClass, errorReporter);
            metaData.getInputDocument().setUrl(src.toString());

            Parse parser = createParser(docClass, errorReporter, null);
            XMLCDocument xmlcDoc = parser.parse(metaData);

            EditDOM domEditor = createDOMEditor(docClass, metaData);
            domEditor.edit(xmlcDoc);

            if (xmlcDoc.getDocument() instanceof LazyDocument) {
                LazyDocument lazyDoc = (LazyDocument)xmlcDoc.getDocument();
                //TODO - Do we really need a clone here, or just a way to set the templateDOM and reset flags to use it properly?
                //The nodes are the same.  The problem is that they end up not being owned by the new document, but by the
                //delegate document inside the TemplateDOM.  This requires extended documents to override cloneAsTemplateDocument()
                //as well as LazyDocument overriding getIdentifier() because it compares the element's parent to "this" which
                //fails because of the delegation.  However, we might need to override anyway to avoid unnecessary expansion of nodes.
                xmlcDoc.setDocument(lazyDoc.cloneAsTemplateDocument(new TemplateDOM(lazyDoc)));
            }

            return xmlcDoc.getDocument();
        } catch (Exception e) {
            XMLCRuntimeException except;
            if (e instanceof XMLCRuntimeException) {
                except = (XMLCRuntimeException) e;
            } else {
                except = new XMLCRuntimeException(e);
            }
            throw except;
        }
    }

    /** Create the parser to use. Override this if you need a special
     * version */
    protected Parse createParser(Class docClass,
                                 ErrorReporter errorReporter,
                                 PrintWriter verboseOut) {
        return new Parse(errorReporter, verboseOut, new SSISystemIdResolver(getSourceFileName(docClass), getResourceLoader(), getFactory()));
    }

    /** Create the DOM editor to use. Override this if you need a special
     * version */
    protected EditDOM createDOMEditor(Class docClass, MetaData metaData) {
        return new EditDOM(metaData);
    }

    /**
     * Bind to a factory.
     *
     * @param factory Factory that is creating the object.
     */
    public void init(final XMLCDeferredParsingFactory factory) {
        fFactory = factory;
    }

    /** Get the associated factory */
    protected XMLCDeferredParsingFactory getFactory() {
        return fFactory;
    }

    /** Get the associated logger */
    protected XMLCLogger getLogger() {
        return fFactory.getLogger();
    }

}
