/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is David
 * Li.
 *
 * The Enhydra Application Server and portions created by Lutris
 * Technologies, Inc. are Copyright Lutris Technologies, Inc.  All
 * Rights Reserved.
 * 
 * Contributor(s): 
 * David Li
 * Jacob Kjome
 *
 * $Id: DynamicClassLoader.java,v 1.11 2007-04-22 05:40:59 jkjome Exp $
 */
package org.enhydra.xml.xmlc.deferredparsing;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.enhydra.xml.xmlc.StreamXMLCLogger;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;

public class DynamicClassLoader extends ClassLoader {

    private ClassLoader delegateLoader = null;
    private XMLCLogger logger = null;
    private Map classMap;

    DynamicClassLoader(ClassLoader parent) {
        this(parent, null);
    }

    DynamicClassLoader(ClassLoader parent, XMLCLogger logger) {
        super(parent);
        if (logger != null) {
            this.logger = logger;
        } else {
            this.logger = new StreamXMLCLogger();
        }
        classMap = Collections.synchronizedMap(new HashMap());
    }

    /**
     * Add a delegate be used in class loading.
     * 
     * @param delegate the loader to be used as delegate
     */
    public void setDelegate(ClassLoader delegate) {
        delegateLoader = delegate;
    }

    /**
     * load class,
     */
    public Class loadClass(String className) throws ClassNotFoundException {
        Class clazz = null;
        ClassLoader loader = null;
        try {
            if (delegateLoader != null) {
                loader = delegateLoader;
            } else {
                loader = getParent();
            }
            clazz = loader.loadClass(className);
        } catch (ClassNotFoundException e) {
            if (loader != getParent()) {
                clazz = getParent().loadClass(className);
            } else {
                throw e;
            }
        }
        return clazz;
    }

    /**
     * Generate a class based on the class name
     * 
     * @param className class name for the class to be generated
     * @param filePath the path to the source file, generally a relative path
     * @param domFactory the domFactory to use providing information about how
     *                   to build the XMLC class
     */
    Class createClass(String className, String filePath, XMLCDomFactory domFactory) {
        if (logger.debugEnabled()) {
            logger.logDebug(">>>Request to create class " + className + " from document " + filePath);
        }
        Class clazz = (Class) classMap.get(className);
        if (clazz == null) {
            synchronized (classMap) {
                clazz = (Class) classMap.get(className);
                if (clazz == null) {
                    DynamicMLCreator dhc = new DynamicMLCreatorASM3Impl(className, filePath, domFactory);
                    byte[] buf = dhc.getByteArrayForClass();
                    clazz = defineClass(null, buf, 0, buf.length);
                    /*
                     * try { //dump class to file system for decompile comparison OutputStream
                     * stream = new FileOutputStream("D:/dev/xmlctest.class"); stream.write(buf);
                     * stream.flush(); stream.close(); } catch (Exception e) {
                     * System.err.println("failed to write class!!!!"); }
                     */
                    classMap.put(className, clazz);
                }
            }
        } else {
            if (logger.debugEnabled()) {
                logger.logDebug(">>>Got class " + className + " from cache");
            }
        }
        return clazz;
    }
}
