/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.deferredparsing;

import java.net.URL;

/**
 * A {@link ResourceLoader resource loader} implementation that looks up
 * candidatePaths from configured resource dirs with fall back to the classpath.
 * 
 * @author Jacob Kjome
 * @since 2.3.2
 */
public class ResourceLoaderImpl implements ResourceLoader {

    protected final XMLCDeferredParsingFactory fFactory;

    /**
     * 
     * @param factory the XMLC deferred parsing factory
     */
    public ResourceLoaderImpl(final XMLCDeferredParsingFactory factory) {        
        this.fFactory = factory;
    }

    /**
     * Delegates to {@link #getPathURLFromResourceDirOrClasspath(String[])}
     * 
     * @param candidatePaths an array of candidate paths from which to load a
     *                      resource in order of preference
     * @return a URL corresponding to one of the candidatePaths or null if not
     *         found
     */
    public URL getResource(final String[] candidatePaths) {
        return getPathURLFromResourceDirOrClasspath(candidatePaths);
    }

    /**
     * Iterates over candidatePaths, where each resource is first searched for
     * in resource dirs with fallback to the classpath.   This allows one to make
     * files available in the classloader, for fallback, while maintaining the
     * capability of overriding classloader-provided files by placing equivalent
     * files in configured resource dirs.  This makes for simple development
     * and flexible deployment.
     * <p>
     * Paths are expected to be relative, i.e., no leading slash, with
     * directory/package names separated by forward slashes regardless of
     * operating system platform.
     * </p>
     * 
     * @param candidatePaths an array of candidate paths from which to load a
     *                      resource in order of preference
     * @return a URL corresponding to one of the candidatePaths or null if not
     *         found
     * @see #getPathURLFromResourceDir(String)
     * @see #getPathURLFromClasspath(String)
     */
    protected URL getPathURLFromResourceDirOrClasspath(final String[] candidatePaths) {
        URL url = null;
        for (int i = 0; i < candidatePaths.length; i++) {
            final String candidatePath = candidatePaths[i];
            url = getPathURLFromResourceDir(candidatePath);
            if (url == null) url = getPathURLFromClasspath(candidatePath);
            if (url != null) break;
        }
        return url;
    }

    /**
     * @param candidatePath a candidate path from which to load a resource
     * @return a URL corresponding the candidatePath or null if not found
     * @see XMLCDeferredParsingFactory#getPathURLFromResourceDir(String)
     */
    protected URL getPathURLFromResourceDir(final String candidatePath) {
        return fFactory.getPathURLFromResourceDir(candidatePath);
    }

    /**
     * @param candidatePath a candidate path from which to load a resource
     * @return a URL corresponding the candidatePath or null if not found
     * @see XMLCDeferredParsingFactory#getPathURLFromClasspath(String)
     */
    protected URL getPathURLFromClasspath(final String candidatePath) {
        return fFactory.getPathURLFromClasspath(candidatePath);
    }

}
