package org.enhydra.xml.xmlc.deferredparsing;

import java.util.Locale;

/**
 * Provides options for {@link org.enhydra.xml.xmlc.XMLObject}
 * {@link XMLCDeferredParsingFactory#create(XMLCCreateOptions) creation}.
 * <p>
 * Of the available options, at least <em>one</em> of the following three are
 * required to be set...
 * </p>
 * <dl>
 * <dt>Options to be used for deferred parsing</dt>
 * <dd>{@link #setXmlcClass(Class) xmlcClass}</dd>
 * <dd>{@link #setXmlcClassName(String) xmlcClassName}</dd>
 * <dt>Option to be used for dynamic loading</dt>
 * <dd>{@link #setDynMarkupFilePath(String) dynMarkupFilePath}</dd>
 * </dl>
 * <p>
 * All other options are entirely optional.  Options starting with "dyn", such as
 * {@link #setDynDomFactoryClassName(String) dynDomFactoryClassName} apply only
 * when dynamic loading is used.  All other options apply to both deferred parsing
 * and dynamic loading.
 * </p>
 */
public class XMLCCreateOptions {
    private Class xmlcClass;
    private String xmlcClassName, dynMarkupFilePath, dynDomFactoryClassName;//, dynMetaDataFilePath, defaultDynMetaDataFilePath;
    private Locale userLocale, fallbackLocale, defaultFallbackLocale;

    /**
     * @return the value applied to {@link #setXmlcClass(Class)} or null
     *         if not set
     */
    public Class getXmlcClass() {
        return xmlcClass;
    }

    /**
     * Use this option for deferred parsing; that is with build-time-generated
     * XMLC classes.  Alternatively, use {@link #setXmlcClassName(String)}.
     * This option is required unless one of
     * {@link #setXmlcClassName(String) xmlcClassName} or
     * {@link #setDynMarkupFilePath(String) dynMarkupFilePath} is set.
     * 
     * @param xmlcClass the XMLC class, either interface or implementation
     * @return the current object, fluent style
     * @see #setXmlcClassName(String)
     */
    public XMLCCreateOptions setXmlcClass(Class xmlcClass) {
        this.xmlcClass = xmlcClass;
        return this;
    }

    /**
     * @return the value applied to {@link #setXmlcClassName(String)} or null
     *         if not set
     */
    public String getXmlcClassName() {
        return xmlcClassName;
    }

    /**
     * Use this option for deferred parsing; that is with build-time-generated
     * XMLC classes.  Alternatively, use {@link #setXmlcClass(Class)}.
     * This option is required unless one of
     * {@link #setXmlcClass(Class) xmlcClass} or
     * {@link #setDynMarkupFilePath(String) dynMarkupFilePath} is set.
     * 
     * @param xmlcClassName the fully qualified name of the XMLC class, either
     *                      interface or implementation
     * @return the current object, fluent style
     * @see #setXmlcClass(Class)
     */
    public XMLCCreateOptions setXmlcClassName(String xmlcClassName) {
        this.xmlcClassName = xmlcClassName;
        return this;
    }

    /**
     * @return the value applied to {@link #setDynMarkupFilePath(String)} or
     *         null if not set
     */
    public String getDynMarkupFilePath() {
        return dynMarkupFilePath;
    }

    /**
     * Use this option for dynamic loading; that is without build-time-generated
     * XMLC classes.  This option is required unless one of
     * {@link #setXmlcClass(Class) xmlcClass} or
     * {@link #setXmlcClassName(String) xmlcClassName} is set.
     * 
     * @param dynMarkupFilePath the path to the source file, generally a
     *                          relative path (relative to resource dirs, the
     *                          classpath, etc...), trimmed if non-null
     * @return the current object, fluent style
     * @see #setDynDomFactoryClassName(String)
     */
    public XMLCCreateOptions setDynMarkupFilePath(String dynMarkupFilePath) {
        this.dynMarkupFilePath = dynMarkupFilePath != null ? dynMarkupFilePath.trim() : null;
        return this;
    }

    /**
     * @return the value applied to {@link #setDynDomFactoryClassName(String)}
     *         or null if not set
     */
    public String getDynDomFactoryClassName() {
        return dynDomFactoryClassName;
    }

    /**
     * @param dynDomFactoryClassName the fully qualified name of the DOM factory
     *                               to use, optional and only relevant when
     *                               using dynamic loading
     * @return the current object, fluent style
     * @see #setDynMarkupFilePath(String)
     */
    public XMLCCreateOptions setDynDomFactoryClassName(String dynDomFactoryClassName) {
        this.dynDomFactoryClassName = dynDomFactoryClassName;
        return this;
    }

    /*
     * @return the value applied to {@link #setDynMetaDataFilePath(String)}
     *         or null if not set
     */
//    public String getDynMetaDataFilePath() {
//        return dynMetaDataFilePath;
//    }

    /*
     * @param dynMetaDataFilePath the path to a metadata file relative to either
     *                            resource dirs or the classpath, either periods
     *                            (package style) or forward slashes (URL style)
     *                            may be used to define the path, optional and
     *                            only relevant when using dynamic loading
     * @return the current object, fluent style
     * @see #setDynMarkupFilePath(String)
     */
//    public XMLCCreateOptions setDynMetaDataFilePath(String dynMetaDataFilePath) {
//        this.dynMetaDataFilePath = dynMetaDataFilePath;
//        return this;
//    }

    /**
     * @return the value applied to {@link #setUserLocale(Locale)} or null if
     *         not set
     */
    public Locale getUserLocale() {
        return userLocale;
    }

    /**
     * @param userLocale the user locale, e.g., the Locale defined in an
     *                   HttpSerlvetRequest
     * @return the current object, fluent style
     * @see #setFallbackLocale(Locale)
     */
    public XMLCCreateOptions setUserLocale(Locale userLocale) {
        this.userLocale = userLocale;
        return this;
    }

    /**
     * @return the value applied to {@link #setFallbackLocale(Locale)} or null
     *         if not set
     */
    public Locale getFallbackLocale() {
        return fallbackLocale;
    }

    /**
     * @param fallbackLocale the fallback locale to be used if the user locale
     *                       is null or user locale-named markup file is not
     *                       found, i.e., the developer preferred fallback
     * @return the current object, fluent style
     * @see #setUserLocale(Locale)
     */
    public XMLCCreateOptions setFallbackLocale(Locale fallbackLocale) {
        this.fallbackLocale = fallbackLocale;
        return this;
    }

    //Methods below are intentionally package-private and are for internal use only

//    XMLCCreateOptions setDefaultDynMetaDataFilePath(String defaultDynMetaDataFilePath) {
//        this.defaultDynMetaDataFilePath = defaultDynMetaDataFilePath;
//        return this;
//    }

//    String getPrimaryDynMetaDataFilePath() {
//        return dynMetaDataFilePath != null ? dynMetaDataFilePath ? defaultDynMetaDataFilePath;
//    }

    XMLCCreateOptions setDefaultFallbackLocale(Locale defaultFallbackLocale) {
        this.defaultFallbackLocale = defaultFallbackLocale;
        return this;
    }

    Locale getPrimaryLocale() {
        return userLocale != null ? userLocale : fallbackLocale != null ? fallbackLocale : defaultFallbackLocale;
    }

    Locale getSecondaryLocale() {
        Locale tmpLocale = userLocale == null ? null : fallbackLocale != null ? fallbackLocale : defaultFallbackLocale;
        return tmpLocale == null ? null : tmpLocale.equals(getPrimaryLocale()) ? null : tmpLocale;
    }

}
