/*
 * Copyright (C) 2003, 2005, 2006, 2007, 2009, 2010 Richard Kunze, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.deferredparsing;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.XMLCRuntimeException;
import org.enhydra.xml.xmlc.XMLCStdFactory;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.commands.xmlc.XMLCOptionsParser;
import org.enhydra.xml.xmlc.compiler.Parse;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;
import org.enhydra.xml.xmlc.dom.XMLCDomFactoryCache;
import org.enhydra.xml.xmlc.metadata.DocumentClass;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.misc.I18nUtil;

/**
 * Factory class to create instances of an XMLC-generated class with
 * "deferred parsing". Deferred parsing means that the DOM is parsed
 * from a file on first access and re-parsed on subsequent accesses
 * whenever the DOM is out of date with regard to the source file.
 */
public class XMLCDeferredParsingFactory extends XMLCStdFactory {

    /** The suffix of meta file names */
    private static final String META_FILE_EXT = ".xmlc",
        XMLC_GENERATED_PACKAGE_PREFIX = "$$XMLC_GENERATED$$";

    /**
     * Provides ThreadLocal storage of two Locales within a double element
     * Locale[] array (used to reduce ThreadLocal lookups).
     * @since 2.3.2
     */
    private static final ThreadLocal threadLocalDoubleElementLocaleArray = new ThreadLocal();

    /**
     * Argument type for object constructor.
     */
    private static final Class[] CONSTRUCTOR_ARG_TYPES = {
        DocumentLoader.class
    };

    /**
     * Argument for object constructor.
     */
    private final Object[] fConstructorArgs;

    /**
     * The document loader, used here for both XMLObject construction and
     * resource loading.
     */
    private final DocumentLoader fDocLoader;

    /**
     * A list of resource paths to search for the source files
     */
    private List fResourceDirList = new LinkedList();
    
    /**
     * A list of package prefix to removed while searching source
     * files under the resource path in fResourceDirList
     */
    private List fPackagePrefixList = new LinkedList();

    /**
     * default meta data file. This file is used as the meta file for
     * the dynamically added new pages
     */
    private String fDefaultMetaDataFile = null;

    /**
     * The class loader to create and load class for newly added
     * template to the system without pre-compiled class
     */
    private final DynamicClassLoader fDynamicClassLoader;

    /**
     * Whether to perform extra lookups for markup files named in a way to mimic
     * ResourceBundle functionality.  By default, this is false to avoid incurring
     * a performance impact when such functionality is not needed nor desired.
     * @since 2.3.2
     */
    private boolean fEnableI18n;

    /**
     * The default fallback Locale, used in the case where I18n is enabled and
     * no fallbackLocale (or userLocale) is specified on the create(XMLCCreateOptions)
     * method.
     * @since 2.3.2
     */
    private Locale fDefaultFallbackLocale;

    /**
     * @param docLoader Document loader to use or <code>null</code>
     * @param classLoader Classloader used to load classes when a class name
     *  is specified.  If null, the system classload is used.
     * for the {@link StandardDocumentLoader}
     * @param logger XMLC logger or null for no logging.
     */
    public XMLCDeferredParsingFactory(DocumentLoader docLoader,
				      ClassLoader classLoader,
				      XMLCLogger logger) {
        super(classLoader, logger);
        if (docLoader == null) {
            docLoader = new StandardDocumentLoader();
        }
        fConstructorArgs = new Object[] { docLoader };
        fDocLoader = docLoader;
        fDynamicClassLoader = new DynamicClassLoader(classLoader, logger);
        // Initialize the document loader...
        fDocLoader.init(this);
    }

    /**
     * Create an object from a XMLC-generate class or subclass.
     * Don't call buildDocument();
     */
    private XMLObject createObject(Class xmlcBasedClass) {
        try {
            try {
                Constructor constructor = xmlcBasedClass.getConstructor(CONSTRUCTOR_ARG_TYPES);
                return (XMLObject)constructor.newInstance(fConstructorArgs);
            } catch (NoSuchMethodException e) {
                // Class has not been compiled for reloading - try to load it as a standalone class
                return (XMLObject)xmlcBasedClass.newInstance();
            }
        } catch (Exception except) {
            throw new XMLCRuntimeException(except);
        }
    }

    /**
     * Create a new document instance.
     */
    protected XMLObject doCreate(Class xmlcBasedClass)
            throws ClassNotFoundException, IllegalAccessException, InstantiationException {
        if (xmlcBasedClass.isInterface()) {
            // Convert interface to class name.
            return doCreate(xmlcBasedClass.getName() + "Impl",
                            getDefaultClassLoader());
        } else {
            if (!XMLObject.class.isAssignableFrom(xmlcBasedClass)) {
                throw new XMLCError("class \"" + xmlcBasedClass.getName()
                                    + "\" does not implement \""
                                    + XMLObject.class.getName() + "\"");
            }
            return createObject(xmlcBasedClass);
        }
    }

    /**
     * Do the actual work of creating a new object given a name.
     */
    protected XMLObject doCreate(String xmlcClassName,
				 ClassLoader classLoader)
	throws ClassNotFoundException, IllegalAccessException, InstantiationException {

        if (classLoader != null) {
	    fDynamicClassLoader.setDelegate(classLoader);
        } else {
	    fDynamicClassLoader.setDelegate(getClass().getClassLoader());
	}

	return doCreate(fDynamicClassLoader.loadClass(xmlcClassName));
    }

    /**
     * Creates an instance of an XMLC-generated class, or a class derived
     * from one, using the associated classloader, given one of the following...
     * <ol>
     * <li>{@link XMLCCreateOptions#setXmlcClass(Class) xmlcClass} (implementation or interface)</li>
     * <li>{@link #setXmlcClassName(String) xmlcClassName} (implementation or interface)</li>
     * <li>{@link #setDynMarkupFilePath(String) dynMarkupFilePath}</li>
     * </ol>
     * <p>
     * When using
     * {@link XMLCCreateOptions#setDynMarkupFilePath(String) dynamic loading},
     * one may define the optional
     * {@link XMLCCreateOptions#setDynDomFactoryClassName(String) dynDomFactoryClassName}
     * , define it in the default metadata, or just fall back to the
     * {@link XMLCDomFactoryCache} class default, in that order.
     * </p>
     * <p>
     * The optional {@link XMLCCreateOptions#setUserLocale(Locale) user} and
     * {@link XMLCCreateOptions#setFallbackLocale(Locale) fallback}
     * {@link Locale Locales} are utilized in choosing the markup file associated
     * with the class.  Locale-named markup files, if found, are used in
     * preference to the base markup file.  If the user Locale is null, the
     * fallback Locale is used in its place.  If the fallback Locale is null,
     * the {@link #setDefaultFallbackLocale defaultFallbackLocale} is used in
     * its place (as well as the user Locale if it is also null).
     * </p>
     * <p>
     * Markup files are looked up based first on user Locale and, if not found,
     * then on fallback Locale and, if not found, finally the base markup file
     * is used. 
     * </p>
     * <p>
     * Note that the Locales are ignored if {@link #isEnableI18n} returns false
     * or the Locales are null.
     * </p>
     * 
     * @param xmlcCreateOptions the XMLC create options providing information on
     *                          how to load the XMLObject
     * @return an XMLObject document instance
     * @see #setEnableI18n
     * @see #setDefaultFallbackLocale
     * @since 2.3.2
     */
    public XMLObject create(XMLCCreateOptions xmlcCreateOptions) {
        Class xmlcClass = xmlcCreateOptions.getXmlcClass();
        String xmlcClassName = xmlcCreateOptions.getXmlcClassName();

        if (xmlcClass == null && xmlcClassName == null) {
            String filePath = xmlcCreateOptions.getDynMarkupFilePath();
            filePath = filePath.replace('\\', '/');
            while (true) {
                if (filePath.startsWith("/")) {
                    filePath = filePath.substring(1);
                    continue;
                }
                break;
            }
            String classname = filePath.replace('/', '.');
            classname = classname.replace(' ', '_');
            classname = classname.replace('-', '_');
            classname = XMLC_GENERATED_PACKAGE_PREFIX + "." + classname;

            String domFactoryClassName = xmlcCreateOptions.getDynDomFactoryClassName();
            XMLCDomFactory domFactory = null;
            if (domFactoryClassName != null) {
                domFactory = XMLCDomFactoryCache.createFactory(domFactoryClassName, fDynamicClassLoader);
            } else {
                String[] paths = getXmlcSourceMetaDataFilePaths(filePath);
                URL metaDataURL = fDocLoader.getResourceLoader().getResource(paths);
                if (metaDataURL == null) {
                    metaDataURL = getDefaultMetaDataURL();
                }
                MetaData metaData = null;
                if (metaDataURL != null) {
                    try {
                        metaData = loadMetaData(metaDataURL, new ErrorReporter());
                    } catch (XMLCException e) {
                        //squash - TODO - should we report something here? - maybe log messages from the ErrorReporter?
                    }
                }
                boolean isHtmlDocument;
                if (metaData != null) {
                    try {
                        isHtmlDocument = !Parse.isXMLDocument(metaData) || (filePath.endsWith("html") || filePath.endsWith("htm"));
                    } catch (IOException ioe) {
                        isHtmlDocument = filePath.endsWith("html") || filePath.endsWith("htm");
                    }
                    DocumentClass documentClass = metaData.getDocumentClass();
                    domFactoryClassName = documentClass.getDomFactoryClass(isHtmlDocument);
                } else {
                    isHtmlDocument = filePath.endsWith("html") || filePath.endsWith("htm");
                }
                domFactory = XMLCDomFactoryCache.createFactory(domFactoryClassName, isHtmlDocument, fDynamicClassLoader);
            }

            xmlcClass = fDynamicClassLoader.createClass(classname, filePath, domFactory); 
        }

        Locale primaryLocale;
        if (fEnableI18n && (primaryLocale = xmlcCreateOptions.setDefaultFallbackLocale(fDefaultFallbackLocale).getPrimaryLocale()) != null) {
            final Locale[] l = new Locale[2];
            threadLocalDoubleElementLocaleArray.set(l);
            try {
                l[0] = primaryLocale;
                l[1] = xmlcCreateOptions.getSecondaryLocale();
                return xmlcClass != null ? create(xmlcClass) : create(xmlcClassName);
            } finally {
                threadLocalDoubleElementLocaleArray.set(null);
            }
        }

        return xmlcClass != null ? create(xmlcClass) : create(xmlcClassName);
    }

    /**
     * @see #createFromFile(String, String)
     * @see #create(XMLCCreateOptions)
     * @deprecated use {@link #create(XMLCCreateOptions) instead
     */
    public XMLObject createFromFile(String filename) {
        return createFromFile(filename, null);
    }

    /**
     * Same as
     * create(new XMLCCreateOptions().setMarkupFilePath(filename).setDomFactoryClassName(domFactoryClassName))
     * 
     * @see #create(XMLCCreateOptions)
     * @deprecated use {@link #create(XMLCCreateOptions) instead
     */
    public XMLObject createFromFile(String filename, String domFactoryClassName) {
        return create(new XMLCCreateOptions().setDynMarkupFilePath(filename).setDynDomFactoryClassName(domFactoryClassName));
    }

    /**
     * @param path a path to add as a resource directory
     */
    public void addResourceDir(String path) {
        synchronized(fResourceDirList) {
            File dir = new File(path);
            if (!fResourceDirList.contains(dir)) {
                try {
                    if (dir.exists() && dir.isDirectory()) {
                        fResourceDirList.add(dir);
                    }
                } catch (java.security.AccessControlException ace) {
                    //Don't allow file system lookup under a strict security
                    //manager, such as Google AppEngine, to bomb the application.
                    getLogger().logInfo(">>>Unable to add resource dir", ace);
                }
            }
        }
    }

    /**
     * @param path a path to remove as a resource directory
     */
    public void removeResourceDir(String path) {
        synchronized(fResourceDirList) {
            fResourceDirList.remove(path);
        }
    }

    /**
     * @return the resource directory list
     */
    List getResourceDirList() {
        return fResourceDirList;
    }

    /**
     * @param prefix a package prefix to add
     */
    public void addPackagePrefix(String prefix) {
        synchronized (fPackagePrefixList) {
            prefix = prefix.replace('.', '/');
            if (!fPackagePrefixList.contains(prefix)) {
                fPackagePrefixList.add(prefix);
            }
        }
    }

    /**
     * @param prefix a package prefix to remove
     * @since 2.3.2
     */
    public void removePackagePrefix(String prefix) {
        synchronized (fPackagePrefixList) {
            prefix = prefix.replace('.', '/');
            fPackagePrefixList.remove(prefix);
        }
    }

    /**
     * @param prefix a package prefix to remove
     * @deprecated
     * @see #removePackagePrefix
     */
    public void remove(String prefix) {
        removePackagePrefix(prefix);
    }

    /**
     * @return the package prefix list, the list of package prefixes is separated
     * with File.pathSeparator instead of '.'
     */
    List getPackagePrefixList() {
        return fPackagePrefixList;
    }

    /**
     * @param path the path to the meta data file
     */
    public void setDefaultMetaDataPath(String path) {
        fDefaultMetaDataFile = path;
    }

    /**
     * @return the default meta data file to be used by the dynamically added pages
     */
    public String getDefaultMetaDataFile() {
        return fDefaultMetaDataFile;
    }

    /** 
     * @return true if I18n is enabled, false if disabled
     * @see #setEnableI18n
     * @since 2.3.2
     */
    public final boolean isEnableI18n() {
        return fEnableI18n;
    }

    /**
     * Set the flag to enable Internationalization (I18n), or Localization (L10n).
     * When true, extra lookups are performed for markup files named in a way to
     * mimick ResourceBundle functionality.  By default, this is false to avoid
     * incurring a performance impact when such functionality is not needed nor desired.
     * <p>
     * Note it's capital "I", and it's "I18n", not "Il8n" (Number "1", not letter "l").
     * See http://www.i18nguy.com/origini18n.html
     * </p>
     * 
     * @param enableI18n true to enable, false to disable (default)
     * @see #getLocales()
     * @since 2.3.2
     */
    public final void setEnableI18n(final boolean enableI18n) {
        this.fEnableI18n = enableI18n;
    }

    /**
     * @param defaultFallbackLocale the default fallback Locale, used in the case
     *                              where I18n is enabled and no fallbackLocale
     *                              (or user locale) is specified on the
     *                              create(XMLCCreateOptions) method.
     * @see #setEnableI18n
     * @since 2.3.2
     */
    public final void setDefaultFallbackLocale(Locale defaultFallbackLocale) {
        this.fDefaultFallbackLocale = defaultFallbackLocale;
    }

    /**
     * Meant to provide the DocumentLoader instance, associated with the current
     * instance of this class, access to user-defined Locale objects (user and
     * developer-preferred fallback) for use in loading localized markup files.
     * <p>
     * Note that this method is guaranteed to return null unless
     * {@link #setEnableI18n setEnableI18n(true)} is called.
     * </p>
     * <p>
     * Note also that the value returned by this method is time/unit-of-work
     * context dependent on a per/thread basis.  The value will be null when
     * called out of context or even when in context but no non-null user-defined
     * Locales are provided.  As such, take care to check for null prior to using
     * the returned value and, of course, don't cache it either.
     * </p>
     * 
     * @return the user-defined Locales ([0] == locale, [1] == fallbackLocale),
     *         available for the duration of calls to
     *         {@link #create(XMLCCreateOptions)}
     * invocation on a per/thread thread basis.
     * @see #setEnableI18n
     * @see #setDefaultFallbackLocale
     * @since 2.3.2
     */
    public final Locale[] getLocales() {
    	return fEnableI18n ? ((Locale[]) threadLocalDoubleElementLocaleArray.get()) : null;
    }

    /**
     * @param metaDataURL the URL of the metadata file
     * @param errorReporter used to report metadata parse errors
     * @return the metadata represented by the URL
     * @throws XMLCException if something goes wrong loading metadata
     * @since 2.3.2
     */
    public final MetaData loadMetaData(URL metaDataURL, ErrorReporter errorReporter) throws XMLCException {
        try {
            //NOTE: Use of "-version" gets around an implementation detail of
            //XMLCOptionsParser.parse() where no filename is required if
            //"-version" is provided.  Might want to revisit this later to make
            //it less hackish.
            return new XMLCOptionsParser().parse(new String[]{"-version", metaDataURL.toString()}, errorReporter, fDynamicClassLoader);
        } catch (IOException ioe) {
            throw new XMLCException(ioe);
        }
    }

    /**
     * @return the default metadata URL, accounting for configured package
     *         prefixes, with dots normalized to forward slashes and looked up
     *         by the configured {@link DocumentLoader#getResource(String[]) document loader}
     *         implementation
     * @since 2.3.2
     * @see #getDefaultMetaDataFile()
     * @see #getPathsFromPackagePrefixes(String)
     */
    public final URL getDefaultMetaDataURL() {
        String defaultMetaFile = getDefaultMetaDataFile();
        if (defaultMetaFile == null) return null;
        if (defaultMetaFile.indexOf('/') == -1) {
            //convert dots to slashes while preserving the file extension
            int lastIndex = defaultMetaFile.lastIndexOf(".");
            if (lastIndex != -1) {
                String defaultPath = defaultMetaFile.substring(0, lastIndex);
                defaultPath = defaultPath.replace('.', '/');
                defaultMetaFile = defaultPath + defaultMetaFile.substring(lastIndex);
            }
        }
        String[] defaultPaths = getPathsFromPackagePrefixes(defaultMetaFile);
        return fDocLoader.getResourceLoader().getResource(defaultPaths);
    }

    /**
     * Get the metadata file paths.  Provides an array containing possible paths
     * to an XMLC class metadata file including every combination of paths
     * corresponding to configured package prefixes.  Paths represent metadata
     * files for both interface and "Impl" classes as well as source files.  This
     * covers all cases of XMLC metadata file generation.
     * <p>
     * An ascending sort is applied resulting in a consistent order when dealing
     * with configured package prefix paths and interface -vs- "Impl" extensions.
     * </p>
     * 
     * @param xmlcClassMetaDataFilePath a relative path (no leading slash) to
     *                                  the meta data file for an XMLC class
     * @return a sorted array of possible paths to an XMLC class metadata file
     *         accounting for interface and Impl classes, both within configured
     *         package prefixes
     * @since 2.3.2
     * @see #getXmlcClassMetaDataFilePath(Class)
     * @see #getXmlcSourceMetaDataFilePaths(String)
     * @see #getPathsFromPackagePrefixes(String)
     */
    public final String[] getXmlcClassMetaDataFilePaths(final Class docClass) {        
        String xmlcClassMetaDataFilePath = getXmlcClassMetaDataFilePath(docClass);

        String[] Aa = getXmlcSourceMetaDataFilePaths(getSourceFilePath(docClass));
        if (xmlcClassMetaDataFilePath.startsWith(XMLC_GENERATED_PACKAGE_PREFIX)) {
            return Aa;
        }
        String[] A = getPathsFromPackagePrefixes(xmlcClassMetaDataFilePath);
        String[] B = getPathsFromPackagePrefixes(xmlcClassMetaDataFilePath.substring(0, xmlcClassMetaDataFilePath.lastIndexOf(META_FILE_EXT)) + DocumentClass.IMPLEMENTATION_SUFFIX + META_FILE_EXT);
        String[] C = new String[A.length + B.length + Aa.length];
        System.arraycopy(A, 0, C, 0, A.length);
        System.arraycopy(B, 0, C, A.length, B.length);
        System.arraycopy(Aa, 0, C, A.length + B.length, Aa.length);
        Arrays.sort(C);
        return C;
    }

    /**
     * Get the metadata file paths.  Provides an array containing possible paths
     * to an XMLC source (as opposed to class) metadata file including every
     * combination of paths corresponding to configured package prefixes.
     * 
     * @param xmlcSourceFilePath a relative path (no leading slash) to a source
     *                           markup file
     * @return an array of possible paths to an XMLC source metadata file,
     *         corresponding to the provided xmlcSourceFilePath, accounting for
     *         configured package prefixes
     * @since 2.3.2
     * @see #getPathsFromPackagePrefixes(String)
     */
    public final String[] getXmlcSourceMetaDataFilePaths(final String xmlcSourceFilePath) {
        String xmlcSourceMetaDataFilePath = xmlcSourceFilePath.substring(0, xmlcSourceFilePath.lastIndexOf('.')) + META_FILE_EXT;
        return getPathsFromPackagePrefixes(xmlcSourceMetaDataFilePath);
    }

    /**
     * Get one of the compiled-in constants in a XMLC-generated class (or
     * subclass).
     *
     * @since 2.3.2
     */
    public static Object getClassConstant(Class xmlcBasedClass,
                                          String constName) {
        // Use generated constant to find the class.
        try {
            Field field = xmlcBasedClass.getField(constName);
            return field.get(null);
        } catch (NoSuchFieldException except) {
            throw new XMLCRuntimeException("Couldn't find class constant \""
                                           + constName
                                           + "\" in class \""
                                           + xmlcBasedClass.getName()
                                           + "\", maybe be a XMLC generated class or descendent",
                                           except);
        } catch (IllegalAccessException except) {
            throw new XMLCRuntimeException(except);
        }
    }

    /** 
     * Get the name of the source file associated with <code>docClass</code> 
     *
     * @since 2.3.2
     */
    public static String getSourceFilePath(Class docClass)
        throws XMLCRuntimeException {
        String srcPath =
            (String) getClassConstant(docClass,
                                      XMLObject.XMLC_SOURCE_FILE_FIELD_NAME);
        if (srcPath == null) {
            throw new XMLCRuntimeException(XMLObject.XMLC_SOURCE_FILE_FIELD_NAME
                                           + " is null in class "
                                           + docClass.getName());
        }
        if (srcPath.startsWith("/")) {
            srcPath = srcPath.substring(1);
        }
        return srcPath;
    }

    /**
     * Get possible file paths by subtracting each respective package prefix
     * from the path prefix.  The original path is included last in the returned
     * value.  Both locale and fallbackLocale are accounted for in the paths
     * except when the path represents an XMLC metadata file.
     * 
     * @param path a relative path (no leading slash) to either a source markup
     *             file or XMLC metadata file
     * @return an array of possible paths to a file, accounting for configured
     *         package prefixes
     * @since 2.3.2
     */
    public final String[] getPathsFromPackagePrefixes(String path) {
        List list = new ArrayList();
        String suffix = null, pathLessSuffix = null;
        Locale[] locales = null;
        boolean doLocalizePaths = isEnableI18n() && !path.endsWith(META_FILE_EXT);
        if (doLocalizePaths) {
            locales = getLocales();
            if (locales != null) suffix = path.substring(path.lastIndexOf('.'));
        }
        for (Iterator iter = getPackagePrefixList().iterator(); iter.hasNext();) {
            String prefix = (String) iter.next();
            if (path.startsWith(prefix)) {
                String candidatePath = path.substring(prefix.length() + 1);
                if (doLocalizePaths && locales != null) {
                    pathLessSuffix = candidatePath.substring(0, candidatePath.lastIndexOf('.'));
                    for (int i = 0; i < locales.length; i++) {
                        Locale locale = locales[i];
                        if (locale != null) {
                            list.addAll(I18nUtil.localeStrings(pathLessSuffix, suffix, locale));
                        }
                    }
                }
                list.add(candidatePath);
                break;
            }
        }
        //Add the full path last so that package prefix paths can be evaluated first
        if (doLocalizePaths && locales != null) {
            pathLessSuffix = path.substring(0, path.lastIndexOf('.'));
            for (int i = 0; i < locales.length; i++) {
                Locale locale = locales[i];
                if (locale != null) {
                    list.addAll(I18nUtil.localeStrings(pathLessSuffix, suffix, locale));
                }
            }
        }
        list.add(path);
        return (String[]) list.toArray(new String[list.size()]);
    }

    /**
     * @param path the relative path to the source file
     * @return the URL for a path relative to a configured resource dir
     * @since 2.3.2
     */
    public final URL getPathURLFromResourceDir(String path) {
        File srcFile = null;
    
        for (Iterator iter = getResourceDirList().iterator(); iter.hasNext();) {
            File dir = (File) iter.next();
            srcFile = new File(dir, path);
            /* checking under dir using long path */
            if (srcFile.exists()) {
                break;
            }
            srcFile = null;
        }
    
        if (srcFile != null) {
            try {
                URL srcURL = srcFile.toURL(); //TODO - replicate file.toURI().toURL() (for proper escaping), only jdk1.3 compatible
                if (getLogger().debugEnabled()) {
                    getLogger().logDebug(">>>Get document '"
                                         + srcURL
                                         + "' from resource dir");
                }
                return srcURL;                
            } catch (MalformedURLException e) {
                throw new XMLCError("Cannot construct URL for file " + srcFile, e);
            }
        }
        return null;        
    }

    /**
     * @param path the fully qualified path to the source file (with no leading slashes)
     * @return the URL for a path relative to the classpath
     * @since 2.3.2
     */
    public final URL getPathURLFromClasspath(String path) {
        URL srcURL = fDynamicClassLoader.getResource(path);
    
        if ((srcURL != null) && getLogger().debugEnabled()) {
            getLogger().logDebug(">>>Get document '"
                                 + srcURL
                                 + "' from classpath");
        }
        return srcURL;
    }

    /**
     * @param xmlcClass an XMLC class
     * @return a path corresponding to to the FQCN of the provided XMLC class,
     *         with periods replaced by slashes and ".class" replaced by ".xmlc",
     *         representing the metadata path (less the "Impl" extension if it exists)
     * @since 2.3.2
     */
    public static String getXmlcClassMetaDataFilePath(Class xmlcClass) {
        //Convert class to file path, drop Impl and add extension.
        String fileName = xmlcClass.getName().replace('.', '/');
        if (fileName.endsWith(DocumentClass.IMPLEMENTATION_SUFFIX)) {
            fileName =
                fileName.substring(0,
                                   fileName.length()
                                   - DocumentClass.IMPLEMENTATION_SUFFIX.length());
        }
        return fileName + META_FILE_EXT;
    }

}
