/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLCDocument.java,v 1.5 2006-08-10 23:57:26 jkjome Exp $
 */

package org.enhydra.xml.xmlc.dom;

import java.util.TreeMap;
import java.util.TreeSet;

import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.Encodings;
import org.enhydra.xml.xmlc.XMLCException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.html.HTMLDocument;

/*
 * NB: This has some HTML specific data in it.  At one time, it was split
 * into two classes, however it proved simpler to store the data here.
 *
 * NB: The DTD data kept here is specialized to collect the minimum amount of
 * information on elements and attributes needed by XMLC.  The API reflects
 * this, which is a kludge, however its only used internally.  This should
 * be more general and interface to the DTD/Schema; however until this
 * information is actually part of the DOM spec, the parsers just
 * fill in the minimum.
 */


/**
 * Class that is a container for the parsed Document DOM. It contains
 * information obtained from that parsers that is not part of the DOM. It also
 * provides various operations on that document.
 */
final public class XMLCDocument {
    /**
     * The DTD-specific DOM Factory.
     */
    private XMLCDomFactory fDomFactory;

    /**
     * Is this an HTML document we are dealing with?
     */
    private boolean fIsHtmlDocument;

    /**
     * Is this a HTML frameset?
     */
    private boolean fIsHtmlFrameSet;
    
    /**
     * Information from XML header.
     */
    private String fXMLVersion;
    private String fEncoding;

    /**
     * The DOM Document.
     */
    private Document fDocument;

    /**
     * The DOM Document type (null for HTML).
     */
    private DocumentType fDocType;

    /**
     * Table of element to id attribute name mappings.
     */
    private TreeMap fElementIdAttrs = new TreeMap();

    /**
     * Table of id attribute names.
     */
    private TreeSet fIdAttrNames = new TreeSet();

    /**
     * Last id attribute name; provides easy access to fIdAttrNames
     * when there is only one.
     */
    private String fLastIdAttrName;

    /**
     * Table of elements that can have PCDATA children.
     */
    private TreeSet fPCDataElements = new TreeSet();

    /**
     * Constructor.  A document is not initially associated with this
     * object.
     *
     * @param domFactory The DOM factory to create and get information about
     * DOMs.
     * @param isHtmlDocument Is this an HTML document?
     */
    public XMLCDocument(XMLCDomFactory domFactory) throws XMLCException {
        fDomFactory = domFactory;
    }

    /**
     * Get the DOM factory associated with this object.
     */
    public XMLCDomFactory getDomFactory() {
        return fDomFactory;
    }

    /**
     * Creates an empty <code>DocumentType</code> node.
     */
    public DocumentType createDocumentType(String qualifiedName,
                                           String publicID,
                                           String systemID,
                                           String internalSubset) {
        fDocType = fDomFactory.createDocumentType(qualifiedName, 
                                                  publicID, systemID,
                                                  internalSubset);
        return fDocType;
    }

    /**
     * Create the document.  All arguments are null for HTML.
     */
    public Document createDocument(String namespaceURI,
                                   String qualifiedName) {
        fDocument = fDomFactory.createDocument(namespaceURI, qualifiedName,
                                               fDocType);
        fIsHtmlDocument = (fDocument instanceof HTMLDocument);
        return fDocument;
    }

    /**
     * Get the document associated with this object.
     */
    public Document getDocument() {
        return fDocument;
    }

    public void setDocument(Document doc) {
        fDocument = doc;
        fIsHtmlDocument = (fDocument instanceof HTMLDocument);
    }

    /*
     * Is this an HTML document?
     */
    public boolean isHtmlDocument() {
        return fIsHtmlDocument;
    }

    /**
     * Is the document as a HTML frameset?
     */
    public boolean isHtmlFrameSet() {
        return fIsHtmlFrameSet;
    }

    /**
     * Flat the document as a HTML frameset.
     */
    public void setIsHtmlFrameSet() {
        fIsHtmlFrameSet = true;
    }

    /**
     * Set the XML version.
     *
     * @param xmlVersion XML version string.
     */
    public void setXMLVersion(String xmlVersion) {
        fXMLVersion = xmlVersion;
    }

    /**
     * Get the XML version.
     */
    public String getXMLVersion() {
        return fXMLVersion;
    }

    /*
     * Get the default encoding or null if none.
     */
    public String getEncoding() {
        return fEncoding;
    }

    /*
     * Set the default encoding.  This will be converted to the MIME-prefered
     * name if not null.
     */
    public void setEncoding(String encoding) {
        if (encoding == null) {
            fEncoding = null;
        } else {
            Encodings encodings = Encodings.getEncodings();
            String mimeEncoding = encodings.getMIMEPreferred(encoding);
            if (mimeEncoding != null) {
                fEncoding = mimeEncoding;
            } else {
                fEncoding = encoding;
            }
        }
    }

    /**
     * Flag an element as having #PCDATA as part of its content model.
     *
     * @param elementName The name of the element.
     */
    public void addPCDataContentElement(String elementName) {
        fPCDataElements.add(elementName);
    }

    /**
     * Define an element id attribute.  For XML, all ID attributes
     * must be defined, even if not used.  This is need to make
     * global id attribute mode work.
     *
     * @param elementName The name of the element.
     * @param attributeName The name of the ID attribute.
     */
    public void addIdAttribute(String elementName,
                               String attributeName) {
        fElementIdAttrs.put(elementName, attributeName);
        if (!(attributeName.equals(fLastIdAttrName)
              || fIdAttrNames.contains(attributeName))) {
            fIdAttrNames.add(attributeName);
            fLastIdAttrName = attributeName;
        }
    }

    /**
     * Convert an implementation-specific DOM node class name to the
     * external interface or class name.
     *
     * @see XMLCDomFactory
     */
    public String nodeClassToInterface(Node node) {
        return fDomFactory.nodeClassToInterface(node);
    }

    /**
     * Get the global id attribute name.  A DTD supports global id mode if a
     * single attribute name is used for all elements that support ids and
     * never used as a non-id attribute name.  This is intended for the HTML
     * `id' attribute.  However, if a DTD conforms to this requirement, more
     * efficient code for cloning the object can be generated.   It has been
     * suggested that `id' be a defacto standard for XML as well.
     *
     * This method will not be called until after the DTD has been parsed, so
     * this can be determined.
     *
     * See the <A href="http://www.xml.com/axml/axml.html">
     * XML Annotated Specification</A>.
     *
     * @return The global id attribute name or null if there is not a global
     * ID attribute.
     */
    public String getGlobalIdAttribute() {
        if (fIsHtmlDocument) {
            return "id";
        } else {
            if (fIdAttrNames.size() == 1) {
                return fLastIdAttrName;
            } else {
                return null;
            }
        }
    }

    /**
     * Get the id attribute name for an element.  The attribute
     * doesn't need to actually exist in this element, this just gets
     * the name for the attribute.
     *
     * @return The attribute name or null if the element does not have
     *  an id attribute.
     */
    public String getIdAttrName(Element element) {
        if (fIsHtmlDocument) {
            return "id";
        } else {
            return (String)fElementIdAttrs.get(element.getTagName());
        }
    }

    /**
     * Extract the unique id for an element object.  This is specified with
     * an attribute of type <CODE>ID</CODE>.
     *
     * @return The id or null if the element does not have an id.
     */
    public String getElementId(Element element) {
        String idAttrName = getIdAttrName(element);
        if (idAttrName == null) {
            return null;  // element doesn't have an id attribute
        }
        String id = element.getAttribute(idAttrName);
        if ((id == null) || (id.length() == 0)) {
            return null;  // No id.
        }
        return id;
    }

    /**
     * Extract the class names for an element.
     * @see XMLCDomFactory#getElementClassNames
     */
    public String[] getElementClassNames(Element element) {
        return fDomFactory.getElementClassNames(element);
    }
    
    /**
     * Extract the name for an element.  This is a non-unique name for an
     * element.  In HTML, the name is specified with the <CODE>name</CODE>
     * attribute.  It's not supported for XML.
     *
     * @return The element name or null if the node has no name.
     *  XML returns null.
     */
    public String getElementName(Element element) {
        if (fIsHtmlDocument) {
            String name = element.getAttribute("name");
            if (name.length() == 0) {
                return null;
            } else {
                return name;
            }
        } else {
            return null;
        }
    }

    /**
     * Determine if an an attribute of an element may contain a URL.
     * @see XMLCDomFactory#isURLAttribute
     */
    public boolean isURLAttribute(Element element,
                                  String attrName) {
        return fDomFactory.isURLAttribute(element, attrName);
    }

    /**
     * Check if an element has #PCDATA as part of it's content model.
     */
    public boolean hasPCDataInContentModel(Element element) {
        return fPCDataElements.contains(element.getTagName());
    }

    /**
     * Convert the document to a string representation of the document.  The
     * results are parsable into the same DOM hierarchy
     */
    public String toDocument() {
        return new DOMFormatter(DOMFormatter.getDefaultOutputOptions(getDocument())).toString(getDocument());
    }
}
