/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDomFactory.java,v 1.5 2006-09-04 05:56:06 jkjome Exp $
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.lazydom.LazyDocument;
import org.enhydra.xml.lazydom.LazyDocumentType;
import org.enhydra.xml.lazydom.TemplateDOM;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;
import org.enhydra.xml.xmlc.dom.xerces.XercesDomFactory;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Node;

/**
 * Lazy DOM factory for creating DocumentType and Document objects.
 */
public class LazyDomFactory extends XercesDomFactory
    implements XMLCDomFactory {
    /*
     * Prefix attached to interface name to form the implementation names
     * (not use on HTML DOM).
     */
    private static final String IMPL_CLASS_PREFIX = "Lazy";

    /*
     * Suffix attached to interface name to form the implementation names
     * for the HTML DOM.
     */
    private static final String IMPL_CLASS_SUFFIX = "Impl";

    /*
     * Lazy DOM to W3C package name map.
     * WARNING: Order is important here to handle most significant match.
     */
    private static final String[][] DOM_PACKAGES_DOT_MAP = {
        {"org.enhydra.xml.lazydom.html.", "org.w3c.dom.html."},
        {"org.enhydra.xml.lazydom.", "org.w3c.dom."},
    };

    /*
     * Special-case element names to interface name that can't be
     * mapped just by changing the package.
     */
    private static final String[][] CLASS_INTERFACE_MAP = {
        {"org.enhydra.xml.lazydom.LazyElementNS", "org.w3c.dom.Element"},
        {"org.enhydra.xml.lazydom.LazyElementNoNS", "org.w3c.dom.Element"},
        {"org.enhydra.xml.lazydom.LazyAttrNS", "org.w3c.dom.Attr"},
        {"org.enhydra.xml.lazydom.LazyAttrNoNS", "org.w3c.dom.Attr"},
    };

    /**
     * @see XMLCDomFactory#createDocumentType
     */
    public DocumentType createDocumentType(String qualifiedName,
                                           String publicId, 
                                           String systemId,
                                           String internalSubset) {
    	return new LazyDocumentType(null, null, qualifiedName,
                                    publicId, systemId,
                                    internalSubset);
    }

    /**
     * @see XMLCDomFactory#createDocument
     */
    public Document createDocument(String namespaceURI,
                                   String qualifiedName,
                                   DocumentType docType) {
        Document doc = new LazyDocument(docType, null);
        // Create the document element
        doc.appendChild(doc.createElementNS(namespaceURI, qualifiedName));
        return doc;
    }

    /**
     * Create a LazyDocument from a template.
     * @see XMLCDomFactory#createDocument
     */
    public Document createDocument(TemplateDOM templateDOM) {
        return new LazyDocument(null, templateDOM);
    }

    /**
     * @see XMLCDomFactory#getDocumentClassName
     */
    public String getDocumentClassName() {
        return LazyDocument.class.getName();
    }

    /**
     * @see XMLCDomFactory#nodeClassToInterface
     */
    public String nodeClassToInterface(Node node) {
        String className = node.getClass().getName();

        // Search explict class name mappings.
        for (int idx = 0; idx < CLASS_INTERFACE_MAP.length; idx++) {
            if (className.equals(CLASS_INTERFACE_MAP[idx][0])) {
                return CLASS_INTERFACE_MAP[idx][1];
            }
        }

        // Search for package name mapping
        int packageIdx;
        for (packageIdx = 0; packageIdx < DOM_PACKAGES_DOT_MAP.length; packageIdx++) {
            if (className.startsWith(DOM_PACKAGES_DOT_MAP[packageIdx][0])) {
                break;
            }
        }
        if (packageIdx == DOM_PACKAGES_DOT_MAP.length) {
            throw new XMLCError("Can't determine DOM interface for node of class \""
                                + className
                                + "\" (maybe be mismatch between XMLCDomFactory DOM implementation)");
        }

        String unqualClassName = className.substring(DOM_PACKAGES_DOT_MAP[packageIdx][0].length());

        // Remove the prefix and suffix to get the base class name.
        int prefixIdx = unqualClassName.lastIndexOf(IMPL_CLASS_PREFIX);
        if (prefixIdx >= 0) {
            unqualClassName = unqualClassName.substring(IMPL_CLASS_PREFIX.length());
        }
        int suffixIdx = unqualClassName.lastIndexOf(IMPL_CLASS_SUFFIX);
        if (suffixIdx >= 0) {
            unqualClassName = unqualClassName.substring(0, suffixIdx);
        }
        
        return DOM_PACKAGES_DOT_MAP[packageIdx][1] + unqualClassName;
    }

}
