/*
 * Copyright (C) 2009  Jacob Kjome [hojuATvisi.com]
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: I18nUtil.java 204 2008-07-23 08:36:28Z jkjome $
 */
package org.enhydra.xml.xmlc.misc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import jregex.Matcher;
import jregex.Pattern;


public class I18nUtil {

    private static final String GRP_language = "language", GRP_country = "country", GRP_variant = "variant";
    private static final Pattern RE_LOCALE_GROUPER = new Pattern("^[\\s_]*?(?:_?({" + GRP_language + "}[a-z]{2})_?)?(?:_?(?:({" + GRP_country + "}[A-Z]{2})_?)?(?:_+?({" + GRP_variant + "}\\S+?))?)?\\s*?$");

    /**
     * @param baseNameLessSuffix the base name path, less any suffix, defaults
     *                           to empty string if null
     * @param suffix the suffix to append to the baseNameLessSuffix + Locale
     *               information, e.g., ".html".  May be null, in which case it
     *               is ignored.
     * @param locale the Locale
     * @return the canonical, or longest, string generated based on the given
     *         Locale information in the form
     *         baseNameLessSuffix + "_[language]_[country]_[variant]" + suffix,
     *         or null if locale is empty, i.e., new Locale("", "", "")
     * @throws NullPointerException if locale is null
     */
    public static String canonicalLocaleString(String baseNameLessSuffix, final String suffix, final Locale locale) {
    	try {
    	    return (String) localeStrings(baseNameLessSuffix, suffix, locale).get(0);
    	} catch (final IndexOutOfBoundsException e) {
    	    return null;
    	}
    }

    /**
     * @param baseNameLessSuffix the base name path, less any suffix, defaults
     *                           to empty string if null
     * @param suffix the suffix to append to the baseNameLessSuffix + Locale
     *               information, e.g., ".html".  May be null, in which case it
     *               is ignored.
     * @param locale the Locale
     * @return a list of Locale-contextualized strings in the form
     *         baseNameLessSuffix + "_[language]_[country]_[variant]" + suffix,
     *         max size 3 elements, in the order of longest path to shortest path.
     * @throws NullPointerException if locale is null
     */
    public static List localeStrings(String baseNameLessSuffix, final String suffix, final Locale locale) {
        final String l = locale.getLanguage()
            , c = locale.getCountry()
            , v = locale.getVariant();
        final int lLen = l.length()
            , cLen = c.length()
            , vLen = v.length();

        final List paths = new ArrayList(3);
        if (lLen + cLen + vLen == 0) {
            return paths;
        }

        if (baseNameLessSuffix == null) baseNameLessSuffix = "";
        //Add 22 to cover the extra long ResourceBundle Javadoc example... "_es_ES_Traditional_WIN"
        final StringBuffer buff = new StringBuffer(baseNameLessSuffix.length() + 22).append(baseNameLessSuffix);
        final boolean suffixNotNull = suffix != null;
        String tmpString;

        buff.append('_');
        buff.append(l);
        tmpString = buff.toString();
        if (suffixNotNull) tmpString += suffix;
        paths.add(tmpString);
        if (cLen + vLen == 0) {
            return paths;
        }
        buff.append('_');
        buff.append(c);
        tmpString = buff.toString();
        if (suffixNotNull) tmpString += suffix;
        paths.add(0, tmpString);
        if (vLen == 0) {
            return paths;
        }
        buff.append('_');
        buff.append(v);
        tmpString = buff.toString();
        if (suffixNotNull) tmpString += suffix;
        paths.add(0, tmpString);

        return paths;
    }

    /**
     * Translates a Locale String into a Locale, e.g....
     * <p>
     * "_en", "en", "en_", "en__", "_en__", "en_US", "en__US", "en_POSIX",
     * "en__POSIX", "en___POSIX", "en_US_POSIX", "en__US__POSIX", "US", "US_",
     * "_US", "_US_", "US_POSIX", "US__POSIX", "_US_POSIX", "_US__POSIX", "_POSIX",
     * "__POSIX"
     * </p>
     * <p>
     * Matches any Locale.toString() output plus more using the following regular
     * expression...
     * </p>
     * <pre>
     * ^[\s_]*?(?:_?([a-z]{2})_?)?(?:_?(?:([A-Z]{2})_?)?(?:_+?(\S+?))?)?\s*?$
     * --OR--
     * ^[\s_]*?(?:_?({language}[a-z]{2})_?)?(?:_?(?:({country}[A-Z]{2})_?)?(?:_+?({variant}\S+?))?)?\s*?$
     * </pre>
     * <p>
     * Preceding and trailing whitespace is ignored.  Single preceding
     * underscores, connected or not to locale text, are ignored (though at least
     * one is required to define a stand-alone variant).  Two preceding contiguous
     * underscores define a variant when either alone or connected to language
     * or country locale text (or variant, obviously). Trailing underscores
     * connected to the locale text are ignored unless part of the variant:
     * three contiguous underscores trailing language, two contiguous underscores
     * trailing country, or any trailing variant, e.g., ...
     * </p>
     * <ul>
     * <li>" _ _ _ " --> match, but empty locale which will be ignored</li>
     * <li>" _ _ __ " --> variant: "_"</li>
     * <li>" _ _ _en__ _ " --> no match</li>
     * <li>" _ _ _en__ " --> language: "en"</li>
     * <li>" _ _ _en___ " --> language: "en", variant: "_"</li>
     * <li>" _ _ _en___POSIX " --> language: "en", variant: "POSIX"</li>
     * <li>" _ _ _en____POSIX " --> language: "en", variant: "_POSIX"</li>
     * <li>" _ _ __en_ " --> variant: "en_"</li>
     * <li>" _ _ _US_ _ " --> no match</li>
     * <li>" _ _ _US_ " --> country: "US"</li>
     * <li>" _ _ _US__ " --> country: "US", variant: "_"</li>
     * <li>" _ _ _US__POSIX " --> country: "US", variant: "POSIX"</li>
     * <li>" _ _ _US___POSIX " --> country: "US", variant: "_POSIX"</li>
     * <li>" _ _ __US_ " --> variant: "US_"</li>
     * <li>" _ _ POSIX_ " --> no match</li>
     * <li>" _ _ _POSIX_ _ " --> no match</li>
     * <li>" _ _ _POSIX_ " --> variant: "POSIX_"</li>
     * <li>" _ _ __POSIX_ " --> variant: "POSIX_"</li>
     * <li>" _ _ ___POSIX_ " --> variant: "_POSIX_"</li>
     * </ul>
     * 
     * @param localeStr a String representing a Locale
     * @return the Locale corresponding to the provided localeStr or null if the
     *         localeStr is null or doesn't match any Locale parts
     */
    public static Locale locale(final String localeStr) {
        if (localeStr == null) return null;
        final Matcher m = RE_LOCALE_GROUPER.matcher(localeStr);
        if (m.find()) {
            final Locale usrLocale = new Locale(emptyIfNull(m.group(GRP_language)), emptyIfNull(m.group(GRP_country)), emptyIfNull(m.group(GRP_variant)));
            if (usrLocale.toString().length() > 0 || usrLocale.getVariant().length() > 0) {
                return usrLocale;
            }
        }
        return null;
    }

    /**
     * Transforms any String literal into a valid regex pattern by escaping
     * reserved regex characters.
     * 
     * <p>
     * In lieu of (PatTwo supports this, but not JRegex)...
     * String literal = "^some $ literal \\ [string] {containing} (various) * <regex> ? special | meta + characters.";
     * literal = (literal = new Pattern("(\\S+)\\s*").replacer("\\\\Q$1\\\\E\\\\s+").replace(literal)).substring(0, literal.lastIndexOf("\\s+"));
     * System.err.println(literal); //yields... \Q^some\E\s+\Q$\E\s+\Qliteral\E\s+\Q\\E\s+\Q[string]\E\s+\Q{containing}\E\s+\Q(various)\E\s+\Q*\E\s+\Q<regex>\E\s+\Q?\E\s+\Qspecial\E\s+\Q|\E\s+\Qmeta\E\s+\Q+\E\s+\Qcharacters.\E
     * </p>
     * 
     * @param literal a String literal
     * @return a valid regex pattern translation of the String literal 
     */
    public static String escapeRegexMetaChars(final String literal) {
    	final int literalLen = literal != null ? literal.length() : 0;
    	if (literalLen == 0) return literal;
    	final int[] regexMetaChars = {'\\', '^', '$', '(', ')', '<', '>', '[', '{', '|', '.', '*', '+', '?'};
    	final List idxs = new ArrayList();
    	for (int i = 0; i < regexMetaChars.length; i++) {
    	    final int regexMetaChar = regexMetaChars[i];
    	    int idx = 0;
    	    while (idx < literalLen && (idx = literal.indexOf(regexMetaChar, idx)) > -1) {
    	        idxs.add(new Integer(idx++));
    	    }
    	}
    	if (!idxs.isEmpty()) {
    	    Collections.sort(idxs);
    	    final StringBuffer sb = new StringBuffer(literal);
    	    int i = 0;
    	    for (final Iterator iter = idxs.iterator(); iter.hasNext(); i++) {
    	        final int idx = ((Integer) iter.next()).intValue();
    	        sb.insert(idx + i, '\\');
    	    }
    	    return sb.toString();
    	}
    	return literal;
    }

    private static String emptyIfNull(final String str) {
        return str != null ? str : "";
    }

}
