/*
 * Copyright (C) 2003, 2005, 2006, 2007, 2008, 2011 David Li, Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.misc;

import java.io.IOException;
import java.io.Reader;

import org.xml.sax.InputSource;

//FIXME: Keeping the LineNumberMap maybe inefficient.  We might want to
//drop this in favor of just tracking the current line number (and maybe
//one back.  Need to see how the parser behaves.  Plus, with this class
//reading everything into memory, we could just lines when actually
//needed.  Also, all of the line number map stuff isn't abstracted.

/**
 * A reader that implements Server-Side Includes (SSI).  This used the syntax
 * defined by Apache mod_include, however it only implements the `include'
 * directive.
 * <p>
 * This input stream automatically closes on reaching EOF.
 */
public final class SSIReader extends Reader {
    /**
     * Listener for SSI open/close events.
     */
    public class Listener {
        /** Classed when a new include file is opened. */

    }

    /**
     * Include command and arguments.
     */
    private static final String INCLUDE = "include";
    private static final String INCLUDE_FILE = "file";
    private static final String INCLUDE_VIRTUAL = "virtual";
    private static final String[] INCLUDE_VALID_ARGS = {
        INCLUDE_FILE,
        INCLUDE_VIRTUAL
    };

    /**
     * Input stream we are reading from.  Also the top of the stack,
     * if others have been included.
     */
    private SSIParsedStream fIn;

    private SSISystemIdResolver fSSISystemIdResolver;
    
    /**
     * Map of stream position to source file and line number.
     */
    private LineNumberRecorder fLineNumbers;

    /**
     * Same as
     * {@link #SSIReader(InputSource, String, SSISystemIdResolver) SSIReader(source, null, null)}
     * 
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public SSIReader(final InputSource source) throws IOException {
        this(source, null, null);
    }

    /**
     * Same as
     * {@link #SSIReader(InputSource, String, SSISystemIdResolver) SSIReader(source, ssiBase, null)}
     * 
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @param ssiBase a path for which a root-relative ssiPath is to be resolved
     *                against, optional, may be null
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public SSIReader(final InputSource source, String ssiBase) throws IOException {
        this(source, ssiBase, null);
    }

    /**
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @param ssiBase a path for which a root-relative ssiPath is to be resolved
     *                against, optional, may be null
     * @param ssiSystemIdResolver a system Id resolver with which to resolve SSIs,
     *                            optional, if null one will be internally created
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public SSIReader(final InputSource source, final String ssiBase, final SSISystemIdResolver ssiSystemIdResolver) throws IOException {
        fLineNumbers = new LineNumberRecorder();
        fIn = new SSIParsedStream(source, fLineNumbers);
        this.fSSISystemIdResolver = ssiSystemIdResolver != null ? ssiSystemIdResolver : new SSISystemIdResolver();
        this.fSSISystemIdResolver.setSsiBase(ssiBase);
    }

    /**
     * @return the system id of the currently opened file
     */
    public String getSystemId() {
        if (fIn != null) {
            return fIn.getSystemId();
        } else {
            return null;
        }
    }

    /**
     * Open a SSI include file, pushing on the stack.
     */
    private void openSSIInclude(final String includingFileSystemId,
                                final String ssiPath) throws IOException {

        final String systemId = fSSISystemIdResolver.getSystemId(includingFileSystemId, ssiPath);
        
        //Open new reader, push current one
        //No need to wrap in ClosingInputSource, because SSIParsedStream manages this
        fIn = new SSIParsedStream(new InputSource(systemId), 
                                  fLineNumbers, fIn);
    }

    /**
     * Process a SSI include directive.
     */
    private void processSSIInclude(final SSIDirective directive) throws IOException {
        // Validate
        directive.validateArgumentNames(INCLUDE_VALID_ARGS);
        if (directive.getNumArgs() < 1) {
            throw new IOException("SSI include must have at least one argument");
        }
        
        // Open includes. Done in reverse, as they are pushed on a stack
        for (int idx = directive.getNumArgs()-1; idx >= 0; idx--) {
            openSSIInclude(directive.getSystemId(), directive.getArgValue(idx));
        }
    }

    /**
     * Process the SSI directive that is next in the stream.
     */
    private void processSSIDirective() throws IOException {
        SSIDirective directive = fIn.parseSSIDirective();
        if (directive.getCmd().equals(INCLUDE)) {
            processSSIInclude(directive);
        } else {
            throw new IOException("Invalid or unsupport SSI command \""
                                  + directive.getCmd() + "\": "
                                  + fIn.getSystemId());
        }
    }

    /**
     * Read a character.
     * @return The character, -1 if no more characters are available.
     */
    public int read() throws IOException {
        int ch;
        do {
            ch = fIn.read();
            if (ch == SSIParsedStream.AT_EOF) {
                fIn = fIn.pop();
                if (fIn == null) {
                    return -1;  // EOF
                }
            } else if (ch == SSIParsedStream.AT_SSI) {
                processSSIDirective();
            }
        } while (ch < 0);
        return ch;
    }

    /**
     * Read characters into a portion of an array.  To simplify the
     * handling of an SSI directive, a directive is only processed if
     * it's at the start of the buffer.  SSIs in the middle of the
     * read result in a partial read.
     *
     * @see java.io.FilterReader#read
     * throws IOException if an I/O error occurs
     */
    public int read(final char cbuf[], final int off, final int len) throws IOException {
        if (fIn == null) return -1; //fIn sometimes null when using NekoHTML. Returning -1 immediately seems to solve the issue with no obvious side effects
        int readLen;
        do {
            readLen = fIn.read(cbuf, off, len);
            if (readLen == SSIParsedStream.AT_EOF) {
                fIn = fIn.pop();
                if (fIn == null) {
                    return -1;  // EOF
                }
            } else if (readLen == SSIParsedStream.AT_SSI) {
                // Directive might change fIn to a new file.
                processSSIDirective();
            }
        } while (readLen < 0);
        return readLen;
    }

    /**
     * Close the stream.
     *
     * @exception  IOException  If an I/O error occurs
     */
    public void close() throws IOException {
        fIn = null;
    }

    /**
     * Get the line number map.
     */
    public final LineNumberMap getLineNumberMap() {
        return fLineNumbers;
    }

    /**
     * Same as
     * {@link #create(InputSource, String, SSISystemIdResolver) create(source, null, null)}
     * 
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public static InputSource create(final InputSource source) throws IOException {
        return create(source, null, null);
    }

    /**
     * Same as
     * {@link #create(InputSource, String, SSISystemIdResolver) create(source, ssiBase, null)}
     * 
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @param ssiBase a path for which a root-relative ssiPath is to be resolved
     *                against, optional, may be null
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public static InputSource create(final InputSource source, final String ssiBase) throws IOException {
        return create(source, ssiBase, null);
    }

    /**
     * Construct an InputSource containing an SSI Reader.
     * <p>Note: does not wrap SSI Reader in ClosingInputSource, as is common in
     * other cases to guarantee that no file handles will be left open (on Windows).
     * The reason is so consumers can access the unwrapped reader in order to
     * obtain the LineNumberMap.  Fortunately, experimentation shows that wrapping
     * in this case is unnecessary; not even by downstream consumers.</p>
     * 
     * @param source the InputSource representing the including file, required,
     *               must not be null
     * @param ssiBase a path for which a root-relative ssiPath is to be resolved
     *                against, optional, may be null
     * @param ssiSystemIdResolver a system Id resolver with which to resolve SSIs,
     *                            optional, if null one will be internally created
     * @throws IOException if unable to load including file and/or SSIs
     * @throws NullPointerException if source is null
     */
    public static InputSource create(final InputSource source, final String ssiBase, final SSISystemIdResolver ssiSystemIdResolver) throws IOException {
        final InputSource ssiSource
            = new InputSource(new SSIReader(source, ssiBase, ssiSystemIdResolver));
        ssiSource.setPublicId(source.getPublicId());
        ssiSource.setSystemId(source.getSystemId());
        ssiSource.setEncoding(source.getEncoding());
        return ssiSource;
    }

}
