/*
 * Copyright (C) 2011 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.misc;

import java.io.File;
import java.io.IOException;
import java.net.URL;

import org.enhydra.xml.xmlc.deferredparsing.ResourceLoader;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;

/**
 * Resolves SSI systemIds relative to either resource loader contexts (when
 * a resource loader is provided) or the parent systemId.
 * <p>
 * Note: providing the 'includingFilePath' to the constructor is useful for
 * determining the operative root to use for relative SSI paths.  And when
 * moving back directories,  it is used to avoid moving past the operative
 * root of 'file:'-based SystemIds.  For instance...
 * </p>
 * <dl>
 * <dt>systemId</dt>
 * <dd>"file:/C:///apache-tomcat/webapps/app/WEB-INF/classes/org/company/app/package/path/to/somefile.html"</dd>
 * <dt>ssiPath (irrational value that moves back waaayyy too many directories)</dt>
 * <dd>"../../../../../../../../../../../../../../../../../../../../../includes/include.ssi"</dd>
 * <dt>includingFilePath</dt>
 * <dd>"org/company/app/package/path/to/somefile.html"</dd>
 * <dt>Resolved SSI systemId using above includingFilePath...</dt>
 * <dd>file:/C:///apache-tomcat/webapps/app/WEB-INF/classes/includes/include.ssi</dd>
 * <dt>Resolved SSI systemId NOT using above includingFilePath...</dt>
 * <dd>file:/C:///includes/include.ssi</dd>
 * </dl>
 * <p>
 * Notice that with 'includingFilePath', the operative root is "file:/C:///..../WEB-INF/classes/"
 * and without 'includingFilePath' the operative root is "file:/C:///".  The first
 * case is probably more useful, but notice that in both cases the value is
 * rational.  That is, the result remains a valid URL even in the face of the
 * irrational "ssiPath" value moving back waaayyy to many directories.  Both
 * "jar:file:/" and "http://" URLs are handled, respectively, in rational ways as well.
 * </p>
 * <p>
 * Note: paths containing "." or ".." starting in the middle/end of the path (e.g.,
 * "/somedir/../../include.ssi") are not resolved to the corresponding directories
 * and, instead, remain as-is in the systemId.  Therefore, when defining SSI include
 * paths or the ssiBase path, any "." or ".." should start at the beginning of
 * the path (e.g., "../include.ssi").
 * </p>
 * 
 * @author Jacob Kjome
 * @since 2.3.3
 */
public class SSISystemIdResolver {

    private final String fIncludingFilePathDir; //this always needs to be non-null, may be empty string
    private final ResourceLoader fResourceLoaderDelegate;
    private final XMLCDeferredParsingFactory fFactory;
    private String ssiBase;

    /**
     * Same as
     * {@link #SSISystemIdResolver(String, ResourceLoader, XMLCDeferredParsingFactory) SSISystemIdResolver(null, null, null)}
     */
    public SSISystemIdResolver() {
        this(null, null, null);
    }

    /**
     * Same as
     * {@link #SSISystemIdResolver(String, ResourceLoader, XMLCDeferredParsingFactory) SSISystemIdResolver(parentFilePath, null, null)}
     * 
     * @param includingFilePath the relative (to a context, such as class loader,
     *                          reparse resource dir, servlet context, etc...)
     *                          path of the including, or parent, file with any
     *                          directories in the path separated by '/', optional,
     *                          may be null
     */
    public SSISystemIdResolver(final String includingFilePath) {
        this(includingFilePath, null, null);
    }

    /**
     * @param includingFilePath the relative (to a context, such as class loader,
     *                          reparse resource dir, servlet context, etc...)
     *                          path of the including, or parent, file with any
     *                          directories in the path separated by '/', optional,
     *                          may be null
     * @param resourceLoader a resource loader, allows for pluggable control over
     *                       loading SSI resources, optional, may be null 
     * @param factory the XMLC deferred parsing factory, used to obtain paths
     *                from package prefixes when a ResourceLoader is also provided,
     *                optional, may be null
     */
    public SSISystemIdResolver(String includingFilePath, final ResourceLoader resourceLoader, final XMLCDeferredParsingFactory factory) {        
        if (includingFilePath != null && (includingFilePath = includingFilePath.trim()).length() > 0) {
            final int pDirIdx = includingFilePath.lastIndexOf('/');
            this.fIncludingFilePathDir = pDirIdx > -1 ? includingFilePath.substring(0, pDirIdx) : "";
        } else {
            this.fIncludingFilePathDir = "";
        }
        this.fResourceLoaderDelegate = resourceLoader;
        this.fFactory = factory;
    }

    void setSsiBase(String ssiBase) {
        if (ssiBase != null && (ssiBase = ssiBase.trim()).length() > 0) {
            ssiBase = ssiBase.replace('\\', '/');
            this.ssiBase = ssiBase.endsWith("/") ? ssiBase.substring(0, ssiBase.length() - 1) : ssiBase;
        } else {
            this.ssiBase = null;
        }
    }

    /**
     * @param includingFileSystemId the fully qualified path of the including file
     * @param ssiPath the path declared in the server-side-include directive
     * @return the resolved fully qualified path of the server-side-include
     * @throws IOException if unable to resolve the server-side-include path
     * @throws NullPointerException if either argument is null
     */
    public String getSystemId(final String includingFileSystemId, String ssiPath)
        throws IOException {

        ssiPath = ssiPath.trim().replace('\\', '/');
        if (ssiBase != null && ssiPath.startsWith("/")) {
            ssiPath = ssiBase + ssiPath;
        }
        return fResourceLoaderDelegate != null
            ? getSystemIdUsingResourceLoader(includingFileSystemId, ssiPath)
            : getSystemIdRelativeToIncludingFileSystemId(includingFileSystemId, ssiPath);
    }

    private String getSystemIdUsingResourceLoader(final String includingFileSystemId, String ssiPath)
        throws IOException {

        final String ssiPathUnmodified = ssiPath;
        if (ssiPath.startsWith("/")) {
            ssiPath = ssiPath.substring(1);
        } else {
            ssiPath = canonicalPath(fIncludingFilePathDir, -1, ssiPath);
        }

        final String[] paths = fFactory != null ? fFactory.getPathsFromPackagePrefixes(ssiPath) : new String[]{ssiPath};
        final URL srcURL = fResourceLoaderDelegate.getResource(paths);
        if (srcURL == null) {
//            throw new XMLCRuntimeException("Source path(s) '"
//                                           + Arrays.asList(paths)
//                                           + "' not found for parent file "
//                                           + includingFileSystemId);
            //fallback...
            return getSystemIdRelativeToIncludingFileSystemId(includingFileSystemId, ssiPathUnmodified);
        }
        return srcURL.toString();
    }

    private String getSystemIdRelativeToIncludingFileSystemId(String includingFileSystemId, String ssiPath)
        throws IOException {
        
        String systemId = "";
        try {
            final File includingFile = new File(includingFileSystemId);
            includingFile.getCanonicalPath(); //Fail fast (IOException) if we are dealing with a URL
            final File inclFile = new File(ssiPath);
            File baseDir = null;
            if (!inclFile.isAbsolute()) {
                baseDir = includingFile.getParentFile();
            }
            final File inclPath = new File(baseDir, ssiPath);
            systemId = inclPath.toString(); // Problem with toURL().getExternalForm(). It returns strange URL
        } catch (final IOException ioe) {
            try {
                includingFileSystemId = includingFileSystemId.replace('\\', '/');
                int idx = includingFileSystemId.indexOf("file:");
                if (ssiPath.startsWith("/")) {
                    if (idx > -1) {
                        ssiPath = includingFileSystemId.substring(0, idx(includingFileSystemId, idx)) + ssiPath.substring(1);
                    } else {
                        ssiPath = includingFileSystemId.substring(0, includingFileSystemId.indexOf('/', idx(includingFileSystemId, idx) + 3)) + ssiPath;
                    }
                } else {
                    ssiPath = canonicalPath(includingFileSystemId.substring(0, includingFileSystemId.lastIndexOf('/')), idx(includingFileSystemId, idx) - (idx > -1 ? 2 : -3), ssiPath);
                }
                systemId = new URL(ssiPath).toString();
            } catch (final IOException ioe1) {
                throw ioe1;
            }
        }
        return systemId;
    }

    private int idx(final String includingFileSystemId, final int idxFileProtocol) {
        int idx = idxFileProtocol;
        if (idx > -1) {
            final int idxJar = includingFileSystemId.indexOf("!/");
            if (idxJar > -1) {
                idx = idxJar + 2;
            } else {
                final int idxIFPD = fIncludingFilePathDir.length() > 0 ? includingFileSystemId.indexOf(fIncludingFilePathDir) : -1;
                if (idxIFPD > -1) {
                    idx = idxIFPD;
                } else {
                    idx = idx + 5;
                    int idxWinColon = includingFileSystemId.indexOf(':', idx);
                    if (idxWinColon > -1) idx = idxWinColon + 1;
                    while (includingFileSystemId.charAt(idx) == '/') idx++;
                }
            }
            return idx;
        } else {
            return includingFileSystemId.indexOf("://");
        }
    }

    private static String canonicalPath(String parentFilePathDir, final int idxParentFilePathDirMin, String ssiPath) {
        if (ssiPath.startsWith("./")) ssiPath = ssiPath.substring(2);
        if (ssiPath.startsWith("..")) {
            int idx = 0;
            while((idx = ssiPath.indexOf("..")) > -1) {
                ssiPath = ssiPath.substring(idx + 3);
                final int idxSlash = parentFilePathDir.lastIndexOf('/');
                if (idxSlash > idxParentFilePathDirMin) {
                    parentFilePathDir = idxSlash > -1 ? parentFilePathDir.substring(0, idxSlash) : "";
                }
            }
        }
        if (parentFilePathDir.length() > 0) parentFilePathDir = parentFilePathDir + '/';
        return parentFilePathDir + ssiPath;
    }

//    public static void main(String[] args) throws Exception {
//        final String
//            includingFilePath = "org/company/app/package/path/to/somefile.html",
//            includingFilePathForConstructor = includingFilePath,
//            includingFileSystemId = /*"http://domain:8080/" + includingFilePath,*/ "file:/C:///apache-tomcat/webapps/app/WEB-INF/classes/" + includingFilePath,
//            ssiPath = "/includes/include.ssi";
//        final SSISystemIdResolver r = new SSISystemIdResolver(includingFilePathForConstructor);
//        r.setSsiBase("../../");
//        System.out.println(r.getSystemId(includingFileSystemId, ssiPath));
//    }

}
