package org.enhydra.xml.xmlc.parsers.xerces;

import java.io.IOException;
import org.apache.xerces.xni.Augmentations;
import org.apache.xerces.xni.XMLResourceIdentifier;
import org.apache.xerces.xni.XMLString;
import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.parser.XMLErrorHandler;
import org.apache.xerces.xni.parser.XMLInputSource;
import org.apache.xerces.xni.parser.XMLParseException;
import org.apache.xerces.xni.parser.XMLParserConfiguration;
import org.enhydra.xml.io.ClosingInputSource;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.XMLEntityResolver;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.metadata.Parser;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;


public class XercesDOMParser extends org.apache.xerces.parsers.DOMParser implements XMLCParser, org.apache.xerces.xni.parser.XMLEntityResolver {

    /**
     * Verbose tracing.
     */
    private XercesTracer fTracer;

    /**
     * The document builder.
     */
    private XMLCDocument fXMLCDocument;

    /**
     * Meant to store encoding from the InputSource passed to parse() and make
     * it accessible to XNI event methods
     */
    private String fEncoding;
    
    /**
     * Classpath/XCatalog entity resolver.
     */
    private XMLEntityResolver fResolver;

    /**
     * Is an element with #PCDATA in its content model being processed?
     */
    private boolean fProcessingPCDATAElement = false;


    public XercesDOMParser() {
        super();
    }

    public XercesDOMParser(XMLParserConfiguration config) {
        super(config);
    }

    /**
     * Parses using the Xerces {@link org.apache.xerces.parsers.DOMParser DOMParser}
     * and the XMLC {@link org.enhydra.xml.io.XMLEntityResolver XMLEntityResolver}.
     * The resulting DOM document is set on the {@link XMLCDocument XMLCDocument}
     * after a successful parse. The type of DOM bound to by the DOM parser is
     * defined by setting the Xerces property
     * "http://apache.org/xml/properties/dom/document-class-name" to the fully
     * qualified DOM Document name returned by the {@link XMLCDomFactory#getDocumentClassName getDocumentClassName()}
     * method of the current XMLCDomFactory. 
     * 
     * @return a populated XMLCDocument
     * @see org.enhydra.xml.xmlc.parsers.XMLCParser#parse(org.xml.sax.InputSource, org.enhydra.xml.xmlc.misc.LineNumberMap, org.enhydra.xml.xmlc.dom.XMLCDomFactory, org.enhydra.xml.xmlc.metadata.MetaData, org.enhydra.xml.io.ErrorReporter, org.enhydra.xml.xmlc.parsers.ParseTracer)
     */
    public XMLCDocument parse(InputSource input, LineNumberMap lineNumberMap, XMLCDomFactory domFactory, MetaData metaData, ErrorReporter errorReporter, ParseTracer tracer) throws IOException, XMLCException, SAXException {
        //fConfiguration.setDocumentHandler(this);
        //fConfiguration.setDTDContentModelHandler(this);
        //fConfiguration.setDTDHandler(this);
        fConfiguration.setEntityResolver(this);
        if (lineNumberMap != null) {
            fConfiguration.setErrorHandler(new MappingErrorHandler(errorReporter,
                                                    lineNumberMap));
        } else {
            fConfiguration.setErrorHandler(errorReporter);
        }
        fConfiguration.setProperty("http://apache.org/xml/properties/dom/document-class-name", domFactory.getDocumentClassName());

        features(metaData);

        //Setup entity resolver
        fResolver = new XMLEntityResolver();
        if (tracer.enabled()) {
            fResolver.setDebugWriter(tracer);
        }
        //Add defaults before adding catalogs so defaults can be overridden.
        fResolver.addClassLoader(Thread.currentThread().getContextClassLoader());
        fResolver.setDefaultResolving();

        //Add configured catalogs
        String[] xCatalog = metaData.getParser().getXCatalogURLs();
        for (int idx = 0; idx < xCatalog.length; idx++) {
            fResolver.loadCatalog(new ClosingInputSource(xCatalog[idx]));
        }

        fTracer = new XercesTracer(tracer);
        fXMLCDocument = new XMLCDocument(domFactory);
        fEncoding = input.getEncoding();

        super.parse(input);

        if (fXMLCDocument.getEncoding() == null) {
            fXMLCDocument.setEncoding(fEncoding != null ? fEncoding : "UTF-8");
        }
        fXMLCDocument.setDocument(getDocument());

        return fXMLCDocument;
    }

    /**
     * Called by {@link #parse(InputSource, LineNumberMap, XMLCDomFactory, MetaData, ErrorReporter, ParseTracer) parse()}
     * to allow DOMParser features to be set.  Also provides an opportunity to
     * validate metadata, if need be.
     * 
     * @param metaData the XMLC metadata for the markup to be parsed
     * @throws XMLCException if metadata is found to be invalid for the current parser
     */
    protected void features(MetaData metaData) throws XMLCException {
        Parser parser = metaData.getParser();
        Boolean validate = parser.getValidate();

        fConfiguration.setFeature("http://apache.org/xml/features/allow-java-encodings", true);
        fConfiguration.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false);
        //fConfiguration.setFeature("http://apache.org/xml/features/dom/create-entity-ref-nodes", false);
        fConfiguration.setFeature("http://xml.org/sax/features/namespaces", true);
        fConfiguration.setFeature("http://xml.org/sax/features/validation", validate == null ? true : validate.booleanValue());
        fConfiguration.setFeature("http://apache.org/xml/features/dom/include-ignorable-whitespace", false);
    }

    /**
     * Sets the version and encoding values to the {@link XMLCDocument xmlcDocument}.
     * Execution is then passed to the superclass.
     * 
     * @see org.apache.xerces.xni.XMLDocumentHandler#xmlDecl(String, String, String, org.apache.xerces.xni.Augmentations)
     */
    public void xmlDecl(String version, String encoding, String standalone, Augmentations augs) throws XNIException {
        fTracer.xmlDecl(version, encoding, standalone);
        fXMLCDocument.setXMLVersion(version);
        String prefEncoding = fEncoding != null ? fEncoding : (encoding != null ? encoding : "UTF-8");
        fXMLCDocument.setEncoding(prefEncoding);
        super.xmlDecl(version, prefEncoding, standalone, augs);
    }

    /**
     * Squashes ignorable whitespace.  Does not pass execution to the superclass.
     * TODO - Still not sure if this is adviseable?  Leave here, but commented out until further investigation determines whether and how it can be useful to discard ignorable whitespace at parse time
     * 
     * @see org.apache.xerces.xni.XMLDocumentHandler#ignorableWhitespace(org.apache.xerces.xni.XMLString, org.apache.xerces.xni.Augmentations)
     */    
    /*public void ignorableWhitespace(XMLString text, Augmentations augs) throws XNIException {
        //super.ignorableWhitespace(text, augs);
    }*/

    /**
     * Checks if the type of the current attribute is of type "ID". If so, the
     * attribute is recorded as such (along with the element it belongs to) in
     * the {@link XMLCDocument xmlcDocument}.
     * Execution is then passed to the superclass.
     * 
     * @see org.apache.xerces.xni.XMLDTDHandler#attributeDecl(String, String, String, String[], String, org.apache.xerces.xni.XMLString, org.apache.xerces.xni.XMLString, org.apache.xerces.xni.Augmentations)
     */
    public void attributeDecl(String elementName, String attributeName, String type, String[] enumeration, String defaultType, XMLString defaultValue, XMLString nonNormalizedDefaultValue, Augmentations augs) throws XNIException {
        fTracer.attributeDecl(elementName, attributeName, type, enumeration, defaultType, defaultValue != null ? defaultValue.toString() : null);
        if ("ID".equals(type)) {
            fXMLCDocument.addIdAttribute(elementName, attributeName);
        }
        super.attributeDecl(elementName, attributeName, type, enumeration, defaultType, defaultValue, nonNormalizedDefaultValue, augs);
    }

    /**
     * The pcData(Augmentations) method sets a flag saying whether the current
     * element can contain PCDATA.  This method, which is called shortly after
     * pcdata() method is called, checks that flag and, if true, stores the
     * element as a PCDATA element in the {@link XMLCDocument xmlcDocument},
     * and resets the flag.
     * Execution is then passed to the superclass.
     * 
     * @see org.apache.xerces.xni.XMLDTDHandler#elementDecl(String, String, org.apache.xerces.xni.Augmentations)
     * @see #pcdata(Augmentations)
     */
    public void elementDecl(String name, String contentModel, Augmentations augs) throws XNIException {
        fTracer.elementDecl(name, contentModel);
        if (fProcessingPCDATAElement) {
            fXMLCDocument.addPCDataContentElement(name);
            fProcessingPCDATAElement = false;
        }
        super.elementDecl(name, contentModel, augs);
    }

    /**
     * Sets a flag saying that the current element can contain PCDATA.
     * Execution is then passed to the superclass.
     * 
     * @see org.apache.xerces.xni.XMLDTDContentModelHandler#pcdata(org.apache.xerces.xni.Augmentations)
     */
    public void pcdata(Augmentations augs) throws XNIException {
        fProcessingPCDATAElement = true;
        super.pcdata(augs);
    }

    /**
     * Uses XMLC's own catalog resolver.
     * Execution is NOT passed to the superclass.
     * 
     * @see org.enhydra.xml.io.XMLEntityResolver
     * @see org.apache.xerces.xni.parser.XMLEntityResolver#resolveEntity(org.apache.xerces.xni.XMLResourceIdentifier)
     */
    public XMLInputSource resolveEntity(XMLResourceIdentifier resourceIdentifier) throws XNIException, IOException {
        XMLInputSource xmlSource = null;
        if (fResolver != null) {
            try {
                InputSource source = fResolver.resolveEntity(resourceIdentifier.getPublicId(), resourceIdentifier.getExpandedSystemId());
                if (source != null) {
                    source = new ClosingInputSource(source, true);
                    if (source.getCharacterStream() != null) {
                        xmlSource =  new XMLInputSource(source.getPublicId(), source.getSystemId(), resourceIdentifier.getBaseSystemId(), source.getCharacterStream(), source.getEncoding());
                    } else {
                        xmlSource = new XMLInputSource(source.getPublicId(), source.getSystemId(), resourceIdentifier.getBaseSystemId(), source.getByteStream(), source.getEncoding());
                    }
                }
                fTracer.resolveEntity(resourceIdentifier.getPublicId(), resourceIdentifier.getExpandedSystemId(), source);
            } catch (SAXException se) {
                throw new XNIException(se);
            }
        }
        return xmlSource;
    }

    /**
     * Error handler class that handles mapping of line numbers
     */
    private class MappingErrorHandler implements ErrorHandler, XMLErrorHandler {
        private ErrorReporter fErrorReporter;
        private LineNumberMap fLineNumberMap;

        /**
         * Constructor.
         */
        public MappingErrorHandler(ErrorReporter errorReporter,
                                   LineNumberMap lineNumberMap) {
            fErrorReporter = errorReporter;
            fLineNumberMap = lineNumberMap;
        }

        /**
         * Wrap a SAXParseException with a SAXParseException with the
         * location mapped.  If there is no line number map, return the
         * exception unchanged.
         */
        private SAXParseException mapException(SAXParseException exception) {
            if (fLineNumberMap == null) {
                return exception;
            } else {
                LineNumberMap.Line line
                    = fLineNumberMap.getLineFromLineNum(exception.getLineNumber());
                return new SAXParseException(exception.getMessage(),
                                             exception.getPublicId(),
                                             line.getFileName(),
                                             line.getLineNum(),
                                             exception.getColumnNumber(),
                                             exception);
            }
        }

        /**
         * Receive notification of a SAX recoverable error.
         */
        public void error(SAXParseException exception) throws SAXException {
            fErrorReporter.error(mapException(exception));
        }

        /**
         * Receive notification of a SAX warning.
         */
        public void warning(SAXParseException exception) throws SAXException {
            fErrorReporter.warning(mapException(exception));
        }

        /**
         * Receive notification of a SAX non-recoverable error.
         */
        public void fatalError(SAXParseException exception) throws SAXException {
            fErrorReporter.fatalError(mapException(exception));
        }
        
        /* (non-Javadoc)
         * @see org.apache.xerces.xni.parser.XMLErrorHandler#error(String, String, org.apache.xerces.xni.parser.XMLParseException)
         */
        public void error(String domain, String key, XMLParseException exception) throws XNIException {
            try {
                error(new SAXParseException(exception.getLocalizedMessage(), exception.getPublicId(), exception.getExpandedSystemId(), exception.getLineNumber(), exception.getColumnNumber(), exception));
            } catch (SAXException se) {
                ;
            }
        }

        /* (non-Javadoc)
         * @see org.apache.xerces.xni.parser.XMLErrorHandler#fatalError(String, String, org.apache.xerces.xni.parser.XMLParseException)
         */
        public void fatalError(String domain, String key, XMLParseException exception) throws XNIException {
            try {
                fatalError(new SAXParseException(exception.getLocalizedMessage(), exception.getPublicId(), exception.getExpandedSystemId(), exception.getLineNumber(), exception.getColumnNumber(), exception));
            } catch (SAXException se) {
                ;
            }
            
        }

        /* (non-Javadoc)
         * @see org.apache.xerces.xni.parser.XMLErrorHandler#warning(String, String, org.apache.xerces.xni.parser.XMLParseException)
         */
        public void warning(String domain, String key, XMLParseException exception) throws XNIException {
            try {
                warning(new SAXParseException(exception.getLocalizedMessage(), exception.getPublicId(), exception.getExpandedSystemId(), exception.getLineNumber(), exception.getColumnNumber(), exception));
            } catch (SAXException se) {
                ;
            }
        }
    }

}
