package org.enhydra.xml.xmlc.parsers.xerces;

import java.io.IOException;
import org.apache.xerces.impl.Constants;
import org.apache.xerces.xni.Augmentations;
import org.apache.xerces.xni.QName;
import org.apache.xerces.xni.XMLAttributes;
import org.apache.xerces.xni.XNIException;
import org.apache.xerces.xni.parser.XMLDocumentFilter;
import org.cyberneko.html.HTMLConfiguration;
import org.cyberneko.html.filters.DefaultFilter;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.dom.XMLCDomFactory;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.metadata.Parser;
import org.enhydra.xml.xmlc.metadata.ParserType;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


public class XercesHTMLDOMParser extends XercesDOMParser {

    /**
     * Table of HTML 4.0 elements that have a content model including #PCDATA.
     */
    private static final String[] PCDATA_ELEMENTS = {
        // From strict DTD, transitional, and frameset DTDs
        "A", "ABBR", "ACRONYM", "ADDRESS", "APPLET", "B", "BDO", "BIG", "BLOCKQUOTE", "BODY", "BUTTON",
        "CAPTION", "CENTER", "CITE", "CODE", "DD", "DEL", "DFN", "DIV", "DT", "EM",
        "FIELDSET", "FONT", "FORM", "H1", "H2", "H3", "H4", "H5", "H6", "I", "IFRAME", "INS", "KBD",
        "LABEL", "LEGEND", "LI", "NOFRAMES", "NOSCRIPT", "OBJECT", "OPTION", "P", "PRE", "Q", "S", "SAMP",
        "SCRIPT", "SMALL", "SPAN", "STRIKE", "STRONG", "STYLE", "SUB", "SUP", "TD",
        "TEXTAREA", "TH", "TITLE", "TT", "U", "VAR"
    };

    /*private static final String[] PCDATA_ELEMENTS = {
        // From strict DTD
        "A", "ABBR", "ACRONYM", "ADDRESS", "B", "BDO", "BIG", "BUTTON",
        "CAPTION", "CITE", "CODE", "DD", "DEL", "DFN", "DIV", "DT", "EM",
        "FIELDSET", "H1", "H2", "H3", "H4", "H5", "H6", "I", "INS", "KBD",
        "LABEL", "LEGEND", "LI", "OBJECT", "OPTION", "P", "PRE", "Q", "SAMP",
        "SCRIPT", "SMALL", "SPAN", "STRONG", "STYLE", "SUB", "SUP", "TD",
        "TEXTAREA", "TH", "TITLE", "TT", "VAR",

        // Added by transitional and frameset DTDs
        "APPLET", "BLOCKQUOTE", "BODY", "CENTER", "FONT", "FORM", "IFRAME",
        "NOSCRIPT", "S", "STRIKE", "U",

        // Added by transitional DTD
        "NOFRAMES"
    };*/

    //private String fCurrentElement;
    
    public XercesHTMLDOMParser() {
        super(new HTMLConfiguration());
    }

    public XMLCDocument parse(InputSource input, LineNumberMap lineNumberMap, XMLCDomFactory domFactory, MetaData metaData, ErrorReporter errorReporter, ParseTracer tracer) throws IOException, XMLCException, SAXException {
        XMLCDocument xmlcDocument = super.parse(input, lineNumberMap, domFactory, metaData, errorReporter, tracer);
        addPCDataContentElements(xmlcDocument);
        return xmlcDocument;
    }

    protected void features(MetaData metaData) throws XMLCException {        
        Parser parser = metaData.getParser();

        Boolean validate = parser.getValidate();
        if ((validate != null) && (!validate.booleanValue())) {
            throw new XMLCException(ParserType.NEKOHTML.getName() + " parser can't disable document validation");
        }

        if (parser.getXCatalogURLs().length != 0) {
            throw new XMLCException("XCatalog not supported for HTML");
        }

        //fConfiguration.setFeature("http://apache.org/xml/features/allow-java-encodings", true);
        fConfiguration.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false);
        //fConfiguration.setFeature("http://apache.org/xml/features/dom/create-entity-ref-nodes", false);
        fConfiguration.setFeature("http://xml.org/sax/features/namespaces", false);
        fConfiguration.setFeature("http://xml.org/sax/features/validation", validate == null ? true : validate.booleanValue());
        //fConfiguration.setFeature("http://apache.org/xml/features/dom/include-ignorable-whitespace", false);

        //fConfiguration.setFeature("http://cyberneko.org/html/features/insert-doctype", true);
        //fConfiguration.setFeature("http://cyberneko.org/html/features/override-doctype", true);
        //fConfiguration.setProperty("http://cyberneko.org/html/properties/names/elems", "lower");

        XMLDocumentFilter idEnhancer = new DefaultFilter() {
            public void startElement(QName element, XMLAttributes attributes, Augmentations augs) throws XNIException {
                int idx = attributes.getIndex("id");
                if (idx > -1) {
                    attributes.setType(idx, "ID");
                    Augmentations attrsAugs = attributes.getAugmentations(idx);
                    attrsAugs.putItem(Constants.ATTRIBUTE_DECLARED, Boolean.TRUE);
                }
                super.startElement(element, attributes, augs);
            }

            //TODO - instead of using attrsAugs above to cheat in order to satisfy Xerces current implementation, spoof DTD events and make "id" attributes of type type "ID" the standard XNI way
            /*public void startDocument(XMLLocator locator, String encoding, NamespaceContext namespaceContext, Augmentations augs) throws XNIException {
                XMLDTDHandler dtdHandler = fConfiguration.getDTDHandler();
                dtdHandler.startDTD(null, null);
                dtdHandler.attributeDecl(arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7);
                dtdHandler.endDTD(null);
                super.startDocument(locator, encoding, namespaceContext, augs);
            }*/
        };
        XMLDocumentFilter[] filters = { idEnhancer };
        fConfiguration.setProperty("http://cyberneko.org/html/properties/filters", filters);
    }

    /**
     * Add elements that have #PCDATA as part of its content model
     * to the XMLCDocument description.
     */
    private void addPCDataContentElements(XMLCDocument xmlcDocument) {
        for (int idx = 0; idx < PCDATA_ELEMENTS.length; idx++) {
            xmlcDocument.addPCDataContentElement(PCDATA_ELEMENTS[idx]);
        }
    }

    /**
     * Execution is passed to the superclass if it's not just whitespace.  See:
     * http://www.cafeconleche.org/books/xmljava/chapters/ch06s10.html
     * 
     * @see org.apache.xerces.xni.XMLDocumentHandler#characters(org.apache.xerces.xni.XMLString, org.apache.xerces.xni.Augmentations)
     */
    /*public void characters(XMLString text, Augmentations augs) throws XNIException {
        if (Arrays.binarySearch(PCDATA_ELEMENTS, fCurrentElement) != -1) {
            //TODO - validate that this doesn't mess up <pre> stuff. This processing makes us match up with Xerces1 parsing, but is that a good thing?  DTD-based XHTML and XML can use "include-ignorable-whitespace" feature, but we need to do it manually for HTML.
            if (text.toString().trim().length() > 0) {
                super.characters(text, augs);
            }
        }
    }*/

    /**
     * Capture the current element for comparison in {@link #characters(XMLString, Augmentations) characters()}
     * with the pre-defined array of PCDATA HTML elements.
     * 
     * @see org.apache.xerces.xni.XMLDocumentHandler#startElement(org.apache.xerces.xni.QName, org.apache.xerces.xni.XMLAttributes, org.apache.xerces.xni.Augmentations)
     */
    /*public void startElement(QName element, XMLAttributes attributes, Augmentations augs) {
        fCurrentElement = element.localpart.toUpperCase();
        super.startElement(element, attributes, augs);
    }*/

}
