/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XercesTracer.java,v 1.5 2007-04-22 05:40:59 jkjome Exp $
 */

package org.enhydra.xml.xmlc.parsers.xerces;

import org.apache.xerces.xni.QName;
import org.apache.xerces.xni.XMLAttributes;
import org.enhydra.xml.io.InputSourceOps;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.xml.sax.InputSource;

/**
 * Tracing for parser callbacks.  Isolated in a class because it
 * was getting too complex and crowding the actual handlers.
 */
class XercesTracer {

    //FIXME - Make sure this all matches up to Xerces2 XNI.  It was basically ported to XNI from Xerces1, but most of it goes unused (since we let Xerces2 DOMParser do most of the work, so there's nothing to trace) and some was just updated to make it compile, not necessarily to make it valid to trace XNI methods

    /**
     * Verbose tracing.
     */
    private ParseTracer fTracer;

    /**
     * Is the tracer enabled?
     */
    private boolean fIsEnabled;

    /**
     * Constructor.
     */
    protected XercesTracer(ParseTracer tracer) {
        fTracer = tracer;
        fIsEnabled = (fTracer != null) && fTracer.enabled();
    }

    public void resolveEntity(String publicId,
                              String systemId,
                              InputSource source) {
        if (fIsEnabled) {
            fTracer.trace("resolveEntity: \""
                         + publicId + "\" \""
                         + systemId + "\" => "
                         + InputSourceOps.getName(source));
        }
        
    }

    public void startDocument() {
        if (fIsEnabled) {
            fTracer.trace("startDocument");
        }
    }

    public void endDocument() {
        if (fIsEnabled) {
            fTracer.trace("endDocument");
        }
    }

    public void xmlDecl(String version, String encoding, String standalone) {
        if (fIsEnabled) {
            fTracer.trace("xmlDecl: " + version
                          + " " + encoding
                          + " " + standalone);
        }
    }

    public void textDecl(String version, String encoding) {
        if (fIsEnabled) {
            fTracer.trace("textDecl: " + version
                          + " " + encoding);
        }
    }

    public void startElement(QName element,
                             XMLAttributes attributes) {
        if (fIsEnabled) {
            StringBuffer buf = new StringBuffer("startElement: " + 
                                                element.rawname);
            for (int i = 0; i < attributes.getLength(); i++) {
                if (!attributes.isSpecified(i)) continue;
                String uri = attributes.getURI(i);
                buf.append(" ");
                if (uri != null) {
                    buf.append("<");
                    buf.append(uri);
                    buf.append(">");
                }
                buf.append(attributes.getLocalName(i));
                buf.append("=\"");
                buf.append(attributes.getValue(i));
                buf.append("\"");
            }
            fTracer.trace(buf.toString());
            fTracer.enter();
        }
    }

    public void endElement(QName element) {
        if (fIsEnabled) {
            fTracer.leave();
            fTracer.trace("endElement: " + element.rawname);
        }
    }

    public void startGeneralEntity(String name,
                                     String namespace,
                                     String publicId,
                                     String systemId,
                                     String encoding) {
        if (fIsEnabled) {
            fTracer.trace("startGeneralEntity: "
                          + name
                          + " " +  namespace
                          + " " +  publicId
                          + " " +  systemId
                          + " " +  encoding);
            fTracer.enter();
        }
    }

    public void endGeneralEntity(String name) {
        if (fIsEnabled) {
            fTracer.leave();
            fTracer.trace("endGeneralEntity: " + name);

        }
    }

    public void startCDATA() {
        if (fIsEnabled) {
            fTracer.trace("startCDATA");
            fTracer.enter();
        }
    }

    public void endCDATA() {
        if (fIsEnabled) {
            fTracer.leave();
            fTracer.trace("endCDATA");
        }
    }

    public void processingInstruction(String target,
                                      String data) {
        if (fIsEnabled) {
            fTracer.trace("processingInstruction: " + target);
        }
    }

    public void comment(String comment) {
        if (fIsEnabled) {
            fTracer.trace("comment: " + comment);
        }
    }

    public void characters(String text) {
        if (fIsEnabled) {
            fTracer.trace("characters: \"" + text + "\"");
        }
    }

    public void ignorableWhitespace(char ch[],
                                    int start,
                                    int length) {
        if (fIsEnabled) {
            fTracer.trace("ignorableWhitespace: " + length);
        }
    }

    public void doctypeDecl(String rootElement, String publicId, String systemId) {
        if (fIsEnabled) {
            fTracer.trace("startDTD: " + rootElement
                    + " " + publicId
                    + " " + systemId);
        }
    }

    public void startDTD() {
        if (fIsEnabled) {
            fTracer.enter();
        }
    }

    /**
     * This has no corresponding event method in XNI.  It is derived between
     * the beginning of the doctypeDecl() and endDTD() when, in fact, there
     * is an internal subset in existence.
     */
    public void internalSubset(String internalSubset) {
        if (fIsEnabled) {
            fTracer.trace("internalSubset: " + internalSubset);
        }
    }

    public void endDTD() {
        if (fIsEnabled) {
            fTracer.leave();
        }
    }

    /**
     * <!ELEMENT Name contentspec>
     */
    public void elementDecl(String name, String contentModel) {
        if (fIsEnabled) {
            fTracer.trace("elementDecl: " + name
                          + " " + contentModel);
        }
    }

    /**
     * <!ATTLIST Name AttDef>
     */
    public void attributeDecl(String elementName,
                            String attributeName,
                            String type,
                            String[] enumeration,
                            String defaultType,
                            String defaultValue) {
        if (fIsEnabled) {
            fTracer.trace("attlistDecl: " + elementName
                          + " " + attributeName
                          + " " + type
                          + " " + enumeration
                          + " " + defaultType
                          + " " + defaultValue);
        }
    }

    /**
     * <!ENTITY Name EntityValue>
     */
    public void internalEntityDecl(String entityName,
                                   String entityValue) {
        if (fIsEnabled) {
            fTracer.trace("internalEntityDecl: " + entityName
                         + " \"" + entityValue + "\"");
        }
    }

    /**
     * <!ENTITY Name ExternalID>
     */
    public void externalEntityDecl(String entityName,
                                   String publicId,
                                   String systemId) {
        if (fIsEnabled) {
            fTracer.trace("externalEntityDecl: " + entityName
                          + " " + publicId
                          + " " + systemId);
        }
    }

    /**
     * <!ENTITY Name ExternalID NDataDecl>
     */
    public void unparsedEntityDecl(String entityName,
                                   String publicId,
                                   String systemId,
                                   String notationName) {
        if (fIsEnabled) {
            fTracer.trace("unparsedEntityDecl: " + entityName
                          + " " + publicId
                          + " " + systemId
                          + " " + notationName);
        }
    }

    /**
     * <!NOTATION Name ExternalId>
     */
    public void notationDecl(String notationName,
                             String publicId,
                             String systemId) {
        if (fIsEnabled) {
            fTracer.trace("notationDecl: " + notationName
                          + " " + publicId
                          + " " + systemId);
        }
    }

}
