/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.servlet;

import java.util.Map;

import javax.servlet.ServletContext;

import org.enhydra.xml.xmlc.deferredparsing.DocumentLoaderImpl;
import org.enhydra.xml.xmlc.deferredparsing.ResourceLoader;

/**
 * A ServletContext aware DocumentLoader implementation.  The only behavior
 * modification from that of the superclass is the installation of a different
 * {@link #getResourceLoader() resource loader} capable of loading resources from
 * the servlet context.
 * <p>
 * Some thought was given to making this the default document loader implementation
 * for {@lnk XMLCContext}.  However, because of the possibility of candidate paths
 * happening not to match resources in resource dirs or the classpath, then
 * happening, by accident, to match resources stored in the servlet context not
 * meant to be loaded by XMLC, it was decided to leave the existing default
 * ({@link DocumentLoaderImpl}) in place.  To configure this class as the document
 * loader, simply add the following to web.xml...
 * </p>
 * <pre>
 * &lt;context-param&gt;
 *   &lt;param-name&gt;xmlcReparseDocumentLoader&lt;/param-name&gt;
 *   &lt;param-value&gt;org.enhydra.xml.xmlc.servlet.ServletDocumentLoaderImpl&lt;/param-value&gt;
 * &lt;/context-param&gt;
 * </pre>
 * <p>
 * For users wishing to serve resources primarily (or exclusively) from the
 * servlet context, it may make sense to extend this class, override
 * {@link #getResourceLoader()}, and provide a custom resource loader.
 * </p>
 * 
 * @author Jacob Kjome
 * @since 2.3.2
 */
public class ServletDocumentLoaderImpl extends DocumentLoaderImpl implements ServletDocumentLoader {

    private ResourceLoader fResourceLoader;
    protected ServletContext fServletContext;

    /**
     * @see DocumentLoaderImpl#DocumentLoaderImpl()
     */
    public ServletDocumentLoaderImpl() {
        this(null);
    }

    /**
     * @param cache a custom cache
     * @see DocumentLoaderImpl#DocumentLoaderImpl(Map)
     */
    public ServletDocumentLoaderImpl(final Map cache) {
        super(cache);
    }

    /**
     * @param servletContext the current servlet context
     */
    public void setServletContext(final ServletContext servletContext) {
        this.fServletContext = servletContext;
    }

    /**
     * See {@link ServletResourceLoaderImpl} for resource loading behavior.
     * 
     * @return a resource loader implementation
     * @see ServletResourceLoaderImpl
     */
    public ResourceLoader getResourceLoader() {
        if (fResourceLoader == null) {
            fResourceLoader = new ServletResourceLoaderImpl(getFactory(), fServletContext);
        }
        return fResourceLoader;
    }

}
