/*
 * Copyright (C) 2010 Jacob Kjome
 * 
 * This file is part of XMLC.
 *
 * XMLC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 2.1
 * of the License, or (at your option) any later version.
 *
 * XMLC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with XMLC.  If not,
 * see <http://www.gnu.org/licenses/>.
 */
package org.enhydra.xml.xmlc.servlet;

import java.net.MalformedURLException;
import java.net.URL;

import javax.servlet.ServletContext;

import org.enhydra.xml.xmlc.deferredparsing.DocumentLoaderImpl;
import org.enhydra.xml.xmlc.deferredparsing.ResourceLoader;
import org.enhydra.xml.xmlc.deferredparsing.ResourceLoaderImpl;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;

/**
 * A {@link ResourceLoader resource loader} implementation that provides the
 * capability of loading resources from the ServletContext.
 * 
 * @author Jacob Kjome
 * @since 2.3.2
 */
public class ServletResourceLoaderImpl extends ResourceLoaderImpl {

    protected final ServletContext fServletContext;

    /**
     * @param factory the XMLC deferred parsing factory
     * @param servletContext the servlet context
     */
    public ServletResourceLoaderImpl(final XMLCDeferredParsingFactory factory, final ServletContext servletContext) {
        super(factory);
        fServletContext = servletContext;
    }

    /**
     * First looks up candidatePaths according to {@link super#getResource(String[])}.
     * If not found, the {@link #getPathURLFromServletContext(String[]) servlet context}
     * is used to look up each candidate path until one is found.  If none is
     * found, null is returned.
     * <p>
     * By performing servlet context lookup last, as a fallback, there should be no
     * noticeable difference in performance for users serving resources from the
     * traditional resource dir and classpath locations.  And resources may be served
     * from a mix of locations.
     * </p>
     * <p>
     * Paths are expected to be relative, i.e., no leading slash, with directory/package
     * names separated by forward slashes regardless of operating system platform.
     * </p>
     * 
     * @param candidatePaths an array of candidate paths from which to load a
     *                      resource in order of preference
     * @return a URL corresponding to one of the candidatePaths or null if not
     *         found
     * @see DocumentLoaderImpl#getResource(String[])
     * @see #getPathURLFromServletContext(String[])
     */
    public URL getResource(final String[] candidatePaths) {
        final URL url = super.getResource(candidatePaths);
        return url != null ? url : getPathURLFromServletContext(candidatePaths);
    }

    /**
     * Iterates over candidatePaths looking up each path from the servlet context.
     * <p>
     * Prior to servlet context resource lookup, each candidate path is prefixed
     * by a forward slash ensuring it is properly formatted.  Even so if a
     * MalformedURLException still occurs, it is squashed and treated merely as
     * a candidate path which was not found.
     * </p>
     * 
     * @param candidatePaths an array of candidate paths from which to load a
     *                      resource in order of preference
     * @return a URL corresponding to one of the candidatePaths or null if not
     *         found
     * @see #getPathURLFromServletContext(String)
     */
    protected URL getPathURLFromServletContext(final String[] candidatePaths) {
        if (fServletContext == null) return null;
        URL url = null;
        for (int i = 0; i < candidatePaths.length; i++) {
            try {
                url = getPathURLFromServletContext("/" + candidatePaths[i]);
                if (url != null) break;
            } catch (final MalformedURLException e) {
                //squash
            }
        }
        return url;
    }

    /**
     * @param candidatePath a candidate path from which to load a resource (with
     *                      a leading slash)
     * @return a URL corresponding the candidatePath or null if not found
     * @throws MalformedURLException if the path is not given in the correct
     *                               form (with a leading slash)
     * @throws NullPointerException if the constructor provided servletContext is null
     */
    protected URL getPathURLFromServletContext(final String candidatePath) throws MalformedURLException {
        final URL url = fServletContext.getResource(candidatePath);
        if (url != null && fFactory.getLogger().debugEnabled()) {
            fFactory.getLogger().logDebug(">>>Get document '"
                                 + url
                                 + "' from servlet context");
        }
        return url;
    }

}
