/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLCContext.java,v 1.7 2005-09-03 05:05:56 jkjome Exp $
 */

package org.enhydra.xml.xmlc.servlet;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.zip.GZIPOutputStream;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.enhydra.xml.dom.DOMStats;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.URLRewriter;
import org.enhydra.xml.xmlc.XMLCFactory;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;

//TODO - Make XMLCLogger configuration more useful than forcing ServletXMLCLogger constructor parameters on all configured loggers
//TODO - STATS logging goes directly to servlet context log.  Probably should change to go through logger.
/**
 * Class to facility the use of XMLC in a HTTP servlet.  This object contains
 * a XMLC factory to create instances of XMLC document classes and methods to
 * output documents.
 *
 * <P>
 * An instance of this object is associated with the ServletContext object
 * for an application (one per WebApp).  The object is created on first use by
 * {@link #getContext getContext}.  It provides a factory for creating XMLC
 * document classes.  Also provided are methods to output a XMLC document as an
 * HTTP response, with automatic setting of response content type.
 * <P>
 * Note: that {@link XMLCContextInit} supports a context-param named
 * <tt>xmlcContextImpl</tt>, which allows for custom XMLCContext implementations
 * to be defined.  The value must be a FQCN of the custom XMLCContext, which must
 * extend XMLCContext and define a constructor with the same signature as XMLCContext,
 * e.g.,...
 * <PRE>
 * public MyCustomXMLCContext(ServletContext servletContext,
 *               XMLCFactory factory,
 *               SessionURLEncodingMode sessionURLEncodingMode,
 *               boolean logStats,
 *               int compressionType)
 * </PRE>
 * <P>
 * Instances of this class are configured using the following Servlet
 * context-wide init parameters:
 * <UL>
 * <LI> <tt>xmlcReparseResourceDirs</tt> - A list of additional directories to 
 *      load the html and meta data files, separated by any of "\t\n\r\f,;".
 *      Space characters can't be used as delimiters because there might be
 *      spaces in the path.  Colons can be used only on Unix because on Windows
 *      they are used when specifying drive letters.  A couple examples are
 *      "/my/project/resources:/other/project/resources" on Unix or
 *      "D:\my\project\resources;E:\other\project\resources" on Windows.
 * <LI> <tt>xmlcReparsePackagePrefixes</tt> - A list of package prefixes to be
 *      removed while searching for markup and metadata files, separated by
 *      any of " \t\n\r\f,;:", such as "my.package, other.package".
 * <LI> <tt>xmlcReparseDefaultMetaDataPath</tt> - A path to the default meta data
 *      file relative to the resource dirs.  Generally, this would just be
 *      "options.xmlc".  Deeper paths would be specified using either dot
 *      ("my.package.options.xmlc") or slash ("my/package/options.xmlc") notation.
 * <LI> <tt>xmlcReparseDocumentLoader</tt> (was <tt>xmlcRecompilationHandler</tt>
 *      which is now deprecated) - Defines the class of the document loader.
 *      Defaults to <tt>org.enhydra.xml.xmlc.deferredparsing.DocumentLoaderImpl</tt> 
 *      if not set.  The specified class must implement the 
 *      <tt>org.enhydra.xml.xmlc.deferredparsing.DocumentLoader</tt> interface.
 * <LI> <tt>xmlcLogger</tt> - Defines the logger class. Defaults to
 *      <tt>org.enhydra.xml.xmlc.servlet.ServletXMLCLogger</tt> if not
 *      set. The specified class must implement
 *      {@link org.enhydra.xml.xmlc.XMLCLogger} and
 *      define a constructor with a signature of <code>MyXMLCLogger(
 *      ServletContext servletContext, boolean enableInfoLogging, boolean
 *      enableDebugLogging)</code>, same as {@link ServletXMLCLogger#ServletXMLCLogger(ServletContext, boolean, boolean)}.
 * <LI> <tt>xmlcLogging</tt> - Control's what runtime information XMLC logs.
 *    Use one or more of the following values, separated by any one of " \t\n\r\f,;:":
 *    <UL>
 *    <LI> <tt>INFO</tt> - Log basic information about notable events, such as
 *         parsing/reparsing markup files or warnings.
 *    <LI> <tt>DEBUG</tt> - Log debugging information.  Mostly related to
 *         parsing/reparsing markup files.
 *    <LI> <tt>STATS</tt> - Log statistics information useful in debugging
 *         performance problems.  Currently writes information about each DOM
 *         that is written.  This is especially useful in looking at the
 *         how much of a Lazy DOM has been expanded.
 *    </UL>
 *    The default is <tt>INFO</tt>.
 * <LI> <tt>xmlcSessionURLEncoding</tt> - Controls encoding of session ID in URLs.
 *    <UL>
 *    <LI> <tt>auto</tt> - Automatically enable session URL encoding as needed
 *         (default)
 *    <LI> <tt>always</tt> - Always enable session URL encoding.  This does not
 *         work on most servers, as they will always try to send a cookie and
 *         not encode a URL if the session is from a cookie.
 *    <LI> <tt>never</tt> - Never enable session URL encoding.
 *    </UL>
 * <LI> <tt>xmlcCompression</tt> - Controls compression of the response
 * sent to the client. The following values are recognized:
 *    <UL>
 *    <LI> <tt>none</tt> - Responses are never compressed
 *    <LI> <tt>gzip</tt> - Responses are compressed with gzip if the client
 *         indicates it can handle gzip compressed content by sending an appropriate
 *         <em>Accept-Encoding</em> HTTP header.
 *    </UL>
 *    The default is <em>none</em>
 * <LI> <tt>xmlcEnableI18n</tt> - Enables Internationalization (I18n), or
 * Localization (L10n), functionality.  Set to <tt>true</tt> to enable and anything
 * else to leave disabled.  By default, this functionality is disabled as it may
 * entail a small performance impact, which makes little sense to incur when no
 * I18n capability is required.  When "true", given markup file foo.html and
 * Locale("en", "US"), attempts will be made to lookup foo_en_US.html, foo_en.html,
 * and finally fall back to the default fallback locale, set using
 * <code>xmlcDefaultFallbackLocale</code> as below, set directly using
 * <code>XMLCDeferredParsingFactory.setDefaultFallbackLocale(Locale)<code>, or
 * using <code>XMLCCreateOptions.setFallbackLocale(Locale)</code> passed to the
 * create(XMLCCreateOptions) method, which takes precedence.  If no template is
 * found for the user-defined or fallback locale, the ultimate fallback is the
 * base template; foo.html.  The
 * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions#setUserLocale(java.util.Locale) user-defined}
 * and/or
 * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions#setFallbackLocale(java.util.Locale) fallback}
 * Locale objects are
 * provided to the deferred parsing factory
 * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory#create(org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions)  create(XMLCCreateOptions)}
 * method.  The markup file base name/path is either
 * {@link org.enhydra.xml.xmlc.deferredparsing.XMLCCreateOptions#setDynMarkupFilePath(String) given directly}
 * or taken from the <tt>'XMLC_SOURCE_FILE'</tt> constant in an XMLC-compiled class.
 * <LI> <tt>xmlcDefaultFallbackLocale</tt> - The default fallback Locale, used
 * in the case where I18n is enabled and no locale-named template is found to
 * match the user-defined locale and no non-null fallbackLocale is provided to
 * XMLCCreateOptions passed to the create(XMLCCreateOptions) method, e.g., en_US.
 * If no fallback is specified here or on the create(XMLCCreateOptions) method,
 * then the final fallback is the base template.
 * </UL>
 *
 * @see ServletContext
 */
public class XMLCContext {
    /**
     * Value indicating that URL session encoding should be detected
     * automatically based on if it was used for the current requested.
     */
    public static final SessionURLEncodingMode URL_ENCODING_AUTO
        = new SessionURLEncodingMode("auto");
    
    /**
     * Value indicating that URL session encoding should always be used.
     */
    public static final SessionURLEncodingMode URL_ENCODING_ALWAYS
        = new SessionURLEncodingMode("always");
    
    /**
     * Value indicating that URL session encoding should never be used.
     */
    public static final SessionURLEncodingMode URL_ENCODING_NEVER
        = new SessionURLEncodingMode("never");

    /**
     * Enumerated constant indicating how URL session encoding is to
     * be handled.
     */
    static public class SessionURLEncodingMode {
        /** String name of the mode */
        private final String fModeName;
        
        /** Constructor */
        private SessionURLEncodingMode(String modeName) {
            fModeName = modeName;
        }

        /** Get the string value of the mode */
        public String toString() {
            return fModeName;
        }
    }

    /**
     * Name of ServletContext attribute that contains the instance of this class
     * for a WebApp.
     */
    public static final String CONTEXT_ATTRIBUTE
        = "http://www.enhydra.org/xmlc/XMLCContext";

    /**
     * The servlet context we are associated with.
     */
    private ServletContext fServletContext;

    /*
     * XMLC factory to use.
     */
    private XMLCFactory fFactory;

    /*
     * URL encoding of session key configuration. 
     */
    private SessionURLEncodingMode fSessionURLEncodingMode;

    /*
     * Log various statistics.
     */
    private boolean fLogStats;

    /*
     * Allowed compression types
     */
    private int fCompressionType;

    /**
     * Definition of xmlcCompression parameter values as a bit set.
     */
    static final int XMLC_COMPRESS_NONE  = 0x00;
    static final int XMLC_COMPRESS_GZIP  = 0x01;

    /**
     * Constructor. Must be public because it is accessed via
     * reflection, but instances should only be created through
     * {@link #getContext getContext}.
     */
    public XMLCContext(ServletContext servletContext,
                XMLCFactory factory,
                SessionURLEncodingMode sessionURLEncodingMode,
                boolean logStats,
		int compressionType) {
        fServletContext = servletContext;
        fFactory = factory;
        fSessionURLEncodingMode = sessionURLEncodingMode;
        fLogStats = logStats;
        fCompressionType = compressionType;
    }

    /**
     * Get the XMLC factory object associated with the context.
     * @deprecated
     * @see #getXMLCDeferredParsingFactory
     */
    public XMLCFactory getXMLCFactory() {
        return fFactory;
    }

    /**
     * Explicitly set XMLC factory.  The XMLC factory is normally defined base
     * on the servlet configuration parameters. This allows the user to
     * override the standard factories.
     * @throws ClassCastException if the factory is not an instance of XMLCDeferredParsingFactory
     * @deprecated
     * @see #setXMLCDeferredParsingFactory
     */
    public void setXMLCFactory(XMLCFactory factory) {
        setXMLCDeferredParsingFactory((XMLCDeferredParsingFactory) factory);
    }

    /**
     * Get the XMLC factory object associated with the context.
     * @since 2.3.2
     */
    public XMLCDeferredParsingFactory getXMLCDeferredParsingFactory() {
        return (XMLCDeferredParsingFactory) fFactory;
    }

    /**
     * Explicitly set XMLC factory.  The XMLC factory is normally defined base
     * on the servlet configuration parameters. This allows the user to
     * override the standard factories.
     * @since 2.3.2
     */
    public void setXMLCDeferredParsingFactory(XMLCDeferredParsingFactory factory) {
        fFactory = factory;
    }

    /**
     * Set the session URL encoding mode.  This option is normally set from
     * the servlet configuration parameters.  This allows the explicit setting
     * of this option.
     *
     * @param mode Constant indicating what mode to use 
     * @see SessionURLEncodingMode
     */
    public void setSessionURLEncoding(SessionURLEncodingMode mode) {
        fSessionURLEncodingMode = mode;
    }

    /*
     * Get the session URL encoding mode.
     *
     * @return The constant indicating what mode should be used for encoding.
     * @see SessionURLEncodingMode
     */
    public SessionURLEncodingMode getSessionURLEncoding() {
        return fSessionURLEncodingMode;
    }

    /**
     * Determine if a session id should be encoded URLs.
     */
    protected boolean shouldEncodeSessionId(HttpServletRequest request) {
        // This checks if the session is from a cookie.  This causes these
        // session id to always be encoded on the first request, as well as
        // sent in a cookied.  If it doesn't come back in a cookie, we keep
        // encoding it.
        return (((fSessionURLEncodingMode == URL_ENCODING_AUTO)
                 && request.isRequestedSessionIdValid()
                 && !request.isRequestedSessionIdFromCookie())
                || (fSessionURLEncodingMode == URL_ENCODING_ALWAYS));
    }

    /**
     * Setup the session URL encoder as required by configuration and
     * request.
     */
    protected URLRewriter setupURLRewriter(HttpServletRequest request,
                                         final HttpServletResponse response) {
        if (shouldEncodeSessionId(request)) {
            return new URLRewriter() {
                    public String rewriteURL(String urlAttrValue) {
                        return response.encodeURL(urlAttrValue);
                    }
                };
        } else {
            return null;
        }
    }

    /**
     * Set up the encoding using the encoding specified, OutputOptions,
     * the response, or the XMLC default encoding.  Modifies OutputOptions. 
     */
    private void setupEncoding(HttpServletResponse response,
                               XMLObject document,
                               OutputOptions options) {
        
        // If encoding is already specified, we don't override.
        if (options.getEncoding() == null) {
            // Use document default encoding.
            String encoding = document.getInputEncoding() != null ? document.getInputEncoding() : document.getXmlEncoding();
            if (encoding != null) {
                options.setEncoding(encoding);
            }
        }
    }

    /**
     * Create an OutputOptions object for a document.  Options are defaulted for
     * the specified document.  The object maybe then modified as needed to
     * override the default values.
     * <P>
     * The following attributes are set in the object:
     * <UL>
     * <LI> encoding
     * <LI> MIME type - Defaults for document.
     * <LI> URLRewriter - Set if URL encoding of sessions is enabled.
     * </UL>
     *
     * @param request The servlet request object.  The request is required
     *  to determine if URL encoding should be done on the document.
     * @param response The servlet response object.  The response is required
     *  to set up the URL encoder.
     * @param document The DOM object to be returned as response.
     */
    public OutputOptions createOutputOptions(HttpServletRequest request,
                                             HttpServletResponse response,
                                             XMLObject document) {
        OutputOptions options = DOMFormatter.getDefaultOutputOptions(document);

        setupEncoding(response, document, options);

        // Get the default MIME type from object
        options.setMIMEType(document.getMIMEType());

        // Enable encoding of session id in URLs, if needed.
        options.setURLRewriter(setupURLRewriter(request, response));
        return options;
    }

    /**
     * Write headers, disabling caching if other policy is not already
     * set.
     */
    private void outputHeaders(HttpServletResponse response,
                               int docLength,
                               String mimeEncoding,
                               String mimeType,
			       boolean compressed) throws IOException {

        // Set headers
        response.setContentLength(docLength);
        if (mimeType != null) {
            if (mimeEncoding != null) {
                response.setContentType(mimeType + "; charset=" + mimeEncoding);
            } else {
                response.setContentType(mimeType);
            }
        }
	if (compressed) {
	    response.setHeader("Content-Encoding","gzip");
	}
        
        // Disable caching of page unless otherwise specified.
        if (response.containsHeader("Cache-Control")) {
            response.setHeader("Cache-Control", "no-cache");
        }
        if (response.containsHeader("Expires")) {
            response.setHeader("Expires", "0");
        }
    }

    /**
     * Output document statistics.
     */
    private void logDocumentStats(XMLObject document) {
        // Buffer results so it gets logged as a single entry
        StringWriter buf = new StringWriter(2048);
        PrintWriter writer = new PrintWriter(buf);
        DOMStats.printStats("Write " + document.getClass().getName(),
                            document, 0, writer);
        writer.flush();
        fServletContext.log(buf.toString());
    }

    /**
     * Do common operations for writing DOM document.  Sets approriate HTTP
     * header, including setting of cache-control headers, if not already set.
     * Handles encoding conversion and output.
     */
    private void outputDocument(HttpServletResponse response,
                                XMLObject document,
                                OutputOptions outputOptions,
				boolean compress) throws IOException {

        // Convert to bytes and output
        DOMFormatter formatter = new DOMFormatter(outputOptions);
        byte[] docBytes = null;
	if (compress) {
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    GZIPOutputStream os = new GZIPOutputStream(bos);
	    formatter.write(document, os);
	    os.finish();
	    docBytes = bos.toByteArray();
	} else {
	    docBytes = formatter.toBytes(document);
	}

        outputHeaders(response, docBytes.length,
                      outputOptions.getMIMEEncoding(),
                      outputOptions.getMIMEType(),
		      compress);

        // Log if requested
        if (fLogStats) {
            logDocumentStats(document);
        }

        // Output
        OutputStream out = response.getOutputStream();
        out.write(docBytes);
        out.flush();
    }

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to it's type.  The MIME type of the response is automatically
     * set.
     *
     * @param request The servlet request object.  The request is required
     *  to determine if URL encoding should be done on the document.
     * @param response The servlet response object.  The 
     * <tt>Content-Type</tt> and <tt>Content-Length</tt> will be set from
     * the document..
     * @param outputFormat Object use to specify options controlling the 
     *  formatting of the document.
     * @param document The DOM object to be returned as response.
     */
    public void writeDOM(HttpServletRequest request,
                         HttpServletResponse response,
                         OutputOptions outputOptions,
                         XMLObject document) throws IOException {
        // Copy options so as to not modify when changing encoding.
        OutputOptions options = new OutputOptions(outputOptions);
        setupEncoding(response, document, options);
        // Get the default MIME type from object if no mimetype is set
        if (options.getMIMEType() == null) {
            options.setMIMEType(document.getMIMEType());
        }

        // Enable encoding of session id in URLs, if needed and no URL 
        // rewriter has been specified in the options
        if (options.getURLRewriter() == null) {
            options.setURLRewriter(setupURLRewriter(request, response));
        }
        outputDocument(response, document, options, useCompression(request));
    }

    /**
     * Check if the output should be compressed for this particular request
     */
    private boolean useCompression(HttpServletRequest request) {
	// check for output compression if it is enabled
	boolean compress = false;
	if ((fCompressionType & XMLC_COMPRESS_GZIP) != 0) {
	    String enc = request.getHeader("Accept-Encoding");
	    if (enc != null) {
		int pos = enc.indexOf("gzip");
		// FIXME: check for a value of 0 and disable compression if found.
		// The current implementation violates the HTTP 1.1 protocol
		// by sending gzip compressed content when the client explicitly
		// *excluded* gzip compression :-(
		// See RFC 2616, section 14.3 for details
		compress = (pos > -1);
	    }
	}
	return compress;
    }

    /**
     * Output an an XMLC document object (DOM). The document is formatted
     * according to its type.  The MIME type of the response is automatically
     * set.
     *
     * @param request The servlet request object.  The request is required
     *  to determine if URL encoding should be done on the document.
     * @param response The servlet response object.  The 
     * <tt>Content-Type</tt> and <tt>Content-Length</tt> will be set from
     * the document..
     * @param document The DOM object to be returned as response.
     */
    public void writeDOM(HttpServletRequest request,
                         HttpServletResponse response,
                         XMLObject document) throws IOException {
        OutputOptions options
            = createOutputOptions(request, response, document);
        outputDocument(response, document, options, useCompression(request));
    }

    /**
     * Obtain the XMLCContext for the current application, creating it if
     * it doesn't exist.
     *
     * @param servlet A servlet in the application.  Used to obtain the servlet context.
     * @return The XMLC context object for the application.
     * @see #getContext(ServletContext)
     */
    public static XMLCContext getContext(HttpServlet servlet) {
        ServletContext servletContext = servlet.getServletContext();
        return getContext(servletContext);
    }

    /**
     * Obtain the XMLCContext for the current application, creating it if
     * it doesn't exist.
     *
     * @param servletContext The servlet context of this application.
     * @return The XMLC context object for the application.
     */
    public static XMLCContext getContext(ServletContext servletContext) {
        //We used to synchronize here, but that causes problems in some environments (JBoss).
        //When the application is reloaded the servlet context object changes, but the static
        //variables used to avoid unnecessary re-checks and synchronization aren't cleared.  As
        //such, after the first initialization (at first use after app server startup), the
        //XMLCContext object would never be re-set, even if the servlet context object is re-set.
        //The fix is to forget the initialized check, forget the synchronization, and take the
        //hit on looking up the XMLCContext object every time this method is called.  This is
        //reasonable and shouldn't cause problems since the only worry about resetting the
        //XMLCContext object is performance.  No real need for synchronization. --Jake
        XMLCContext context = (XMLCContext)servletContext.getAttribute(CONTEXT_ATTRIBUTE);
        if (context != null) {
            return context;
        }
        context = new XMLCContextInit().createContext(servletContext);
        servletContext.setAttribute(CONTEXT_ATTRIBUTE, context);
        return context;
    }
    

    /** Get the associated servlet context */
    protected ServletContext getServletContext() { return fServletContext; }
}
