/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLParse.java,v 1.4 2007-04-22 05:40:59 jkjome Exp $
 */

package org.enhydra.xml.devtools;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.apache.xerces.parsers.DOMParser;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.io.XMLEntityResolver;
import org.w3c.dom.Document;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Parse files with the Xerces parser and optionally display information.
 *   XMLParse [options] in.xml [output]
 *     -dump
 *     -ns
 *     -no-validation
 *     -cp-resolve
 *     -debug
 *     -grammar
 */
public class XMLParse implements ErrorHandler {
    /** Count of errors */
    private int fErrorCnt = 0;

    /** Arguments parsed from the command line */
    private boolean fDump = false;
    private boolean fNameSpaces = false;
    private boolean fClassPathResolve = false;
    private boolean fDebugFlag = false;
    private boolean fGrammar = false;
    private boolean fValidate = true;

    /**
     * Print an error message.
     */
    private void printError(String msg,
                            SAXParseException ex) {
        System.err.println(ex.getSystemId() + ":"
                           + ex.getLineNumber() + ":"
                           + ex.getColumnNumber() + ":"
                           + msg + ":"
                           + ex.toString());
    }
    
    /** 
     * @see ErrorHandler.warning
     */
    public void warning(SAXParseException ex) {
        printError("Warning", ex);
    }
    
    /** 
     * @see ErrorHandler.error
     */
    public void error(SAXParseException ex) {
        printError("Error", ex);
        fErrorCnt++;
    }
    
    /** 
     * @see ErrorHandler.fatalError
     */
    public void fatalError(SAXParseException ex) throws SAXException {
        printError("Fatal error", ex);
        throw ex;
    }

    /**
     * DOM parser class.  Allows control over features of Xerces parser.
     */
    class XercesDOMParser extends DOMParser {
        /**
         * Constructor.
         */
        public XercesDOMParser() throws SAXException {
            super.setFeature("http://xml.org/sax/features/namespaces", fNameSpaces);
            super.setFeature("http://xml.org/sax/features/validation", fValidate);
            super.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false);

            if (fGrammar) {
                super.setFeature("http://apache.org/xml/features/domx/grammar-access", true);
            }
        }
    }

    /*
     * Parse and XML object and dump information about it.
     */
    private Document parse(String xmlFile) throws IOException, SAXException {
        XercesDOMParser parser = new XercesDOMParser();
        parser.setErrorHandler(this);
        if (fClassPathResolve) {
            XMLEntityResolver resolver = new XMLEntityResolver();
            resolver.setDefaultResolving();
            parser.setEntityResolver(resolver);
            if (fDebugFlag) {
                resolver.setDebugWriter(new PrintWriter(System.err, true));
            }
        }

        parser.parse(xmlFile);
        return parser.getDocument();
    }

    /**
     * Parse and dump the document.
     */
    private void process(String xmlFile,
                         OutputStream output) throws IOException, SAXException {
        Document doc = parse(xmlFile);
        if (fDump) {
            DOMInfo.printTree(null, doc, output);
        }
    }

    /*
     * Parse arguments and execute request.
     */
    private void run(String[] args) throws Throwable {
        int argIdx = 0;
        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-dump")) {
                fDump = true;
            } else if (args[argIdx].equals("-ns")) {
                fNameSpaces = true;
            } else if (args[argIdx].equals("-cp-resolve")) {
                fClassPathResolve = true;
            } else if (args[argIdx].equals("-debug")) {
                fDebugFlag = true;
            } else if (args[argIdx].equals("-grammar")) {
                fGrammar = true;
            } else if (args[argIdx].equals("-no-validation")) {
                fValidate = false;
            } else {
                System.err.println("Error: invalid argument \"" + args[argIdx]
                                   + "\", expected one of -dump, -ns, -cp-resolve, -debug, -grammar");
                System.exit(1);
            }
            argIdx++;
        }

        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            System.err.println("wrong # args: xml-parse [options] xml.file [output]");
            System.exit(1);
        }
        String xmlFile = args[argIdx];
        String outFile = (argsLeft > 1) ? args[argIdx+1] : null;

        OutputStream out;
        if (argsLeft == 1) {
            out = System.out;
        } else {
            out = new FileOutputStream(args[argIdx+1]);
        }
        process(xmlFile, out);
        out.flush();
    }

    /*
     * Entry point.
     */
    public static void main(String[] args) throws Throwable {
        try {
            new XMLParse().run(args);
        } catch (Throwable error) {
            if (error instanceof ExceptionInInitializerError) {
                error = ((ExceptionInInitializerError)error).getException();
            }
            System.err.println(error.toString());
            error.printStackTrace();
            System.exit(2);
        }
        
    }
}
