/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: RunTests.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.driver;

import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.HashSet;

import junit.framework.TestSuite;

/**
 * Program for running tests without junit.  This is a much easier environment
 * for debugging.
 * <PRE>
 *    run-tests [options] testcase1 ...
 * </PRE>
 * <UL>
 * <LI> <CODE>-test-root <EM>dir</EM></CODE> - Specifyies test case root.
 * <LI> <CODE>-parser <EM>name</EM></CODE> - Specify the parser to use.
 * <LI> <CODE>-dom <EM>name</EM></CODE> - Specify the DOM to run the tests on.
 * <LI> <CODE>-reloading</CODE> - Enable reloading testing.
 * <LI> <CODE>-no-reloading</CODE> - Do normal loading testing.
 * <LI> <CODE>-update</CODE> - Updated the expected test results.
 * <LI> <CODE>-test <EM>method</EM></CODE> - Restrict test to listed methods.
 *       These are unqualified method names. This option maybe repeated.
 * </UL>
 */
public class RunTests implements TestCaseBase.TestSelector {
    /** Compile-time debugging of tests */
    private static boolean DEBUG = false;

    /** Code for finding the static factory */
    private Class[] FACTORY_SIG = {
        Class.class, TestCaseBase.TestSelector.class
    };
    
    /** Parameters */
    private HashSet fStepRestrictions;

    /** Output stream */
    private PrintWriter fOut = new PrintWriter(System.err, true);

    /** 
     * Determine if a test method should be executed
     * @see TestCaseBase.TestSelector
     */
    public boolean select(Method method) {
        if (fStepRestrictions == null) {
            return true;
        } else {
            return fStepRestrictions.contains(method.getName());
        }
    }

    /** 
     * Create a test suite.
     */
    private TestSuite createTestSuite(String className) {
        // Create the suite though a reflective call to the static
        // createSuite() method.  This object servers as a selection
        // filter to limit the tests actually choosen.
        try {
            Class testClass = getClass().getClassLoader().loadClass(className);
            Method createSuite = testClass.getMethod("createSuite",
                                                     FACTORY_SIG);
            Object[] args = new Object[]{testClass, this};
            return (TestSuite)createSuite.invoke(null, args);
        } catch (Exception except) {
            throw new TestException(except);
        }
    }

    /** Run a test */
    private void runTest(TestCaseBase test) {
        fOut.println("RUN: " + test);
        if (DEBUG) {
            test.dumpInfo(fOut);
        }
        boolean succeed = false;
        try {
            test.runTest();
            succeed = true;
        } catch (Throwable except) {
            fOut.println("FAILED: " + test + ": " + except);
            if (!(except instanceof DiffException)) {
                except.printStackTrace();
            }
        }
        if (succeed) {
            fOut.println("SUCCEED: " + test);
        }
    }  

    /** All tests in a suite */
    private void runTestSuite(TestSuite suite) {
        Enumeration tests = suite.tests();
        while (tests.hasMoreElements()) {
            runTest((TestCaseBase)tests.nextElement());
        }
    }

    /*
     * Parse arguments and run
     */
    private void run(String[] args) {
        int argi = 0;
        while ((argi < args.length) && args[argi].startsWith("-")) {
            if (args[argi].equals("-test-root")) {
                TestProperties.setTestRoot(args[++argi]);
            } else if (args[argi].equals("-dom")) {
                TestProperties.setDom(args[++argi]);
            } else if (args[argi].equals("-parser")) {
                TestProperties.setParser(args[++argi]);
            } else if (args[argi].equals("-reloading")) {
                TestProperties.setReloading(true);
            } else if (args[argi].equals("-no-reloading")) {
                TestProperties.setReloading(false);
            } else if (args[argi].equals("-update")) {
                TestProperties.setUpdate(true);
            } else if (args[argi].equals("-test")) {
                if (fStepRestrictions == null) {
                    fStepRestrictions = new HashSet();
                }
                fStepRestrictions.add(args[++argi]);
            } else {
                fOut.println("invalid opt: " + args[argi]);
            }
            argi++;
        }
        TestProperties.setVerbose(true);

        for (; argi < args.length; argi++) {
            runTestSuite(createTestSuite(args[argi]));
        }
    }

    /** entry */
    public static void main(String[] args) {
        new RunTests().run(args);
    }
}

