/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TestCaseBase.java,v 1.5 2006-09-04 05:38:21 jkjome Exp $
 */

package org.enhydra.xml.driver;

import java.io.File;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Base class of all test classes.  Derived class implement
 * <CODE>testXXX()</CODE> methods for each test.  Each generated method is run
 * in an instance of this class.  Set package overview doc for why this 
 * approach was chosen.
 * <P>
 * The derived test case class should defined a method:
 * <PRE>
 *   public static Test suite();
 * </PRE>
 * that create methods by calling <CODE>createSuite()</CODE>.
 * <p>
 * Note: The key thing to remember is that a instance of this object
 * is created foreach test method!!!
 */
public class TestCaseBase extends TestCase {
    /** Derived class constructor signature */
    private static final Class[] CONST_SIGN = {
        Method.class
    };
    
    /** Empty arguments used in call test method */
    private static final Object[] NO_ARGS = new Object[0];

    /** Standard input directory */
    private static final File INPUT_DIR = new File("input");

    /** Output directory */
    private static final File OUTPUT_DIR = new File("output");

    /** Classes directory in output */
    private static final File CLASSES_DIR = new File(OUTPUT_DIR, "classes");

    /** Results directory in output */
    private static final File RESULTS_DIR = new File(OUTPUT_DIR, "results");

    /** Expected results directory */
    private static final File EXPECTED_DIR = new File("expected");

    /**
     * Part of package name that is skipped when generating result name.  This
     * is to avoid creating a bunch of directories in results and expected 
     * that don't really do much
     */
    private static final String RESULT_PKG_SKIP = "org.enhydra.xml.";

    /** Standard replacement patterns for diffing input */
    public static final String INPUT_REPLACE = "${testCaseRoot}/input";
    public static final String OUTPUT_REPLACE = "${testCaseRoot}/output";
        
    /** 
     * Class used in selecting tests to run.  If an instance of this is
     * supplied, it can be used to choose a subset of tests.
     */
    public interface TestSelector {
        /** Determine if a test method should be executed */
        public boolean select(Method method);
    }

    /** Method that this instace of the test will executed */
    private Method fMethod;

    /** 
     * Full test id, default is test case class plus method name, but
     * this may have more info in it.
     */
    private String fTestId;

    /** Writer for logging messages */
    private PrintWriter fMsgWriter;

    /** Write tracing if not null */
    private PrintWriter fVerboseOut;

    /** Should missing or miss-matched expected files be updated? */
    private boolean fUpdateExpected;

    /** 
     * Test case root directory.  This is the the path to the directory
     * containing this test case.  It's either the current directory or set
     * from a property.  All files are relative to this location.
     */
    private File fTestCaseRoot;

    /** Standard input directory, relative to test case root */
    private File fInputRoot;

    /** Output directory, relative to test case root */
    private File fOutputRoot;

    /** Classes directory in output */
    private File fClassesRoot;

    /** Results directory in output */
    private File fResultsRoot;

    /** Expected results directory */
    private File fExpectedRoot;

    /** Package directory (lazy) */
    private File fPackageDir;

    /** Relative result dir (lazy) */
    private File fRelResultsDir;

    /** Result dir (lazy) */
    private File fResultsDir;

    /** Expected result dir (lazy) */
    private File fExpectedDir;

    /** Class loader use to load compiled objects */
    private ClassLoader fTestClassLoader;

    /** Extra diff filters to use */
    private ArrayList fExtraFilters;

    /** Differ object that queues exceptions. */
    private TestDiff fDiffer;

    /**  Determine if a method is a test method */
    private static boolean isTestMethod(Method method) {
        // check name an signature
        return method.getName().startsWith("test")
            && Modifier.isPublic(method.getModifiers())
            && (method.getParameterTypes().length == 0)
            && method.getReturnType().equals(Void.TYPE);
    }

    /**  Determine if a selected test method */
    private static boolean isTestMethod(Method method,
                                        TestSelector selector) {
        return (isTestMethod(method)
                && ((selector == null) || selector.select(method)));
    }

    /** Create a test instance */
    private static Test createTest(Constructor constr,
                                   Method method) {
        try {
            return (Test)constr.newInstance(new Object[]{method});
        } catch (InstantiationException except) {
            throw new TestError(except);
        } catch (IllegalAccessException except) {
            throw new TestError(except);
        } catch (InvocationTargetException except) {
            throw new TestError(except);
        }
    }

    /**
     * Build the test suite
     */
    public static Test createSuite(Class testCaseClass,
                                   TestSelector selector) {
        Constructor constr;
        try {
            constr = testCaseClass.getConstructor(CONST_SIGN);
        } catch (NoSuchMethodException except) {
            throw new TestError(except);
        }
        TestSuite suite = new TestSuite();
        Method[] methods= testCaseClass.getDeclaredMethods();
        
        for (int idx = 0; idx < methods.length; idx++) {
            if (isTestMethod(methods[idx], selector)) {
                suite.addTest(createTest(constr, methods[idx]));
            }
        }
        return suite;
    }

    /**
     * Constructor for a specific test instance, specifying the unqualTestId.
     * Used when more than the test method name must identify the tests.
     */
    protected TestCaseBase(String unqualTestId,
                           Method method) {
        super(unqualTestId);
        fMethod = method;
        fTestId = method.getDeclaringClass().getName() + "."
            + unqualTestId;
        fMsgWriter = new PrintWriter(System.err, true);
        if (TestProperties.getVerbose()) {
            fVerboseOut = fMsgWriter;
        }
        fUpdateExpected = TestProperties.getUpdate();

        // get root and files relative to it
        fTestCaseRoot = TestProperties.getTestRoot();
        fInputRoot = new File(fTestCaseRoot, INPUT_DIR.getPath());
        fOutputRoot = new File(fTestCaseRoot, OUTPUT_DIR.getPath());
        fClassesRoot = new File(fTestCaseRoot, CLASSES_DIR.getPath());
        fResultsRoot = new File(fTestCaseRoot, RESULTS_DIR.getPath());
        fExpectedRoot = new File(fTestCaseRoot, EXPECTED_DIR.getPath());
    }

    /** 
     * Constructor for a specific test instance using method name as the
     * unqualTestId.
     */
    protected TestCaseBase(Method method) {
        this(method.getName(), method);
    }

    /**
     * Called by junit to run the test.
     */
    public void runTest() {
        try {
            fMethod.invoke(this, NO_ARGS);
        } catch (IllegalAccessException except) {
            throw new TestError(except);
        } catch (InvocationTargetException except) {
            Throwable except2 = except.getTargetException();
            if (except2 instanceof TestException) {
                throw (TestException)except2;
            }
            if (except2 instanceof TestError) {
                throw (TestError)except2;
            }
            throw new TestException(except2);
        }
        // If any diff failures occured,  generate an exception
        if (fDiffer != null) {
            fDiffer.throwPendingDiffFailures();
        }
    }

    /** Get the full test id */
    public String getTestId() {
        return fTestId;
    }

    /** Get the unqualified test id */
    public String getUnqualTestId() {
        return getName();
    }

    /**
     * Get the test method name.
     */
    public String getTestName() {
        return fMethod.getName();
    }
    
    /**
     * Get writer for writing messages
     */
    public PrintWriter getMsgWriter() {
        return fMsgWriter;
    }

    /** print a newline to the message writer */
    public void msgPrintln() {
        fMsgWriter.println();
    }

    /** print a message without newline */
    public void msgPrint(String msg) {
        fMsgWriter.print(msg);
    }

    /** print a message with newline */
    public void msgPrintln(String msg) {
        msgPrint(msg);
        msgPrintln();
    }

    /** Get the verbose writer, or null if not emabled */
    public PrintWriter getVerboseOut() {
        return fVerboseOut;
    }

    /**
     * Get root directory for the test case.  All files are relative to
     * this directory.
     */
    public File getTestCaseRoot() {
        return fTestCaseRoot;
    }

    /**
     * Get the test package.  This is the <em>class</em> name of the
     * concrete test case.  The other directories are derived from this
     * directory, so that overriding this method can change all of the
     * others.
     */
    public String getTestPackage() {
        // Need to add `Pkg' to prevent conflict with class name
        return getClass().getName() + "Pkg";
    }

    /**
     * Get the tail part of the package name that removes parts that are not
     * required to keep tests unique.
     */
    public String getTestPackageTail() {
        return getTestPackage().substring(RESULT_PKG_SKIP.length());
    }

    /**
     * Get the root of the class tree.
     */
    public File getClassRoot() {
        return fClassesRoot;
    }

    /** Get the test package directory, where compiled files are placed */
    public File getPackageDir() {
        if (fPackageDir == null) {
            String pkgPath = getTestPackage().replace('.', File.separatorChar);
            fPackageDir = new File(getClassRoot(), pkgPath);
        }
        return fPackageDir;
    }

    /** Get input file root dir */
    public File getInputRoot() {
        return fInputRoot;
    }

    /** Get path to an input file, give file relative to input/ */
    public File getInputFile(String fileName) {
        return new File(fInputRoot, fileName);
    }

    public URL getInputURL(String fileName) {
        URL url = null;
        try {
            url = getInputFile(fileName).toURL();
        } catch (MalformedURLException mue) {
            //squash
        }
        return url;
    }

    /** 
     * Get path to an input file that is in another test case directory.
     * Given the relative location of that directory and the file relative to
     * input.
     */
    public File getOtherInputFile(String otherTestCaseDir,
                                  String fileName) {
        File otherRoot = new File(fTestCaseRoot, otherTestCaseDir);
        File otherInput = new File(otherRoot, INPUT_DIR.getPath());

        // Add filter for this
        addOtherInputFileDiffFilter(otherTestCaseDir, otherRoot);

        return new File(otherInput, fileName);
    }
    
    /** Get output root dir */
    public File getOutputRoot() {
        return fOutputRoot;
    }

    /** 
     * Get the relative results directory (not including output or expected)
     */
    public File getRelResultsDir() {
        if (fRelResultsDir == null) {
            // Get package with several leading ones removed
            String pkgPath = getTestPackageTail().replace('.', File.separatorChar);
            fRelResultsDir = new File(pkgPath);
        }
        return fRelResultsDir;
    }

    /** Get the results directory for this test */
    public File getResultsDir() {
        if (fResultsDir == null) {
            fResultsDir = new File(fResultsRoot,
                                   getRelResultsDir().getPath());
            fResultsDir.mkdirs();  // always create 
        }
        return fResultsDir;
    }

    /** Get a result file path */
    public File getResultFile(String fname,
                              String suffix) {
        return new File(getResultsDir(), fname + "." + suffix); 
    }
    
    /** Get a result file path */
    public File getResultFile(String suffix) {
        return getResultFile(getTestName(), suffix);
    }
    
    /** Get the test expected root directory */
    public File getExpectedRoot() {
        return fExpectedRoot;
    }

    /** Get the test expected results directory */
    public File getExpectedDir() {
        if (fExpectedDir == null) {
            fExpectedDir = new File(fExpectedRoot,
                                    getRelResultsDir().getPath());
        }
        return fExpectedDir;
    }
    
    /** Get an expected file name. */
    public File getExpectedFile(String fname,
                                String suffix) {
        return new File(getExpectedDir(), fname + "." + suffix); 
    }

    /** Get an expected file name. */
    public File getExpectedFile(String suffix) {
        return getExpectedFile(getTestName(), suffix); 
    }

    /**
     * Add an additional diff filter for handling special-case
     * patterns.
     */
    public void addDiffFilter(TestDiff.LineFilter filter) {
        if (fExtraFilters == null) {
            fExtraFilters = new ArrayList();
        }
        fExtraFilters.add(filter);
    }

    /**
     * Define a diff filter for handling an input file that is coming
     * from an another, as obtained by getOtherInputFile().
     */
    private void addOtherInputFileDiffFilter(String otherTestCaseDir,
                                             File otherRoot) {
        addDiffFilter(new SimpleDiffFilter(otherRoot.getPath(),
                                           "${testCaseRoot}/" + otherTestCaseDir,
                                           '\\',
                                           '/'));
    }

    /**
     * Add filters to diff object.  These allow compares to not
     * fail due to path variations in where the test is run.
     */
    private void addFilters(TestDiff differ) {
        // add standard filters
        differ.addFilter(new SimpleDiffFilter(getInputRoot().getPath(),
                                              INPUT_REPLACE,
                                              '\\',
                                              '/'));
        differ.addFilter(new SimpleDiffFilter(getOutputRoot().getPath(),
                                              OUTPUT_REPLACE,
                                              '\\',
                                              '/'));

        // add extra filters
        if (fExtraFilters != null) {
            for (int idx = 0; idx < fExtraFilters.size(); idx++) {
                differ.addFilter((TestDiff.LineFilter)fExtraFilters.get(idx));
            }
        }
    }

    /** 
     * Get the diff object for the class. Failures are queued until
     * after the test completes.  The runTest method will then throw
     * the exception when complete.  Optionally updates expected.
     */
    public TestDiff getDiffer() {
        if (fDiffer == null) {
            int flags = TestDiff.QUEUE_FAILURES;
            if (fUpdateExpected) {
                flags |= TestDiff.UPDATE_EXPECTED;
            }
            fDiffer = new TestDiff(fMsgWriter, flags);
            addFilters(fDiffer);
        }
        return fDiffer;
    }

    /** 
     * Get a new diff object. If queue failures are requested, they
     * are not automatically thrown  as with <code>getDiffer()</code>.
     */
    public TestDiff createDiffer(boolean queueFailures) {
        int flags = 0;
        if (queueFailures) {
            flags |= TestDiff.QUEUE_FAILURES;
        }
        if (fUpdateExpected) {
            flags |= TestDiff.UPDATE_EXPECTED;
        }
        return new TestDiff(fMsgWriter, flags);
    }

    /**
     * Get a class loader for use in the tests We need to have own instance of
     * the classloader, as system classloader sees to cache file searchs and
     * doesn't see newly created classes.  Note that output directory must not
     * be on the system classpath.
     */
    public ClassLoader createClassLoader() {
        getClassRoot().mkdirs();
        URL pathUrl = TestFileOps.dirToUrl(getClassRoot());
        return new URLClassLoader(new URL[]{pathUrl},
                                  getClass().getClassLoader());
    }

    /**
     * Get the test classloader, creating if it doesn't exist.
     */
    public ClassLoader getTestClassLoader() {
        if (fTestClassLoader == null) {
            fTestClassLoader = createClassLoader();
        }
        return fTestClassLoader;
    }

    /**
     * Print information about the class for debugging.
     */
    public void dumpInfo(PrintWriter out) {
        out.println("testCase: " + getClass().getName());
        out.println("    getName(): " + getName());
        out.println("    getTestId(): " + getTestId());
        out.println("    getTestCaseRoot(): " + getTestCaseRoot());
        out.println("    getInputRoot(): " + getInputRoot());
        out.println("    getUnqualTestId(): " + getUnqualTestId());
        out.println("    getTestName(): " + getTestName());
        out.println("    getTestPackage(): " + getTestPackage());
        out.println("    getClassRoot(): " + getClassRoot());
        out.println("    getPackageDir(): " + getPackageDir());
        out.println("    getRelResultsDir(): " + getRelResultsDir());
        out.println("    getResultsDir(): " + getResultsDir());
        out.println("    getExpectedDir(): " + getExpectedDir());
    }
}

