/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TestFileOps.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.driver;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.net.URL;
import java.util.ArrayList;

/**
 * Class of static methods to do various file related operations.
 */
public class TestFileOps {
    /**
     * Seconds to sleep to allow for system time to advance.  This takes
     * makes sure that we sleep longer than the resolution of the clock
     * used for files.
     * FIXME: Find out the scoop on the resoultion, not sure this is really
     * need, it might be some other bug that was causing the problem.
     */
    public static final double TOUCH_DELAY = 1.0;

    /** Ensure that a file's directory exists. */
    public static void ensureFileDir(File file) {
        File dir = file.getParentFile();
        if (dir != null) {
            dir.mkdirs();
        }
    }

    /** Copy a file another file, creating dir if needed */
    public static void copyFile(File srcFile,
                                File destFile) {
        try {
            ensureFileDir(destFile);
            InputStream src = new FileInputStream(srcFile);
            try {
                OutputStream dest = new FileOutputStream(destFile);
                try {
                    int byteCnt;
                    byte[] buffer = new byte[8*1024];
                    while ((byteCnt = src.read(buffer, 0, buffer.length)) >= 0) {
                         dest.write(buffer, 0, byteCnt);
                    }
                } finally {
                    dest.close();
                }
            } finally {
                src.close();
            }
        } catch (IOException except) {
            throw new TestException("copy of \"" + srcFile + "\" to \""
                                    + destFile + "\" failed", except);
        }
    }

    /** Copy a file to a directory, creating dir if needed */
    public static void copyFileToDir(File srcFile,
                                     File destDir) {
        destDir.mkdirs();
        File destFile = new File(destDir, srcFile.getName());
        copyFile(srcFile, destFile);
    }

    /** Copy a list of files to a directory, creating dir if needed */
    public static void copyFilesToDir(File[] srcFiles,
                                      File destDir) {
        for (int idx = 0; idx < srcFiles.length; idx++) {
            copyFileToDir(srcFiles[idx], destDir);
        }
    }

    /** Copy a list of files to a directory, creating dir if needed */
    public static void copyFilesToDir(String[] srcFiles,
                                      File destDir) {
        for (int idx = 0; idx < srcFiles.length; idx++) {
            copyFileToDir(new File(srcFiles[idx]), destDir);
        }
    }

    /** Copy a list of files to a directory, creating dir if needed */
    public static void copyFilesToDir(ArrayList srcFiles,
                                      File destDir) {
        for (int idx = 0; idx < srcFiles.size(); idx++) {
            Object file = srcFiles.get(idx);
            if (file instanceof File) {
                copyFileToDir((File)file, destDir);
            } else {
                copyFileToDir(new File((String)file), destDir);
            }
        }
    }

    /**
     * Delay for the specified number of seconds.
     */
    public static void sleep(double secs) {
        try {
            Thread.sleep((long)(secs*1000));
        } catch (InterruptedException except) {
            throw new TestError(except);
        }
    }

    /**
     * Touch a file by reading and rewriting one byte.
     */
    public static void touchFile(File file) {
        // Wait, just to let file clock get ahead.
        sleep(TOUCH_DELAY);

        try {
            RandomAccessFile raFile = new RandomAccessFile(file, "rw");
            int fileByte = raFile.read();
            if (fileByte < 0) {
                throw new TestError("unexpected EOF on " + file);
            }
            raFile.seek(0);
            raFile.write(fileByte);
            raFile.close();
        } catch (FileNotFoundException except) {
            throw new TestError("error touching " + file, except);
        } catch (IOException except) {
            throw new TestError("error touching " + file, except);
        }
    }


    /** Create a URL from a directory File */
    static public URL dirToUrl(File file) {
        if (!file.isDirectory()) {
            throw new TestError("file for URL not a directory: \"" + file
                                + "\"");
        }
        try {
            return new URL("file", null, -1, file.getCanonicalPath() + "/");
        } catch (IOException except) {
            throw new TestError("can't construct URL from File", except);
        }
    }

    /** Create a URL from a file File */
    static public URL fileToUrl(File file) {
        try {
            // URL class loader requires a `/' at the  end if it's directory
            String path = file.getAbsolutePath();
            if (!(path.endsWith(".zip") || path.endsWith(".jar"))) {
                path += '/';
            }
            return new URL("file", null, -1, path);
        } catch (IOException except) {
            throw new TestError("can't construct URL from File", except);
        }
    }

    /**
     * Get the file extension.
     */
    public static String getFileExt(String fname) {
        int dotIdx = fname.lastIndexOf('.');
        if (dotIdx < 0) {
            return null;
        } else {
            return fname.substring(dotIdx+1);
        }
    }

    /**
     * Get the file extension.
     */
    public static String getFileExt(File file) {
        return getFileExt(file.getPath());
    }
}
