/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicTests.java,v 1.5 2007-11-24 21:09:54 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;

import junit.framework.Test;

import org.apache.xerces.parsers.DOMParser;
import org.enhydra.xml.driver.TestDiff;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.xmlc.html.HTMLDocumentFactory;
import org.w3c.dom.Document;
import org.w3c.dom.html.HTMLDocument;
import org.w3c.dom.html.HTMLElement;
import org.xml.sax.SAXException;

/**
 * Basic DOMFormatter tests, mostly of special cases.  More extensive 
 * tests are needed, however its well excerised by the XMLC tests.
 */
public class BasicTests extends IOTestCaseBase {
    // Character encodings.
    //private final String UTF8 = "UTF-8";
    private final String ISO8859_1 = "ISO8859-1";

    /** Test input files */ 
    private final String TEST1_WML = "xml/wml/test1.wml";
    private final String TEST2_WML = "xml/wml/test2.wml";

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(BasicTests.class, null);
    }
    
    /** Required constructor */
    public BasicTests(Method method) {
        super(method);
    }

    /**
     * Create skeleton html object with the specified text
     * in the body.
     */
    private HTMLDocument createHtml1(String text) {
        HTMLDocument doc = HTMLDocumentFactory.createBasicDocument(getTestName());
        HTMLElement body = doc.getBody();
        body.appendChild(doc.createTextNode(text));
        return doc;
    }

    /**
     * Parse an XML document
     */
    private Document parseXml(File xmlFile) { 
        try {
            DOMParser parser = new DOMParser();
            parser.parse(xmlFile.getPath());
            return parser.getDocument();
        } catch (IOException except) {
            throw new TestException(except);
        } catch (SAXException except) {
            throw new TestException(except);
        }
    }

    /**
     * Test 1: HTML formatting tests with special characters to a string.
     */
    public void test1() {
        // &#146; = 0122 = 0x96
        String data = "One\u0096s check";
        HTMLDocument doc = createHtml1(data);

        DOMFormatter fmt = new DOMFormatter();  // Defaults
        File outFile = getResultFile("html");
        try {
            FileOutputStream out = new FileOutputStream(outFile);
            try {
                out.write(data.getBytes(ISO8859_1));
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }
        TestDiff differ = getDiffer();
        differ.diff(getExpectedFile("html"), outFile);
    }

    /**
     * Test 2: HTML formatting tests with special characters to a file
     */
    public void test2() {
        // &#146; = 0122 = 0x96
        String data = "One\u0096s check";
        HTMLDocument doc = createHtml1(data);

        OutputOptions opts = new OutputOptions();
        opts.setEncoding(ISO8859_1);
        DOMFormatter fmt = new DOMFormatter(opts);
        File outFile = getResultFile("html");
        try {
            fmt.write(doc, outFile);
        } catch (IOException except) {
            throw new TestException(except);
        }
        TestDiff differ = getDiffer();
        differ.diff(getExpectedFile("html"), outFile);
    }

    /**
     * Test various output options useful for WML.  ASCII and omitting
     * the encoding.
     */
    private void testWML(File xmlFile) {
        Document doc = parseXml(xmlFile);

        OutputOptions opts = new OutputOptions();
        opts.setOmitEncoding(true);
        opts.setEncoding("ASCII");

        File outFile = getResultFile("wml");
        DOMFormatter domf = new DOMFormatter(opts);
        try {
            domf.write(doc, outFile);
        } catch (IOException except) {
            throw new TestException(except);
        }

        getDiffer().diff(getExpectedFile("wml"), outFile);
    }

    /**
     * Test 3: WML formatting options
     */
    public void test3() {
        testWML(getInputFile(TEST1_WML));
    }

    /**
     * Test 4: WML formatting options
     */
    public void test4() {
        testWML(getInputFile(TEST2_WML));
    }

}
