/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EncodingTests.java,v 1.4 2006-03-04 05:56:08 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Method;

import junit.framework.Test;

import org.enhydra.xml.driver.TestException;

/**
 * Test of encodings.
 */
public class EncodingTests extends IOTestCaseBase {
    /** Extension for result file with missing encodings being listed */
    private static final String MISSING_ENCODINGS_EXT
        = "missing.encodings";

    /**
     * Table of java support encodings from JDK documentation.
     * XMLC doesn't yet support all of these.
     */
    private static final String[] JAVA_ENCODINGS = {
        "ASCII", "Cp1252", "ISO8859_1", "UnicodeBig", "UnicodeBigUnmarked",
        "UnicodeLittle", "UnicodeLittleUnmarked", "UTF-8", "UTF-16",
        "Big5", "Cp037", "Cp273", "Cp277", "Cp278",
        "Cp280", "Cp284", "Cp285", "Cp297", "Cp420", "Cp424", "Cp437", "Cp500",
        "Cp737", "Cp775", "Cp838", "Cp850", "Cp852", "Cp855", "Cp856", "Cp857",
        "Cp858", "Cp860", "Cp861", "Cp862", "Cp863", "Cp864", "Cp865", "Cp866",
        "Cp868", "Cp869", "Cp870", "Cp871", "Cp874", "Cp875", "Cp918", "Cp921",
        "Cp922", "Cp930", "Cp933", "Cp935", "Cp937", "Cp939", "Cp942", "Cp942C",
        "Cp943", "Cp943C", "Cp948", "Cp949", "Cp949C", "Cp950", "Cp964", "Cp970",
        "Cp1006", "Cp1025", "Cp1026", "Cp1046", "Cp1097", "Cp1098", "Cp1112",
        "Cp1122", "Cp1123", "Cp1124", "Cp1140", "Cp1141", "Cp1142", "Cp1143",
        "Cp1144", "Cp1145", "Cp1146", "Cp1147", "Cp1148", "Cp1149", "Cp1250",
        "Cp1251", "Cp1253", "Cp1254", "Cp1255", "Cp1256", "Cp1257", "Cp1258",
        "Cp1381", "Cp1383", "Cp33722", "EUC_CN", "EUC_JP", "EUC_KR", "EUC_TW",
        "GBK", "ISO2022CN", "ISO2022CN_CNS", "ISO2022CN_GB", "ISO2022JP",
        "ISO2022KR", "ISO8859_2", "ISO8859_3", "ISO8859_4", "ISO8859_5",
        "ISO8859_6", "ISO8859_7", "ISO8859_8", "ISO8859_9", "ISO8859_13",
        "ISO8859_15_FDIS", "JIS0201", "JIS0208", "JIS0212", "JISAutoDetect",
        "Johab", "KOI8_R", "MS874", "MS932", "MS936", "MS949", "MS950",
        "MacArabic", "MacCentralEurope", "MacCroatian", "MacCyrillic",
        "MacDingbat", "MacGreek", "MacHebrew", "MacIceland", "MacRoman",
        "MacRomania", "MacSymbol", "MacThai", "MacTurkish", "MacUkraine", "SJIS",
        "TIS620"
    };

    /** Count of encodings that were not found in XMLC tables  */
    private int fUnknownEncodingCnt = 0;

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(EncodingTests.class, null);
    }
    
    /** Required constructor */
    public EncodingTests(Method method) {
        super(method);
    }

    /**
     * Check converting to encoding
     */
    private void checkEncodingConvert(String encoding)  {
        try {
            String data = "This is a test";
            byte[] b = data.getBytes(encoding);
        } catch (UnsupportedEncodingException uee) {
            msgPrintln("Note: String.getBytes(\"" + encoding + "\") not supported by this JVM");
            uee.printStackTrace();
        } catch (UnsupportedOperationException uoe) {
            msgPrintln("Note: String.getBytes(\"" + encoding + "\") not supported by this JVM");
            uoe.printStackTrace();
        }
    }
        
    /**
     * Test an encoding.
     */
    private void testEncoding(String encoding,
                              PrintWriter out) {
        Encodings encodings = Encodings.getEncodings();
        if (!encodings.isValid(encoding)) {
            out.println("No entry for \"" + encoding + "\"");
            fUnknownEncodingCnt++;
        }
        checkEncodingConvert(encoding);

        // At least try to get it.
        String mimePrefered = encodings.getMIMEPreferred(encoding);
        if (mimePrefered != null) {
            checkEncodingConvert(mimePrefered);
        }
    }

    /**
     * Test 1: Try converting a bunch of encodings.  This really isn't
     * much of a test
     */
    public void test1() {
        File resultFile = getResultFile(MISSING_ENCODINGS_EXT);
        try {
            PrintWriter out = new PrintWriter(new FileWriter(resultFile));
            try {
                for (int i = 0; i < JAVA_ENCODINGS.length; i++) {
                    testEncoding(JAVA_ENCODINGS[i], out);
                }
                out.println("" + fUnknownEncodingCnt
                            + " encodings not known to XMLC tables");
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }

        getDiffer().diff(getExpectedFile(MISSING_ENCODINGS_EXT),
                         resultFile);
    }
}

