/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: URLRewriteTests.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Method;

import junit.framework.Test;

import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.dom.DOMStats;
import org.enhydra.xml.driver.TestDomOps;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.driver.TestFileOps;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.driver.ExecXmlc;
import org.enhydra.xml.xmlc.driver.TestDOMSource;
import org.enhydra.xml.xmlc.driver.XmlcTestParams;
import org.w3c.dom.Node;

/**
 * Test DOMFormatter URL rewriting funtionality.
 */
public class URLRewriteTests extends IOTestCaseBase {
    /** Flag indicating test of omitAttributeCharEntityRefs */
    private static final int OMIT_ATTR_CHAR_ENTITY_REFS = 0x01;

    /** Traverse to expand Lazy DOM */
    private static final int EXPAND = 0x02;

    /** Optionally start at an element to test the XMLObjectLink stuff */
    private static final int START_AT_NODE = 0x04;

    /** Input file */
    private static final String TEST_DOC_HTML = "html/TestDoc1.html";
    private static final String TEST_DOC_XML = "xml/wml/test1.wml";

    /** 
     * DOM source objects used to create test documents.  These repersent
     * documents compiled with different options.
     */
    private TestDOMSource fXercesDomTestHtml;
    private TestDOMSource fXercesDomKestrelHtml;
    private TestDOMSource fLazyDomTestHtml; 
    private TestDOMSource fLazyDomKestrelHtml;
    private TestDOMSource fXercesDomTestXml;
    private TestDOMSource fLazyDomTestXml; 

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(URLRewriteTests.class, null);
    }
    
    /** Required constructor, params is ignored */
    public URLRewriteTests(Method method) {
        super(method);

        /// FIXME:make these static...
        // Initialize test DOM sources; actual compile happens on demand.
        XmlcTestParams xercesHtmlParams
            = new XmlcTestParams(false, ExecXmlc.TIDY_PARSER,
                                 ExecXmlc.XERCES_DOM, false);
        fXercesDomTestHtml
            = new TestDOMSource(this, xercesHtmlParams,
                                getInputFile(TEST_DOC_HTML),
                                "testHtml", null);
        fXercesDomKestrelHtml
            = new TestDOMSource(this, xercesHtmlParams,
                                getInputFile(TEST_DOC_HTML),
                                "kestrelHtml", Kestrel.class);

        XmlcTestParams lazyHtmlParams
            = new XmlcTestParams(false, ExecXmlc.TIDY_PARSER,
                                 ExecXmlc.LAZY_DOM, false);
        fLazyDomTestHtml
            = new TestDOMSource(this, lazyHtmlParams,
                                getInputFile(TEST_DOC_HTML),
                                "testHtml", null);
        fLazyDomKestrelHtml
            = new TestDOMSource(this, lazyHtmlParams,
                                getInputFile(TEST_DOC_HTML),
                                "kestrelHtml", Kestrel.class);

        XmlcTestParams xercesXmlParams
            = new XmlcTestParams(true, ExecXmlc.XERCES_PARSER,
                                 ExecXmlc.XERCES_DOM, false);
        fXercesDomTestXml
            = new TestDOMSource(this, xercesXmlParams,
                                getInputFile(TEST_DOC_XML),
                                "testXml", null);
        
        XmlcTestParams lazyXmlParams
            = new XmlcTestParams(true, ExecXmlc.XERCES_PARSER,
                                 ExecXmlc.LAZY_DOM, false);
        fLazyDomTestXml
            = new TestDOMSource(this, lazyXmlParams,
                                getInputFile(TEST_DOC_XML),
                                "testXml", null);
        
    }

    /**
     * Output a document with URL rewritting
     */
    private void outputWithRewritting(Node node,
                                      int flags,
                                      File outFile) throws IOException {
        OutputOptions options = new OutputOptions();
        if ((flags & OMIT_ATTR_CHAR_ENTITY_REFS) != 0) {
            options.setOmitAttributeCharEntityRefs(true);
        }

        options.setURLRewriter(new URLRewriter() {
                public String rewriteURL(String urlAttrValue) {
                    return urlAttrValue + "#REWRITTEN";
                }
            });
        DOMFormatter formatter = new DOMFormatter(options);
        formatter.write(node, outFile);
    }
    
    /**
     * Basic test step.  Output and compare with expected.
     */
    private void testStep(TestDOMSource domSource,
                          int flags) {
        String docFileExt = TestFileOps.getFileExt(domSource.getInputFile());
        XMLObject xmlcDoc = domSource.create();
        if ((flags & EXPAND) != 0) {
            TestDomOps.expandTree(xmlcDoc);
        }
        Node startingNode = ((flags & START_AT_NODE) != 0)
            ? (Node)xmlcDoc.getDocumentElement() : (Node)xmlcDoc;

        // Insert a attribute value with quotes:
        if (xmlcDoc instanceof Kestrel) {
            Kestrel kestrel = (Kestrel)xmlcDoc;
            kestrel.getElementKestrel().setHref("javascript:submitForm(\"\",\"folder\",\"/INBOX/\")");
        }

        try {
            // Output and verify
            File outFile = getResultFile(docFileExt);
            outputWithRewritting(startingNode, flags, outFile);
            getDiffer().diff(getExpectedFile(docFileExt), outFile);

            // Print DOM info and verify that DOM is not modified
            File outDom = getResultFile("dom");
            File expectDom = getExpectedFile("dom");
            PrintWriter writer = new PrintWriter(new FileWriter(outDom));
            DOMInfo.printTree("DOM after URL rewritting", xmlcDoc,
                              DOMInfo.PRINT_ATTR_DETAILS, writer);
            DOMStats.printStats("DOM after URL rewritting",
                                xmlcDoc, 0, writer);
            writer.close();
            getDiffer().diff(expectDom, outDom);
        } catch (IOException except) {
            throw new TestException(except);
        }
    }

    /**
     * Test 1: Xerces DOM URL rewriting
     */
    public void test1() {
        testStep(fXercesDomTestHtml, 0);
    }

    /**
     * Test 2: Lazy DOM URL rewriting
     */
    public void test2() {
        testStep(fLazyDomTestHtml, 0);
    }

    /**
     * Test 3: Expanded Lazy DOM URL rewriting
     */
    public void test3() {
        testStep(fLazyDomTestHtml, EXPAND);
    }

    /**
     * Test 4: Xerces XML URL rewriting
     */
    public void test4() {
        testStep(fXercesDomTestXml, 0);
    }
    /**
     * Test 5: Xerces DOM URL rewriting starting below the document
     * (test XMLObjectLink).
     */
    public void test5() {
        testStep(fXercesDomTestHtml, START_AT_NODE);
    }

    /**
     * Test 6: Lazy DOM URL rewriting starting below the document
     * (test XMLObjectLink)
     */
    public void test6() {
        testStep(fLazyDomTestHtml, START_AT_NODE);
    }

    /**
     * Test 7: Xerces XML URL rewriting starting below the document.
     */
    public void test7() {
        testStep(fXercesDomTestXml, START_AT_NODE);
    }

    /**
     * Test 8: LazyDOM XML URL rewriting starting below the document.
     */
    public void test8() {
        testStep(fLazyDomTestHtml, START_AT_NODE);
    }

    /**
     * Test 9: Bug with not substituting URL character entity references, URL
     * rewriting and attributes that contain double quites
     */
    public void test9() {
        testStep(fXercesDomTestHtml, OMIT_ATTR_CHAR_ENTITY_REFS);
    }

    /**
     * Test 10: Above test with LazyDOM
     */
    public void test10() {
        testStep(fLazyDomTestHtml, OMIT_ATTR_CHAR_ENTITY_REFS);
    }
}
