/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DocInfo.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.lazydom;

import java.util.HashMap;

import org.enhydra.xml.dom.SimpleDOMTraversal;
import org.enhydra.xml.driver.TestException;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Information collected on a test document.  Use for various tests to find
 * their way into the DOM.
 */
class DocInfo {
    /**
     * name/value pair to element node id entry.
     */
    private class AttrElementId {
        /* the mapping */
        public final String attrName;
        public final String attrValue;
        public final int elementNodeId;

        /** Constructor */
        public AttrElementId(String name,
                             String value,
                             int id) {
            attrName = name;
            attrValue = value;
            elementNodeId = id;
        }
        
        /** Equality test on name and value */
        public boolean equals(Object obj) {
            if (obj instanceof AttrElementId) {
                AttrElementId entry2 = (AttrElementId)obj;
                return attrName.equals(entry2.attrName)
                    && attrValue.equals(entry2.attrValue);
            } else {
                return false;
            }
        }

        /** Hash code on name and value */
        public int hashCode() {
            return attrName.hashCode() + attrValue.hashCode();
        }
    }

    /**
     * Table of attribute name/value pairs to node id.
     */
    private HashMap fAttrElementIdMap = new HashMap();

    /**
     * Constructor. Collect data from the template document. 
     */
    public DocInfo(LazyDocument document) {

        // Traverse template tree, collecting infomation
        SimpleDOMTraversal traverser
            = new SimpleDOMTraversal(new SimpleDOMTraversal.Handler() {
                    public void handleNode(Node node) {
                        if (node instanceof Attr) {
                            addAttrElementId((LazyAttr)node);
                        }
                    }
                });
        traverser.traverse(document.getTemplateDocument());
    }
    
    /**
     * Add an attribute to the attrElementIdMap.
     */
    private void addAttrElementId(LazyAttr attr) {
        LazyElement element = (LazyElement)attr.getOwnerElement();
        AttrElementId attrElementId = new AttrElementId(attr.getNodeName(),
                                                        attr.getNodeValue(),
                                                        element.getNodeId());
        if (!fAttrElementIdMap.containsKey(attrElementId)) {
            fAttrElementIdMap.put(attrElementId, attrElementId);
        }
    }

    /**
     * Convert a attibute name/value to the containing element's node id.
     */
    public int getElementId(String attrName,
                            String attrValue) {
        // Make a key
        AttrElementId key = new AttrElementId(attrName, attrValue, 0);
        AttrElementId entry = (AttrElementId)fAttrElementIdMap.get(key);
        if (entry == null) {
            throw new TestException("Attribute/value pair not found: "
                                    + attrName + "/" + attrValue);
        }
        return entry.elementNodeId;
    }

    /**
     * Find the first Element in a lazy tree with an attribute with the
     * given name and value.  Uses the proto tree and dom converter to
     * find the node by id.
     */
    public Element getByAttrValue(LazyDocument lazyDoc,
                                  String attrName,
                                  String attrValue) {
        
        return (Element)lazyDoc.getNodeById(getElementId(attrName, attrValue));
    }
}
