/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ExecXmlc.java,v 1.4 2006-08-10 23:57:26 jkjome Exp $
 */

package org.enhydra.xml.xmlc.driver;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;

import org.enhydra.xml.driver.TestCaseBase;
import org.enhydra.xml.driver.TestError;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.driver.TestFileOps;
import org.enhydra.xml.driver.TestProperties;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.commands.xmlc.XMLC;

/**
 * Run XMLC for tests.  This calls XMLC with command line options to
 * test parsing.  Also has various constants useful for calling XMLC.
 */
public class ExecXmlc {
    /** Compile-time force verbose XMLC output, for debugging */
    private static final boolean DEBUG_FORCE_VERBOSE = false;

    /** Constants for XMLC options */
    public static final String OPT_PARSER = "-parser";
    public static final String OPT_DOM_FACTORY = "-domfactory";
    public static final String OPT_DOM = "-dom";
    public static final String OPT_FOR_RECOMP = "-for-recomp";
    public static final String OPT_FOR_DEFERRED_PARSING = "-for-deferred-parsing";
    public static final String OPT_DEST_DIR = "-d";
    public static final String OPT_SOURCE_OUT = "-sourceout";
    public static final String OPT_KEEP = "-keep";
    public static final String OPT_CLASS = "-class";
    public static final String OPT_DUMP = "-dump";
    public static final String OPT_METHODS = "-methods";
    public static final String OPT_NO_COMPILE = "-nocompile";
    public static final String OPT_DOC_OUT = "-docout";
    public static final String OPT_URL_MAPPING = "-urlmapping";
    public static final String OPT_URL_REGEXP_MAPPING = "-urlregexpmapping";
    public static final String OPT_URL_SETTING = "-urlsetting";
    public static final String OPT_JAVAC_G = "-g";
    public static final String OPT_SSI = "-ssi";
    public static final String OPT_WARNINGS = "-warnings";
    public static final String OPT_DELETE_CLASS = "-delete-class";
    public static final String OPT_JAVAC_FLAG = "-javacflag";
    public static final String OPT_JAVAC_OPT = "-javacopt";
    public static final String OPT_VERBOSE = "-verbose";
    public static final String OPT_CREATE_GET_TAG_METHODS
        = "-create-get-tag-methods";
    public static final String OPT_EXTENDS = "-extends";
    public static final String OPT_IMPLEMENTS = "-implements";
    public static final String OPT_GENERATE = "-generate";

    public static final String OPT_HTML_ENCODING = "-html:encoding";
    public static final String OPT_ADD_TAG_SET = "-html:addtagset";
    public static final String OPT_VALIDATE = "-validate";
    public static final String OPT_XCATALOG = "-xcatalog";
    public static final String OPT_PARSEINFO = "-parseinfo";

    /** Constants for various argument values */
    public static final String XERCES_DOM = "xerces";
    public static final String XERCES_HTML_DOM_FACTORY
        = "org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory";
    public static final String XERCES_DOM_FACTORY
        = "org.enhydra.xml.xmlc.dom.xerces.XercesDomFactory";
    public static final String LAZY_DOM = "lazydom";
    public static final String LAZY_HTML_DOM_FACTORY
        = "org.enhydra.xml.xmlc.dom.lazydom.LazyHTMLDomFactory";
    public static final String LAZY_DOM_FACTORY
        = "org.enhydra.xml.xmlc.dom.lazydom.LazyDomFactory";
    public static final String CYBERSTUDIO_TAG_SET = "cyberstudio";
    public static final String SWING_PARSER = "swing";
    public static final String TIDY_PARSER = "tidy";
    public static final String XERCES_PARSER = "xerces";
    public static final String GENERATE_BOTH = "both";
    public static final String GENERATE_INTERFACE = "interface";
    public static final String GENERATE_IMPLEMENTATION = "implementation";
    public static final String YES = "yes";
    public static final String NO = "no";
    
    /** Test case we are assocaited with. */
    private TestCaseBase fTest;

    /** Object containing test parameters */
    private XmlcTestParams fParams;

    /** List of options and arguments */
    private ArrayList fOpts = new ArrayList();

    /** If not null, value to pass to javac -classpath */
    private String fClassPath;

    /** Destination directory */
    private File fDestDir;

    /** Class name */
    private String fClassName;

    /** Option files (created on first use) */
    private ArrayList fOptFiles;

    /** Source file to compile */
    private File fSrcFile;

    /* Automatically set the DOM */
    private boolean fAutoSetDom = true;

    /** Use -for-recomp and copy source to class dir */
    private boolean fForRecomp = false;

    /** Use -for-deferred-parsing and copy source to class dir */
    private boolean fForDeferredParsing = false;

    /** was -verbose set? */
    private boolean fSetVerbose = false;

    /** 
     * List of additional source files that are to be copied to class
     * directory if recompilation is being tested. null if none have been
     * added.
     */
    private ArrayList fExtraSrc;

    /** Constructor */
    public ExecXmlc(XmlcTestCaseBase test) {
	this(test, test.getParams());
    }

    /** Constructor */
    public ExecXmlc(TestCaseBase testCase,
                    XmlcTestParams params) {
        fTest = testCase;
        fParams = params;
	fForDeferredParsing = fParams.getIsDeferredParsing();
    }

    /** Specify the -dom or -domfactory option, preventing automatic setting */
    public void setDomOpt(String opt,
                          String arg) {
        addOpt(opt, arg);
        fAutoSetDom = false;
    }

    /** 
     * Add a element to the -classpath passed to javac.  If this has not been
     * called, then the option is not passed.
     */
    public void addClassPath(File element) {
        if (fClassPath == null) {
            fClassPath = element.getPath();
        } else {
            fClassPath = element.getPath() + File.pathSeparator + fClassPath;
        }
    }

    /** Create classpath to pass to javac, or null if none should be passed */
    private String makeJavacClassPath() {
        String propPath = TestProperties.getJavacClassPath();
        if ((propPath == null) && (fClassPath == null)) {
            return null; // no path to add
        }
        StringBuffer path = new StringBuffer(256);
        if (fClassPath != null) {
            path.append(fClassPath);
        }
        if (propPath != null) {
            if (path.length() > 0) {
                path.append(File.pathSeparator);
            }
            path.append(propPath);
        }
        
        // must add system classpath, as CLASSPATH environment is ignored
        // if -classpath is specified
        path.append(File.pathSeparator);
        path.append(System.getProperty("java.class.path"));

        return path.toString();
    }

    /** Add an option */
    public void addOpt(String opt) {
        fOpts.add(opt);

        // Detect options that need special handling
        if (opt.equals(OPT_FOR_RECOMP)) {
            fForRecomp = true;
        } else if (opt.equals(OPT_FOR_DEFERRED_PARSING)) {
	    fForDeferredParsing = true;
	} else if (opt.equals(OPT_VERBOSE)) {
            fSetVerbose = true;
        } 
    }

    /** Add and option and argument */
    public void addOpt(String opt,
                       String arg) {
        fOpts.add(opt);
        fOpts.add(arg);

        // Detect options that need special handling
        if (opt.equals(OPT_CLASS)) {
            fClassName = arg;
        } else if (opt.equals(OPT_DEST_DIR)) {
            fDestDir = new File(arg);
        } else if (opt.equals(OPT_DOM)) {
            fAutoSetDom = false;
        } else if (opt.equals(OPT_DOM_FACTORY)) {
            fAutoSetDom = false;
        }
    }

    /** Add and option and file argument */
    public void addOpt(String opt,
                       File arg) {
        addOpt(opt, arg.getPath());
    }

    /** Add an option and two arguments */
    public void addOpt(String opt,
                       String arg1,
                       String arg2) {
        fOpts.add(opt);
        fOpts.add(arg1);
        fOpts.add(arg2);
    }

    /** Add an option or metadata file */
    public void addOptionFile(File optFile) {
        if (fOptFiles == null) {
            fOptFiles = new ArrayList();
        }
        fOptFiles.add(optFile);
    }

    /**
     * Add the cyberstudio tag set.
     */
    public void addCyberStudioTags() {
        addOpt(ExecXmlc.OPT_ADD_TAG_SET, ExecXmlc.CYBERSTUDIO_TAG_SET);
    }

    /** Set the source file */
    public void setSrcFile(File srcFile) {
        fSrcFile = srcFile;
    }

    /** 
     * Add an additional source file to are to be copied to class directory if
     * recompilation is being tested.
     */
    public void addExtraSrcFile(File extraSrcFile) {
        if (fExtraSrc == null) {
            fExtraSrc = new ArrayList();
        }
        fExtraSrc.add(extraSrcFile);
    }

    /** Build arguments */
    private String[] buildArgs() {
        if (fSrcFile == null) {
            throw new TestError("bug: no source file specified");
        }
        
        ArrayList args = new ArrayList();
        if (fAutoSetDom && fParams.getDom().equals(ExecXmlc.XERCES_DOM)) {
            args.add(OPT_DOM_FACTORY);
            if (fParams.getIsXml()) {
                args.add(XERCES_DOM_FACTORY);
            } else {
                args.add(XERCES_HTML_DOM_FACTORY);
            }
        }
        if (fForRecomp) {
            args.add(ExecXmlc.OPT_FOR_RECOMP);
        }

        if (fForDeferredParsing) {
            args.add(ExecXmlc.OPT_FOR_DEFERRED_PARSING);
        }

        args.add(OPT_PARSER);
        args.add(fParams.getParser());

        // class path, if specified
        String path = makeJavacClassPath();
        if (path != null) {
            args.add(OPT_JAVAC_OPT);
            args.add("-classpath");
            args.add(path);
        }

        // Add specified opts
        args.addAll(fOpts);

        if (fSetVerbose || DEBUG_FORCE_VERBOSE) {
            args.add(OPT_VERBOSE);
        }

        // Options or metadata files
        if (fOptFiles != null) {
            Iterator iter = fOptFiles.iterator();
            while(iter.hasNext()) {
                args.add(((File)iter.next()).getPath());
            }
        }
        args.add(fSrcFile.getPath());
        return (String[])args.toArray(new String[args.size()]);
    }

    /** Figure out the package directory */
    private File getPkgDir() {
        // If dest dir is not specified, it's current dir
        File dir = fDestDir;
        if (dir == null) {
            dir = new File(".");
        }
        // If class isn't specified, it's the dest dir
        if (fClassName != null) {
            String classNamePath = fClassName.replace('.', '/');
            dir = new File(dir, classNamePath).getParentFile();
        }
        return dir;
    }

    /** copy source files to package directory */
    private void copySrcToPkgDir() {
        File pkgDir = getPkgDir();
        TestFileOps.copyFileToDir(fSrcFile, pkgDir);
        if (fExtraSrc != null) {
            for (int idx = 0; idx < fExtraSrc.size(); idx++) {
                TestFileOps.copyFileToDir((File)fExtraSrc.get(idx), pkgDir);
            }
        }
    }

    /** Do work of running xmlc */
    private void doCompile(PrintWriter out) 
        throws XMLCException, IOException  {
        //if (fForRecomp || fForDeferredParsing) {
            copySrcToPkgDir();
        //}
        String[] args = buildArgs();
        PrintWriter verbOut = fTest.getVerboseOut();
        if (verbOut !=  null) {
            verbOut.print("xmlc compile: xmlc ");
            for (int iArg = 0; iArg < args.length; iArg++) {
                verbOut.print(" ");
                verbOut.print(args[iArg]);
            }
            verbOut.println();
        }

	if (verbOut != null) {
	    new XMLC(out, verbOut).compile(args);
	} else {
	    new XMLC(out, out).compile(args);
	}

    }

    /** Execute XMLC */
    public void compile(File outFile) throws TestException {
        try {
            TestFileOps.ensureFileDir(outFile);
            if (fDestDir != null) {
                fDestDir.mkdirs();
            }

            // Force consistent encoding
            PrintWriter out
                = new PrintWriter(new OutputStreamWriter(
                                    new FileOutputStream(outFile), "UTF-8"));
            try {
                doCompile(out);
            } catch (XMLCException except) {
                // Write exception that may or may not be normal to file
                out.println(except.toString());
                throw except;
            } finally {
                out.close();
            }
        } catch (Throwable err) {
            throw new TestException("xmlc compile failed", err);
        }
    }
    
}
