/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TestClassLoaderFactory.java,v 1.2 2005-01-26 08:29:24 jkjome Exp $
 */

package org.enhydra.xml.xmlc.driver;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;

import org.enhydra.xml.driver.TestError;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.driver.TestFileOps;

/**
 * Create class loaders for the tests. This provides an abstraction for
 * creating various classloaders that XMLC reloading is expected to work
 * under.  It creates them via reflection, allowing this file to be compiled
 * even if supported class loaders are not available.
 */
public class TestClassLoaderFactory {
    /**
     * Identifiers for valid class loaders.
     */
    private static final String LOADER_ENHYDRA = "enhydra";
    private static final String LOADER_MULTI = "multi";
    private static final String LOADER_URL = "url";

    /** Load a class using the classloader that loaded this class */
    private Class loadClass(String className)
        throws ClassNotFoundException {
        return getClass().getClassLoader().loadClass(className);
    }

    /** Create a MultiClassLoader for tests. Used with Enhydra 3.*. */
    private ClassLoader mkMultiClassLoader(URL[] classPath,
                                           ClassLoader parent)
        throws NoSuchMethodException, InvocationTargetException,
               IllegalAccessException, InstantiationException,
               ClassNotFoundException {
        // Need LogChannel object to find constructor
        Class logClass = loadClass("com.lutris.logging.LogChannel");
        Class loaderClass
            = loadClass("com.lutris.classloader.MultiClassLoader");

        // new MultiClassLoader(URL[] path, LogChannel loadLogChannel);
        Constructor constr
            = loaderClass.getConstructor(new Class[]{URL[].class, logClass});
        ClassLoader loader
            = (ClassLoader)constr.newInstance(new Object[]{classPath, null});

        // setParent(parent)
        Method setParent = loaderClass.getMethod("setParent",
                                                 new Class[]{ClassLoader.class});
        setParent.invoke(loader, new Object[]{parent});
        return loader;
    }

    /** Create an EnhydraClassLoader for tests. Used with EAS 4.*. */
    private ClassLoader mkEnhydraClassLoader(URL[] classPath,
                                             ClassLoader parent)
        throws NoSuchMethodException, InvocationTargetException,
               IllegalAccessException, InstantiationException,
               ClassNotFoundException {
        Class loaderClass
            = loadClass("org.enhydra.classloader.EnhydraClassLoader");

        // new EnhydraClassLoader(ClassLoader parent, URL[] classPath)
        Constructor constr
            = loaderClass.getConstructor(new Class[]{ClassLoader.class,
                                                     URL[].class});
        return (ClassLoader)constr.newInstance(new Object[]{parent, classPath});
    }

    /**
     * Create a URLClassLoader for tests.
     */
    private ClassLoader mkURLClassLoader(URL[] classPath,
                                         ClassLoader parent) {
        return URLClassLoader.newInstance(classPath, parent);
    }

    /**
     * Create a class loader of the specified type.
     */
    public ClassLoader mkTestClassLoader(String classLoaderName,
                                         URL[] classPath,
                                         ClassLoader parent) {
        try {
            if (classLoaderName.equals(LOADER_MULTI)) {
                return mkMultiClassLoader(classPath, parent);
            } else if (classLoaderName.equals(LOADER_ENHYDRA)) {
                return mkEnhydraClassLoader(classPath, parent);
            } else if (classLoaderName.equals(LOADER_URL)) {
                return mkURLClassLoader(classPath, parent);
            } else {
                throw new TestError("invalid classloader name \""
                                    + classLoaderName + "\"");
            }
        } catch (NoSuchMethodException except) {
            throw new TestException(except);
        } catch (InvocationTargetException except) {
            throw new TestException(except);
        } catch (ClassNotFoundException except) {
            throw new TestException(except);
        } catch (IllegalAccessException except) {
            throw new TestException(except);
        } catch (InstantiationException except) {
            throw new TestException(except);
        }
    }

    /**
     * Create a class loader of the type associated with this factory
     */
    public ClassLoader mkTestClassLoader(URL[] classPath,
                                         ClassLoader parent) {
        // FIXME: need to parameterize type
        return mkTestClassLoader(LOADER_URL, classPath, parent);
    }

    /**
     * Create a class loader of the type associated with this factory
     */
    public ClassLoader mkTestClassLoader(File[] classPath,
                                         ClassLoader parent) {
        URL[] urlClassPath = new URL[classPath.length];
        for (int idx = 0; idx < classPath.length; idx++) {
            urlClassPath[idx] = TestFileOps.dirToUrl(classPath[idx]);
        }

        // FIXME: need to parameterize type
        return mkTestClassLoader(LOADER_URL, urlClassPath, parent);
    }
}
