/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XmlcTestParams.java,v 1.2 2005-01-26 08:29:25 jkjome Exp $
 */

package org.enhydra.xml.xmlc.driver;

import java.io.File;

import org.enhydra.xml.driver.TestProperties;
import org.enhydra.xml.xmlc.commands.xmlc.XMLC;

/**
 * Object to hold parameters that are globally applied to the tests.
 * Instances are immutable.  Also contains methods for getting
 * the paths to directories defined by these paramameters.
 */
public class XmlcTestParams {
    /** Constants for test names */
    public static final String RELOADING = "reload";
    public static final String STD_LOADING = "stdload";

    /** Is this an XML or HTML test? */
    private final boolean fIsXml;

    /** Whether to use deferred parsing for reloading */
    private final boolean fIsDeferredParsing;

    /** Parser to use */
    private final String fParser;

    /** The DOM to use */
    private final String fDom;
    
    /** Is jikes being used? */
    private final boolean fUsingJikes;

    /** Doing reloading/recomp testing? */
    private final boolean fReloading;
    private final String fReloadingStr;

    /** Parameters, seperated by `.' */
    private final String fParamsStr;

    /** Get the basic parameters as a string (less reloading) */
    private final String fBasicParamsStr;

    /**
     * Constructor.
     */
    public XmlcTestParams(boolean isXml,
                          String parser,
                          String dom,
                          boolean reloading) {
	this(isXml, parser, dom, reloading, false);
    }

    /**
     * Constructor.
     */
    public XmlcTestParams(boolean isXml,
                          String parser,
                          String dom,
                          boolean reloading,
			  boolean isDeferringParsing) {
        fIsXml = isXml;
	fIsDeferredParsing = isDeferringParsing;

        fParser = parser;
        fDom = dom;
        fReloading = reloading;
        fReloadingStr = (fReloading ? RELOADING : STD_LOADING);

        // Determine if the property is set to use jikes as the 
        // default XMLC parser
        String defaultJavac = System.getProperty(XMLC.JAVAC_PROPERTY);
        boolean usingJikes = false;
        if (defaultJavac != null) {
            File javac = new File(defaultJavac);
            if (javac.getName().indexOf("jikes") >= 0) {
                usingJikes = true;
            }
        }
        fUsingJikes = usingJikes;
        fBasicParamsStr = fParser + "." + fDom;
        fParamsStr = fBasicParamsStr + "." + fReloadingStr;
    }

    /**
     * Factory method to obtain an instance of the parameters with parser,
     * DOM, and reload set from properties.
     */
    public static XmlcTestParams getParams(boolean isXml) {
        String defParser = (isXml ? ExecXmlc.XERCES_PARSER
                            : ExecXmlc.TIDY_PARSER);
	
        return new XmlcTestParams(isXml,
                                  TestProperties.getParser(defParser),
                                  TestProperties.getDom(ExecXmlc.XERCES_DOM),
                                  TestProperties.getReloading(),
				  TestProperties.getDeferredParsing());
    }


    /** Is this an XML or HTML test */
    public boolean getIsXml() {
        return fIsXml;
    }

    /** Is this a deferred parsing test */
    public boolean getIsDeferredParsing() {
        return fIsDeferredParsing;
    }

    /** Get the parser to use */
    public String getParser() {
        return fParser;
    }

    /** The DOM to use */
    public String getDom() {
        return fDom;
    }

    /** Doing reloading testing? */
    public boolean getReloading() {
        return fReloading;
    }

    /** Is jikes the default compiler? */
    public boolean getUsingJikes() {
        return fUsingJikes;
    }

    /** Get the parameters as a string, seperated by `.' */
    public String getParamsStr() {
        return fParamsStr;
    }

    /** 
     * Get the basic parameters as a string (less reloading),
     * seperated by `.'
     */
    public String getBasicParamsStr() {
        return fBasicParamsStr;
    }
}
