/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ApiTests.java,v 1.2 2005-01-26 08:29:25 jkjome Exp $
 */

package org.enhydra.xml.xmlc.html;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

import junit.framework.Test;

import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.dom.DOMStats;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.xmlc.XMLCUtil;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.dom.xerces.XercesHTMLDomFactory;
import org.enhydra.xml.xmlc.driver.ExecXmlc;
import org.enhydra.xml.xmlc.driver.TestDOMSource;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.html.HTMLAnchorElement;
import org.w3c.dom.html.HTMLBodyElement;
import org.w3c.dom.html.HTMLCollection;
import org.w3c.dom.html.HTMLDocument;
import org.w3c.dom.html.HTMLElement;
import org.w3c.dom.html.HTMLFormElement;
import org.w3c.dom.html.HTMLHeadElement;
import org.w3c.dom.html.HTMLHtmlElement;
import org.w3c.dom.html.HTMLLIElement;
import org.w3c.dom.html.HTMLTableElement;
import org.w3c.dom.html.HTMLTableRowElement;
import org.w3c.dom.html.HTMLTitleElement;
import org.w3c.dom.html.HTMLUListElement;

/**
 * Tests of various things at the API level.  Interfaces or reflection is used
 * to access to compiled objects as we want this test to work with documents
 * compiled in different ways.
 */
public class ApiTests extends HtmlTestCaseBase {
    /** Object for creating instances of FrameTreeHTML. */
    private TestDOMSource fFrameTreeSource;

    /** Object for creating instances of ShelfHTML. */
    private TestDOMSource fShelfSource;

    /** Object for creating instances of GetElementByIdHTML. */
    private TestDOMSource fGetElementByIdSource;

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(ApiTests.class, null);
    }
    
    /** Constructor */
    public ApiTests(Method method) {
        super(method);
        fFrameTreeSource = new TestDOMSource(this, getParams(),
                                             getInputFile("FrameTree.html"),
                                             "FrameTreeHTMLImpl",
                                             FrameTreeHTML.class);
        fShelfSource = new TestDOMSource(this, getParams(),
                                         getInputFile("Shelf.html"),
                                         "ShelfHTMLImpl",
                                         ShelfHTML.class);
        fGetElementByIdSource = new TestDOMSource(this, getParams(),
                                                  getInputFile("GetElementById.html"),
                                                  "GetElementByIdHTMLImpl",
                                                  GetElementByIdHTML.class);
    }

    /**
     * Verify various attributes of a compiled HTML document object.
     */
    private void verifyHtmlDoc(XMLObject htmlObj) {
        if (!(htmlObj instanceof HTMLObject)) {
            throw new TestException("Document not an instance of HTMLObject: "
                                    + htmlObj.getClass());
        }
        if (!(htmlObj instanceof HTMLDocument)) {
            throw new TestException("Document not an instance of HTMLDocument: "
                                    + htmlObj.getClass());
        }
        Object docObj = htmlObj.getDocument();
        if (!(docObj instanceof HTMLDocument)) {
            throw new TestException("Document contained in XMLObject is not instance of HTMLDocument: "
                                    + docObj.getClass());
        }
    }

    /**
     * Get and instance of the FrameTreeHTML object.
     */
    private FrameTreeHTML createFrameTreeDoc() {
        FrameTreeHTML htmlObj = (FrameTreeHTML)fFrameTreeSource.create();
        verifyHtmlDoc(htmlObj);
        return htmlObj;
    }

    /*
     * Find an element by class attribute name.  Makes the incorrect
     * assumption that there is only one class name per element.
     */
    private Element getByClassAttrName(Node node,
                                       String className) {
        if (node instanceof HTMLElement) {
            String elemClassName = ((HTMLElement)node).getClassName();
            if ((elemClassName != null)
                && elemClassName.equalsIgnoreCase(className)) {
                return (HTMLElement)node;
            }
        }
        NodeList children = node.getChildNodes();
        for (int i=0; i<children.getLength();i++) {
            Element ret = getByClassAttrName(children.item(i), className);
            if (ret != null) {
                return ret;
            }
        }
        return null;
    }
    
    /*
     * Find an element by class attribute name.  Generates error if not found.
     */
    private Element getFirstByClassAttrName(Node node,
                                            String className) {
        Element elem = getByClassAttrName(node, className);
        if (elem == null) {
            throw new TestException("Could't find class \"" + className
                                    + "\" starting at: " + node.toString());
        }
        return elem;
    }

    /* Delete an element of the table */
    private void deleteElement(HTMLTableElement table,
                               String className) {
        Element elem = getFirstByClassAttrName(table, className);
        table.removeChild(elem);
    }

    /*
     * Test 1: deleting of children. Taken from code that (I thought) caused
     * grief in the past.  In fact, there it was a bug in the XMLC
     * application, but this test will be left in.
     */
    public void test1() {
        FrameTreeHTML htmlObj = createFrameTreeDoc();

        // Use reflection here, as we want to make sure the
        // constants are in the generated file.
        HTMLTableElement table = htmlObj.getElementTable();

        // deleteElement(table, htmlObj.CLASS_customerExpanded);
        deleteElement(table, getStringField(htmlObj, "CLASS_customerExpanded"));
        deleteElement(table, getStringField(htmlObj, "CLASS_projectNoSite"));
        deleteElement(table, getStringField(htmlObj, "CLASS_projectWithSite"));
        deleteElement(table, getStringField(htmlObj, "CLASS_selectedProjectNoSite"));
        deleteElement(table, getStringField(htmlObj, "CLASS_selectedProjectWithSite"));
        deleteElement(table, getStringField(htmlObj, "CLASS_newProject"));
        deleteElement(table, getStringField(htmlObj, "CLASS_customerWithProject"));
        deleteElement(table, getStringField(htmlObj, "CLASS_customerNoProject"));
        dumpVerifyDom(htmlObj, DOC_DOM_DUMP_EXT);
    }

    /** Clone a document via the copy constructor */
    private HTMLObject copyConstruct(HTMLObject srcDoc) {
        Class srcClass = srcDoc.getClass();
        try {
            Constructor constr = srcClass.getConstructor(new Class[]{srcClass});
            return (HTMLObject)constr.newInstance(new Object[]{srcDoc});
        } catch (Throwable except) {
            throw new TestException("call of copy constructe failed on "
                                    + srcClass, except);
        }
    }

    /**
     * Test 2: copy constructor of generate object.
     */
    public void test2() {
        FrameTreeHTML htmlObj = createFrameTreeDoc();

        // Set fields values in the original document
        htmlObj.getElementIMG1().setSrc("image1.gif");
        htmlObj.getElementIMG2().setSrc("image2.gif");

        // Clone object and set other field values.
        FrameTreeHTML html2Obj = (FrameTreeHTML)copyConstruct(htmlObj);
        html2Obj.getElementIMG2().setSrc("image2-clone.gif");
        html2Obj.getElementIMG3().setSrc("image3-clone.gif");

        // Modify the original document 
        htmlObj.getElementIMG3().setSrc("image3.gif");

        // Write em out and verify
        String name1 = "1.doc.dom";
        dumpVerifyDom(htmlObj, getResultFile(name1),
                      getExpectedFile(name1));
        String name2 = "2.doc.dom";
        dumpVerifyDom(htmlObj, getResultFile(name2),
                      getExpectedFile(name2));
    }

    /*
     * Test 3: setting title
     */
    public void test3() {
        FrameTreeHTML htmlObj = createFrameTreeDoc();
        htmlObj.setTitle("This is a new title");
        dumpVerifyDom(htmlObj, DOC_DOM_DUMP_EXT);
    }

    /* Create a new HTML document. */
    private HTMLDocument createHTMLDocument() {
        // Use the right method, depending on the DOM.
        if (getParams().getDom().equals(ExecXmlc.LAZY_DOM)) {
            return HTMLDocumentFactory.createDocument();
        } else {
            return (HTMLDocument)new XercesHTMLDomFactory().createDocument(null, null, null);
        }
    }

    /** Append a test special character to the list. */
    private void appendChar(HTMLUListElement list,
                            String entityName,
                            char entityChar) {
        Document doc = list.getOwnerDocument();
        HTMLLIElement li = (HTMLLIElement)doc.createElement("li");
        list.appendChild(li);

        String txt = entityName + " `" + entityChar + "'";
        li.appendChild(doc.createTextNode(txt));
    }

    /*
     * Test 4: special characters.
     */
    public void test4() {
        HTMLDocument doc = createHTMLDocument();

        HTMLHtmlElement htmlElem = (HTMLHtmlElement)doc.getDocumentElement();

        HTMLHeadElement headElem = (HTMLHeadElement)doc.createElement("head");
        htmlElem.appendChild(headElem);

        HTMLTitleElement title = (HTMLTitleElement)doc.createElement("title");
        headElem.appendChild(title);
        title.appendChild(doc.createTextNode("Test of inserting special characeters into DOM"));

        HTMLBodyElement body = (HTMLBodyElement)doc.createElement("body");
        htmlElem.appendChild(body);

        HTMLUListElement list = (HTMLUListElement)doc.createElement("ul");
        body.appendChild(list);
                
        appendChar(list, "nbsp", '\u00a0');
        appendChar(list, "gt", '\u003e');
        appendChar(list, "pound", '\u00a3');

        dumpVerifyDom(doc, DOC_DOM_DUMP_EXT);
    }

    /*
     * Test 5: bug where createElement didn't work right when called on XMLC
     * document object.
     */
    public void test5() {
        FrameTreeHTML htmlObj = createFrameTreeDoc();

        // This created a class cast exception in the reported bug.
        HTMLAnchorElement anchor = (HTMLAnchorElement)htmlObj.createElement("A");
    }

    /*
     * Do most of the work of getElementByIdTest().
     */
    private void doGetElementByIdTest(String id,
                                      boolean modifyAttr,
                                      boolean addChild,
                                      PrintWriter out) {
        FrameTreeHTML htmlObj = createFrameTreeDoc();

        HTMLElement element = (HTMLElement)htmlObj.getElementById(id);
                
        DOMInfo.printTree("Element returned for id=\"" + id + "\"",
                          element, out);
        out.println();
        if ((modifyAttr || addChild) && (element != null)) {
            if (modifyAttr) {
                element.setClassName("modified");
            }
            if (addChild) {
                element.appendChild(htmlObj.createComment("this was modified"));
            }
            DOMInfo.printTree("Element after modification",
                              element, out);
            out.println();
        }
        DOMInfo.printTree("Tree after getElementById(\"" + id + "\")",
                          htmlObj, out);
    }

    /*
     * Method used to implements tests of getElementById().
     */
    private void getElementByIdTest(String id,
                                    boolean modifyAttr,
                                    boolean addChild) {
        try {
            PrintWriter out = new PrintWriter(new FileOutputStream(getResultFile(DOC_DOM_DUMP_EXT)));
            try {
                doGetElementByIdTest(id, modifyAttr, addChild, out);
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }
                
        getDiffer().diff(getExpectedFile(DOC_DOM_DUMP_EXT),
                         getResultFile(DOC_DOM_DUMP_EXT));
    }

    /*
     * Do work of getElementsByNameTest()
     */
    private void doGetElementsByNameTest(String name,
                                         PrintWriter out) {
        FrameTreeHTML htmlObj = createFrameTreeDoc();

        NodeList elements = htmlObj.getElementsByName(name);
        out.println("Number of elements for name=\"" + name + "\": "
                    + elements.getLength());
        out.println();
        for (int i = 0 ; i < elements.getLength(); i++) {
            DOMInfo.printTree("Element " + i, elements.item(i),
                              out);
            out.println();
        }
        DOMInfo.printTree("Tree after getElementsByName(\"" + name + "\")", htmlObj,
                          out);
    }

    /*
     * Method used to implements tests of getElementsByName()
     */
    private void getElementsByNameTest(String name) {
        try {
            PrintWriter out = new PrintWriter(new FileOutputStream(getResultFile(DOC_DOM_DUMP_EXT)));
            try {
                doGetElementsByNameTest(name, out);
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }
        getDiffer().diff(getExpectedFile(DOC_DOM_DUMP_EXT),
                         getResultFile(DOC_DOM_DUMP_EXT));
    }

    /*
     * Tests 6-11: test element lookup functions.  Function print both thes
     * Elements found and the trees after search, so that expansion of lazydom
     * can be detected.
     */
    public void test6() {
        getElementByIdTest("table", false, false);
    }
    public void test7() {
        getElementByIdTest("IMG2", false, false);
    }
    public void test8() {
        getElementByIdTest("table", true, true);
    }
    public void test9() {
        getElementByIdTest("IMG2", true, true);
    }
    public void test10() {
        getElementsByNameTest("CustomerA");
    }
    public void test11() {
        getElementsByNameTest("CustomerB"); 
    }

    /*
     * Test 12: HTMLTableElement operations
     */
    public void test12() {
        try {
            PrintWriter out = new PrintWriter(new FileOutputStream(getResultFile(DOC_DOM_DUMP_EXT)));
            try {
                FrameTreeHTML htmlObj = createFrameTreeDoc();
                HTMLTableElement table = htmlObj.getElementTable();
                HTMLCollection rows = table.getRows();
                out.println("Number of rows for table: " + rows.getLength());
                for (int rowIdx = 0; rowIdx < rows.getLength(); rowIdx++) {
                    HTMLTableRowElement row = (HTMLTableRowElement)rows.item(rowIdx);
                    out.println();
                    DOMInfo.printTree("row " + rowIdx, rows.item(rowIdx),
                                      out);
                }
                out.println();
                DOMInfo.printTree("Table after getRows", table, out);
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }
        getDiffer().diff(getExpectedFile(DOC_DOM_DUMP_EXT),
                         getResultFile(DOC_DOM_DUMP_EXT));
    }

    /*
     * Test 13: setText methods.
     */
    public void test13() {
        ShelfHTML htmlObj = (ShelfHTML)fShelfSource.create();
        verifyHtmlDoc(htmlObj);

        // See if setting to null is detected:
        Throwable gotExcept = null;
        try {
            htmlObj.setTextSampleCatText(null);
        } catch (IllegalArgumentException except) {
            gotExcept = except;
        }
        if (!(gotExcept instanceof IllegalArgumentException)) {
            throw new TestException("didn't get expected IllegalArgumentException");
        }

        htmlObj.setTextSampleCatText("@@@@ This value updated by a setText method @@@@");
        dumpVerifyDom(htmlObj, DOC_DOM_DUMP_EXT);
    }

    /**
     * Test 14: Problem that occured with OpenXML and Xerces where Node
     * implemented NodeList so save creating an object when getChildNodes is
     * called.  However, caused problems with HTMLFormElement and
     * HTMLSelectElement, since they have getLength() methods that override
     * NodeList.getLength().
     */
    public void test14() {
        // Construct an element with at part of a form
        HTMLDocument doc = createHTMLDocument();
        Element body = doc.getBody();
        HTMLFormElement form = (HTMLFormElement)doc.createElement("FORM");
        body.appendChild(form);

        // Add 3 text children and 3 element children
        for (int i = 0; i < 3; i++) {
            form.appendChild(doc.createTextNode("form text child #" + i));
            Element input = doc.createElement("INPUT");
            form.appendChild(input);
            input.appendChild(doc.createTextNode("input #" + i + " text child"));
        }

        dumpVerifyDom(doc, DOC_DOM_DUMP_EXT);
        if (form.getLength() != 3) {
            throw new TestException("form.getLength() != 3");
        }
        if (form.getChildNodes().getLength() != 6) {
            throw new TestException("form.getChildNodes().getLength() != 6");
        }
    }

    /*
     * Test 15: reproduce bug with calling XMLCUtil.getFirstText() to set a
     * text node followed by calling HTMLDocument.getElementById().  This
     * caused the LazyDOM to hang.
     */
    public void test15() {
        try {
            PrintWriter out = new PrintWriter(new FileOutputStream(getResultFile(DOC_DOM_DUMP_EXT)));
            try {
                GetElementByIdHTML htmlObj = (GetElementByIdHTML)fGetElementByIdSource.create();

                XMLCUtil.getFirstText(htmlObj.getElementTime()).setData("the time is now");
                Element testElem = htmlObj.getElementById("test");

                DOMInfo.printTree("test element", testElem, out);
                DOMInfo.printTree("tree", htmlObj, out);
                DOMStats.printStats(null, htmlObj,
                                    DOMStats.SIMPLE_CLASS_NAMES, out);
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }
        getDiffer().diff(getExpectedFile(DOC_DOM_DUMP_EXT),
                         getResultFile(DOC_DOM_DUMP_EXT));
    }
}
