/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MiscOptionsTests.java,v 1.2 2005-01-26 08:29:25 jkjome Exp $
 */

package org.enhydra.xml.xmlc.html;
import java.io.File;
import java.lang.reflect.Method;

import junit.framework.Test;

import org.enhydra.xml.driver.TestError;
import org.enhydra.xml.driver.TestFileOps;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.driver.ExecXmlc;

/**
 * Tests of various compilation options not tested elsewhere.
 */
public class MiscOptionsTests extends HtmlTestCaseBase {
    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(MiscOptionsTests.class, null);
    }
    
    /** Constructor */
    public MiscOptionsTests(Method method) {
        super(method);
    }

    /** 
     * Class used to implement tests that run the compiler with different
     * options.  Extended to implement the test.
     */
    private abstract class OptTest {
        protected ExecXmlc fExecXmlc;

        /** Constructor, set up ExecXmlc object for test */
        protected OptTest() {
            fExecXmlc = new ExecXmlc(MiscOptionsTests.this);
            fExecXmlc.addOpt(ExecXmlc.OPT_DEST_DIR,
                             getClassRoot().getPath());
        }

        /** Add the standard -class option */
        protected void addClass() {
            fExecXmlc.addOpt(ExecXmlc.OPT_CLASS, getTestClass());
        }

        /** Use input/testPage1.html as the imput file. */
        protected void setTestPage1() {
            fExecXmlc.setSrcFile(getInputFile("testPage1.html"));
        }

        /** Compile the file and verify the output */
        protected void compile() {
            fExecXmlc.compile(getResultFile(COMPILE_OUTPUT_EXT));
            getDiffer().diff(getExpectedFile(COMPILE_OUTPUT_EXT),
                             getResultFile(COMPILE_OUTPUT_EXT));
        }
        
        /** 
         * Setup and run the basic test; should add specific options before
         * calling.
         */
        public void basicTest() {
            addClass();
            setTestPage1();
            compile();
        }

        /** This method is implemented to run the test */
        public abstract void run();
    }

    /*
     * Test 1: proprietary tags
     */
    public void test1() {
        HtmlBasicTest test = createTest("proptags1.html");
        test.addOptionFile(getInputFile("proptags1.xmlc"));
        test.basicTest();
    }

    /*
     * Test 2: Keep, outdir, etc.
     */
    public void test2() {
        new OptTest() {
            public void run() {
                // Assemble output file paths, and remove existing files
                File genSrcFile = getGenSrcFile();
                removeFile(genSrcFile);
                File genClassFile = getGenClassFile();
                removeFile(genClassFile);

                fExecXmlc.addOpt(ExecXmlc.OPT_SOURCE_OUT,
                                 getGenSourceDir().getPath());
                fExecXmlc.addOpt(ExecXmlc.OPT_KEEP);
                basicTest();
                verifyExists(genSrcFile);
                verifyExists(genClassFile);
            }
        }.run();
    }

    /*
     * Test 3: Default package, class name from source file
     */
    public void test3() {
        new OptTest() {
            public void run() {
                // Assemble class and remove existing files
                File genClassFile = new File(getClassRoot(),
                                             "testPage1.class");
                removeFile(genClassFile);
                setTestPage1();
                compile();
                verifyExists(genClassFile);
            }
        }.run();
    }

    /*
     * Test 4: Two .xmlc files
     */
    public void test4() {
        HtmlBasicTest test = createTest("testPage1.html");
        test.addCyberStudioTags();
        test.addOptionFile(getInputFile("test-11.1a.xmlc"));
        test.addOptionFile(getInputFile("test-11.1b.xmlc"));
        test.basicTest();
    }

    /*
     * Test 5: HTML Compatibility options for older versions of XMLC.
     */
    public void test5() {
        HtmlBasicTest test = createLutrisIndexTest();
        test.addCyberStudioTags();
        test.addBoth("-html:old-class-constants");
        test.addBoth("-html:old-name-constants");
        test.addOptionFile(getInputFile("lutris-index1.xmlc"));
        test.basicTest();
    }

    /*
     * Test 6: -javacflag
     */
    public void test6() {
        new OptTest() {
            public void run() {
                fExecXmlc.addOpt(ExecXmlc.OPT_JAVAC_FLAG, "-nowarn");
                basicTest();
            }
        }.run();
    }

    /*
     * Test 7: -javacopt
     */
    public void test7() {
        // -encoding doesn't work for jikes, so skip
        if (!fParams.getUsingJikes()) {
            new OptTest() {
                public void run() {
                    fExecXmlc.addOpt(ExecXmlc.OPT_JAVAC_OPT);
                    fExecXmlc.addOpt("-encoding", "US-ASCII");
                    basicTest();
                }
            }.run();
        }
    }

    /*
     * Test 8: -dom
     */
    public void test8() {
        new OptTest() {
            public void run() {
                fExecXmlc.setDomOpt(ExecXmlc.OPT_DOM, getParams().getDom());
                basicTest();
                XMLObject doc = loadTestDocument(getTestClass());
                dumpVerifyDom(doc, DOC_DOM_DUMP_EXT);
            }
        }.run();
    }

    /*
     * Test 9: -domfactory
     */
    public void test9() {
        new OptTest() {
            public void run() {
                if (getParams().getDom().equals(ExecXmlc.XERCES_DOM)) {
                    fExecXmlc.setDomOpt(ExecXmlc.OPT_DOM_FACTORY, 
                                        ExecXmlc.XERCES_HTML_DOM_FACTORY);
                } else if (getParams().getDom().equals(ExecXmlc.LAZY_DOM)) {
                    fExecXmlc.setDomOpt(ExecXmlc.OPT_DOM_FACTORY, 
                                        ExecXmlc.LAZY_HTML_DOM_FACTORY);
                } else {
                    throw new TestError("bug in test: unknown DOM: \""
                                        + getParams().getDom() + "\"");
                }
                basicTest();
                XMLObject doc = loadTestDocument(getTestClass());
                dumpVerifyDom(doc, DOC_DOM_DUMP_EXT);
            }
        }.run();
    }

    /*
     * Test 10: -nocompile (-verbose will indicate if the compiler is
     * started).
     */
    public void test10() {
        new OptTest() {
            public void run() {
                fExecXmlc.addOpt(ExecXmlc.OPT_NO_COMPILE);
                fExecXmlc.addOpt(ExecXmlc.OPT_VERBOSE);
                basicTest();
            }
        }.run();
    }

    /*
     * Test 11: -create-get-tag-methods
     */
    public void test11() {
        HtmlBasicTest test = createLutrisIndexTest();
        test.addBoth(ExecXmlc.OPT_CREATE_GET_TAG_METHODS);
        test.basicTest();
    }

    /*
     * Test 12: -create-get-tag-methods along with setTextXXX() methods
     */
    public void test12() {
        HtmlBasicTest test = createTest("testPage1.html");
        test.addBoth(ExecXmlc.OPT_CREATE_GET_TAG_METHODS);
        test.addBoth(ExecXmlc.OPT_METHODS);
        test.basicTest();
    }

    /*
     * Test 13: -warnings
     */
    public void test13() {
        HtmlBasicTest test = createLutrisIndexTest();
        test.addBoth(ExecXmlc.OPT_WARNINGS, "false");
        test.basicTest();
    }

    /*
     * Test 14: -docout
     */
    public void test14() {
        new OptTest() {
            public void run() {
                File docOut = getResultFile("docout.html");
                TestFileOps.ensureFileDir(docOut);
                fExecXmlc.addOpt(ExecXmlc.OPT_DOC_OUT, docOut.getPath());
                fExecXmlc.addOpt(ExecXmlc.OPT_NO_COMPILE);
                setTestPage1();
                compile();
                parseVerifyDom(docOut);
            }
        }.run();
    }

    /*
     * Test 16: Depreciated -html:frameset option.
     */
    public void test16() {
        HtmlBasicTest test = createTest("FrameSet.html");
        test.addBoth("-html:frameset");
        test.basicTest();
    }

}
