/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PerfRunner.java,v 1.2 2005-01-26 08:29:25 jkjome Exp $
 */

package org.enhydra.xml.xmlc.perf;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.math.BigDecimal;

import org.enhydra.xml.xmlc.XMLCFactory;
import org.enhydra.xml.xmlc.XMLCStdFactory;
import org.enhydra.xml.xmlc.XMLObject;

/**
 * Run XMLC generated objects for performance testing.
 *   PerfRunner class
 */
public class PerfRunner {
    /**
     * Constants controling number of tests.
     */
    private static final int TEST_COUNT = 6; //4;
    private static final int REPEAT_COUNT = 10; //10;
    private static final String TEST_RESULTS_FILE = "perf.out";
    private static final String NOOP_EDITOR =
	"tests.org.enhydra.xml.xmlc.perf.NoOpEditor";

    class TimeResult {
	// in seconds...
	private double createTime;
	private double toDocumentTime;
	private double editTime;
	private int iterations;
	private int editedNodeCount;
	TimeResult(long createTime, long editTime,
		   long toDocumentTime, int iterations,
		   int editedNodeCount) {
	    this.createTime = createTime;
	    this.editTime = editTime;
	    this.toDocumentTime = toDocumentTime;
	    this.iterations = iterations;
	    this.editedNodeCount = editedNodeCount;
	}
	double getTotalTime() {
	    return getCreateTime() + getToDocumentTime() + getEditTime();
	}
	Number getTotalTime(int scale) {
	    return getScaledNumber(getTotalTime(), scale);
	}
	Number getAverageTime(int scale) {
	    return getScaledNumber(getTotalTime()/iterations, scale);
	}
	double getCreateTime() {
	    return createTime/1000.0;
	}
	Number getAverageCreateTime(int scale) {
	    return getScaledNumber(getCreateTime()/iterations, scale);
	}
	double getToDocumentTime() {
	    return toDocumentTime/1000.0;
	}
	Number getAverageToDocumentTime(int scale) {
	    return getScaledNumber(getToDocumentTime()/iterations, scale);
	}
	double getEditTime() {
	    return editTime/1000.0;
	}
	Number getAverageEditTime(int scale) {
	    return getScaledNumber(getEditTime()/iterations, scale);
	}
	int getEditedNodeCount() {
	    return editedNodeCount/iterations;
	}
    }

    /**
     * @return a scaled version of a number.  For presentation purposes.
     */
    Number getScaledNumber(double num, int scale)
	throws NumberFormatException {
	BigDecimal bd = new BigDecimal(num);
	return bd.setScale(scale, BigDecimal.ROUND_HALF_UP);
    }

    /**
     * @return a scaled version of a number.  For presentation purposes.
     */
    Number calcRate(double num, int scale)
	throws NumberFormatException {
	if (num == 0) {
	    return new BigDecimal(-1);
	}
	BigDecimal bd = new BigDecimal(1.0/num);
	return bd.setScale(scale, BigDecimal.ROUND_HALF_UP);
    }

    /**
     * Loads an instance of a DOM editor.
     *
     * @param className the editor's class name.
     * @return the editor.
     */
    static DOMEditor loadEditor(String className)
	throws ClassNotFoundException, InstantiationException, IllegalAccessException {
	return (DOMEditor)Class.forName(className).newInstance();
    }

    /**
     * Create a XMLC document object and convert it to a string a specified
     * number of times.
     */
    private TimeResult runTestIterations(Class docClass, DOMEditor editor,
					 int repeatCount)
	throws Exception {
        XMLCFactory factory = new XMLCStdFactory(null, null);
	long checkPoint;
	long deltaCreate = 0, deltaEdit = 0, deltaToDocument = 0;
        int editedNodeCount = 0; 

        for (int cnt = 0; cnt < repeatCount; cnt++) {

	    checkPoint = System.currentTimeMillis();
            XMLObject xmlcObj = factory.create(docClass);
	    deltaCreate
		+= (System.currentTimeMillis() - checkPoint);

	    checkPoint = System.currentTimeMillis();
            editedNodeCount += editor.edit(xmlcObj);
	    deltaEdit
		+= (System.currentTimeMillis() - checkPoint);

	    checkPoint = System.currentTimeMillis();
            String docStr = xmlcObj.toDocument();
	    deltaToDocument
		+= (System.currentTimeMillis() - checkPoint);
        }
	return new TimeResult(deltaCreate, deltaEdit, deltaToDocument,
			      repeatCount, editedNodeCount);
    }

    /**
     *  Run runTestIterations a number of times.
     */
    private void runTest(String docClassName,
			 int testCount,
			 int repeatCount,
			 DOMEditor editor,
			 PrintStream out) 
        throws Exception {
	do {
	    Class docClass = Class.forName(docClassName);
	    TimeResult timeSec[] = new TimeResult[testCount];
        
	    for (int cnt = 0; cnt < testCount; cnt++) {
		timeSec[cnt] = runTestIterations(docClass, editor, repeatCount);
	    }

	    if (out != null) {
		out.println("**Test class = " + docClass.getName());
		out.println("**Test editor = " + editor);
		out.println("**Test = " + TEST_COUNT
			    + ", Repeat = " + REPEAT_COUNT);
		out.println();
		double totalSec = 0;
		double totalCreateSec = 0;
		double totalEditSec = 0;
		double totalToDocumentSec = 0;

		for (int cnt = 0; cnt < testCount; cnt++) {
		    out.println("Pass " + cnt + ":\t"
				+ timeSec[cnt].getAverageTime(4)
				+ " sec/iteration "
				+ "(create="
				+ timeSec[cnt].getAverageCreateTime(4)
				+ ", edit="
				+ timeSec[cnt].getAverageEditTime(5)
				+ " (" + timeSec[cnt].getEditedNodeCount()
				+ " nodes)"
				+ ", toDocument="
				+ timeSec[cnt].getAverageToDocumentTime(4)
				+ ")");
		    totalSec += timeSec[cnt].getTotalTime();
		    totalCreateSec += timeSec[cnt].getCreateTime();
		    totalEditSec += timeSec[cnt].getEditTime();
		    totalToDocumentSec += timeSec[cnt].getToDocumentTime();
		}
	    
		double totalAveTimeSec =
		    totalSec/(repeatCount*testCount);
		double totalAveCreateSec =
		    totalCreateSec/(repeatCount*testCount);
		double totalAveEditSec =
		    totalEditSec/(repeatCount*testCount);
		double totalAveToDocumentSec =
		    totalToDocumentSec/(repeatCount*testCount);
		out.println();
		out.println("Average: "
			    + getScaledNumber(totalAveTimeSec, 4)
			    + " sec/iteration "
			    + "(create="
			    + getScaledNumber(totalAveCreateSec, 4)
			    + ", edit="
			    + getScaledNumber(totalAveEditSec, 4)
			    + ", toDocument="
			    + getScaledNumber(totalAveToDocumentSec, 4)
			    + ")");

		out.println("Rate:    "
			    + calcRate(totalAveTimeSec, 2)
			    + " iterations/sec "
			    + "(create="
			    + calcRate(totalAveCreateSec, 2)
			    + ", edit="
			    + calcRate(totalAveEditSec, 2)
			    + ", toDocument="
			    + calcRate(totalAveToDocumentSec, 2)
			    + ")");

		out.println();
		out.println();
	    }
	} while (editor.next());
    }

    public static void main(String[] args) throws Exception  {
	PrintStream out = System.out;

        if (args.length < 1) {
            System.err.println("wrong # args: PerfRunner class [editors]");
            System.exit(1);
        }

	try {
	    FileOutputStream fout = new FileOutputStream(TEST_RESULTS_FILE, true);
	    out = new PrintStream(fout);
	} catch (SecurityException e) {
	    System.err.println(TEST_RESULTS_FILE + ": permission denied: "
			       + e.getMessage());
	} catch (FileNotFoundException e) {
	    System.err.println(TEST_RESULTS_FILE + ": file not found: "
			       + e.getMessage());
	}	    
	
        PerfRunner tests = new PerfRunner();

	// First time get JIT going...
  	tests.runTest(args[0], TEST_COUNT,
    		      REPEAT_COUNT,
    		      loadEditor(NOOP_EDITOR),
    		      null);

	if (args.length == 1) {
	    // Now perform the real tests...
	    tests.runTest(args[0], TEST_COUNT,
			  REPEAT_COUNT,
			  loadEditor(NOOP_EDITOR),
			  out);
	} else {
	    // Run test with each specified editor...
	    for (int i=1; i<args.length; i++) {
		tests.runTest(args[0], TEST_COUNT,
			      REPEAT_COUNT,
			      loadEditor(args[i]),
			      out);
	    }
	}
    }
}






