/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SSIRead.java,v 1.4 2007-11-24 22:00:51 jkjome Exp $
 */

package org.enhydra.xml.xmlc.ssi;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import org.enhydra.xml.driver.TestCaseBase;
import org.enhydra.xml.driver.TestDiff;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.xmlc.misc.SSIReader;
import org.xml.sax.InputSource;

/**
 * Class to used in implementing SSI processing tests.
 */
public class SSIRead {
    /** 
     * Print various debugging info.  Some tests will not pass if this
     * is true, as it prints stack on IOExceptions.
     */
    //private static boolean DEBUG = false;

    /** Buffer size to use if not specified */
    private static final int DEFAULT_BUF_SIZE = 13;

    /** Test case we are assocaited with. */
    private TestCaseBase fTest;

    /** Size of read to do */
    int fReadSize;

    /** Constructor */
    public SSIRead(TestCaseBase test,
                   int readSize) {
        fTest = test;
        fReadSize = readSize;
    }

    /** Constructor */
    public SSIRead(TestCaseBase test) {
        this(test, DEFAULT_BUF_SIZE);
    }

    /**
     * Read file with SSIReader.
     */
    private void ssiCopyBuffered(File inFile,
                                 PrintWriter out) throws IOException {
        char[] buf = new char[fReadSize];
        SSIReader ssiReader = new SSIReader(new InputSource(inFile.getPath()), null);
        int readLen;
        try {
            do {
                readLen = ssiReader.read(buf, 0, fReadSize);
                if (readLen > 0) {
                    out.write(buf, 0, readLen);                    
                }
            } while (readLen != -1);
        } finally {
            ssiReader.close();
        }
    }

    /**
     * Read file with SSIReader, one char at a tupe
     */
    private void ssiCopyChar(File inFile,
                             PrintWriter out) throws IOException {
        int ch;
        SSIReader ssiReader = new SSIReader(new InputSource(inFile.getPath()), null);
        try {
            do {
                ch = ssiReader.read();
                if (ch >= 0) {
                    out.write(ch);
                }
            } while (ch != -1);
        } finally {
            ssiReader.close();
        }
    }
    
    /**
     * Read and output the file.
     */
    private void ssiCopyWrite(File inFile,
                              PrintWriter out) throws IOException {
        if (fReadSize == 1) {
            ssiCopyChar(inFile, out);
        } else {
            ssiCopyBuffered(inFile, out);
        }
    }

    /** Get the output extension */
    private String mkOutExt(File inFile) {
        String path = inFile.getPath();

        // Use same extension as input file.
        int dotIdx = path.lastIndexOf('.');
        return path.substring(dotIdx+1);
    }

    /**
     * Read file with specified buffer size and verify.
     */
    public void runOkTest(String relInFile)  {
        File inFile = fTest.getInputFile(relInFile);
        String outExt = mkOutExt(inFile);
        File outFile = fTest.getResultFile(outExt);
        File expectFile = fTest.getExpectedFile(outExt);

        try {
            PrintWriter out = new PrintWriter(new FileWriter(outFile), true);
            try {
                ssiCopyWrite(inFile, out);
            } finally {
                out.close();
            }
        } catch (IOException except) {
            throw new TestException(except);
        }

        TestDiff differ = fTest.getDiffer();
        differ.diff(expectFile, outFile);
    }

    /**
     * Run a test that should fail.
     */
    public void runFailTest(String relInFile)  {
        File inFile = fTest.getInputFile(relInFile);
        String outExt = mkOutExt(inFile);
        File outFile = fTest.getResultFile(outExt);
        File expectFile = fTest.getExpectedFile(outExt);

        try {
            PrintWriter out = new PrintWriter(new FileWriter(outFile), true);
            boolean failed = false;
            try {
                ssiCopyWrite(inFile, out);
            } catch (IOException except) {
                out.println("Test is expected to fail: " + except);
                failed = true;
            } finally {
                out.close();
            }
            if (!failed) {
                throw new TestException("test should have failed");
            }
        } catch (IOException except) {
            throw new TestException(except);
        }

        fTest.getDiffer().diff(expectFile, outFile);
    }
}
